@@ -10,19 +10,19 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.4.2.2"
+(defvar agda2-version "2.4.2.5"
   "The version of the Agda mode.
 Note that the same version of the Agda executable must be used.")
 
 (require 'cl)
-(set (make-local-variable 'lisp-indent-function)
-     'common-lisp-indent-function)
 (require 'compile)
 (require 'pp)
 (require 'time-date)
 (require 'eri)
 (require 'annotation)
+(require 'fontset)
 (require 'agda-input)
+(require 'agda2)
 (require 'agda2-highlight)
 (require 'agda2-abbrevs)
 (require 'agda2-queue)
@@ -33,6 +33,8 @@ Note that the same version of the Agda executable must be used.")
     (error nil))
   (unless (fboundp 'overlays-in) (load "overlay")) ; for Xemacs
   (unless (fboundp 'propertize)                    ; for Xemacs 21.4
+    ;; FIXME: XEmacs-21.4 (patch 22) does have `propertize' and so does Emacs-22
+    ;; (and agda2-mode doesn't work in Emacs-21, AFAICT).
     (defun propertize (string &rest properties)
       "Return a copy of STRING with text properties added.
 First argument is the string to copy.
@@ -41,12 +43,8 @@ properties to add to the result."
       (let ((str (copy-sequence string)))
         (add-text-properties 0 (length str) properties str)
         str)))
-  (unless (fboundp 'run-mode-hooks)
-    (fset 'run-mode-hooks 'run-hooks))  ; For Emacs versions < 21.
-  (unless (fboundp 'cl-labels)
-    (fset 'cl-labels 'labels))  ; For Emacs versions < 24.2.
-  (unless (fboundp 'cl-flet)
-    (fset 'cl-flet 'flet)))  ; For Emacs versions < 24.2.
+  (unless (fboundp 'prog-mode)          ;For Emacs<24.
+    (defalias 'prog-mode 'fundamental-mode)))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; Utilities
@@ -56,9 +54,10 @@ properties to add to the result."
 Or possibly (let* VARBIND (labels FUNCBIND BODY...))."
   (declare (debug ((&rest [&or symbolp (symbolp form)])
                    (&rest (cl-defun))
-                   body)))
-  `(let* ,varbind (cl-labels ,funcbind ,@body)))
-(put 'agda2-let 'lisp-indent-function 2)
+                   body))
+           (indent 2))
+  ;; Use cl-labels if available to avoid obsolescence warnings.
+  `(let* ,varbind (,(if (fboundp 'cl-labels) 'cl-labels 'labels) ,funcbind ,@body)))
 
 (defun agda2-chunkify (n xs)
   "Returns a list containing chunks of XS of length at most N.
@@ -252,6 +251,7 @@ constituents.")
     (describe-char                           nil                  (global)       "Information about the character at point")
     (agda2-comment-dwim-rest-of-buffer       ,(kbd "C-c C-x M-;") (global)       "Comment/uncomment the rest of the buffer")
     (agda2-display-program-version           nil                  (global)       "Version")
+    (agda2-set-program-version               nil                  (global)       "Switch to another version of Agda")
     (eri-indent                  ,(kbd "TAB"))
     (eri-indent-reverse          [S-iso-lefttab])
     (eri-indent-reverse          [S-lefttab])
@@ -381,9 +381,7 @@ Note that this variable is not buffer-local.")
 ;;;###autoload
 (modify-coding-system-alist 'file "\\.l?agda\\'" 'utf-8)
 ;;;###autoload
-(eval `(define-derived-mode agda2-mode
-  ,(if (fboundp 'prog-mode) 'prog-mode)
-  "Agda"
+(define-derived-mode agda2-mode prog-mode "Agda"
   "Major mode for Agda files.
 
 The following paragraph does not apply to Emacs 23 or newer.
@@ -444,7 +442,7 @@ Special commands:
  ;; including "mode: latex" is loaded chances are that the Agda mode
  ;; is activated before the LaTeX mode, and the LaTeX mode does not
  ;; seem to remove the text properties set by the Agda mode.
- (add-hook 'change-major-mode-hook 'agda2-quit nil 'local)))
+ (add-hook 'change-major-mode-hook 'agda2-quit nil 'local))
 
 (defun agda2-restart ()
   "Kill and restart the *agda2* buffer and load `agda2-toplevel-module'."
@@ -573,7 +571,13 @@ arrived. Otherwise highlighting annotations are
 reloaded from `agda2-highlighting-file', unless
 `agda2-highlighting-in-progress' is nil."
 
-
+  ;; Beware: the buffer may have been killed in the mean time.  E.g. when
+  ;; viewing an attachment containing Agda code in Gnus, Gnus will
+  ;; create a temp buffer, set it in agda2-mode, call font-lock-ensure on it
+  ;; (which won't know that it needs to wait for some process to reply), then
+  ;; extract the fontified text and kill the temp buffer; so when Agda
+  ;; finally answers, the temp buffer is long gone.
+  (when (buffer-live-p agda2-file-buffer)
   (let (;; The input lines in the current chunk.
         (lines (split-string chunk "\n"))
 
@@ -652,7 +656,7 @@ reloaded from `agda2-highlighting-file', unless
               (setq agda2-measure-data nil)
               (message "Load time: %s." elapsed)
               (when continuation
-                (funcall continuation elapsed)))))))))
+                (funcall continuation elapsed))))))))))
 
 (defun agda2-run-last-commands nil
   "Execute the last commands in the right order.
@@ -988,11 +992,12 @@ is inserted, and point is placed before this text."
 (defun agda2-term ()
   "Send a SIGTERM signal to the Agda2 process, then kill its buffer."
   (interactive)
-  (with-current-buffer agda2-process-buffer
-    (condition-case nil
-        (signal-process agda2-process 'SIGTERM)
-      (error nil))
-    (kill-buffer)))
+  (when (buffer-live-p agda2-process-buffer)
+    (with-current-buffer agda2-process-buffer
+      (condition-case nil
+          (signal-process agda2-process 'SIGTERM)
+        (error nil))
+      (kill-buffer))))
 
 (defmacro agda2-maybe-normalised (name comment cmd want)
   "This macro constructs a function NAME which runs CMD.
@@ -1220,6 +1225,15 @@ If there is any to load."
   "Is the current buffer a literate Agda buffer?"
   (equal (file-name-extension (buffer-name)) "lagda"))
 
+(defmacro agda2--case (exp &rest branches) ;FIXME: Use `pcase' instead!
+  (declare (debug t) (indent 1))
+  (let ((s (make-symbol "v")))
+    `(let ((,s ,exp))
+       (cond
+         ,@(mapcar (lambda (branch)
+                     `((equal ,s ,(car branch)) ,@(cdr branch)))
+                   branches)))))
+
 (defun agda2-goals-action (goals)
   "Annotates the goals in the current buffer with text properties.
 GOALS is a list of the buffer's goal numbers, in the order in
@@ -1267,34 +1281,33 @@ ways."
       (if literate (push 'outside stk))
       (goto-char (point-min))
       (while (and goals (safe-delims))
-        (cl-labels ((c (s) (equal s (match-string 0))))
-          (cond
-           ((c "\\begin{code}") (when (outside-code)               (pop stk)))
-           ((c "\\end{code}")   (when (not stk)                    (push 'outside stk)))
-           ((c "--")            (when (and (not stk)
-                                           (is-proper "--" t))     (end-of-line)))
-           ((c "{-")            (when (and (inside-code)
-                                           (not (inside-goal)))    (push nil           stk)))
-           ((c "-}")            (when (inside-comment)             (pop stk)))
-           ((c "{!")            (when (and (inside-code)
-                                           (not (inside-comment))) (push (- (point) 2) stk)))
-           ((c "!}")            (when (inside-goal)
-                                  (setq top (pop stk))
-                                  (unless stk (make top))))
-           ((c "?")             (progn
-                                  (when (and (not stk) (is-proper "?" nil))
-                                    (delete-char -1)
-                                    (insert "{!!}")
-                                    (make (- (point) 4)))))))))))
+        (agda2--case (match-string 0)
+          ("\\begin{code}" (when (outside-code)               (pop stk)))
+          ("\\end{code}"   (when (not stk)                    (push 'outside stk)))
+          ("--"            (when (and (not stk)
+                                      (is-proper "--" t))     (end-of-line)))
+          ("{-"            (when (and (inside-code)
+                                      (not (inside-goal)))    (push nil           stk)))
+          ("-}"            (when (inside-comment)             (pop stk)))
+          ("{!"            (when (and (inside-code)
+                                      (not (inside-comment))) (push (- (point) 2) stk)))
+          ("!}"            (when (inside-goal)
+                             (setq top (pop stk))
+                             (unless stk (make top))))
+          ("?"             (progn
+                             (when (and (not stk) (is-proper "?" nil))
+                               (delete-char -1)
+                               (insert "{!!}")
+                               (make (- (point) 4))))))))))
 
 (defun agda2-make-goal (p q n)
   "Make a goal with number N at <P>{!...!}<Q>.  Assume the region is clean."
   (annotation-preserve-mod-p-and-undo
-   (cl-flet ((atp (x ps) (add-text-properties x (1+ x) ps)))
-     (atp p       '(category agda2-delim1))
-     (atp (1+ p)  '(category agda2-delim2))
-     (atp (- q 2) '(category agda2-delim3))
-     (atp (1- q)  '(category agda2-delim4)))
+   (let ((atp (lambda (x ps) (add-text-properties x (1+ x) ps))))
+     (funcall atp p       '(category agda2-delim1))
+     (funcall atp (1+ p)  '(category agda2-delim2))
+     (funcall atp (- q 2) '(category agda2-delim3))
+     (funcall atp (1- q)  '(category agda2-delim4)))
    (let ((o (make-overlay p q nil t nil)))
      (overlay-put o 'modification-hooks '(agda2-protect-goal-markers))
      (overlay-put o 'agda2-gn           n)
@@ -1561,12 +1574,10 @@ to display the given position."
   "Go to the definition site of the name clicked on, if any.
 Otherwise, yank (see `mouse-yank-primary')."
   (interactive "e")
-  (let ((pos (posn-point (event-end ev))))
-    (if (annotation-goto-possible pos)
-        (annotation-goto-indirect pos)
-      ;; FIXME: Shouldn't we use something like
-      ;; (call-interactively (key-binding ev))?  --Stef
-      (mouse-yank-primary ev))))
+  (unless (annotation-goto-indirect ev)
+    ;; FIXME: Shouldn't we use something like
+    ;; (call-interactively (key-binding ev))?  --Stef
+    (mouse-yank-primary ev)))
 
 (defun agda2-go-back nil
   "Go back to the previous position in which
@@ -1601,5 +1612,90 @@ the argument is a positive number, otherwise turn it off."
            (call-interactively
             (lookup-key agda2-goal-map (apply 'vector choice)))))))
 
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;; Switching to a different version of Agda
+
+;; Note that other versions of Agda may use different protocols, so
+;; this function unloads the Emacs mode.
+
+(defun agda2-set-program-version (version)
+  "Tries to switch to Agda version VERSION.
+
+This command assumes that the agda and agda-mode executables for
+Agda version VERSION are called agda-VERSION and
+agda-mode-VERSION, and that they are located on the PATH. (If
+VERSION is empty, then agda and agda-mode are used instead.)"
+  (interactive "M")
+
+  (let*
+      ((agda-buffers
+        (mapcan (lambda (buf)
+                  (with-current-buffer buf
+                    (when (equal major-mode 'agda2-mode)
+                      (list buf))))
+                (buffer-list)))
+
+       (version-suffix (if (or (equal version "")
+                               (equal version nil))
+                           ""
+                         (concat "-" version)))
+
+       ;; Run agda-mode<version-suffix> and make sure that it returns
+       ;; successfully.
+       (coding-system-for-read 'utf-8)
+       (agda-mode-prog (concat "agda-mode" version-suffix))
+       (agda-mode-path
+        (condition-case nil
+            (with-temp-buffer
+              (unless
+                  (equal 0 (call-process agda-mode-prog
+                                         nil (current-buffer) nil
+                                         "locate"))
+                (error "%s" (concat "Error when running "
+                                    agda-mode-prog)))
+              (buffer-string))
+          (file-error
+           (error "%s" (concat "Could not find " agda-mode-prog))))))
+
+    ;; Make sure that agda-mode<version-suffix> returns a valid file.
+    (unless (file-readable-p agda-mode-path)
+      (error "%s" (concat "Could not read " agda-mode-path)))
+
+    ;; Kill some processes/buffers related to Agda.
+    (when (and agda2-process
+               (process-status agda2-process))
+      (kill-process agda2-process))
+    (when (buffer-live-p agda2-process-buffer)
+      (kill-buffer agda2-process-buffer))
+    (when (buffer-live-p agda2-info-buffer)
+      (kill-buffer agda2-info-buffer))
+    (when (and agda2-debug-buffer-name
+               (get-buffer agda2-debug-buffer-name))
+      (kill-buffer agda2-debug-buffer-name))
+
+    ;; Remove the Agda mode directory from the load path.
+    (setq load-path (delete agda2-directory load-path))
+
+    ;; Unload the Agda mode and its dependencies.
+    (unload-feature 'agda2-mode      'force)
+    (unload-feature 'agda2           'force)
+    (unload-feature 'eri             'force)
+    (unload-feature 'annotation      'force)
+    (unload-feature 'agda-input      'force)
+    (unload-feature 'agda2-highlight 'force)
+    (unload-feature 'agda2-abbrevs   'force)
+    (unload-feature 'agda2-queue     'force)
+
+    ;; Load the new version of Agda.
+    (load-file agda-mode-path)
+    (require 'agda2-mode)
+    (setq agda2-program-name (concat "agda" version-suffix))
+
+    ;; Restart the Agda mode in all former Agda mode buffers.
+    (mapc (lambda (buf)
+            (with-current-buffer buf
+              (agda2-mode)))
+          agda-buffers)))
+
 (provide 'agda2-mode)
 ;;; agda2-mode.el ends here
