@@ -1,5 +1,9 @@
 {-# LANGUAGE CPP #-}
 
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
+
 module Agda.Auto.Convert where
 
 import Control.Applicative hiding (getConst, Const(..))
@@ -41,6 +45,7 @@ import Agda.Utils.Except
   , ExceptT
   , MonadError(catchError, throwError)
   )
+import Agda.Utils.Lens
 
 import Agda.Utils.Impossible
 #include "undefined.h"
@@ -462,7 +467,7 @@ fmLevel :: I.MetaId -> I.PlusLevel -> Bool
 fmLevel m I.ClosedLevel{} = False
 fmLevel m (I.Plus _ l) = case l of
   I.MetaLevel m' _   -> m == m'
-  I.NeutralLevel v   -> fmExp m v
+  I.NeutralLevel _ v -> fmExp m v
   I.BlockedLevel _ v -> fmExp m v
   I.UnreducedLevel v -> fmExp m v
 
@@ -667,7 +672,7 @@ frommyClause (ids, pats, mrhs) = do
         let (Just ndrop, name) = cdorigin cdef
         ps' <- cnvps ndrop ps
         let con = I.ConHead name Common.Inductive [] -- TODO: restore record fields!
-        return (I.ConP con Nothing ps')
+        return (I.ConP con I.noConPatternInfo ps')
        CSPatExp e -> do
         e' <- frommyExp e {- renm e -} -- renaming before adding to clause below
         return (I.DotP e')
@@ -749,7 +754,7 @@ findClauseDeep :: I.MetaId -> MB.TCM (Maybe (AN.QName, I.Clause, Bool))
 findClauseDeep m = do
   sig <- getImportedSignature
   let res = do
-        def <- HMap.elems $ MB.sigDefinitions sig
+        def <- HMap.elems $ sig ^. MB.sigDefinitions
         MB.Function{MB.funClauses = cs} <- [MB.theDef def]
         c <- cs
         unless (peelbinds False findMeta $ I.clauseBody c) []
