@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP                 #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.Compiler.Epic.Forcing where
@@ -47,7 +47,7 @@ dataParameters = lift . dataParametersTCM
 -- | Returns how many parameters a datatype has
 dataParametersTCM :: QName -> TCM Nat
 dataParametersTCM name = do
-    m <- (sigDefinitions <$> use stImports)
+    m <- use $ stImports . sigDefinitions
     return $ maybe __IMPOSSIBLE__ (defnPars . theDef) (HM.lookup name m)
   where
     defnPars :: Defn -> Nat
@@ -123,7 +123,7 @@ insertTele x 0 ins term (ExtendTel t to) = do
 -}
     when (genericLength arg < pars) __IMPOSSIBLE__
     -- we deal with absBody to directly since we remove t
-    return ( ctele +:+  (S.subst term $ S.raiseFrom 1 (size ctele) (unAbs to))
+    return ( ctele +:+  (S.subst 0 term $ S.raiseFrom 1 (size ctele) (unAbs to))
            , (ctele, S.raise (size ctele) $ unDom t , ctyp)
            )
   where
@@ -154,7 +154,7 @@ takeTele _ _ = __IMPOSSIBLE__
 -- | Main function for removing pattern matching on forced variables
 remForced :: [Fun] -> Compile TCM [Fun]
 remForced fs = do
-    defs <- lift (sigDefinitions  <$> use stImports)
+    defs <- lift $ use $ stImports . sigDefinitions
     forM fs $ \f -> case f of
         Fun{} -> case funQName f >>= flip HM.lookup defs of
             Nothing -> __IMPOSSIBLE__
