@@ -1,9 +1,6 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE TypeOperators        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
-{-# LANGUAGE OverlappingInstances #-}
+{-# LANGUAGE CPP               #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards     #-}
 
 module Agda.Compiler.Epic.Injection where
 
@@ -26,7 +23,7 @@ import Agda.Syntax.Literal
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Pretty hiding (empty)
+import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.Utils.List
@@ -69,8 +66,8 @@ findInjection defs = do
 
 replaceFunCC :: QName -> CompiledClauses -> Compile TCM ()
 replaceFunCC name cc = lift $ do
-    stSignature %= \sig -> sig {sigDefinitions = HM.adjust replaceDef name (sigDefinitions sig)}
-    stImports   %= \imp -> imp {sigDefinitions = HM.adjust replaceDef name (sigDefinitions imp)}
+    stSignature %= updateDefinition name replaceDef
+    stImports   %= updateDefinition name replaceDef
   where
     replaceDef :: Definition -> Definition
     replaceDef def = case theDef def of
@@ -123,7 +120,7 @@ substForDot = makeSubst 0 0 . reverse . calcDots
   where
     makeSubst i accum [] = raiseS (i + accum)
     makeSubst i accum (True  : ps) = makeSubst i (accum +1) ps
-    makeSubst i accum (False : ps) = var (i + accum) :# makeSubst (i+1) accum ps
+    makeSubst i accum (False : ps) = consS (var $ i + accum) $ makeSubst (i+1) accum ps
 
     calcDots = concatMap calcDots' . map namedArg
     calcDots' p = case p of
@@ -186,8 +183,8 @@ litInt LitInt{} = True
 litInt _        = False
 
 insertAt :: (Nat,Term) -> Term -> Term
-insertAt (index, ins) =
-  applySubst ([var i | i <- [0 .. index - 1]] ++# ins :# raiseS (index + 1))
+insertAt (index, ins) = applySubst $
+  [var i | i <- [0 .. index - 1]] ++# consS ins (raiseS $ index + 1)
 
 solve :: [QName] -> [((QName, InjectiveFun), [(QName,QName)])] -> Compile TCM [(QName, InjectiveFun)]
 solve newNames xs = do
@@ -216,7 +213,7 @@ solve newNames xs = do
         let (hasTags, eqs) = Map.partition isTag (constrGroup tags)
             isTag (IsTag _) = True
             isTag _         = False
-        forM (Map.toList hasTags) $ \ (c, tagged) -> case tagged of
+        forM_ (Map.toList hasTags) $ \ (c, tagged) -> case tagged of
             IsTag tag -> putCon c tag
             _         -> __IMPOSSIBLE__
         case Map.toList eqs of
