@@ -7,8 +7,8 @@ module Agda.Compiler.Epic.Primitive where
 import Control.Applicative
 import Control.Monad
 import Control.Monad.Trans
-import Data.Map(Map)
-import qualified Data.Map as M
+import Data.Map (Map)
+import qualified Data.Map as Map
 import Data.Maybe
 
 import Agda.Syntax.Internal(QName)
@@ -22,9 +22,11 @@ import Agda.Compiler.Epic.CompileState
 import Agda.Compiler.Epic.Interface
 import Agda.Compiler.Epic.NatDetection
 
-#include "../../undefined.h"
+import Agda.Utils.List
 import Agda.Utils.Impossible
 
+#include "undefined.h"
+
 {- Stacken, Heapen -- Optimizern -}
 
 data PrimTransform = PrimTF
@@ -97,31 +99,29 @@ getBuiltins =
                 if b then return $ Just (transf names) else return Nothing
            else return Nothing
 
+defName :: T.Term -> QName
 defName (T.Def q []) = q
 defName (T.Con q []) = T.conName q
 defName _            = __IMPOSSIBLE__
 
-head'' (x:xs) e = x
-head'' _      e = e
-
 -- | Translation to primitive integer functions
 natPrimTF :: ForcedArgs -> [QName] -> PrimTransform
 natPrimTF filt [zero, suc] = PrimTF
-  { mapCon = M.fromList [(zero, prZero), (suc, prSuc)]
+  { mapCon = Map.fromList [(zero, prZero), (suc, prSuc)]
   , translateCase = \ce brs -> case brs of
         -- Assuming only the first two branches are relevant when casing on Nats
         (Branch _ n vs e:Branch _ _n' vs'' e'':_) ->
             if n == zero
-               then primNatCaseZS ce e  (head'' vs'' __IMPOSSIBLE__) e''
-               else primNatCaseZS ce e'' (head'' vs __IMPOSSIBLE__) e
+               then primNatCaseZS ce e  (headWithDefault __IMPOSSIBLE__ vs'') e''
+               else primNatCaseZS ce e'' (headWithDefault __IMPOSSIBLE__ vs) e
         (Branch _ n vs e:Default e'':_) ->
             if n == zero
                then primNatCaseZD ce e e'' -- zero
-               else primNatCaseZS ce e'' (head'' vs __IMPOSSIBLE__) e -- suc
+               else primNatCaseZS ce e'' (headWithDefault __IMPOSSIBLE__ vs) e -- suc
         [ Branch _ n vs e ] ->
             if n == zero
               then e
-              else lett (head'' vs __IMPOSSIBLE__) (App prPred [ce]) e
+              else lett (headWithDefault __IMPOSSIBLE__ vs) (App prPred [ce]) e
         _ -> __IMPOSSIBLE__
   }
 natPrimTF _ _ = __IMPOSSIBLE__
@@ -145,7 +145,7 @@ primNatCaseZD n zeroBr defBr = If (App prNatEquality [n, Var prZero]) zeroBr def
 -- | Translation to primitive bool functions
 boolPrimTF :: [QName] -> PrimTransform
 boolPrimTF [true, false] = PrimTF
-  { mapCon = M.fromList [(true, prTrue), (false, prFalse)]
+  { mapCon = Map.fromList [(true, prTrue), (false, prFalse)]
   , translateCase = \ce brs ->
     case brs of
         (Branch _ n _vs e:b':_) ->
@@ -189,7 +189,7 @@ primExpr prim ex = case ex of
     --   case, otherwise Nothing.
     testCon :: [PrimTransform] -> QName -> Maybe Var
     testCon [] _ = Nothing
-    testCon (p : ps) k = M.lookup k (mapCon p) `mplus` testCon ps k
+    testCon (p : ps) k = Map.lookup k (mapCon p) `mplus` testCon ps k
 
     -- | Test if we should transform the case, based on the branches. Returns
     --   the (first) PrimTransform that is applicable.
@@ -201,7 +201,7 @@ primExpr prim ex = case ex of
     --   Returns the PrimTransform in that case.
     check :: PrimTransform -> Branch -> Maybe PrimTransform
     check p br = case br of
-        Branch  _ n _ _ -> fmap (const p) $ M.lookup n (mapCon p)
+        Branch  _ n _ _ -> fmap (const p) $ Map.lookup n (mapCon p)
         BrInt _ _       -> Nothing
         Default _       -> Nothing
 
