@@ -1,6 +1,5 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
 
 -- | Find the places where the builtin static is used and do some normalisation
 --   there.
@@ -45,7 +44,8 @@ evaluateCC ccs = case ccs of
             Nothing -> return Nothing
             Just cc -> Just <$> evaluateCC cc
         return $ Case n Branches
-            { conBranches    = Map.fromList cbrs
+            { projPatterns   = projPatterns brs
+            , conBranches    = Map.fromList cbrs
             , litBranches    = Map.fromList lbrs
             , catchAllBranch = cab
             }
@@ -54,7 +54,7 @@ evaluateCC ccs = case ccs of
 
 etaExpand :: Term -> Compile TCM Term
 etaExpand def@(Def n ts) = do
-    defs <- lift (sigDefinitions <$> use stImports)
+    defs <- lift $ use $ stImports . sigDefinitions
     let f   = maybe __IMPOSSIBLE__ theDef (HM.lookup n defs)
         len = length . clausePats . head .  funClauses $ f
         toEta :: Num a => a
@@ -115,7 +115,7 @@ instance Evaluate Term where
 -}
     isStatic :: QName -> Compile TCM Bool
     isStatic q = do
-      defs <- lift (sigDefinitions <$> use stImports)
+      defs <- lift $ use $ stImports . sigDefinitions
       return $ case fmap theDef $ HM.lookup q defs of
           Nothing -> False
           Just (f@Function{}) -> funStatic f
