@@ -54,10 +54,10 @@ curHsMod :: TCM HS.ModuleName
 curHsMod = mazMod <$> curMName
 
 curDefs :: TCM Definitions
-curDefs = sigDefinitions <$> curSig
+curDefs = (^. sigDefinitions) <$> curSig
 
 sigMName :: Signature -> ModuleName
-sigMName sig = case Map.keys (sigSections sig) of
+sigMName sig = case Map.keys (sig ^. sigSections) of
   []    -> __IMPOSSIBLE__
   m : _ -> m
 
@@ -128,7 +128,6 @@ conhqn :: QName -> TCM HS.QName
 conhqn q = do
     cq  <- canonicalName q
     def <- getConstInfo cq
-    hsr <- compiledHaskell . defCompiledRep <$> getConstInfo cq
     case (compiledHaskell (defCompiledRep def), theDef def) of
       (Just (HsDefn _ hs), Constructor{}) -> return $ HS.UnQual $ HS.Ident hs
       _                                   -> xhqn "C" cq
@@ -204,7 +203,7 @@ unsafeCoerceMod = HS.ModuleName "Unsafe.Coerce"
 fakeD :: HS.Name -> String -> HS.Decl
 fakeD v s = HS.FunBind [ HS.Match dummy v [] Nothing
                            (HS.UnGuardedRhs $ hsVarUQ $ HS.Ident $ s)
-                           (HS.BDecls [])
+                           emptyBinds
                        ]
 
 fakeDS :: String -> String -> HS.Decl
@@ -221,3 +220,15 @@ fakeExp = HS.Var . HS.UnQual . HS.Ident
 
 dummy :: a
 dummy = error "MAlonzo : this dummy value should not have been eval'ed."
+
+--------------------------------------------------
+-- Auxiliary definitions
+--------------------------------------------------
+
+#if MIN_VERSION_haskell_src_exts(1,17,0)
+emptyBinds :: Maybe HS.Binds
+emptyBinds = Nothing
+#else
+emptyBinds :: HS.Binds
+emptyBinds = HS.BDecls []
+#endif
