@@ -1,9 +1,11 @@
-{-# LANGUAGE CPP                   #-}
-{-# LANGUAGE DeriveFunctor         #-}
-{-# LANGUAGE FlexibleInstances     #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE TupleSections         #-}
-{-# LANGUAGE UndecidableInstances  #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE UndecidableInstances #-}
+
+{-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.Interaction.BasicOps where
 
@@ -95,30 +97,30 @@ giveExpr mi e = do
       ctx <- getContextArgs
       let t' = t `piApply` permute (takeP (length ctx) $ mvPermutation mv) ctx
       traceCall (CheckExprCall e t') $ do
-      reportSDoc "interaction.give" 20 $
-        TP.text "give: instantiated meta type =" TP.<+> prettyTCM t'
-      v <- checkExpr e t'
-      case mvInstantiation mv of
-        InstV xs v' -> unlessM ((Irrelevant ==) <$> asks envRelevance) $ do
-          reportSDoc "interaction.give" 20 $ TP.sep
-            [ TP.text "meta was already set to value v' = " TP.<+> prettyTCM v'
-              TP.<+> TP.text " with free variables " TP.<+> return (fsep $ map pretty xs)
-            , TP.text "now comparing it to given value v = " TP.<+> prettyTCM v
-            , TP.text "in context " TP.<+> inTopContext (prettyTCM ctx)
-            ]
-          -- The number of free variables should be at least the size of the context
-          -- (Ideally, if we implemented contextual type theory, it should be the same.)
-          when (length xs < size ctx) __IMPOSSIBLE__
-          -- if there are more free variables than the context has
-          -- we need to abstract over the additional ones (xs2)
-          let (_xs1, xs2) = splitAt (size ctx) xs
-          v' <- return $ foldr (\ (Arg ai x) -> I.Lam ai . I.Abs x) v' xs2
-          reportSDoc "interaction.give" 20 $ TP.sep
-            [ TP.text "in meta context, v' = " TP.<+> prettyTCM v'
-            ]
-          equalTerm t' v v'  -- Note: v' now lives in context of meta
-        _ -> updateMeta mi v
-      reify v
+        reportSDoc "interaction.give" 20 $
+          TP.text "give: instantiated meta type =" TP.<+> prettyTCM t'
+        v <- checkExpr e t'
+        case mvInstantiation mv of
+          InstV xs v' -> unlessM ((Irrelevant ==) <$> asks envRelevance) $ do
+            reportSDoc "interaction.give" 20 $ TP.sep
+              [ TP.text "meta was already set to value v' = " TP.<+> prettyTCM v'
+                TP.<+> TP.text " with free variables " TP.<+> return (fsep $ map pretty xs)
+              , TP.text "now comparing it to given value v = " TP.<+> prettyTCM v
+              , TP.text "in context " TP.<+> inTopContext (prettyTCM ctx)
+              ]
+            -- The number of free variables should be at least the size of the context
+            -- (Ideally, if we implemented contextual type theory, it should be the same.)
+            when (length xs < size ctx) __IMPOSSIBLE__
+            -- if there are more free variables than the context has
+            -- we need to abstract over the additional ones (xs2)
+            let (_xs1, xs2) = splitAt (size ctx) xs
+            v' <- return $ foldr (\ (Arg ai x) -> I.Lam ai . I.Abs x) v' xs2
+            reportSDoc "interaction.give" 20 $ TP.sep
+              [ TP.text "in meta context, v' = " TP.<+> prettyTCM v'
+              ]
+            equalTerm t' v v'  -- Note: v' now lives in context of meta
+          _ -> updateMeta mi v
+        reify v
 
 -- | Try to fill hole by expression.
 --
@@ -136,7 +138,7 @@ give ii mr e = liftTCM $ do
   reportSDoc "interaction.give" 10 $ TP.text "giving expression" TP.<+> prettyTCM e
   reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnScope e
   -- Try to give mi := e
-  giveExpr mi e `catchError` \ err -> case err of
+  _ <- catchError (giveExpr mi e) $ \ err -> case err of
     -- Turn PatternErr into proper error:
     PatternErr{} -> do
       err <- withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
@@ -230,7 +232,9 @@ data OutputConstraint a b
       | JustSort b | CmpSorts Comparison b b
       | Guard (OutputConstraint a b) ProblemId
       | Assign b a | TypedAssign b a a | PostponedCheckArgs b [a] a a
-      | IsEmptyType a | FindInScopeOF b a [(a,a)]
+      | IsEmptyType a
+      | SizeLtSat a
+      | FindInScopeOF b a [(a,a)]
   deriving (Functor)
 
 -- | A subset of 'OutputConstraint'.
@@ -258,6 +262,7 @@ outputFormId (OutputForm _ _ o) = out o
       TypedAssign i _ _          -> i
       PostponedCheckArgs i _ _ _ -> i
       IsEmptyType _              -> __IMPOSSIBLE__   -- Should never be used on IsEmpty constraints
+      SizeLtSat{}                -> __IMPOSSIBLE__
       FindInScopeOF _ _ _        -> __IMPOSSIBLE__
 
 instance Reify ProblemConstraint (Closure (OutputForm Expr Expr)) where
@@ -277,44 +282,49 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
         return $ Guard o pid
     reify (UnBlock m) = do
         mi <- mvInstantiation <$> lookupMeta m
+        m' <- reify (MetaV m [])
         case mi of
           BlockedConst t -> do
             e  <- reify t
-            m' <- reify (MetaV m [])
             return $ Assign m' e
           PostponedTypeCheckingProblem cl _ -> enterClosure cl $ \p -> case p of
             CheckExpr e a -> do
                 a  <- reify a
-                m' <- reify (MetaV m [])
                 return $ TypedAssign m' e a
+            CheckLambda (Arg ai (xs, mt)) body target -> do
+              domType <- maybe (return underscore) reify mt
+              target  <- reify target
+              let bs = TypedBindings noRange $ Arg (mapArgInfoColors (const []) ai) $
+                       TBind noRange xs domType
+                  e  = A.Lam Info.exprNoRange (DomainFull bs) body
+              return $ TypedAssign m' e target
             CheckArgs _ _ _ args t0 t1 _ -> do
               t0 <- reify t0
               t1 <- reify t1
-              m  <- reify (MetaV m [])
-              return $ PostponedCheckArgs m (map (namedThing . unArg) args) t0 t1
+              return $ PostponedCheckArgs m' (map (namedThing . unArg) args) t0 t1
           Open{}  -> __IMPOSSIBLE__
           OpenIFS{}  -> __IMPOSSIBLE__
           InstS{} -> __IMPOSSIBLE__
           InstV{} -> __IMPOSSIBLE__
-    reify (FindInScope m mcands) = do
-      let cands = caseMaybe mcands [] (\ x -> x)
-      m' <- reify (MetaV m [])
-      ctxArgs <- getContextArgs
-      t <- getMetaType m
-      t' <- reify t
-      cands' <- mapM (\(tm,ty) -> (,) <$> reify tm <*> reify ty) cands
-      return $ FindInScopeOF m' t' cands' -- IFSTODO
+    reify (FindInScope m mcands) = FindInScopeOF
+      <$> (reify $ MetaV m [])
+      <*> (reify =<< getMetaType m)
+      <*> (forM (fromMaybe [] mcands) $ \ (tm, ty) -> do
+            (,) <$> reify tm <*> reify ty)
     reify (IsEmpty r a) = IsEmptyType <$> reify a
+    reify (CheckSizeLtSat a) = SizeLtSat  <$> reify a
 
+-- ASR TODO (28 December 2014): This function will be unnecessary when
+-- using a Pretty instance for OutputConstraint instead of the Show
+-- instance.
 showComparison :: Comparison -> String
-showComparison CmpEq  = " = "
-showComparison CmpLeq = " =< "
+showComparison cmp = " " ++ prettyShow cmp ++ " "
 
 instance (Show a,Show b) => Show (OutputForm a b) where
   show o =
     case o of
       OutputForm r 0   c -> show c ++ range r
-      OutputForm r pid c -> "[" ++ show pid ++ "] " ++ show c ++ range r
+      OutputForm r pid c -> "[" ++ prettyShow pid ++ "] " ++ show c ++ range r
     where
       range r | null s    = ""
               | otherwise = " [ at " ++ s ++ " ]"
@@ -330,7 +340,7 @@ instance (Show a,Show b) => Show (OutputConstraint a b) where
     show (CmpLevels cmp t t')   = show t ++ showComparison cmp ++ show t'
     show (CmpTeles  cmp t t')   = show t ++ showComparison cmp ++ show t'
     show (CmpSorts cmp s s')    = show s ++ showComparison cmp ++ show s'
-    show (Guard o pid)          = show o ++ " [blocked by problem " ++ show pid ++ "]"
+    show (Guard o pid)          = show o ++ " [blocked by problem " ++ prettyShow pid ++ "]"
     show (Assign m e)           = show m ++ " := " ++ show e
     show (TypedAssign m e a)    = show m ++ " := " ++ show e ++ " :? " ++ show a
     show (PostponedCheckArgs m es t0 t1) = show m ++ " := (_ : " ++ show t0 ++ ") " ++ unwords (map (paren . show) es)
@@ -338,9 +348,12 @@ instance (Show a,Show b) => Show (OutputConstraint a b) where
       where paren s | elem ' ' s = "(" ++ s ++ ")"
                     | otherwise  = s
     show (IsEmptyType a)        = "Is empty: " ++ show a
-    show (FindInScopeOF s t cs) = "Resolve instance argument " ++ showCand (s,t) ++ ". Candidates: [" ++
-                                    intercalate ", " (map showCand cs) ++ "]"
-      where showCand (tm,ty) = show tm ++ " : " ++ show ty
+    show (SizeLtSat a)          = "Not empty type of sizes: " ++ show a
+    show (FindInScopeOF s t cs) = "Resolve instance argument " ++ showCand (s,t) ++ ".\n  Candidates:\n    [ " ++
+                                    intercalate "\n    , " (map showCand cs) ++ " ]"
+      where showCand (tm,ty) = indent 6 $ show tm ++ " : " ++ show ty
+            indent n s = intercalate ("\n" ++ replicate n ' ') (l:ls)
+              where l:ls = lines s
 
 instance (ToConcrete a c, ToConcrete b d) =>
          ToConcrete (OutputForm a b) (OutputForm c d) where
@@ -370,6 +383,7 @@ instance (ToConcrete a c, ToConcrete b d) =>
     toConcrete (PostponedCheckArgs m args t0 t1) =
       PostponedCheckArgs <$> toConcrete m <*> toConcrete args <*> toConcrete t0 <*> toConcrete t1
     toConcrete (IsEmptyType a) = IsEmptyType <$> toConcreteCtx TopCtx a
+    toConcrete (SizeLtSat a) = SizeLtSat <$> toConcreteCtx TopCtx a
     toConcrete (FindInScopeOF s t cs) =
       FindInScopeOF <$> toConcrete s <*> toConcrete t
                     <*> mapM (\(tm,ty) -> (,) <$> toConcrete tm <*> toConcrete ty) cs
@@ -381,18 +395,6 @@ instance (ToConcrete a c, ToConcrete b d) =>
             ToConcrete (OutputConstraint' a b) (OutputConstraint' c d) where
   toConcrete (OfType' e t) = OfType' <$> toConcrete e <*> toConcreteCtx TopCtx t
 
---ToDo: Move somewhere else
-instance ToConcrete InteractionId C.Expr where
-    toConcrete (InteractionId i) = return $ C.QuestionMark noRange (Just i)
-
-instance ToConcrete NamedMeta C.Expr where
-    toConcrete i = do
-      return $ C.Underscore noRange (Just $ show i)
-
-judgToOutputForm :: Judgement a c -> OutputConstraint a c
-judgToOutputForm (HasType e t) = OfType e t
-judgToOutputForm (IsSort  s t) = JustSort s
-
 getConstraints :: TCM [OutputForm C.Expr C.Expr]
 getConstraints = liftTCM $ do
     cs <- M.getAllConstraints
@@ -445,7 +447,7 @@ typeOfMetaMI norm mi =
         withMetaInfo (getMetaInfo mv) $
           rewriteJudg mv (mvJudgement mv)
    where
-    rewriteJudg :: MetaVariable -> Judgement Type MetaId ->
+    rewriteJudg :: MetaVariable -> Judgement MetaId ->
                    TCM (OutputConstraint Expr NamedMeta)
     rewriteJudg mv (HasType i t) = do
       ms <- getMetaNameSuggestion i
@@ -458,9 +460,10 @@ typeOfMetaMI norm mi =
           [ TP.text "len  =" TP.<+> TP.text (show $ length vs)
           , TP.text "args =" TP.<+> prettyTCM vs
           , TP.text "t    =" TP.<+> prettyTCM t
-          , TP.text "x    =" TP.<+> TP.text (show x)
+          , TP.text "x    =" TP.<+> TP.pretty x
           ]
         ]
+      reportSDoc "interactive.meta.scope" 20 $ TP.text $ show $ getMetaScope mv
       OfType x <$> reify (t `piApply` permute (takeP (size vs) $ mvPermutation mv) vs)
     rewriteJudg mv (IsSort i t) = do
       ms <- getMetaNameSuggestion i
@@ -483,9 +486,14 @@ typesOfHiddenMetas norm = liftTCM $ do
   store <- Map.filterWithKey (openAndImplicit is) <$> getMetaStore
   mapM (typeOfMetaMI norm) $ Map.keys store
   where
-  openAndImplicit is x (MetaVar{mvInstantiation = M.Open}) = x `notElem` is
-  openAndImplicit is x (MetaVar{mvInstantiation = M.BlockedConst _}) = True
-  openAndImplicit _  _ _                                    = False
+  openAndImplicit is x m =
+    case mvInstantiation m of
+      M.InstV{} -> False
+      M.InstS{} -> False
+      M.Open    -> x `notElem` is
+      M.OpenIFS -> x `notElem` is  -- OR: True !?
+      M.BlockedConst{} -> True
+      M.PostponedTypeCheckingProblem{} -> False
 
 metaHelperType :: Rewrite -> InteractionId -> Range -> String -> TCM (OutputConstraint' Expr Expr)
 metaHelperType norm ii rng s = case words s of
@@ -502,15 +510,9 @@ metaHelperType norm ii rng s = case words s of
       -- Remember the arity of a
       TelV atel _ <- telView a
       let arity = size atel
-          fv = allVars $ freeVars (vs, as)
-          SplitTel delta1 delta2 perm = splitTelescope fv tel
-          a' = renameP (reverseP perm) a
-      (vs, as) <- do
-        let -- We know that as does not depend on Δ₂
-            rho = compactS __IMPOSSIBLE__ (replicate (size delta2) Nothing)
-        return $ applySubst rho $ renameP (reverseP perm) (vs, as)
+          (delta1, delta2, _, a', as', vs') = splitTelForWith tel a as vs
       a <- local (\e -> e { envPrintDomainFreePi = True }) $ do
-        reify =<< cleanupType arity args =<< normalForm norm =<< withFunctionType delta1 vs as delta2 a'
+        reify =<< cleanupType arity args =<< normalForm norm =<< withFunctionType delta1 vs' as' delta2 a'
       return (OfType' h a)
   where
     cleanupType arity args t = do
@@ -720,17 +722,14 @@ introTactic pmLambda ii = do
         Left err -> return []
         Right cov -> mapM showTCM $ concatMap (conName . scPats) $ splitClauses cov
 
+    introRec :: QName -> TCM [String]
     introRec d = do
       hfs <- getRecordFieldNames d
       fs <- ifM showImplicitArguments
             (return $ map unArg hfs)
             (return [ unArg a | a <- hfs, getHiding a == NotHidden ])
-      return
-        [ concat $
-            "record {" :
-            intersperse ";" (map (\ f -> show f ++ " = ?") fs) ++
-            ["}"]
-        ]
+      let e = C.Rec noRange $ map (, C.QuestionMark noRange Nothing) fs
+      return [ prettyShow e ]
 
 -- | Runs the given computation as if in an anonymous goal at the end
 --   of the top-level module.
@@ -799,4 +798,3 @@ whyInScope s = do
   return ( lookup x $ map (first C.QName) $ scopeLocals scope
          , scopeLookup x scope
          , scopeLookup x scope )
-
