@@ -1,6 +1,6 @@
-{-# LANGUAGE CPP              #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE RelaxedPolyRec   #-}
+{-# LANGUAGE RelaxedPolyRec #-}
 
 -- | Generates data used for precise syntax highlighting.
 
@@ -20,41 +20,15 @@ module Agda.Interaction.Highlighting.Generate
 
 import Prelude hiding (null)
 
-import Agda.Interaction.FindFile
-import Agda.Interaction.Response (Response(Resp_HighlightingInfo))
-import Agda.Interaction.Highlighting.Precise hiding (tests)
-import Agda.Interaction.Highlighting.Range   hiding (tests)
-import qualified Agda.TypeChecking.Errors as E
-import Agda.TypeChecking.MetaVars (isBlockedTerm)
-import Agda.TypeChecking.Monad
-  hiding (MetaInfo, Primitive, Constructor, Record, Function, Datatype)
-import qualified Agda.TypeChecking.Monad as M
-import Agda.TypeChecking.Pretty
-import qualified Agda.TypeChecking.Reduce as R
-import qualified Agda.Syntax.Abstract as A
-import Agda.Syntax.Common (Delayed(..))
-import qualified Agda.Syntax.Common as SC
-import qualified Agda.Syntax.Concrete as C
-import qualified Agda.Syntax.Info as SI
-import qualified Agda.Syntax.Internal as I
-import qualified Agda.Syntax.Literal as L
-import qualified Agda.Syntax.Parser as Pa
-import qualified Agda.Syntax.Parser.Tokens as T
-import qualified Agda.Syntax.Position as P
-import Agda.Utils.Lens
-import Agda.Utils.List
-import Agda.Utils.TestHelpers
-import Agda.Utils.HashMap (HashMap)
-import qualified Agda.Utils.HashMap as HMap
 import Control.Monad
 import Control.Monad.Trans
 import Control.Monad.State
 import Control.Monad.Reader
 import Control.Applicative
 import Control.Arrow ((***), first, second)
+
 import Data.Monoid
 import Data.Generics.Geniplate
-import Agda.Utils.FileName
 import Data.HashSet (HashSet)
 import qualified Data.HashSet as HSet
 import qualified Data.Map as Map
@@ -64,9 +38,39 @@ import qualified Data.Foldable as Fold (toList, fold, foldMap)
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
 
+import Agda.Interaction.FindFile
+import Agda.Interaction.Response (Response(Resp_HighlightingInfo))
+import Agda.Interaction.Highlighting.Precise hiding (tests)
+import Agda.Interaction.Highlighting.Range   hiding (tests)
+
+import qualified Agda.TypeChecking.Errors as E
+import Agda.TypeChecking.MetaVars (isBlockedTerm)
+import Agda.TypeChecking.Monad
+  hiding (MetaInfo, Primitive, Constructor, Record, Function, Datatype)
+import qualified Agda.TypeChecking.Monad as M
+import Agda.TypeChecking.Pretty
+import qualified Agda.TypeChecking.Reduce as R
+
+import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.Common (Delayed(..))
+import qualified Agda.Syntax.Common as Common
+import qualified Agda.Syntax.Concrete as C
+import qualified Agda.Syntax.Info as SI
+import qualified Agda.Syntax.Internal as I
+import qualified Agda.Syntax.Literal as L
+import qualified Agda.Syntax.Parser as Pa
+import qualified Agda.Syntax.Parser.Tokens as T
+import qualified Agda.Syntax.Position as P
+
+import Agda.Utils.FileName
 import Agda.Utils.Functor
+import Agda.Utils.Lens
+import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Null
+import Agda.Utils.TestHelpers
+import Agda.Utils.HashMap (HashMap)
+import qualified Agda.Utils.HashMap as HMap
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -142,9 +146,9 @@ data Level
 -- 'stTokens'.
 
 generateAndPrintSyntaxInfo :: A.Declaration -> Level -> TCM ()
-generateAndPrintSyntaxInfo decl _ | P.noRange == P.getRange decl = return ()
+generateAndPrintSyntaxInfo decl _ | null $ P.getRange decl = return ()
 generateAndPrintSyntaxInfo decl hlLevel = do
-  file <- envCurrentPath <$> ask
+  file <- fromMaybe __IMPOSSIBLE__ <$> asks envCurrentPath
 
   reportSLn "import.iface.create" 15 $
       "Generating syntax info for " ++ filePath file ++ ' ' :
@@ -223,8 +227,10 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     bound n = nameToFile modMap file [] (A.nameConcrete n)
                          (\isOp -> mempty { aspect = Just $ Name (Just Bound) isOp })
                          (Just $ A.nameBindingSite n)
+
     patsyn n = nameToFileA modMap file n True $ \isOp ->
-                  mempty { aspect = Just $ Name (Just $ Constructor SC.Inductive) isOp }
+                  mempty { aspect = Just $ Name (Just $ Constructor Common.Inductive) isOp }
+
     field m n = nameToFile modMap file m n
                            (\isOp -> mempty { aspect = Just $ Name (Just Field) isOp })
                            Nothing
@@ -232,6 +238,7 @@ generateAndPrintSyntaxInfo decl hlLevel = do
                           n
                           (\isOp -> mempty { aspect = Just $ Name (Just Module) isOp })
                           Nothing
+
     mod isTopLevelModule n =
       nameToFile modMap file []
                  (A.nameConcrete n)
@@ -246,7 +253,7 @@ generateAndPrintSyntaxInfo decl hlLevel = do
 
     -- Ulf, 2014-04-09: It would be nicer to have it on Named_ a, but
     -- you can't have polymorphic functions in universeBi.
-    getNamedArg :: SC.RString -> File
+    getNamedArg :: Common.RString -> File
     getNamedArg x = singleton (rToR $ P.getRange x) mempty{ aspect = Just $ Name (Just Argument) False }
 
     getLet :: A.LetBinding -> File
@@ -260,11 +267,11 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     getLam (A.DomainFull {})  = mempty
 
     getTyped :: A.TypedBinding -> File
-    getTyped (A.TBind _ xs _) = mconcat $ map bound xs
+    getTyped (A.TBind _ xs _) = mconcat $ map (bound . dget) xs
     getTyped A.TLet{}         = mempty
 
     getPatSynArgs :: A.Declaration -> File
-    getPatSynArgs (A.PatternSynDef _ xs _) = mconcat $ map (bound . SC.unArg) xs
+    getPatSynArgs (A.PatternSynDef _ xs _) = mconcat $ map (bound . Common.unArg) xs
     getPatSynArgs _                        = mempty
 
     getPattern :: A.Pattern -> File
@@ -380,7 +387,7 @@ nameKinds hlLevel decl = do
   let merged = HMap.unions [local, imported, syntax]
   return (\n -> HMap.lookup n merged)
   where
-  fix = HMap.map (defnToKind . theDef) . sigDefinitions
+  fix = HMap.map (defnToKind . theDef) . (^. sigDefinitions)
 
   -- | The 'M.Axiom' constructor is used to represent various things
   -- which are not really axioms, so when maps are merged 'Postulate's
@@ -416,20 +423,20 @@ nameKinds hlLevel decl = do
   declToKind (A.Pragma {})          = id
   declToKind (A.ScopedDecl {})      = id
   declToKind (A.Open {})            = id
-  declToKind (A.PatternSynDef q _ _) = insert q (Constructor SC.Inductive)
+  declToKind (A.PatternSynDef q _ _) = insert q (Constructor Common.Inductive)
   declToKind (A.FunDef  _ q _ _)    = insert q Function
   declToKind (A.UnquoteDecl _ _ q _) = insert q Function
   declToKind (A.DataSig _ q _ _)    = insert q Datatype
   declToKind (A.DataDef _ q _ cs)   = \m ->
                                       insert q Datatype $
                                       foldr (\d -> insert (A.axiomName d)
-                                                          (Constructor SC.Inductive))
+                                                          (Constructor Common.Inductive))
                                             m cs
   declToKind (A.RecSig _ q _ _)     = insert q Record
   declToKind (A.RecDef _ q _ c _ _ _) = insert q Record .
                                       case c of
                                         Nothing -> id
-                                        Just q  -> insert q (Constructor SC.Inductive)
+                                        Just q  -> insert q (Constructor Common.Inductive)
 
 -- | Generates syntax highlighting information for all constructors
 -- occurring in patterns and expressions in the given declaration.
@@ -450,18 +457,18 @@ generateConstructorInfo modMap file kinds decl = do
   -- @noRange@ should be impossible, but in case of @noRange@
   -- it makes sense to return the empty File.
   ifNull (P.getRange decl) (return mempty) $ \ (P.Range is) -> do
-  let start = fromIntegral $ P.posPos $ P.iStart $ head is
-      end   = fromIntegral $ P.posPos $ P.iEnd   $ last is
+    let start = fromIntegral $ P.posPos $ P.iStart $ head is
+        end   = fromIntegral $ P.posPos $ P.iEnd   $ last is
 
-  -- Get all disambiguated names that fall within the range of decl.
-  m0 <- use stDisambiguatedNames
-  let (_, m1) = IntMap.split (pred start) m0
-      (m2, _) = IntMap.split end m1
-      constrs = IntMap.elems m2
+    -- Get all disambiguated names that fall within the range of decl.
+    m0 <- use stDisambiguatedNames
+    let (_, m1) = IntMap.split (pred start) m0
+        (m2, _) = IntMap.split end m1
+        constrs = IntMap.elems m2
 
-  -- Return suitable syntax highlighting information.
-  let files = for constrs $ \ q -> generate modMap file kinds $ A.AmbQ [q]
-  return $ Fold.fold files
+    -- Return suitable syntax highlighting information.
+    let files = for constrs $ \ q -> generate modMap file kinds $ A.AmbQ [q]
+    return $ Fold.fold files
 
 
 -- | Prints syntax highlighting info for an error.
