@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP          #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE ViewPatterns #-}
 
 -- | Function for generating highlighted and aligned LaTeX from literate
@@ -413,7 +413,7 @@ fixity = do
 
     -- Operations followed by newlines.
     (ops, nls) | otherwise      -> do
-        output $ escape ops
+        output $ (T.pack " " <+>) $ T.unwords $ map ((cmdPrefix <+> T.pack "FixityOp" <+>) . cmdArg . escape) $ T.words ops
         spaces (T.group nls)
 
 
