@@ -31,7 +31,7 @@ import Agda.Utils.QuickCheck
 --
 -- Invariant: @'from' '<=' 'to'@.
 
-data Range = Range { from, to :: Integer }
+data Range = Range { from, to :: Int }
              deriving (Eq, Ord, Show, Typeable)
 
 -- | The 'Range' invariant.
@@ -72,14 +72,15 @@ empty r = to r <= from r
 
 -- | Converts a range to a list of positions.
 
-rangeToPositions :: Range -> [Integer]
+rangeToPositions :: Range -> [Int]
 rangeToPositions r = [from r .. to r - 1]
 
 -- | Converts several ranges to a list of positions.
 
-rangesToPositions :: Ranges -> [Integer]
+rangesToPositions :: Ranges -> [Int]
 rangesToPositions (Ranges rs) = concatMap rangeToPositions rs
 
+prop_rangesToPositions :: Ranges -> Bool
 prop_rangesToPositions rs = sorted (rangesToPositions rs)
 
 -- | Converts a 'P.Range' to a 'Ranges'.
@@ -90,7 +91,7 @@ rToR (P.Range is) = Ranges (map iToR is)
   iToR (P.Interval { P.iStart = P.Pn { P.posPos = pos1 }
                    , P.iEnd   = P.Pn { P.posPos = pos2 }
                    }) =
-    Range { from = toInteger pos1, to = toInteger pos2 }
+    Range { from = fromIntegral pos1, to = fromIntegral pos2 }
 
 ------------------------------------------------------------------------
 -- Operations
@@ -115,6 +116,7 @@ minus (Ranges rs1) (Ranges rs2) = Ranges (m rs1 rs2)
     | to y < to x     = m (Range { from = to y, to = to x } : xs) ys
     | otherwise       = m xs (y:ys)
 
+prop_minus :: Ranges -> Ranges -> Bool
 prop_minus xs ys =
   rangesToPositions (xs `minus` ys) ==
   rangesToPositions xs \\ rangesToPositions ys
