@@ -1,13 +1,19 @@
-{-# LANGUAGE CPP           #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE TupleSections #-}
 
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
+
 {-| This module deals with finding imported modules and loading their
     interface files.
 -}
 module Agda.Interaction.Imports where
 
-import Prelude
+import Prelude hiding (null)
 
+import Control.Arrow
+import Control.DeepSeq
 import Control.Monad.Reader
 import Control.Monad.State
 import qualified Control.Exception as E
@@ -17,7 +23,7 @@ import qualified Data.Map as Map
 import qualified Data.List as List
 import qualified Data.Set as Set
 import qualified Data.Foldable as Fold (toList)
-import Data.List
+import Data.List hiding (null)
 import Data.Maybe
 import Data.Monoid (mempty, mappend)
 import Data.Map (Map)
@@ -28,7 +34,7 @@ import System.FilePath ((</>))
 
 import qualified Text.PrettyPrint.Boxes as Boxes
 
-import Paths_Agda (getDataFileName)
+import Agda.Benchmarking
 
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Concrete as C
@@ -42,16 +48,15 @@ import Agda.Syntax.Internal
 import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Base.KillRange  -- killRange for Signature
 import Agda.TypeChecking.Serialise
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Primitive
-import Agda.TypeChecking.Monad.Benchmark (billTop, reimburseTop)
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
 import Agda.TheTypeChecker
 
 import Agda.Interaction.FindFile
+import {-# SOURCE #-} Agda.Interaction.InteractionTop (showOpenMetas)
 import Agda.Interaction.Options
 import qualified Agda.Interaction.Options.Lenses as Lens
 import Agda.Interaction.Highlighting.Precise (HighlightingInfo)
@@ -62,7 +67,7 @@ import Agda.Utils.Except ( MonadError(catchError, throwError) )
 import Agda.Utils.FileName
 import Agda.Utils.Lens
 import Agda.Utils.Monad
-import Agda.Utils.Null (unlessNullM)
+import Agda.Utils.Null
 import Agda.Utils.IO.Binary
 import Agda.Utils.Pretty
 import Agda.Utils.Time
@@ -124,16 +129,18 @@ addImportedThings isig ibuiltin hsImports patsyns = do
 
 scopeCheckImport :: ModuleName -> TCM (ModuleName, Map ModuleName Scope)
 scopeCheckImport x = do
-    reportSLn "import.scope" 5 $ "Scope checking " ++ show x
+    reportSLn "import.scope" 5 $ "Scope checking " ++ prettyShow x
     verboseS "import.scope" 10 $ do
       visited <- Map.keys <$> getVisitedModules
       reportSLn "import.scope" 10 $
-        "  visited: " ++ intercalate ", " (map (render . pretty) visited)
+        "  visited: " ++ intercalate ", " (map prettyShow visited)
     -- Since scopeCheckImport is called from the scope checker,
     -- we need to reimburse her account.
-    i <- reimburseTop Bench.Scoping $ getInterface x
+    i <- Bench.billTo [] $ getInterface x
     addImport x
-    return (iModuleName i `withRangesOfQ` mnameToConcrete x, iScope i)
+    -- let s = publicModules $ iInsideScope i
+    let s = iScope i
+    return (iModuleName i `withRangesOfQ` mnameToConcrete x, s)
 
 data MaybeWarnings = NoWarnings | SomeWarnings Warnings
 
@@ -155,12 +162,12 @@ alreadyVisited x getIface = do
         -- A module with warnings should never be allowed to be
         -- imported from another module.
         Just mi | not (miWarnings mi) -> do
-          reportSLn "import.visit" 10 $ "  Already visited " ++ render (pretty x)
+          reportSLn "import.visit" 10 $ "  Already visited " ++ prettyShow x
           return (miInterface mi, NoWarnings)
         _ -> do
-          reportSLn "import.visit" 5 $ "  Getting interface for " ++ render (pretty x)
+          reportSLn "import.visit" 5 $ "  Getting interface for " ++ prettyShow x
           r@(i, wt) <- getIface
-          reportSLn "import.visit" 5 $ "  Now we've looked at " ++ render (pretty x)
+          reportSLn "import.visit" 5 $ "  Now we've looked at " ++ prettyShow x
           visitModule $ ModuleInfo
             { miInterface  = i
             , miWarnings   = hasWarnings wt
@@ -176,14 +183,14 @@ alreadyVisited x getIface = do
 
 typeCheckMain :: AbsolutePath -> TCM (Interface, MaybeWarnings)
 typeCheckMain f = do
-  -- liftIO $ putStrLn $ "This is typeCheckMain " ++ show f
+  -- liftIO $ putStrLn $ "This is typeCheckMain " ++ prettyShow f
   -- liftIO . putStrLn . show =<< getVerbosity
   reportSLn "import.main" 10 $ "Importing the primitive modules."
-  libpath <- liftIO $ getDataFileName "lib"
-  reportSLn "import.main" 20 $ "Library path = " ++ show libpath
+  libdir <- liftIO defaultLibDir
+  reportSLn "import.main" 20 $ "Library dir = " ++ show libdir
   -- To allow posulating the built-ins, check the primitive module
   -- in unsafe mode
-  bracket_ (gets $ Lens.getSafeMode) Lens.putSafeMode $ do
+  _ <- bracket_ (gets $ Lens.getSafeMode) Lens.putSafeMode $ do
     Lens.putSafeMode False
     -- Turn off import-chasing messages.
     -- We have to modify the persistent verbosity setting, since
@@ -194,7 +201,7 @@ typeCheckMain f = do
       withHighlightingLevel None $
         getInterface_ =<< do
           moduleName $ mkAbsolute $
-            libpath </> "prim" </> "Agda" </> "Primitive.agda"
+            libdir </> "prim" </> "Agda" </> "Primitive.agda"
   reportSLn "import.main" 10 $ "Done importing the primitive modules."
 
   -- Now do the type checking via getInterface.
@@ -233,64 +240,67 @@ getInterface'
   -> TCM (Interface, MaybeWarnings)
 getInterface' x isMain = do
   withIncreasedModuleNestingLevel $ do
-  -- Preserve the pragma options unless includeStateChanges is True.
-  bracket_ (use stPragmaOptions)
-           (unless includeStateChanges . setPragmaOptions) $ do
-   -- Forget the pragma options (locally).
-   setCommandLineOptions . stPersistentOptions . stPersistentState =<< get
+    -- Preserve the pragma options unless includeStateChanges is True.
+    bracket_ (use stPragmaOptions)
+             (unless includeStateChanges . setPragmaOptions) $ do
+     -- Forget the pragma options (locally).
+     setCommandLineOptions . stPersistentOptions . stPersistentState =<< get
 
-   alreadyVisited x $ addImportCycleCheck x $ do
-    file <- findFile x  -- requires source to exist
+     alreadyVisited x $ addImportCycleCheck x $ do
+      file <- findFile x  -- requires source to exist
 
-    reportSLn "import.iface" 10 $ "  Check for cycle"
-    checkForImportCycle
+      reportSLn "import.iface" 10 $ "  Check for cycle"
+      checkForImportCycle
 
-    uptodate <- billTop Bench.Import $ do
-      ignore <- ignoreInterfaces
-      cached <- isCached file -- if it's cached ignoreInterfaces has no effect
-                              -- to avoid typechecking a file more than once
-      sourceH <- liftIO $ hashFile file
-      ifaceH  <-
-        case cached of
-          Nothing -> fmap fst <$> getInterfaceFileHashes (filePath $ toIFile file)
-          Just i  -> return $ Just $ iSourceHash i
-      let unchanged = Just sourceH == ifaceH
-      return $ unchanged && (not ignore || isJust cached)
+      uptodate <- Bench.billTo [Bench.Import] $ do
+        ignore <- ignoreInterfaces
+        cached <- isCached file -- if it's cached ignoreInterfaces has no effect
+                                -- to avoid typechecking a file more than once
+        sourceH <- liftIO $ hashFile file
+        ifaceH  <-
+          case cached of
+            Nothing -> fmap fst <$> getInterfaceFileHashes (filePath $ toIFile file)
+            Just i  -> return $ Just $ iSourceHash i
+        let unchanged = Just sourceH == ifaceH
+        return $ unchanged && (not ignore || isJust cached)
 
-    reportSLn "import.iface" 5 $
-      "  " ++ render (pretty x) ++ " is " ++
-      (if uptodate then "" else "not ") ++ "up-to-date."
+      reportSLn "import.iface" 5 $
+        "  " ++ prettyShow x ++ " is " ++
+        (if uptodate then "" else "not ") ++ "up-to-date."
 
-    -- Andreas, 2014-10-20 AIM XX:
-    -- Always retype-check the main file to get the iInsideScope
-    -- which is no longer serialized.
-    (stateChangesIncluded, (i, wt)) <-
-      if uptodate && isMain == NotMainInterface then skip file else typeCheckThe file
+      (stateChangesIncluded, (i, wt)) <- do
+        -- -- Andreas, 2014-10-20 AIM XX:
+        -- -- Always retype-check the main file to get the iInsideScope
+        -- -- which is no longer serialized.
+        -- let maySkip = isMain == NotMainInterface
+        -- Andreas, 2015-07-13: Serialize iInsideScope again.
+        let maySkip = True
+        if uptodate && maySkip then skip file else typeCheckThe file
 
-    -- Ensure that the given module name matches the one in the file.
-    let topLevelName = toTopLevelModuleName $ iModuleName i
-    unless (topLevelName == x) $ do
-      -- Andreas, 2014-03-27 This check is now done in the scope checker.
-      -- checkModuleName topLevelName file
-      typeError $ OverlappingProjects file topLevelName x
+      -- Ensure that the given module name matches the one in the file.
+      let topLevelName = toTopLevelModuleName $ iModuleName i
+      unless (topLevelName == x) $ do
+        -- Andreas, 2014-03-27 This check is now done in the scope checker.
+        -- checkModuleName topLevelName file
+        typeError $ OverlappingProjects file topLevelName x
 
-    visited <- isVisited x
-    reportSLn "import.iface" 5 $ if visited then "  We've been here. Don't merge."
-                                 else "  New module. Let's check it out."
-    unless (visited || stateChangesIncluded) $ do
-      mergeInterface i
-      billTop Bench.Highlighting $
-        ifTopLevelAndHighlightingLevelIs NonInteractive $
-          highlightFromInterface i file
+      visited <- isVisited x
+      reportSLn "import.iface" 5 $ if visited then "  We've been here. Don't merge."
+                                   else "  New module. Let's check it out."
+      unless (visited || stateChangesIncluded) $ do
+        mergeInterface i
+        Bench.billTo [Bench.Highlighting] $
+          ifTopLevelAndHighlightingLevelIs NonInteractive $
+            highlightFromInterface i file
 
-    stCurrentModule .= Just (iModuleName i)
+      stCurrentModule .= Just (iModuleName i)
 
-    -- Interfaces are only stored if no warnings were encountered.
-    case wt of
-      SomeWarnings w -> return ()
-      NoWarnings     -> storeDecodedModule i
+      -- Interfaces are only stored if no warnings were encountered.
+      case wt of
+        SomeWarnings w -> return ()
+        NoWarnings     -> storeDecodedModule i
 
-    return (i, wt)
+      return (i, wt)
 
     where
       includeStateChanges = isMain == MainInterface
@@ -311,7 +321,7 @@ getInterface' x isMain = do
       chaseMsg kind file = do
         nesting <- envModuleNestingLevel <$> ask
         let s = genericReplicate nesting ' ' ++ kind ++
-                " " ++ render (pretty x) ++
+                " " ++ prettyShow x ++
                 case file of
                   Nothing -> "."
                   Just f  -> " (" ++ f ++ ")."
@@ -324,7 +334,7 @@ getInterface' x isMain = do
         let ifile = filePath $ toIFile file
         h <- fmap snd <$> getInterfaceFileHashes ifile
         mm <- getDecodedModule x
-        (cached, mi) <- billTop Bench.Deserialization $ case mm of
+        (cached, mi) <- Bench.billTo [Bench.Deserialization] $ case mm of
           Just mi ->
             if Just (iFullHash mi) /= h
             then do dropDecodedModule x
@@ -460,7 +470,7 @@ readInterface :: FilePath -> TCM (Maybe Interface)
 readInterface file = do
     -- Decode the interface file
     (s, close) <- liftIO $ readBinaryFile' file
-    do  i <- liftIO . E.evaluate =<< decodeInterface s
+    do  mi <- liftIO . E.evaluate =<< decodeInterface s
 
         -- Close the file. Note
         -- ⑴ that evaluate ensures that i is evaluated to WHNF (before
@@ -469,7 +479,7 @@ readInterface file = do
         -- so it is safe to close the file here.
         liftIO close
 
-        return i
+        return $ constructIScope <$> mi
       -- Catch exceptions and close
       `catchError` \e -> liftIO close >> handler e
   -- Catch exceptions
@@ -489,10 +499,15 @@ readInterface file = do
 writeInterface :: FilePath -> Interface -> TCM ()
 writeInterface file i = do
     reportSLn "import.iface.write" 5  $ "Writing interface file " ++ file ++ "."
-    -- Andreas, Makoto, 2014-10-18 AIM XX:
-    -- iInsideScope is bloating the interface files, so we do not serialize it?
+    -- Andreas, 2015-07-13
+    -- After QName memoization (AIM XXI), scope serialization might be cheap enough.
+    -- -- Andreas, Makoto, 2014-10-18 AIM XX:
+    -- -- iInsideScope is bloating the interface files, so we do not serialize it?
+    -- i <- return $
+    --   i { iInsideScope  = emptyScopeInfo
+    --     }
     i <- return $
-      i { iInsideScope  = emptyScopeInfo
+      i { iInsideScope  = removePrivates $ iInsideScope i
         }
     encodeFile file i
     reportSLn "import.iface.write" 5 $ "Wrote interface file."
@@ -504,6 +519,9 @@ writeInterface file i = do
       whenM (doesFileExist file) $ removeFile file
     throwError e
 
+removePrivates :: ScopeInfo -> ScopeInfo
+removePrivates si = si { scopeModules = restrictPrivate <$> scopeModules si }
+
 -- | Tries to type check a module and write out its interface. The
 -- function only writes out an interface file if it does not encounter
 -- any warnings.
@@ -516,22 +534,23 @@ createInterface
   -> C.TopLevelModuleName  -- ^ The expected module name.
   -> TCM (Interface, MaybeWarnings)
 createInterface file mname =
-  local (\e -> e { envCurrentPath = file }) $ do
+  local (\e -> e { envCurrentPath = Just file }) $ do
     modFile       <- use stModuleToSource
-    fileTokenInfo <- billTop Bench.Highlighting $ generateTokenInfo file
+    fileTokenInfo <- Bench.billTo [Bench.Highlighting] $
+                       generateTokenInfo file
     stTokens .= fileTokenInfo
 
     reportSLn "import.iface.create" 5 $
-      "Creating interface for " ++ render (pretty mname) ++ "."
+      "Creating interface for " ++ prettyShow mname ++ "."
     verboseS "import.iface.create" 10 $ do
       visited <- Map.keys <$> getVisitedModules
       reportSLn "import.iface.create" 10 $
-        "  visited: " ++ intercalate ", " (map (render . pretty) visited)
+        "  visited: " ++ intercalate ", " (map prettyShow visited)
 
     previousHsImports <- getHaskellImports
 
     -- Parsing.
-    (pragmas, top) <- billTop Bench.Parsing $
+    (pragmas, top) <- Bench.billTo [Bench.Parsing] $
       liftIO $ parseFile' moduleParser file
 
     pragmas <- concat <$> concreteToAbstract_ pragmas
@@ -542,20 +561,26 @@ createInterface file mname =
     mapM_ setOptionsFromPragma options
 
     -- Scope checking.
-    topLevel <- billTop Bench.Scoping $
+    reportSLn "import.iface.create" 7 $ "Starting scope checking."
+    topLevel <- Bench.billTo [Bench.Scoping] $
       concreteToAbstract_ (TopLevel file top)
+    reportSLn "import.iface.create" 7 $ "Finished scope checking."
 
     let ds = topLevelDecls topLevel
 
     -- Highlighting from scope checker.
-    billTop Bench.Highlighting $ do
+    reportSLn "import.iface.create" 7 $ "Starting highlighting from scope."
+    Bench.billTo [Bench.Highlighting] $ do
       ifTopLevelAndHighlightingLevelIs NonInteractive $ do
-      -- Generate and print approximate syntax highlighting info.
-      printHighlightingInfo fileTokenInfo
-      mapM_ (\ d -> generateAndPrintSyntaxInfo d Partial) ds
+        -- Generate and print approximate syntax highlighting info.
+        printHighlightingInfo fileTokenInfo
+        mapM_ (\ d -> generateAndPrintSyntaxInfo d Partial) ds
+    reportSLn "import.iface.create" 7 $ "Finished highlighting from scope."
 
     -- Type checking.
-    billTop Bench.Typing $ checkDecls ds
+    reportSLn "import.iface.create" 7 $ "Starting type checking."
+    Bench.billTo [Bench.Typing] $ checkDecls ds
+    reportSLn "import.iface.create" 7 $ "Finished type checking."
 
     -- Ulf, 2013-11-09: Since we're rethrowing the error, leave it up to the
     -- code that handles that error to reset the state.
@@ -573,7 +598,8 @@ createInterface file mname =
       tickN "metas" (fromIntegral n)
 
     -- Highlighting from type checker.
-    billTop Bench.Highlighting $ do
+    reportSLn "import.iface.create" 7 $ "Starting highlighting from type info."
+    Bench.billTo [Bench.Highlighting] $ do
 
       -- Move any remaining token highlighting to stSyntaxInfo.
       toks <- use stTokens
@@ -584,28 +610,34 @@ createInterface file mname =
       whenM (optGenerateVimFile <$> commandLineOptions) $
         -- Generate Vim file.
         withScope_ (insideScope topLevel) $ generateVimFile $ filePath file
+    reportSLn "import.iface.create" 7 $ "Finished highlighting from type info."
 
     setScope $ outsideScope topLevel
-
     reportSLn "scope.top" 50 $ "SCOPE " ++ show (insideScope topLevel)
 
     -- Serialization.
+    reportSLn "import.iface.create" 7 $ "Starting serialization."
     syntaxInfo <- use stSyntaxInfo
-    i <- billTop Bench.Serialization $ do
+    i <- Bench.billTo [Bench.Serialization] $ do
       buildInterface file topLevel syntaxInfo previousHsImports options
+    reportSLn "import.iface.create" 7 $ "Finished serialization."
 
     -- TODO: It would be nice if unsolved things were highlighted
     -- after every mutual block.
 
-    unsolvedMetas       <- List.nub <$> (mapM getMetaRange =<< getOpenMetas)
+    openMetas           <- getOpenMetas
+    unless (null openMetas) $ do
+      reportSLn "import.metas" 10 . unlines =<< showOpenMetas
+    unsolvedMetas       <- List.nub <$> mapM getMetaRange openMetas
     unsolvedConstraints <- getAllConstraints
     interactionPoints   <- getInteractionPoints
 
     ifTopLevelAndHighlightingLevelIs NonInteractive $
       printUnsolvedInfo
 
+    reportSLn "import.iface.create" 7 $ "Starting writing to interface file."
     r <- if and [ null unsolvedMetas, null unsolvedConstraints, null interactionPoints ]
-     then billTop Bench.Serialization $ do
+     then Bench.billTo [Bench.Serialization] $ do
       -- The file was successfully type-checked (and no warnings were
       -- encountered), so the interface should be written out.
       let ifile = filePath $ toIFile file
@@ -613,6 +645,7 @@ createInterface file mname =
       return (i, NoWarnings)
      else do
       return (i, SomeWarnings $ Warnings unsolvedMetas unsolvedConstraints)
+    reportSLn "import.iface.create" 7 $ "Finished writing to interface file."
 
     -- Profiling: Print statistics.
     printStatistics 30 (Just mname) =<< getStatistics
@@ -624,7 +657,11 @@ createInterface file mname =
     verboseS "profile" 1 $ do
       reportSLn "import.iface" 5 $ "Accumulated statistics."
 
-    return r
+    return $ first constructIScope r
+
+-- constructIScope :: ScopeInfo -> Map ModuleName Scope
+constructIScope :: Interface -> Interface
+constructIScope i = i{ iScope = billToPure [ Deserialization ] $ publicModules $ iInsideScope i }
 
 -- | Builds an interface for the current module, which should already
 -- have been successfully type checked.
@@ -655,7 +692,10 @@ buildInterface file topLevel syntaxInfo previousHsImports pragmas = do
     ms      <- getImports
     mhs     <- mapM (\ m -> (m,) <$> moduleHash m) $ Set.toList ms
     hsImps  <- getHaskellImports
-    patsyns <- getPatternSyns
+    -- Andreas, 2015-02-09 kill ranges in pattern synonyms before
+    -- serialization to avoid error locations pointing to external files
+    -- when expanding a pattern synoym.
+    patsyns <- killRange <$> getPatternSyns
     h       <- liftIO $ hashFile file
     let builtin' = Map.mapWithKey (\ x b -> (x,) . primFunName <$> b) builtin
     reportSLn "import.iface" 7 "  instantiating all meta variables"
@@ -663,7 +703,7 @@ buildInterface file topLevel syntaxInfo previousHsImports pragmas = do
       { iSourceHash      = h
       , iImportedModules = mhs
       , iModuleName      = m
-      , iScope           = publicModules scope
+      , iScope           = empty -- publicModules scope
       , iInsideScope     = insideScope topLevel
       , iSignature       = sig
       , iBuiltin         = builtin'
@@ -685,12 +725,6 @@ getInterfaceFileHashes ifile = do
     liftIO $ maybe 0 (uncurry (+)) hs `seq` close
     return hs
 
-safeReadInterface :: FilePath -> TCM (Maybe Interface)
-safeReadInterface ifile = do
-  exist <- liftIO $ doesFileExist ifile
-  if exist then readInterface ifile
-           else return Nothing
-
 moduleHash :: ModuleName -> TCM Hash
 moduleHash m = iFullHash <$> getInterface m
 
