@@ -1,21 +1,24 @@
-{-# LANGUAGE CPP                   #-}
-{-# LANGUAGE DeriveFoldable        #-}
-{-# LANGUAGE DeriveFunctor         #-}
-{-# LANGUAGE DeriveTraversable     #-}
-{-# LANGUAGE FlexibleInstances     #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE Rank2Types            #-}
-{-# LANGUAGE ScopedTypeVariables   #-}
-{-# LANGUAGE TupleSections         #-}
-{-# LANGUAGE TypeSynonymInstances  #-}
-
 {-# OPTIONS -fno-cse #-}
 
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE Rank2Types #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TupleSections #-}
+
+{-# OPTIONS_GHC -fno-warn-orphans #-}
+
 module Agda.Interaction.InteractionTop
   ( module Agda.Interaction.InteractionTop
   )
   where
 
+import Prelude hiding (null)
+
 import Control.Applicative hiding (empty)
 import qualified Control.Exception as E
 import Control.Monad.Identity
@@ -25,7 +28,7 @@ import Control.Monad.State
 import qualified Data.Char as Char
 import Data.Foldable (Foldable)
 import Data.Function
-import Data.List as List
+import Data.List as List hiding (null)
 import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Map as Map
@@ -58,6 +61,7 @@ import Agda.Syntax.Scope.Base
 
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
+import Agda.Interaction.Options.Lenses as Lenses
 import Agda.Interaction.MakeCase
 import Agda.Interaction.Response hiding (Function, ExtendedLambda)
 import qualified Agda.Interaction.Response as R
@@ -84,7 +88,9 @@ import Agda.Utils.FileName
 import Agda.Utils.Hash
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Lens
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Pretty
 import Agda.Utils.String
 import Agda.Utils.Time
@@ -217,7 +223,7 @@ runInteraction (IOTCM current highlighting highlightingMethod cmd)
             when (not (independent cmd) && Just current /= (fst <$> cf)) $
                 lift $ typeError $ GenericError "Error: First load the file."
 
-            interpret cmd
+            withCurrentFile $ interpret cmd
 
             cf <- gets theCurrentFile
             when (Just current == (fst <$> cf)) $
@@ -234,7 +240,10 @@ runInteraction (IOTCM current highlighting highlightingMethod cmd)
       -- If an independent command fails we should reset theCurrentFile (Issue853).
       let sErr | independent cmd = s { theCurrentFile = Nothing }
                | otherwise       = s
-      (x, s') <- lift (runStateT m s `catchError_` \e -> runStateT (h e) sErr)
+      (x, s') <- lift $ do
+           disableDestructiveUpdate (runStateT m s)
+         `catchError_` \ e ->
+           runStateT (h e) sErr
       put s'
       return x
 
@@ -246,8 +255,9 @@ runInteraction (IOTCM current highlighting highlightingMethod cmd)
     -- | Handle nasty errors like stack space overflow (issue 637)
     -- We assume that the input action handles other kind of errors.
     handleNastyErrors :: CommandM () -> CommandM ()
-    handleNastyErrors m = commandMToIO $ \toIO ->
-        toIO m `E.catch` (toIO . handleErr . Exception noRange . (show :: E.SomeException -> String))
+    handleNastyErrors m = commandMToIO $ \ toIO ->
+        toIO m `E.catch` \ (e :: E.SomeException) ->
+          toIO $ handleErr $ Exception noRange $ text $ show e
 
     -- | Displays an error and instructs Emacs to jump to the site of the
     -- error. Because this function may switch the focus to another file
@@ -517,31 +527,11 @@ interpret Cmd_constraints =
     display_info . Info_Constraints . unlines . map show =<< lift B.getConstraints
 
 interpret Cmd_metas = do -- CL.showMetas []
-  ims <- lift $ B.typesOfVisibleMetas B.AsIs
-  -- Show unsolved implicit arguments simplified.
-  hms <- lift $ B.typesOfHiddenMetas B.Simplified
-  if not $ null ims && null hms
-    then do
-      di <- lift $ forM ims $ \i -> B.withInteractionId (B.outputFormId $ B.OutputForm noRange 0 i) (showATop i)
-      dh <- lift $ mapM showA' hms
-      display_info $ Info_AllGoals $ unlines $ di ++ dh
-    else do
-      cs <- lift B.getConstraints
-      if null cs
-        then display_info $ Info_AllGoals ""
-        else interpret Cmd_constraints
-  where
-    metaId (B.OfType i _) = i
-    metaId (B.JustType i) = i
-    metaId (B.JustSort i) = i
-    metaId (B.Assign i e) = i
-    metaId _ = __IMPOSSIBLE__
-    showA' :: B.OutputConstraint A.Expr NamedMeta -> TCM String
-    showA' m = do
-      let i = nmid $ metaId m
-      r <- getMetaRange i
-      d <- B.withMetaId i (showATop m)
-      return $ d ++ "  [ at " ++ show r ++ " ]"
+  ms <- lift $ showOpenMetas
+  -- If we do not have open metas, but open constaints, display those.
+  ifM (return (null ms) `and2M` do not . null <$> lift B.getConstraints)
+    {-then-} (interpret Cmd_constraints)
+    {-else-} (display_info $ Info_AllGoals $ unlines ms)
 
 interpret (Cmd_show_module_contents_toplevel norm s) =
   liftCommandMT B.atTopLevel $ showModuleContents norm noRange s
@@ -606,7 +596,7 @@ interpret (Cmd_load_highlighting_info source) = do
                 return Nothing
     mapM_ putResponse resp
 
-interpret (Cmd_highlight ii rng s) = withCurrentFile $ do
+interpret (Cmd_highlight ii rng s) = do
   scope <- getOldInteractionScope ii
   removeOldInteractionScope ii
   handle $ do
@@ -647,7 +637,8 @@ interpret (Cmd_intro pmLambda ii rng _) = do
             ]
 
 interpret (Cmd_refine_or_intro pmLambda ii r s) = interpret $
-  (if null s then Cmd_intro pmLambda else Cmd_refine) ii r s
+  let s' = trim s
+  in (if null s' then Cmd_intro pmLambda else Cmd_refine) ii r s'
 
 interpret (Cmd_auto ii rng s) = do
   -- Andreas, 2014-07-05 Issue 1226:
@@ -728,15 +719,15 @@ interpret (Cmd_make_case ii rng s) = do
     render = renderStyle (style { mode = OneLineMode })
 
     makeCaseVariant :: CaseContext -> MakeCaseVariant
-    makeCaseVariant FunctionDef          = R.Function
-    makeCaseVariant (ExtendedLambda _ _) = R.ExtendedLambda
+    makeCaseVariant Nothing = R.Function
+    makeCaseVariant Just{}  = R.ExtendedLambda
 
     -- very dirty hack, string manipulation by dropping the function name
     -- and replacing the last " = " with " -> ". It's important not to replace
     -- the equal sign in named implicit with an arrow!
     extlam_dropName :: CaseContext -> String -> String
-    extlam_dropName FunctionDef x = x
-    extlam_dropName (ExtendedLambda _ _) x
+    extlam_dropName Nothing x = x
+    extlam_dropName Just{}  x
         = unwords $ reverse $ replEquals $ reverse $ drop 1 $ words x
       where
         replEquals ("=" : ws) = "→" : ws
@@ -745,9 +736,9 @@ interpret (Cmd_make_case ii rng s) = do
 
     -- Drops pattern added to extended lambda functions when lambda lifting them
     extlam_dropLLifted :: CaseContext -> Bool -> A.Clause -> A.Clause
-    extlam_dropLLifted FunctionDef _ x = x
-    extlam_dropLLifted (ExtendedLambda h nh) hidden (A.Clause (A.LHS info (A.LHSProj{}) ps) rhs decl) = __IMPOSSIBLE__
-    extlam_dropLLifted (ExtendedLambda h nh) hidden (A.Clause (A.LHS info (A.LHSHead name nps) ps) rhs decl)
+    extlam_dropLLifted Nothing _ x = x
+    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden (A.Clause (A.LHS info A.LHSProj{} ps) rhs decl) = __IMPOSSIBLE__
+    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden (A.Clause (A.LHS info (A.LHSHead name nps) ps) rhs decl)
       = let n = if hidden then h + nh else nh
         in
          (A.Clause (A.LHS info (A.LHSHead name (drop n nps)) ps) rhs decl)
@@ -762,10 +753,33 @@ interpret (Cmd_compute ignore ii rng s) = do
 
 interpret Cmd_show_version = display_info Info_Version
 
-type GoalCommand = InteractionId -> Range -> String -> Interaction
+-- | Print open metas nicely.
+showOpenMetas :: TCM [String]
+showOpenMetas = do
+  ims <- B.typesOfVisibleMetas B.AsIs
+  di <- forM ims $ \ i ->
+    B.withInteractionId (B.outputFormId $ B.OutputForm noRange 0 i) $
+      showATop i
+  -- Show unsolved implicit arguments simplified.
+  dh <- mapM showA' =<< B.typesOfHiddenMetas B.Simplified
+  return $ di ++ dh
+  where
+    metaId (B.OfType i _) = i
+    metaId (B.JustType i) = i
+    metaId (B.JustSort i) = i
+    metaId (B.Assign i e) = i
+    metaId _ = __IMPOSSIBLE__
+    showA' :: B.OutputConstraint A.Expr NamedMeta -> TCM String
+    showA' m = do
+      let i = nmid $ metaId m
+      r <- getMetaRange i
+      d <- B.withMetaId i (showATop m)
+      return $ d ++ "  [ at " ++ show r ++ " ]"
 
--- | @cmd_load' m includes cmd cmd2@ loads the module in file @m@,
--- using @includes@ as the include directories.
+
+-- | @cmd_load' file includes unsolvedOk cmd@
+--   loads the module in file @file@,
+--   using @includes@ as the include directories.
 --
 -- If type checking completes without any exceptions having been
 -- encountered then the command @cmd r@ is executed, where @r@ is the
@@ -778,7 +792,7 @@ cmd_load' :: FilePath -> [FilePath]
 cmd_load' file includes unsolvedOK cmd = do
     f <- liftIO $ absolute file
     ex <- liftIO $ doesFileExist $ filePath f
-    lift $ setIncludeDirs includes $
+    lift $ TM.setIncludeDirs includes $
       if ex then ProjectRoot f else CurrentDir
 
     -- Forget the previous "current file" and interaction points.
@@ -795,12 +809,9 @@ cmd_load' file includes unsolvedOK cmd = do
     opts <- lift $ commandLineOptions
     defaultOptions <- gets optionsOnReload
     setCommandLineOptions' $
-      defaultOptions { optIncludeDirs   = optIncludeDirs opts
-                     , optPragmaOptions =
-                         (optPragmaOptions defaultOptions)
-                           { optAllowUnsolved = unsolvedOK
-                           }
-                     }
+      Lenses.setIncludeDirs (optIncludeDirs opts) $
+      mapPragmaOptions (\ o -> o { optAllowUnsolved = unsolvedOK }) $
+      defaultOptions
 
     -- Reset the state, preserving options and decoded modules. Note
     -- that if the include directories have changed, then the decoded
@@ -828,10 +839,11 @@ cmd_load' file includes unsolvedOK cmd = do
 
     cmd ok
 
+-- | Set 'envCurrentPath' to 'theCurrentFile', if any.
 withCurrentFile :: CommandM a -> CommandM a
 withCurrentFile m = do
-  Just (file, _) <- gets $ theCurrentFile
-  local (\e -> e { envCurrentPath = file }) m
+  mfile <- fmap fst <$> gets theCurrentFile
+  local (\ e -> e { envCurrentPath = mfile }) m
 
 -- | Available backends.
 
@@ -857,52 +869,56 @@ give_gen
   -> String
   -> GiveRefine
   -> CommandM ()
-give_gen ii rng s giveRefine = withCurrentFile $ do
+give_gen ii rng s0 giveRefine = do
+  let s = trim s0
   lift $ reportSLn "interaction.give" 20 $ "give_gen  " ++ s
-  let give_ref =
-        case giveRefine of
-          Give   -> B.give
-          Refine -> B.refine
-  -- save scope of the interaction point (for printing the given expr. later)
-  scope     <- lift $ getInteractionScope ii
-  -- parse string and "give", obtaining an abstract expression
-  -- and newly created interaction points
-  (ae, iis) <- lift $ do
-    mis  <- getInteractionPoints
-    reportSLn "interaction.give" 30 $ "interaction points before = " ++ show mis
-    ae   <- give_ref ii Nothing =<< B.parseExprIn ii rng s
-    mis' <- getInteractionPoints
-    reportSLn "interaction.give" 30 $ "interaction points after = " ++ show mis'
-    return (ae, mis' \\ mis)
-  -- favonia: backup the old scope for highlighting
-  insertOldInteractionScope ii scope
-  -- sort the new interaction points and put them into the state
-  -- in replacement of the old interaction point
-  iis       <- lift $ sortInteractionPoints iis
-  modifyTheInteractionPoints $ replace ii iis
-  -- print abstract expr
-  ce        <- lift $ abstractToConcreteEnv (makeEnv scope) ae
-  lift $ reportSLn "interaction.give" 30 $ "ce = " ++ show ce
-  -- if the command was @Give@, use the literal user input;
-  -- Andreas, 2014-01-15, see issue 1020:
-  -- Refine could solve a goal by introducing the sole constructor
-  -- without arguments.  Then there are no interaction metas, but
-  -- we still cannot just `give' the user string (which may be empty).
-  -- WRONG: also, if no interaction metas were created by @Refine@
-  -- WRONG: let literally = (giveRefine == Give || null iis) && rng /= noRange
-  let literally = giveRefine == Give && rng /= noRange
-  -- Ulf, 2014-01-24: This works for give since we're highlighting the string
-  -- that's already in the buffer. Doing it before the give action means that
-  -- the highlighting is moved together with the text when the hole goes away.
-  -- To make it work for refine we'd have to adjust the ranges.
-  when literally $ lift $ do
-    printHighlightingInfo =<< generateTokenInfoFromString rng s
-    highlightExpr ae
-  putResponse $ Resp_GiveAction ii $ mkNewTxt literally ce
-  lift $ reportSLn "interaction.give" 30 $ "putResponse GiveAction passed"
-  -- display new goal set
-  interpret Cmd_metas
-  lift $ reportSLn "interaction.give" 30 $ "interpret Cmd_metas passed"
+  -- Andreas, 2015-02-26 if string is empty do nothing rather
+  -- than giving a parse error.
+  unless (null s) $ do
+    let give_ref =
+          case giveRefine of
+            Give   -> B.give
+            Refine -> B.refine
+    -- save scope of the interaction point (for printing the given expr. later)
+    scope     <- lift $ getInteractionScope ii
+    -- parse string and "give", obtaining an abstract expression
+    -- and newly created interaction points
+    (ae, iis) <- lift $ do
+      mis  <- getInteractionPoints
+      reportSLn "interaction.give" 30 $ "interaction points before = " ++ show mis
+      ae   <- give_ref ii Nothing =<< B.parseExprIn ii rng s
+      mis' <- getInteractionPoints
+      reportSLn "interaction.give" 30 $ "interaction points after = " ++ show mis'
+      return (ae, mis' \\ mis)
+    -- favonia: backup the old scope for highlighting
+    insertOldInteractionScope ii scope
+    -- sort the new interaction points and put them into the state
+    -- in replacement of the old interaction point
+    iis       <- lift $ sortInteractionPoints iis
+    modifyTheInteractionPoints $ replace ii iis
+    -- print abstract expr
+    ce        <- lift $ abstractToConcreteEnv (makeEnv scope) ae
+    lift $ reportSLn "interaction.give" 30 $ "ce = " ++ show ce
+    -- if the command was @Give@, use the literal user input;
+    -- Andreas, 2014-01-15, see issue 1020:
+    -- Refine could solve a goal by introducing the sole constructor
+    -- without arguments.  Then there are no interaction metas, but
+    -- we still cannot just `give' the user string (which may be empty).
+    -- WRONG: also, if no interaction metas were created by @Refine@
+    -- WRONG: let literally = (giveRefine == Give || null iis) && rng /= noRange
+    let literally = giveRefine == Give && rng /= noRange
+    -- Ulf, 2014-01-24: This works for give since we're highlighting the string
+    -- that's already in the buffer. Doing it before the give action means that
+    -- the highlighting is moved together with the text when the hole goes away.
+    -- To make it work for refine we'd have to adjust the ranges.
+    when literally $ lift $ do
+      printHighlightingInfo =<< generateTokenInfoFromString rng s
+      highlightExpr ae
+    putResponse $ Resp_GiveAction ii $ mkNewTxt literally ce
+    lift $ reportSLn "interaction.give" 30 $ "putResponse GiveAction passed"
+    -- display new goal set
+    interpret Cmd_metas
+    lift $ reportSLn "interaction.give" 30 $ "interpret Cmd_metas passed"
   where
     -- Substitutes xs for x in ys.
     replace x xs ys = concatMap (\ y -> if y == x then xs else [y]) ys
@@ -981,7 +997,7 @@ showModuleContents :: B.Rewrite -> Range -> String -> CommandM ()
 showModuleContents norm rng s = do
   (modules, types) <- lift $ B.moduleContents norm rng s
   types' <- lift $ forM types $ \ (x, t) -> do
-     t <- prettyTCM t
+     t <- TCP.prettyTCM t
      return (show x, text ":" <+> t)
   display_info $ Info_ModuleContents $
     text "Modules" $$
@@ -1021,7 +1037,7 @@ whyInScope s = do
           where
             asVar :: TCM Doc
             asVar = do
-              TCP.text "* a variable bound at" TCP.<+> prettyTCM (nameBindingSite $ localVar x)
+              TCP.text "* a variable bound at" TCP.<+> TCP.prettyTCM (nameBindingSite $ localVar x)
             shadowing :: LocalVar -> TCM Doc
             shadowing LocalVar{}    = TCP.text "shadowing"
             shadowing ShadowedVar{} = TCP.text "in conflict with"
@@ -1050,20 +1066,20 @@ whyInScope s = do
           TCP.nest 2 (pWhy (nameBindingSite $ qnameName $ mnameToQName $ amodName a) (amodLineage a))
 
         pWhy :: Range -> WhyInScope -> TCM Doc
-        pWhy r Defined = TCP.text "- its definition at" TCP.<+> prettyTCM r
+        pWhy r Defined = TCP.text "- its definition at" TCP.<+> TCP.prettyTCM r
         pWhy r (Opened (C.QName x) w) | isNoName x = pWhy r w
         pWhy r (Opened m w) =
           TCP.text "- the opening of"
           TCP.<+> TCP.text (show m)
           TCP.<+> TCP.text "at"
-          TCP.<+> prettyTCM (getRange m)
+          TCP.<+> TCP.prettyTCM (getRange m)
           TCP.$$
           pWhy r w
         pWhy r (Applied m w) =
           TCP.text "- the application of"
           TCP.<+> TCP.text (show m)
           TCP.<+> TCP.text "at"
-          TCP.<+> prettyTCM (getRange m)
+          TCP.<+> TCP.prettyTCM (getRange m)
           TCP.$$
           pWhy r w
 
