@@ -1,11 +1,11 @@
-{-# LANGUAGE CPP             #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DoAndIfThenElse #-}
-{-# LANGUAGE TupleSections   #-}
+{-# LANGUAGE TupleSections #-}
 
 module Agda.Interaction.MakeCase where
 
-import Prelude hiding (mapM, mapM_)
-import Control.Applicative
+import Prelude hiding (mapM, mapM_, null)
+import Control.Applicative hiding (empty)
 import Control.Monad hiding (mapM, mapM_, forM)
 import Data.Maybe
 import Data.Traversable
@@ -23,27 +23,30 @@ import Agda.Syntax.Translation.InternalToAbstract
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Coverage
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Irrelevance
+import Agda.TypeChecking.Rules.LHS.Implicit
 import Agda.TheTypeChecker
 
+import Agda.Interaction.Options
 import Agda.Interaction.BasicOps
 
 import Agda.Utils.Functor
+import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import qualified Agda.Utils.Pretty as P
+import Agda.Utils.Singleton
 import Agda.Utils.Size
 import qualified Agda.Utils.HashMap as HMap
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-data CaseContext
-  = FunctionDef
-  | ExtendedLambda Int Int
-  deriving (Eq)
+type CaseContext = Maybe ExtLamInfo
 
 -- | Find the clause whose right hand side is the given meta
 -- BY SEARCHING THE WHOLE SIGNATURE. Returns
@@ -59,11 +62,11 @@ findClause :: MetaId -> TCM (CaseContext, QName, Clause)
 findClause m = do
   sig <- getImportedSignature
   let res = do
-        def <- HMap.elems $ sigDefinitions sig
+        def <- HMap.elems $ sig ^. sigDefinitions
         Function{funClauses = cs, funExtLam = extlam} <- [theDef def]
         c <- cs
         unless (rhsIsm $ clauseBody c) []
-        return (defName def, c, extlam)
+        return (extlam, defName def, c)
   case res of
     []  -> do
       reportSDoc "interaction.case" 10 $ vcat $
@@ -71,7 +74,7 @@ findClause m = do
         , text "expected rhs to be meta var" <+> (text $ show m)
         , text "but could not find it in the signature"
         ]
-      reportSDoc "interaction.case" 100 $ vcat $ map (text . show) (HMap.elems $ sigDefinitions sig)  -- you asked for it!
+      reportSDoc "interaction.case" 100 $ vcat $ map (text . show) (HMap.elems $ sig ^. sigDefinitions)  -- you asked for it!
       ifM (isInstantiatedMeta m)
         -- Andreas, 2012-03-22 If the goal has been solved by eta expansion, further
         -- case splitting is pointless and `smart-ass Agda' will refuse.
@@ -79,8 +82,7 @@ findClause m = do
         -- SUPER UGLY HACK.
         (typeError $ GenericError "Since goal is solved, further case distinction is not supported; try `Solve constraints' instead")
         (typeError $ GenericError "Right hand side must be a single hole when making a case distinction")
-    [(n,c, Just (h, nh))] -> return (ExtendedLambda h nh , n , c)
-    [(n,c, Nothing)]      -> return (FunctionDef , n , c)
+    [triple] -> return triple
     _   -> __IMPOSSIBLE__
   where
     rhsIsm (Bind b)   = rhsIsm $ unAbs b
@@ -102,48 +104,48 @@ parseVariables ii rng ss = do
   mi  <- getMetaInfo <$> lookupMeta mId
   enterClosure mi $ \ r -> do
 
-  -- Get printed representation of variables in context.
-  n  <- getContextSize
-  xs <- forM (downFrom n) $ \ i -> do
-    (,i) . P.render <$> prettyTCM (var i)
+    -- Get printed representation of variables in context.
+    n  <- getContextSize
+    xs <- forM (downFrom n) $ \ i -> do
+      (,i) . P.render <$> prettyTCM (var i)
 
-  -- Get number of module parameters.  These cannot be split on.
-  fv <- getModuleFreeVars =<< currentModule
-  let numSplittableVars = n - fv
+    -- Get number of module parameters.  These cannot be split on.
+    fv <- getCurrentModuleFreeVars
+    let numSplittableVars = n - fv
 
-  -- Resolve each string to a variable.
-  forM ss $ \ s -> do
-    let failNotVar = typeError $ GenericError $ "Not a (splittable) variable: " ++ s
-        done i
-          | i < numSplittableVars = return i
-          | otherwise             = failNotVar
+    -- Resolve each string to a variable.
+    forM ss $ \ s -> do
+      let failNotVar = typeError $ GenericError $ "Not a (splittable) variable: " ++ s
+          done i
+            | i < numSplittableVars = return i
+            | otherwise             = failNotVar
 
-    -- Note: the range in the concrete name is only approximate.
-    resName <- resolveName $ C.QName $ C.Name r $ C.stringNameParts s
-    case resName of
+      -- Note: the range in the concrete name is only approximate.
+      resName <- resolveName $ C.QName $ C.Name r $ C.stringNameParts s
+      case resName of
 
-      -- Fail if s is a name, but not of a variable.
-      DefinedName{}       -> failNotVar
-      FieldName{}         -> failNotVar
-      ConstructorName{}   -> failNotVar
-      PatternSynResName{} -> failNotVar
+        -- Fail if s is a name, but not of a variable.
+        DefinedName{}       -> failNotVar
+        FieldName{}         -> failNotVar
+        ConstructorName{}   -> failNotVar
+        PatternSynResName{} -> failNotVar
 
-      -- If s is a variable name in scope, get its de Bruijn index
-      -- via the type checker.
-      VarName x -> do
-        (v, _) <- getVarInfo x
-        case ignoreSharing v of
-          Var i [] -> done i
-          _        -> failNotVar
+        -- If s is a variable name in scope, get its de Bruijn index
+        -- via the type checker.
+        VarName x -> do
+          (v, _) <- getVarInfo x
+          case ignoreSharing v of
+            Var i [] -> done i
+            _        -> failNotVar
 
-      -- If s is not a name, compare it to the printed variable representation.
-      -- This fallback is to enable splitting on hidden variables.
-      UnknownName -> do
-        case filter ((s ==) . fst) xs of
-          []      -> typeError $ GenericError $ "Unbound variable " ++ s
-          [(_,i)] -> done i
-          -- Issue 1325: Variable names in context can be ambiguous.
-          _       -> typeError $ GenericError $ "Ambiguous variable " ++ s
+        -- If s is not a name, compare it to the printed variable representation.
+        -- This fallback is to enable splitting on hidden variables.
+        UnknownName -> do
+          case filter ((s ==) . fst) xs of
+            []      -> typeError $ GenericError $ "Unbound variable " ++ s
+            [(_,i)] -> done i
+            -- Issue 1325: Variable names in context can be ambiguous.
+            _       -> typeError $ GenericError $ "Ambiguous variable " ++ s
 
 
 -- | Entry point for case splitting tactic.
@@ -164,21 +166,36 @@ makeCase hole rng s = withInteractionId hole $ do
   let vars = words s
   if null vars then do
     -- split result
-    (newPats, sc) <- fixTarget (clauseToSplitClause clause)
-    res <- splitResult f sc
-    scs <- case res of
-      Nothing  -> if newPats then return [sc] else
-        typeError $ GenericError $ "Cannot split on result here"
-      Just cov -> mapM (snd <.> fixTarget) $ splitClauses cov
+    (piTel, sc) <- fixTarget $ clauseToSplitClause clause
+    -- Andreas, 2015-05-05 If we introduced new function arguments
+    -- do not split on result.  This might be more what the user wants.
+    -- To split on result, he can then C-c C-c again.
+    -- Andreas, 2015-05-21 Issue 1516:  However, if only hidden
+    -- arguments are introduced, C-c C-c virtually does nothing
+    -- (as they are not shown and get lost on the way to emacs and back).
+    newPats <- if null piTel then return False else do
+      -- If there were any pattern introduce, they will only have effect
+      -- if any of them is shown by the printer
+      imp <- optShowImplicit <$> pragmaOptions
+      return $ imp || any visible (telToList piTel)
+    scs <- if newPats then return [sc] else do
+      res <- splitResult f sc
+      case res of
+        Nothing  -> typeError $ GenericError $ "Cannot split on result here"
+        Just cov -> ifNotM (optCopatterns <$> pragmaOptions) failNoCop $ {-else-} do
+          mapM (snd <.> fixTarget) $ splitClauses cov
     (casectxt,) <$> mapM (makeAbstractClause f) scs
   else do
     -- split on variables
     vars <- parseVariables hole rng vars
-    (casectxt,) <$> split f vars clause
+    (casectxt,) <$> do split f vars $ clauseToSplitClause clause
   where
-  split :: QName -> [Nat] -> Clause -> TCM [A.Clause]
-  split f [] clause =
-    (:[]) <$> makeAbstractClause f (clauseToSplitClause clause)
+
+  failNoCop = typeError $ GenericError $
+    "OPTION --copatterns needed to split on result here"
+
+  split :: QName -> [Nat] -> SplitClause -> TCM [A.Clause]
+  split f [] clause = singleton <$> makeAbstractClause f clause
   split f (var : vars) clause = do
     z <- splitClauseWithAbsurd clause var
     case z of
@@ -188,19 +205,7 @@ makeCase hole rng s = withInteractionId hole $ do
         | null vars -> mapM (makeAbstractClause f) $ splitClauses cov
         | otherwise -> concat <$> do
             forM (splitClauses cov) $ \ cl ->
-              split f (mapMaybe (newVar cl) vars) $ splitClauseToClause cl
-    where
-    -- Note that the body of the created clause is the body of the
-    -- argument to split.
-    splitClauseToClause :: SplitClause -> Clause
-    splitClauseToClause c = Clause
-      { clauseRange     = noRange
-      , clauseTel       = scTel c
-      , clausePerm      = scPerm c
-      , namedClausePats = scPats c
-      , clauseBody      = clauseBody clause
-      , clauseType      = scTarget c
-      }
+              split f (mapMaybe (newVar cl) vars) cl
 
   -- Finds the new variable corresponding to an old one, if any.
   newVar :: SplitClause -> Nat -> Maybe Nat
@@ -208,7 +213,6 @@ makeCase hole rng s = withInteractionId hole $ do
     Var y [] -> Just y
     _        -> Nothing
 
-  -- NOTE: clauseToSplitClause moved to Coverage.hs
 
 makeAbsurdClause :: QName -> SplitClause -> TCM A.Clause
 makeAbsurdClause f (SClause tel perm ps _ t) = do
@@ -222,9 +226,12 @@ makeAbsurdClause f (SClause tel perm ps _ t) = do
       ]
     ]
   withCurrentModule (qnameModule f) $ do
+    -- Andreas, 2015-05-29 Issue 635
+    -- Contract implicit record patterns before printing.
+    c <- translateRecordPatterns $ Clause noRange tel perm ps NoBody t
     -- Normalise the dot patterns
-    ps <- addCtxTel tel $ normalise ps
-    inContext [] $ reify $ QNamed f $ Clause noRange tel perm ps NoBody t
+    ps <- addCtxTel tel $ normalise $ namedClausePats c
+    inTopContext $ reify $ QNamed f $ c { namedClausePats = ps }
 
 -- | Make a clause with a question mark as rhs.
 makeAbstractClause :: QName -> SplitClause -> TCM A.Clause
