@@ -1,10 +1,5 @@
 {-# LANGUAGE CPP #-}
 
-#if __GLASGOW_HASKELL__ <= 706
-{-# LANGUAGE DeriveFunctor      #-}
-{-# LANGUAGE StandaloneDeriving #-}
-#endif
-
 module Agda.Interaction.Options
     ( CommandLineOptions(..)
     , PragmaOptions(..)
@@ -26,40 +21,51 @@ module Agda.Interaction.Options
     , mapFlag
     , usage
     , tests
+    , defaultLibDir
+    -- Reused by PandocAgda
+    , inputFlag
+    , standardOptions
+    , getOptSimple
     ) where
 
 import Control.Monad            ( when )
-import Data.Maybe               ( isJust )
+
+-- base-4.7 defines the Functor instances for OptDescr and ArgDescr
+#if !(MIN_VERSION_base(4,7,0))
+import Data.Orphans             ()
+#endif
+
+import Data.Maybe
 import Data.List                ( isSuffixOf , intercalate )
-import System.Console.GetOpt    ( getOpt, usageInfo, ArgOrder(ReturnInOrder)
+import System.Console.GetOpt    ( getOpt', usageInfo, ArgOrder(ReturnInOrder)
                                 , OptDescr(..), ArgDescr(..)
                                 )
+import System.Directory         ( doesDirectoryExist )
+
+import Text.EditDistance
 
 import Agda.Termination.CutOff  ( CutOff(..) )
 
 import Agda.Utils.TestHelpers   ( runTests )
 import Agda.Utils.QuickCheck    ( quickCheck' )
-import Agda.Utils.FileName      ( AbsolutePath )
-import Agda.Utils.Monad         ( readM )
-import Agda.Utils.List          ( wordsBy )
+import Agda.Utils.FileName      ( absolute, AbsolutePath, filePath )
+import Agda.Utils.Monad         ( ifM, readM )
+import Agda.Utils.List          ( groupOn, wordsBy )
 import Agda.Utils.String        ( indent )
 import Agda.Utils.Trie          ( Trie )
 import qualified Agda.Utils.Trie as Trie
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
 
+-- Paths_Agda.hs is in $(BUILD_DIR)/build/autogen/.
+import Paths_Agda ( getDataFileName )
+
 -- | This should probably go somewhere else.
 isLiterate :: FilePath -> Bool
 isLiterate file = ".lagda" `isSuffixOf` file
 
 -- OptDescr is a Functor --------------------------------------------------
 
--- base-4.7 defines these
-#if !(MIN_VERSION_base(4,7,0))
-deriving instance Functor OptDescr
-deriving instance Functor ArgDescr
-#endif
-
 type Verbosity = Trie String Int
 
 type IncludeDirs = Either [FilePath] [AbsolutePath]
@@ -122,6 +128,7 @@ data PragmaOptions = PragmaOptions
   , optWithoutK                  :: Bool
   , optCopatterns                :: Bool  -- ^ Allow definitions by copattern matching?
   , optPatternMatching           :: Bool  -- ^ Is pattern matching allowed in the current file?
+  , optRewriting                 :: Bool  -- ^ Can rewrite rules be added and used?
   }
   deriving Show
 
@@ -192,8 +199,9 @@ defaultPragmaOptions = PragmaOptions
   , optGuardingTypeConstructors  = False
   , optUniversePolymorphism      = True
   , optWithoutK                  = False
-  , optCopatterns                = False
+  , optCopatterns                = True
   , optPatternMatching           = True
+  , optRewriting                 = False
   }
 
 -- | The default termination depth.
@@ -255,7 +263,7 @@ checkOpts opts
 
   interactive = [optInteractive, optGHCiInteraction]
 
--- Check for unsafe pramas. Gives a list of used unsafe flags.
+-- | Check for unsafe pramas. Gives a list of used unsafe flags.
 
 unsafePragmaOptions :: PragmaOptions -> [String]
 unsafePragmaOptions opts =
@@ -268,9 +276,10 @@ unsafePragmaOptions opts =
   [ "--injective-type-constructors"              | optInjectiveTypeConstructors opts ] ++
   [ "--guardedness-preserving-type-constructors" | optGuardingTypeConstructors opts  ] ++
   [ "--experimental-irrelevance"                 | optExperimentalIrrelevance opts   ] ++
-  [ "--copatterns"                               | optCopatterns opts   ]
+  [ "--rewriting"                                | optRewriting opts                 ] ++
+  []
 
--- The default pragma options should be considered safe
+-- | The default pragma options should be considered safe.
 
 defaultPragmaOptionsSafe :: IO Bool
 defaultPragmaOptionsSafe
@@ -375,9 +384,15 @@ withoutKFlag o = return $ o { optWithoutK = True }
 copatternsFlag :: Flag PragmaOptions
 copatternsFlag o = return $ o { optCopatterns = True }
 
+noCopatternsFlag :: Flag PragmaOptions
+noCopatternsFlag o = return $ o { optCopatterns = False }
+
 noPatternMatchingFlag :: Flag PragmaOptions
 noPatternMatchingFlag o = return $ o { optPatternMatching = False }
 
+rewritingFlag :: Flag PragmaOptions
+rewritingFlag o = return $ o { optRewriting = True }
+
 interactiveFlag :: Flag CommandLineOptions
 interactiveFlag  o = return $ o { optInteractive    = True
                                 , optPragmaOptions  = (optPragmaOptions o)
@@ -390,8 +405,10 @@ compileFlag o = return $ o { optCompile = True }
 compileFlagNoMain :: Flag CommandLineOptions
 compileFlagNoMain o = return $ o { optCompileNoMain = True }
 
+-- The Epic backend has been removed. See Issue 1481.
 compileEpicFlag :: Flag CommandLineOptions
-compileEpicFlag o = return $ o { optEpicCompile = True}
+-- compileEpicFlag o = return $ o { optEpicCompile = True}
+compileEpicFlag o = throwError "the Epic backend has been disabled"
 
 compileJSFlag :: Flag CommandLineOptions
 compileJSFlag  o = return $ o { optJSCompile = True }
@@ -404,8 +421,10 @@ ghcFlag :: String -> Flag CommandLineOptions
 ghcFlag f o = return $ o { optGhcFlags = optGhcFlags o ++ [f] }
 
 -- NOTE: Quadratic in number of flags.
+-- The Epic backend has been removed. See Issue 1481.
 epicFlagsFlag :: String -> Flag CommandLineOptions
-epicFlagsFlag s o = return $ o { optEpicFlags = optEpicFlags o ++ [s] }
+-- epicFlagsFlag s o = return $ o { optEpicFlags = optEpicFlags o ++ [s] }
+epicFlagsFlag s o = throwError "the Epic backend has been disabled"
 
 htmlFlag :: Flag CommandLineOptions
 htmlFlag o = return $ o { optGenerateHTML = True }
@@ -459,14 +478,23 @@ standardOptions =
                     "compile program using the MAlonzo backend (experimental)"
     , Option []     ["no-main"] (NoArg compileFlagNoMain)
                     "when compiling using the MAlonzo backend (experimental), do not treat the requested module as the main module of a program"
-    , Option []     ["epic"] (NoArg compileEpicFlag) "compile program using the Epic backend"
+
+    -- The Epic backend has been removed. See Issue 1481.
+    , Option []     ["epic"] (NoArg compileEpicFlag)
+    --                "compile program using the Epic backend"
+                    "the Epic backend has been removed"
+
     , Option []     ["js"] (NoArg compileJSFlag) "compile program using the JS backend"
     , Option []     ["compile-dir"] (ReqArg compileDirFlag "DIR")
                     ("directory for compiler output (default: the project root)")
     , Option []     ["ghc-flag"] (ReqArg ghcFlag "GHC-FLAG")
                     "give the flag GHC-FLAG to GHC when compiling using MAlonzo"
+
+    -- The Epic backend has been removed. See Issue 1481.
     , Option []     ["epic-flag"] (ReqArg epicFlagsFlag "EPIC-FLAG")
-                    "give the flag EPIC-FLAG to Epic when compiling using Epic"
+    --                "give the flag EPIC-FLAG to Epic when compiling using Epic"
+                    "the Epic backend has been removed"
+
     , Option []     ["test"] (NoArg runTestsFlag)
                     "run internal test suite"
     , Option []     ["vim"] (NoArg vimFlag)
@@ -543,28 +571,72 @@ pragmaOptions =
     , Option []     ["without-K"] (NoArg withoutKFlag)
                     "disable the K rule in pattern matching"
     , Option []     ["copatterns"] (NoArg copatternsFlag)
-                    "enable definitions by copattern matching"
+                    "enable definitions by copattern matching (default)"
+    , Option []     ["no-copatterns"] (NoArg noCopatternsFlag)
+                    "disable definitions by copattern matching"
     , Option []     ["no-pattern-matching"] (NoArg noPatternMatchingFlag)
                     "disable pattern matching completely"
+    , Option []     ["rewriting"] (NoArg rewritingFlag)
+                    "enable declaration and use of REWRITE rules"
     ]
 
 -- | Used for printing usage info.
 standardOptions_ :: [OptDescr ()]
 standardOptions_ = map (fmap $ const ()) standardOptions
 
--- | Don't export
-parseOptions' ::
-  [String] -> [OptDescr (Flag opts)] -> (String -> Flag opts) -> Flag opts
-parseOptions' argv opts fileArg = \defaults ->
-    case getOpt (ReturnInOrder fileArg) opts argv of
-        (o,_,[])    -> foldl (>>=) (return defaults) o
-        (_,_,errs)  -> throwError $ concat errs
+-- | Simple interface for System.Console.GetOpt
+--   Could be moved to Agda.Utils.Options (does not exist yet)
+getOptSimple
+  :: [String]               -- ^ command line argument words
+  -> [OptDescr (Flag opts)] -- ^ options handlers
+  -> (String -> Flag opts)  -- ^ handler of non-options (only one is allowed)
+  -> Flag opts              -- ^ combined opts data structure transformer
+getOptSimple argv opts fileArg = \ defaults ->
+  case getOpt' (ReturnInOrder fileArg) opts argv of
+    (o, _, []          , [] )  -> foldl (>>=) (return defaults) o
+    (_, _, unrecognized, errs) -> throwError $ umsg ++ emsg
+
+      where
+      ucap = "Unrecognized " ++ plural unrecognized "option" ++ ":"
+      ecap = plural errs "Option error" ++ ":"
+      umsg = if null unrecognized then "" else unlines $
+       ucap : map suggest unrecognized
+      emsg = if null errs then "" else unlines $
+       ecap : errs
+      plural [_] x = x
+      plural _   x = x ++ "s"
+
+      -- Suggest alternatives that are at most 3 typos away
+
+      longopts :: [String]
+      longopts = map ("--" ++) $ concat $ map (\ (Option _ long _ _) -> long) opts
+
+      dist :: String -> String -> Int
+      dist s t = restrictedDamerauLevenshteinDistance defaultEditCosts s t
+
+      close :: String -> String -> Maybe (Int, String)
+      close s t = let d = dist s t in if d <= 3 then Just (d, t) else Nothing
+
+      closeopts :: String -> [(Int, String)]
+      closeopts s = mapMaybe (close s) longopts
+
+      alts :: String -> [[String]]
+      alts s = map (map snd) $ groupOn fst $ closeopts s
+
+      suggest :: String -> String
+      suggest s = case alts s of
+        []     -> s
+        as : _ -> s ++ " (did you mean " ++ sugs as ++ " ?)"
+
+      sugs :: [String] -> String
+      sugs [a] = a
+      sugs as  = "any of " ++ intercalate " " as
 
 -- | Parse the standard options.
 parseStandardOptions :: [String] -> Either String CommandLineOptions
 parseStandardOptions argv =
   checkOpts =<<
-    parseOptions' argv standardOptions inputFlag defaultOptions
+    getOptSimple argv standardOptions inputFlag defaultOptions
 
 -- | Parse options from an options pragma.
 parsePragmaOptions
@@ -574,16 +646,16 @@ parsePragmaOptions
      -- ^ Command-line options which should be updated.
   -> Either String PragmaOptions
 parsePragmaOptions argv opts = do
-  ps <- parseOptions' argv pragmaOptions
+  ps <- getOptSimple argv pragmaOptions
           (\s _ -> throwError $ "Bad option in pragma: " ++ s)
           (optPragmaOptions opts)
-  checkOpts (opts { optPragmaOptions = ps })
+  _ <- checkOpts (opts { optPragmaOptions = ps })
   return ps
 
 -- | Parse options for a plugin.
 parsePluginOptions :: [String] -> [OptDescr (Flag opts)] -> Flag opts
 parsePluginOptions argv opts =
-  parseOptions' argv opts
+  getOptSimple argv opts
     (\s _ -> throwError $
                "Internal error: Flag " ++ s ++ " passed to a plugin")
 
@@ -613,6 +685,18 @@ usage options pluginInfos progName =
                 inheritedOptions pls =
                     "\n    Inherits options from: " ++ unwords pls
 
+------------------------------------------------------------------------
+-- Some paths
+
+-- | Returns the absolute default lib dir. This directory is used to
+-- store the Primitive.agda file.
+defaultLibDir :: IO FilePath
+defaultLibDir = do
+  libdir <- fmap filePath (absolute =<< getDataFileName "lib")
+  ifM (doesDirectoryExist libdir)
+      (return libdir)
+      (error $ "The lib directory " ++ libdir ++ " does not exist")
+
 ------------------------------------------------------------------------
 -- All tests
 
