@@ -7,18 +7,15 @@ module Agda.Main where
 import Control.Monad.State
 import Control.Applicative
 
-import qualified Data.List as List
 import Data.Maybe
 
 import System.Environment
 import System.Exit
 
-import qualified Text.PrettyPrint.Boxes as Boxes
-
 import Agda.Syntax.Concrete.Pretty ()
 import Agda.Syntax.Abstract.Name (toTopLevelModuleName)
 
-import Agda.Interaction.CommandLine.CommandLine
+import Agda.Interaction.CommandLine
 import Agda.Interaction.Options
 import Agda.Interaction.Monad
 import Agda.Interaction.EmacsTop (mimicGHCi)
@@ -29,9 +26,9 @@ import qualified Agda.Interaction.Highlighting.LaTeX as LaTeX
 import Agda.Interaction.Highlighting.HTML
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Benchmark
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import Agda.TypeChecking.Errors
+import Agda.TypeChecking.Pretty
 
 import Agda.Compiler.MAlonzo.Compiler as MAlonzo
 import Agda.Compiler.Epic.Compiler as Epic
@@ -39,9 +36,7 @@ import Agda.Compiler.JS.Compiler as JS
 
 import Agda.Utils.Lens
 import Agda.Utils.Monad
-import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.String
-import qualified Agda.Utils.Trie as Trie
 
 import Agda.Tests
 import Agda.Version
@@ -59,38 +54,32 @@ runAgda = do
   let opts = parseStandardOptions argv
   case opts of
     Left err -> liftIO $ optionError err
-    Right opts
-      | optShowHelp opts    -> liftIO printUsage
-      | optShowVersion opts -> liftIO printVersion
-      | optRunTests opts    -> liftIO $ do
+    Right opts -> runAgdaWithOptions generateHTML progName opts
+
+-- | Run Agda with parsed command line options and with a custom HTML generator
+runAgdaWithOptions
+  :: TCM ()             -- ^ HTML generating action
+  -> String             -- ^ program name
+  -> CommandLineOptions -- ^ parsed command line options
+  -> TCM ()
+runAgdaWithOptions generateHTML progName opts
+      | optShowHelp opts    = liftIO printUsage
+      | optShowVersion opts = liftIO printVersion
+      | optRunTests opts    = liftIO $ do
           ok <- testSuite
           unless ok exitFailure
       | isNothing (optInputFile opts)
           && not (optInteractive opts)
           && not (optGHCiInteraction opts)
-                            -> liftIO printUsage
-      | otherwise           -> do
+                            = liftIO printUsage
+      | otherwise           = do
           setCommandLineOptions opts
           -- Main function.
           -- Bill everything to root of Benchmark trie.
-          billTo [] $ checkFile
+          Bench.billTo [] $ checkFile
 
           -- Print benchmarks.
-          whenM benchmarking $ do
-            (accounts, times) <- List.unzip . Trie.toList <$> getBenchmark
-            -- Generate a table.
-            let showAccount [] = "Total time"
-                showAccount ks = List.concat . List.intersperse "." . map show $ ks
-                -- First column is accounts.
-                col1 = Boxes.vcat Boxes.left $
-                       map (Boxes.text . showAccount) $
-                       accounts
-                -- Second column is times.
-                col2 = Boxes.vcat Boxes.right $
-                       map (Boxes.text . prettyShow) $
-                       times
-                table = Boxes.hsep 1 Boxes.left [col1, col2]
-            reportBenchmarkingLn $ Boxes.render table
+          Bench.print
 
           -- Print accumulated statistics.
           printStatistics 20 Nothing =<< use lensAccumStatistics
@@ -150,6 +139,8 @@ runAgda = do
             SomeWarnings (Warnings _ _)        -> return Nothing
             NoWarnings                         -> return $ Just i
 
+          reportSDoc "main" 50 $ pretty i
+
           whenM (optGenerateHTML <$> commandLineOptions) $
             generateHTML
 
@@ -179,10 +170,10 @@ optionError err = do
   printUsage
   exitFailure
 
--- | Main
-main :: IO ()
-main = do
-    r <- runTCMTop $ runAgda `catchError` \err -> do
+-- | Run a TCM action in IO; catch and pretty print errors.
+runTCMPrettyErrors :: TCM () -> IO ()
+runTCMPrettyErrors tcm = do
+    r <- runTCMTop $ tcm `catchError` \err -> do
       s <- prettyError err
       liftIO $ putStrLn s
       throwError err
@@ -192,3 +183,7 @@ main = do
   `catchImpossible` \e -> do
     putStr $ show e
     exitFailure
+
+-- | Main
+main :: IO ()
+main = runTCMPrettyErrors runAgda
