@@ -6,7 +6,6 @@
 {-# LANGUAGE FlexibleInstances     #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE TemplateHaskell       #-}
-{-# LANGUAGE TypeSynonymInstances  #-}
 
 {-| The abstract syntax. This is what you get after desugaring and scope
     analysis of the concrete syntax. The type checker works on abstract syntax,
@@ -117,7 +116,7 @@ data Axiom
   deriving (Typeable, Eq, Ord, Show)
 
 -- | Renaming (generic).
-type Ren a = Map a a
+type Ren a = [(a, a)]
 
 data Declaration
   = Axiom      Axiom DefInfo ArgInfo QName Expr      -- ^ type signature (can be irrelevant and colored, but not hidden)
@@ -170,6 +169,9 @@ data ModuleApplication
 data Pragma
   = OptionsPragma [String]
   | BuiltinPragma String Expr
+  | BuiltinNoDefPragma String QName
+    -- ^ Builtins that do not come with a definition,
+    --   but declare a name for an Agda concept.
   | RewritePragma QName
   | CompiledPragma QName String
   | CompiledExportPragma QName String
@@ -209,20 +211,25 @@ data TypedBindings = TypedBindings Range (Arg TypedBinding)
             -- ^ . @(xs : e)@ or @{xs : e}@
   deriving (Typeable, Show)
 
--- | A typed binding. Appears in dependent function spaces, typed lambdas, and
---   telescopes. I might be tempting to simplify this to only bind a single
---   name at a time. This would mean that we would have to typecheck the type
---   several times (@(x y : A)@ vs. @(x : A)(y : A)@).
---   In most cases this wouldn't really be a problem, but it's good
---   principle to not do extra work unless you have to.
+-- | A typed binding.  Appears in dependent function spaces, typed lambdas, and
+--   telescopes.  It might be tempting to simplify this to only bind a single
+--   name at a time, and translate, say, @(x y : A)@ to @(x : A)(y : A)@
+--   before type-checking.  However, this would be slightly problematic:
 --
---   (Andreas, 2013-12-10: The more serious problem would that the translation
---   from @(x y : ?)@ to @(x : ?) (y : ?)@ duplicates the hole @?@.
+--     1. We would have to typecheck the type @A@ several times.
+--
+--     2. If @A@ contains a meta variable or hole, it would be duplicated
+--        by such a translation.
+--
+--   While 1. is only slightly inefficient, 2. would be an outright bug.
+--   Duplicating @A@ could not be done naively, we would have to make sure
+--   that the metas of the copy are aliases of the metas of the original.
+
 data TypedBinding
-  = TBind Range [Name] Expr
+  = TBind Range [WithHiding Name] Expr
     -- ^ As in telescope @(x y z : A)@ or type @(x y z : A) -> B@.
   | TLet Range [LetBinding]
-    -- ^
+    -- ^ E.g. @(let x = e)@ or @(let open M)@.
   deriving (Typeable, Show)
 
 type Telescope  = [TypedBindings]
@@ -244,7 +251,7 @@ data RHS
   | AbsurdRHS
   | WithRHS QName [Expr] [Clause]
       -- ^ The 'QName' is the name of the with function.
-  | RewriteRHS [QName] [Expr] RHS [Declaration]
+  | RewriteRHS [(QName, Expr)] RHS [Declaration]
       -- ^ The 'QName's are the names of the generated with functions.
       --   One for each 'Expr'.
       --   The RHS shouldn't be another @RewriteRHS@.
@@ -288,7 +295,7 @@ data LHSCore' e
              , lhsPatsRight  :: [NamedArg (Pattern' e)]
                -- ^ Further applied to patterns.
              }
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  deriving (Eq, Typeable, Show, Functor, Foldable, Traversable)
 
 type LHSCore = LHSCore' Expr
 
@@ -369,12 +376,11 @@ data Pattern' e
     -- ^ Defined pattern: function definition @f ps@ or destructor pattern @d p ps@.
   | WildP PatInfo
     -- ^ Underscore pattern entered by user.
+    --   Or generated at type checking for implicit arguments.
   | AsP PatInfo Name (Pattern' e)
   | DotP PatInfo e
   | AbsurdP PatInfo
   | LitP Literal
-  | ImplicitP PatInfo
-    -- ^ Generated at type checking for implicit arguments.
   | PatternSynP PatInfo QName [NamedArg (Pattern' e)]
   deriving (Typeable, Show, Functor, Foldable, Traversable)
 
@@ -399,6 +405,41 @@ instance IsProjP a => IsProjP (Named n a) where
     Instances
  --------------------------------------------------------------------------}
 
+-- | Literal equality of patterns, ignoring dot patterns
+instance Eq (Pattern' e) where
+  p == p' =
+    case (p,p') of
+      ((VarP x)             , (VarP x')             ) -> x === x'
+      ((ConP _ x ps)        , (ConP _ x' ps')       ) -> x == x' && ps == ps'
+      ((DefP _ x ps)        , (DefP _ x' ps')       ) -> x == x' && ps == ps'
+      ((WildP _)            , (WildP _)             ) -> True
+      ((AsP _ x p)          , (AsP _ x' p')         ) -> x === x' && p == p'
+      ((DotP _ _)           , (DotP _ _)            ) -> True
+      (AbsurdP{}            , AbsurdP{}             ) -> True
+      ((LitP l)             , (LitP l')             ) -> l == l'
+      ((PatternSynP _ x ps) , (PatternSynP _ x' ps')) -> x == x' && ps == ps'
+      (_                    , _                     ) -> False
+    where (Name _ (C.Name _ x) _ _) === (Name _ (C.Name _ x') _ _) = True
+          (Name _ C.NoName{}   _ _) === (Name _ C.NoName{}    _ _) = True
+          _                         === _                          = False
+
+instance Eq LHS where
+  (LHS _ core wps) == (LHS _ core' wps') = core == core' && wps == wps'
+
+instance Underscore Expr where
+  underscore   = Underscore emptyMetaInfo
+  isUnderscore = __IMPOSSIBLE__
+
+instance LensHiding TypedBindings where
+  getHiding   (TypedBindings _ a) = getHiding a
+  mapHiding f (TypedBindings r a) = TypedBindings r $ mapHiding f a
+
+instance LensHiding LamBinding where
+  getHiding   (DomainFree ai _) = getHiding ai
+  getHiding   (DomainFull tb)   = getHiding tb
+  mapHiding f (DomainFree ai x) = mapHiding f ai `DomainFree` x
+  mapHiding f (DomainFull tb)   = DomainFull $ mapHiding f tb
+
 instance HasRange LamBinding where
     getRange (DomainFree _ x) = getRange x
     getRange (DomainFull b)   = getRange b
@@ -464,7 +505,6 @@ instance HasRange (Pattern' e) where
     getRange (ConP i _ _)        = getRange i
     getRange (DefP i _ _)        = getRange i
     getRange (WildP i)           = getRange i
-    getRange (ImplicitP i)       = getRange i
     getRange (AsP i _ _)         = getRange i
     getRange (DotP i _)          = getRange i
     getRange (AbsurdP i)         = getRange i
@@ -488,7 +528,7 @@ instance HasRange RHS where
     getRange AbsurdRHS                = noRange
     getRange (RHS e)                  = getRange e
     getRange (WithRHS _ e cs)         = fuseRange e cs
-    getRange (RewriteRHS _ es rhs wh) = getRange (es, rhs, wh)
+    getRange (RewriteRHS xes rhs wh)  = getRange (map snd xes, rhs, wh)
 
 instance HasRange LetBinding where
     getRange (LetBind  i _ _ _ _     ) = getRange i
@@ -502,7 +542,6 @@ instance SetRange (Pattern' a) where
     setRange r (ConP i ns as)       = ConP (setRange r i) ns as
     setRange r (DefP _ n as)        = DefP (PatRange r) (setRange r n) as
     setRange r (WildP _)            = WildP (PatRange r)
-    setRange r (ImplicitP _)        = ImplicitP (PatRange r)
     setRange r (AsP _ n p)          = AsP (PatRange r) (setRange r n) p
     setRange r (DotP _ e)           = DotP (PatRange r) e
     setRange r (AbsurdP _)          = AbsurdP (PatRange r)
@@ -531,17 +570,15 @@ instance KillRange Expr where
   killRange (App i e1 e2)          = killRange3 App i e1 e2
   killRange (WithApp i e es)       = killRange3 WithApp i e es
   killRange (Lam i b e)            = killRange3 Lam i b e
-  killRange (AbsurdLam i h)        = killRange1 AbsurdLam i h
+  killRange (AbsurdLam i h)        = killRange2 AbsurdLam i h
   killRange (ExtendedLam i n d ps) = killRange4 ExtendedLam i n d ps
   killRange (Pi i a b)             = killRange3 Pi i a b
   killRange (Fun i a b)            = killRange3 Fun i a b
-  killRange (Set i n)              = Set (killRange i) n
+  killRange (Set i n)              = killRange2 Set i n
   killRange (Prop i)               = killRange1 Prop i
   killRange (Let i ds e)           = killRange3 Let i ds e
-  killRange (Rec i fs)             = Rec (killRange i) (map (id -*- killRange) fs)
-  killRange (RecUpdate i e fs)     = RecUpdate (killRange i)
-                                               (killRange e)
-                                               (map (id -*- killRange) fs)
+  killRange (Rec i fs)             = killRange2 Rec i fs
+  killRange (RecUpdate i e fs)     = killRange3 RecUpdate i e fs
   killRange (ETel tel)             = killRange1 ETel tel
   killRange (ScopedExpr s e)       = killRange1 (ScopedExpr s) e
   killRange (QuoteGoal i x e)      = killRange3 QuoteGoal i x e
@@ -549,7 +586,7 @@ instance KillRange Expr where
   killRange (Quote i)              = killRange1 Quote i
   killRange (QuoteTerm i)          = killRange1 QuoteTerm i
   killRange (Unquote i)            = killRange1 Unquote i
-  killRange (DontCare e)           = DontCare e
+  killRange (DontCare e)           = killRange1 DontCare e
   killRange (PatternSyn x)         = killRange1 PatternSyn x
 
 instance KillRange Declaration where
@@ -576,15 +613,11 @@ instance KillRange ModuleApplication where
   killRange (SectionApp a b c  ) = killRange3 SectionApp a b c
   killRange (RecordModuleIFS a ) = killRange1 RecordModuleIFS a
 
-instance KillRange x => KillRange (ThingWithFixity x) where
-  killRange (ThingWithFixity c f) = ThingWithFixity (killRange c) f
-
 instance KillRange e => KillRange (Pattern' e) where
   killRange (VarP x)            = killRange1 VarP x
   killRange (ConP i a b)        = killRange3 ConP i a b
   killRange (DefP i a b)        = killRange3 DefP i a b
   killRange (WildP i)           = killRange1 WildP i
-  killRange (ImplicitP i)       = killRange1 ImplicitP i
   killRange (AsP i a b)         = killRange3 AsP i a b
   killRange (DotP i a)          = killRange2 DotP i a
   killRange (AbsurdP i)         = killRange1 AbsurdP i
@@ -608,7 +641,7 @@ instance KillRange RHS where
   killRange AbsurdRHS                = AbsurdRHS
   killRange (RHS e)                  = killRange1 RHS e
   killRange (WithRHS q e cs)         = killRange3 WithRHS q e cs
-  killRange (RewriteRHS x es rhs wh) = killRange4 RewriteRHS x es rhs wh
+  killRange (RewriteRHS xes rhs wh)  = killRange3 RewriteRHS xes rhs wh
 
 instance KillRange LetBinding where
   killRange (LetBind    i info a b c) = killRange5 LetBind  i info a b c
@@ -685,7 +718,7 @@ instance AllNames RHS where
   allNames (RHS e)                   = allNames e
   allNames AbsurdRHS{}               = Seq.empty
   allNames (WithRHS q _ cls)         = q <| allNames cls
-  allNames (RewriteRHS qs _ rhs cls) = Seq.fromList qs >< allNames rhs >< allNames cls
+  allNames (RewriteRHS qes rhs cls) = Seq.fromList (map fst qes) >< allNames rhs >< allNames cls
 
 instance AllNames Expr where
   allNames Var{}                   = Seq.empty
@@ -783,7 +816,6 @@ patternToExpr (AsP _ _ p)         = patternToExpr p
 patternToExpr (DotP _ e)          = e
 patternToExpr (AbsurdP _)         = Underscore emptyMetaInfo  -- TODO: could this happen?
 patternToExpr (LitP l)            = Lit l
-patternToExpr (ImplicitP _)       = Underscore emptyMetaInfo
 patternToExpr (PatternSynP _ _ _) = __IMPOSSIBLE__
 
 type PatternSynDefn = ([Arg Name], Pattern)
@@ -802,8 +834,7 @@ substPattern s p = case p of
   DotP i e    -> DotP i (substExpr (map (fmap patternToExpr) s) e)
   AbsurdP i   -> p
   LitP l      -> p
-  _           -> __IMPOSSIBLE__ -- Implicits (generated at TC time),
-                                -- pattern synonyms (already gone), and
+  _           -> __IMPOSSIBLE__ -- pattern synonyms (already gone) and
                                 -- @-patterns (not supported anyways).
 
 class SubstExpr a where
@@ -892,4 +923,3 @@ insertImplicitPatSynArgs wild r ns as = matchArgs r ns as
     matchArgs r (n:ns) as = do
       (p, as) <- matchNextArg r n as
       first ((unArg n, p) :) <$> matchArgs (getRange p) ns as
-
