@@ -1,12 +1,11 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE DeriveFunctor        #-}
-{-# LANGUAGE FlexibleContexts     #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE ScopedTypeVariables  #-}
-{-# LANGUAGE StandaloneDeriving   #-}
-{-# LANGUAGE TupleSections        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE TupleSections #-}
 
 module Agda.Syntax.Abstract.Copatterns (translateCopatternClauses) where
 
@@ -263,6 +262,10 @@ class Rename e where
   rename' :: NameMap -> e -> e
   rename' rho = rename (flip lookup rho)
 
+-- | 'QName's are not renamed.
+instance Rename QName where
+  rename _ q = q
+
 instance Rename Expr where
   rename rho e =
     case e of
@@ -325,19 +328,13 @@ instance Rename RHS where
       RHS e                 -> RHS (rename rho e)
       AbsurdRHS             -> e
       WithRHS n es cs       -> WithRHS n (rename rho es) (rename rho cs)
-      RewriteRHS ns es r ds -> RewriteRHS ns (rename rho es) (rename rho r) (rename rho ds)
+      RewriteRHS nes r ds   -> RewriteRHS (rename rho nes) (rename rho r) (rename rho ds)
 
 instance Rename LHS where
   rename rho (LHS i core ps) = LHS i (rename rho core) (rename rho ps)
 
 instance Rename LHSCore where
   rename rho = fmap (rename rho) -- only rename in dot patterns
-{-
-  rename rho = ren where
-    ren e = case e of
-      LHSHead f ps           -> LHSHead f (ren ps)
-      LHSProj d ps1 core ps2 -> LHSProj d (ren ps1) (ren core) (ren ps2)
--}
 
 instance Rename Pattern where
   rename rho = fmap (rename rho) -- only rename in dot patterns
@@ -354,6 +351,8 @@ instance Rename a => Rename (Named n a) where
 instance Rename a => Rename [a] where
   rename rho = map (rename rho)
 
+instance (Rename a, Rename b) => Rename (a, b) where
+  rename rho (a,b) = (rename rho a, rename rho b)
 
 
 
@@ -379,15 +378,12 @@ instance Alpha (Pattern' e) where
       ((DotP _ _)           , (DotP _ _)            ) -> return ()
       (AbsurdP{}            , AbsurdP{}             ) -> return ()
       ((LitP l)             , (LitP l')             ) -> guard (l == l')
-      (ImplicitP{}          , ImplicitP{}           ) -> return ()
       ((PatternSynP _ x ps) , (PatternSynP _ x' ps')) -> guard (x == x') >> alpha' ps ps'
       (_                    , _                     ) -> fail "not alpha equivalent"
 
 tell1 :: (MonadWriter [a] m) => a -> m ()
 tell1 a = tell [a]
 
-deriving instance Eq AmbiguousQName
-
 instance Alpha (LHSCore' e) where
   alpha' (LHSHead f ps) (LHSHead f' ps') = guard (f == f') >> alpha' ps ps'
   alpha' (LHSProj d ps1 lhs ps2) (LHSProj d' ps1' lhs' ps2') =
@@ -405,27 +401,3 @@ instance (Eq n, Alpha a) => Alpha (Named n a) where
 
 instance Alpha a => Alpha [a] where
   alpha' l l' = guard (length l == length l') >> zipWithM_ alpha' l l'
-
--- | Literal equality of patterns, ignoring dot patterns
-instance Eq (Pattern' e) where
-  p == p' =
-    case (p,p') of
-      ((VarP x)             , (VarP x')             ) -> x === x'
-      ((ConP _ x ps)        , (ConP _ x' ps')       ) -> x == x' && ps == ps'
-      ((DefP _ x ps)        , (DefP _ x' ps')       ) -> x == x' && ps == ps'
-      ((WildP _)            , (WildP _)             ) -> True
-      ((AsP _ x p)          , (AsP _ x' p')         ) -> x === x' && p == p'
-      ((DotP _ _)           , (DotP _ _)            ) -> True
-      (AbsurdP{}            , AbsurdP{}             ) -> True
-      ((LitP l)             , (LitP l')             ) -> l == l'
-      (ImplicitP{}          , ImplicitP{}           ) -> True
-      ((PatternSynP _ x ps) , (PatternSynP _ x' ps')) -> x == x' && ps == ps'
-      (_                    , _                     ) -> False
-    where (A.Name _ (C.Name _ x) _ _) === (A.Name _ (C.Name _ x') _ _) = True
-          (A.Name _ C.NoName{}   _ _) === (A.Name _ C.NoName{}    _ _) = True
-          _                           === _                            = False
-
-deriving instance Eq (LHSCore' e)
-
-instance Eq LHS where
-  (LHS _ core wps) == (LHS _ core' wps') = core == core' && wps == wps'
