@@ -1,12 +1,12 @@
-{-# LANGUAGE CPP                        #-}
-{-# LANGUAGE DeriveDataTypeable         #-}
-{-# LANGUAGE DeriveFoldable             #-}
-{-# LANGUAGE DeriveFunctor              #-}
-{-# LANGUAGE DeriveTraversable          #-}
-{-# LANGUAGE FlexibleContexts           #-}
-{-# LANGUAGE FlexibleInstances          #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE TypeSynonymInstances       #-}
+{-# LANGUAGE StandaloneDeriving #-}
 
 {-| Abstract names carry unique identifiers and stuff.
 -}
@@ -24,13 +24,16 @@ import Data.List
 import Data.Function
 import Data.Hashable
 
+import Test.QuickCheck
+
 import Agda.Syntax.Position
 import Agda.Syntax.Common
-import Agda.Syntax.Fixity
+import {-# SOURCE #-} Agda.Syntax.Fixity
 import Agda.Syntax.Concrete.Name (IsNoName(..))
 import qualified Agda.Syntax.Concrete.Name as C
 
--- import Agda.Utils.Function
+import Agda.Utils.List
+import Agda.Utils.Monad
 import Agda.Utils.Pretty
 import Agda.Utils.Size
 import Agda.Utils.Suffix
@@ -64,7 +67,7 @@ data QNamed a = QNamed
   { qname  :: QName
   , qnamed :: a
   }
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  deriving (Typeable, Functor, Foldable, Traversable)
 
 -- | A module name is just a qualified name.
 --
@@ -78,7 +81,7 @@ newtype ModuleName = MName { mnameToList :: [Name] }
 -- Invariant: All the names in the list must have the same concrete,
 -- unqualified name.  (This implies that they all have the same 'Range').
 newtype AmbiguousQName = AmbQ { unAmbQ :: [QName] }
-  deriving (Typeable, Show)
+  deriving (Eq, Typeable)
 
 -- | A module is anonymous if the qualification path ends in an underscore.
 isAnonymousModuleName :: ModuleName -> Bool
@@ -103,12 +106,11 @@ isAnonymousModuleName (MName ms) = isNoName $ last ms
 -- least as large as the length of the list.
 
 withRangesOf :: ModuleName -> [C.Name] -> ModuleName
-MName ms `withRangesOf` ns
-  | length ms < length ns  = __IMPOSSIBLE__
-  | otherwise              = MName $
-      reverse $ zipWith setRange
-                        (reverse (map getRange ns) ++ repeat noRange)
-                        (reverse ms)
+MName ms `withRangesOf` ns = if m < n then __IMPOSSIBLE__ else MName $
+      zipWith setRange (replicate (m - n) noRange ++ map getRange ns) ms
+  where
+    m = length ms
+    n = length ns
 
 -- | Like 'withRangesOf', but uses the name parts (qualifier + name)
 -- of the qualified name as the list of concrete names.
@@ -122,6 +124,10 @@ mnameFromList = MName
 noModuleName :: ModuleName
 noModuleName = mnameFromList []
 
+commonParentModule :: ModuleName -> ModuleName -> ModuleName
+commonParentModule m1 m2 =
+  mnameFromList $ commonPrefix (mnameToList m1) (mnameToList m2)
+
 -- | Make a 'Name' from some kind of string.
 class MkName a where
   -- | The 'Range' sets the /definition site/ of the name, not the use site.
@@ -254,6 +260,12 @@ instance IsNoName Name where
 -- * Show instances
 ------------------------------------------------------------------------
 
+-- deriving instance Show Name
+-- deriving instance Show ModuleName
+-- deriving instance Show QName
+deriving instance Show a => Show (QNamed a)
+deriving instance Show AmbiguousQName
+
 -- | Only use this @show@ function in debugging!  To convert an
 --   abstract 'Name' into a string use @prettyShow@.
 instance Show Name where
@@ -261,19 +273,17 @@ instance Show Name where
   -- Reason: I do not have time just now to properly fix the
   -- use of Show Name for pretty printing everywhere, e.g. in
   -- the Epic backend.  But I want to push the fix for Issue 836 now.
-  show n = show (nameConcrete n)
-  -- show n = show (nameConcrete n) ++ "^" ++ show (nameId n)
-  -- show n = applyWhen (isNoName n) (++ show (nameId n)) $ show (nameConcrete n)
+  show = prettyShow
 
 -- | Only use this @show@ function in debugging!  To convert an
 --   abstract 'ModuleName' into a string use @prettyShow@.
 instance Show ModuleName where
-  show m = concat $ intersperse "." $ map show $ mnameToList m
+  show = prettyShow
 
 -- | Only use this @show@ function in debugging!  To convert an
 --   abstract 'QName' into a string use @prettyShow@.
 instance Show QName where
-  show q = concat $ intersperse "." $ map show $ qnameToList q
+  show = prettyShow
 
 ------------------------------------------------------------------------
 -- * Pretty instances
@@ -288,6 +298,12 @@ instance Pretty ModuleName where
 instance Pretty QName where
   pretty = hcat . punctuate (text ".") . map pretty . qnameToList
 
+instance Pretty AmbiguousQName where
+  pretty (AmbQ qs) = hcat $ punctuate (text " | ") $ map pretty qs
+
+instance Pretty a => Pretty (QNamed a) where
+  pretty (QNamed a b) = pretty a <> text "." <> pretty b
+
 ------------------------------------------------------------------------
 -- * Range instances
 ------------------------------------------------------------------------
@@ -326,23 +342,25 @@ instance SetRange ModuleName where
 -- ** KillRange
 
 instance KillRange Name where
-  killRange x = x { nameConcrete = killRange $ nameConcrete x
-                  -- Andreas, 2014-03-30
-                  -- An experiment: what happens if we preserve
-                  -- the range of the binding site, but kill all
-                  -- other ranges before serialization?
-                  -- Andreas, Makoto, 2014-10-18 AIM XX
-                  -- Kill all ranges in signature, including nameBindingSite.
-                  , nameBindingSite = noRange
-                  }
+  killRange (Name a b c d) = killRange4 Name a b c d
+  -- killRange x = x { nameConcrete = killRange $ nameConcrete x
+  --                 -- Andreas, 2014-03-30
+  --                 -- An experiment: what happens if we preserve
+  --                 -- the range of the binding site, but kill all
+  --                 -- other ranges before serialization?
+  --                 -- Andreas, Makoto, 2014-10-18 AIM XX
+  --                 -- Kill all ranges in signature, including nameBindingSite.
+  --                 , nameBindingSite = noRange
+  --                 }
 
 instance KillRange ModuleName where
   killRange (MName xs) = MName $ killRange xs
 
 instance KillRange QName where
-  killRange q = q { qnameModule = killRange $ qnameModule q
-                  , qnameName   = killRange $ qnameName   q
-                  }
+  killRange (QName a b) = killRange2 QName a b
+  -- killRange q = q { qnameModule = killRange $ qnameModule q
+  --                 , qnameName   = killRange $ qnameName   q
+  --                 }
 
 instance KillRange AmbiguousQName where
   killRange (AmbQ xs) = AmbQ $ killRange xs
@@ -356,3 +374,26 @@ instance Sized QName where
 
 instance Sized ModuleName where
   size = size . mnameToList
+
+------------------------------------------------------------------------
+-- * Arbitrary instances
+------------------------------------------------------------------------
+
+-- | The generated names all have the same 'Fixity'': 'noFixity''.
+
+instance Arbitrary Name where
+  arbitrary =
+    Name <$> arbitrary <*> arbitrary <*> arbitrary
+         <*> return noFixity'
+
+instance CoArbitrary Name where
+  coarbitrary = coarbitrary . nameId
+
+instance Arbitrary QName where
+  arbitrary = do
+    ms <- arbitrary
+    n  <- arbitrary
+    return (QName (MName ms) n)
+
+instance CoArbitrary QName where
+  coarbitrary = coarbitrary . qnameName
