@@ -1,11 +1,11 @@
-{-# LANGUAGE CPP                        #-}
-{-# LANGUAGE DeriveDataTypeable         #-}
-{-# LANGUAGE DeriveFoldable             #-}
-{-# LANGUAGE DeriveFunctor              #-}
-{-# LANGUAGE DeriveTraversable          #-}
-{-# LANGUAGE FlexibleInstances          #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE TypeSynonymInstances       #-}
 
 {-| Some common syntactic entities are defined in this module.
 -}
@@ -17,16 +17,18 @@ import Data.ByteString.Char8 (ByteString)
 import qualified Data.ByteString.Char8 as ByteString
 import Data.Foldable
 import Data.Hashable
+import Data.Monoid
 import Data.Traversable
 import Data.Typeable (Typeable)
 
-import Test.QuickCheck
+import GHC.Generics (Generic)
+
+import Test.QuickCheck hiding (Small)
 
 import Agda.Syntax.Position
 
 import Agda.Utils.Functor
 import Agda.Utils.Pretty
-import Agda.Utils.Size
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -73,9 +75,39 @@ instance CoArbitrary Induction where
 data Hiding  = Hidden | Instance | NotHidden
   deriving (Typeable, Show, Eq, Ord)
 
+-- | 'Hiding' is an idempotent partial monoid, with unit 'NotHidden'.
+--   'Instance' and 'NotHidden' are incompatible.
+instance Monoid Hiding where
+  mempty = NotHidden
+  mappend NotHidden h         = h
+  mappend h         NotHidden = h
+  mappend Hidden    Hidden    = Hidden
+  mappend Instance  Instance  = Instance
+  mappend _         _         = __IMPOSSIBLE__
+
 instance KillRange Hiding where
   killRange = id
 
+-- | Decorating something with 'Hiding' information.
+data WithHiding a = WithHiding Hiding a
+  deriving (Typeable, Eq, Ord, Show, Functor, Foldable, Traversable)
+
+instance Decoration WithHiding where
+  traverseF f (WithHiding h a) = WithHiding h <$> f a
+
+instance Applicative WithHiding where
+  pure = WithHiding mempty
+  WithHiding h f <*> WithHiding h' a = WithHiding (mappend h h') (f a)
+
+instance HasRange a => HasRange (WithHiding a) where
+  getRange = getRange . dget
+
+instance SetRange a => SetRange (WithHiding a) where
+  setRange = fmap . setRange
+
+instance KillRange a => KillRange (WithHiding a) where
+  killRange = fmap killRange
+
 -- | A lens to access the 'Hiding' attribute in data structures.
 --   Minimal implementation: @getHiding@ and one of @setHiding@ or @mapHiding@.
 class LensHiding a where
@@ -93,6 +125,15 @@ instance LensHiding Hiding where
   setHiding = const
   mapHiding = id
 
+instance LensHiding (WithHiding a) where
+  getHiding   (WithHiding h _) = h
+  setHiding h (WithHiding _ a) = WithHiding h a
+  mapHiding f (WithHiding h a) = WithHiding (f h) a
+
+-- | Monoidal composition of 'Hiding' information in some data.
+mergeHiding :: LensHiding a => WithHiding a -> a
+mergeHiding (WithHiding h a) = mapHiding (mappend h) a
+
 -- | @isHidden@ does not apply to 'Instance', only to 'Hidden'.
 isHidden :: LensHiding a => a -> Bool
 isHidden a = getHiding a == Hidden
@@ -119,27 +160,55 @@ makeInstance = setHiding Instance
 -- * Relevance
 ---------------------------------------------------------------------------
 
+-- | An constructor argument is big if the sort of its type is bigger than
+--   the sort of the data type.  Only parameters (and maybe forced arguments)
+--   are allowed to be big.
+--   @
+--      List : Set -> Set
+--      nil  : (A : Set) -> List A
+--   @
+--   @A@ is big in constructor @nil@ as the sort @Set1@ of its type @Set@
+--   is bigger than the sort @Set@ of the data type @List@.
+data Big = Big | Small
+  deriving (Typeable, Show, Eq, Enum, Bounded)
+
+instance Ord Big where
+  Big <= Small = False
+  _   <= _     = True
+
 -- | A function argument can be relevant or irrelevant.
---   See 'Agda.TypeChecking.Irrelevance'.
+--   See "Agda.TypeChecking.Irrelevance".
 data Relevance
   = Relevant    -- ^ The argument is (possibly) relevant at compile-time.
   | NonStrict   -- ^ The argument may never flow into evaluation position.
                 --   Therefore, it is irrelevant at run-time.
                 --   It is treated relevantly during equality checking.
   | Irrelevant  -- ^ The argument is irrelevant at compile- and runtime.
-  | Forced      -- ^ The argument can be skipped during equality checking
+  | Forced Big  -- ^ The argument can be skipped during equality checking
                 --   because its value is already determined by the type.
+                --   If a constructor argument is big, it has to be regarded
+                --   absent, otherwise we get into paradoxes.
   | UnusedArg   -- ^ The polarity checker has determined that this argument
                 --   is unused in the definition.  It can be skipped during
                 --   equality checking but should be mined for solutions
                 --   of meta-variables with relevance 'UnusedArg'
-    deriving (Typeable, Show, Eq, Enum, Bounded)
+    deriving (Typeable, Show, Eq)
+
+allRelevances :: [Relevance]
+allRelevances =
+  [ Relevant
+  , NonStrict
+  , Irrelevant
+  , Forced Small
+  , Forced Big
+  , UnusedArg
+  ]
 
 instance KillRange Relevance where
   killRange rel = rel -- no range to kill
 
 instance Arbitrary Relevance where
-  arbitrary = elements [minBound..maxBound]
+  arbitrary = elements allRelevances
 
 instance Ord Relevance where
   (<=) = moreRelevant
@@ -186,11 +255,76 @@ moreRelevant r r' =
     (UnusedArg, _)  -> True
     (_, UnusedArg)  -> False
     -- third bottom
-    (Forced, _)     -> True
-    (_, Forced)     -> False
+    (Forced{}, _)   -> True
+    (_, Forced{})   -> False
     -- remaining case
     (NonStrict,NonStrict) -> True
 
+irrelevantOrUnused :: Relevance -> Bool
+irrelevantOrUnused r =
+  case r of
+    Irrelevant -> True
+    UnusedArg  -> True
+    NonStrict  -> False
+    Relevant   -> False
+    Forced{}   -> False
+
+-- | @unusableRelevance rel == True@ iff we cannot use a variable of @rel@.
+unusableRelevance :: Relevance -> Bool
+unusableRelevance rel = NonStrict `moreRelevant` rel
+
+-- | 'Relevance' composition.
+--   'Irrelevant' is dominant, 'Relevant' is neutral.
+composeRelevance :: Relevance -> Relevance -> Relevance
+composeRelevance r r' =
+  case (r, r') of
+    (Irrelevant, _) -> Irrelevant
+    (_, Irrelevant) -> Irrelevant
+    (NonStrict, _)  -> NonStrict
+    (_, NonStrict)  -> NonStrict
+    (Forced b, Forced b') -> Forced (max b b')  -- prefer Big over Small
+    (Forced b, _)     -> Forced b
+    (_, Forced b)     -> Forced b
+    (UnusedArg, _)  -> UnusedArg
+    (_, UnusedArg)  -> UnusedArg
+    (Relevant, Relevant) -> Relevant
+
+-- | @inverseComposeRelevance r x@ returns the most irrelevant @y@
+--   such that forall @x@, @y@ we have
+--   @x \`moreRelevant\` (r \`composeRelevance\` y)@
+--   iff
+--   @(r \`inverseComposeRelevance\` x) \`moreRelevant\` y@ (Galois connection).
+inverseComposeRelevance :: Relevance -> Relevance -> Relevance
+inverseComposeRelevance r x =
+  case (r, x) of
+    (Relevant, x)        -> x          -- going to relevant arg.: nothing changes
+    _ | r == x           -> Relevant   -- because Relevant is comp.-neutral
+    (Forced{}, Forced{}) -> Relevant   -- same, but (==) does not ignore Big
+    (UnusedArg, x)       -> x
+    (Forced{}, UnusedArg)  -> Relevant
+    (Forced{}, x)          -> x
+    (Irrelevant, x)      -> Relevant   -- going irrelevant: every thing usable
+    (_, Irrelevant)      -> Irrelevant -- otherwise: irrelevant things remain unusable
+    (NonStrict, _)       -> Relevant   -- but @NonStrict@s become usable
+
+-- | For comparing @Relevance@ ignoring @Forced@ and @UnusedArg@.
+ignoreForced :: Relevance -> Relevance
+ignoreForced Forced{}   = Relevant
+ignoreForced UnusedArg  = Relevant
+ignoreForced Relevant   = Relevant
+ignoreForced NonStrict  = NonStrict
+ignoreForced Irrelevant = Irrelevant
+
+-- | Irrelevant function arguments may appear non-strictly in the codomain type.
+irrToNonStrict :: Relevance -> Relevance
+irrToNonStrict Irrelevant = NonStrict
+-- irrToNonStrict NonStrict  = Relevant -- TODO: is that what we want (OR: NonStrict)  -- better be more conservative
+irrToNonStrict rel        = rel
+
+nonStrictToIrr :: Relevance -> Relevance
+nonStrictToIrr NonStrict = Irrelevant
+nonStrictToIrr rel       = rel
+
 ---------------------------------------------------------------------------
 -- * Argument decoration
 ---------------------------------------------------------------------------
@@ -206,7 +340,7 @@ data ArgInfo c = ArgInfo
 instance KillRange c => KillRange (ArgInfo c) where
   killRange (ArgInfo h r cs) = killRange3 ArgInfo h r cs
 
-{- FAILED to define a less for ArgInfo, since it is parametrized by c
+{- FAILED to define a lens for ArgInfo, since it is parametrized by c
 
    can't instantiate the following to f c = Arg c e
    since Haskell does not have lambda abstraction
@@ -270,12 +404,12 @@ instance Decoration (Arg c) where
 instance HasRange a => HasRange (Arg c a) where
     getRange = getRange . unArg
 
+instance SetRange a => SetRange (Arg c a) where
+  setRange r = fmap $ setRange r
+
 instance (KillRange c, KillRange a) => KillRange (Arg c a) where
   killRange (Arg info a) = killRange2 Arg info a
 
-instance Sized a => Sized (Arg c a) where
-  size = size . unArg
-
 instance (Eq a, Eq c) => Eq (Arg c a) where
   Arg (ArgInfo h1 _ cs1) x1 == Arg (ArgInfo h2 _ cs2) x2 = (h1, cs1, x1) == (h2, cs2, x2)
 
@@ -285,11 +419,13 @@ instance (Show a, Show c) => Show (Arg c a) where
         showH Hidden     s = "{" ++ s ++ "}"
         showH NotHidden  s = "(" ++ s ++ ")"
         showH Instance   s = "{{" ++ s ++ "}}"
-        showR Irrelevant s = "." ++ s
-        showR NonStrict  s = "?" ++ s
-        showR Forced     s = "!" ++ s
-        showR UnusedArg  s = "k" ++ s -- constant
-        showR Relevant   s = "r" ++ s -- Andreas: I want to see it explicitly
+        showR r s = case r of
+          Irrelevant   -> "." ++ s
+          NonStrict    -> "?" ++ s
+          Forced Big   -> "!b" ++ s
+          Forced Small -> "!" ++ s
+          UnusedArg    -> "k" ++ s -- constant
+          Relevant     -> "r" ++ s -- Andreas: I want to see it explicitly
         showC cs         s = show cs ++ s
 
 instance LensHiding (Arg c e) where
@@ -392,9 +528,6 @@ instance HasRange a => HasRange (Dom c a) where
 instance (KillRange c, KillRange a) => KillRange (Dom c a) where
   killRange (Dom info a) = killRange2 Dom info a
 
-instance Sized a => Sized (Dom c a) where
-  size = size . unDom
-
 instance (Show a, Show c) => Show (Dom c a) where
   show = show . argFromDom
 
@@ -447,12 +580,12 @@ instance Decoration (Named name) where
 instance HasRange a => HasRange (Named name a) where
     getRange = getRange . namedThing
 
+instance SetRange a => SetRange (Named name a) where
+  setRange r = fmap $ setRange r
+
 instance (KillRange name, KillRange a) => KillRange (Named name a) where
   killRange (Named n a) = Named (killRange n) (killRange a)
 
-instance Sized a => Sized (Named name a) where
-  size = size . namedThing
-
 instance Show a => Show (Named_ a) where
     show (Named Nothing x)  = show x
     show (Named (Just n) x) = rawNameToString (rangedThing n) ++ " = " ++ show x
@@ -521,6 +654,17 @@ stringToRawName = id
 -- | String with range info.
 type RString = Ranged RawName
 
+---------------------------------------------------------------------------
+-- * Constructor pattern info
+---------------------------------------------------------------------------
+
+-- | Where does the 'ConP' of come from?
+data ConPOrigin
+  = ConPImplicit  -- ^ Expanded from an implicit pattern.
+  | ConPCon       -- ^ User wrote a constructor pattern.
+  | ConPRec       -- ^ User wrote a record pattern.
+  deriving (Typeable, Show, Eq, Ord, Enum, Bounded)
+
 ---------------------------------------------------------------------------
 -- * Infixity, access, abstract, etc.
 ---------------------------------------------------------------------------
@@ -553,7 +697,10 @@ type Arity  = Nat
 -- | The unique identifier of a name. Second argument is the top-level module
 --   identifier.
 data NameId = NameId Integer Integer
-    deriving (Eq, Ord, Typeable)
+    deriving (Eq, Ord, Typeable, Generic)
+
+instance KillRange NameId where
+  killRange = id
 
 instance Show NameId where
   show (NameId x i) = show x ++ "@" ++ show i
@@ -568,6 +715,24 @@ instance Hashable NameId where
   {-# INLINE hashWithSalt #-}
   hashWithSalt salt (NameId n m) = hashWithSalt salt (n, m)
 
+instance Arbitrary NameId where
+  arbitrary = elements [ NameId x y | x <- [-1, 1], y <- [-1, 1] ]
+
+instance CoArbitrary NameId
+
+-- | A meta variable identifier is just a natural number.
+--
+newtype MetaId = MetaId { metaId :: Nat }
+    deriving (Eq, Ord, Num, Real, Enum, Integral, Typeable)
+
+instance Pretty MetaId where
+  pretty (MetaId n) = text $ "_" ++ show n
+
+-- | Show non-record version of this newtype.
+instance Show MetaId where
+  showsPrec p (MetaId n) = showParen (p > 0) $
+    showString "MetaId " . shows n
+
 newtype Constr a = Constr a
 
 ---------------------------------------------------------------------------
@@ -603,4 +768,3 @@ data TerminationCheck m
 instance KillRange m => KillRange (TerminationCheck m) where
   killRange (TerminationMeasure _ m) = TerminationMeasure noRange (killRange m)
   killRange t                        = t
-
