@@ -1,10 +1,9 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE DeriveDataTypeable   #-}
-{-# LANGUAGE DeriveFoldable       #-}
-{-# LANGUAGE DeriveFunctor        #-}
-{-# LANGUAGE DeriveTraversable    #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP                #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable     #-}
+{-# LANGUAGE DeriveFunctor      #-}
+{-# LANGUAGE DeriveTraversable  #-}
+{-# LANGUAGE FlexibleInstances  #-}
 
 {-| The concrete syntax is a raw representation of the program text
     without any desugaring at all.  This is what the parser produces.
@@ -45,7 +44,7 @@ module Agda.Syntax.Concrete
   , ThingWithFixity(..)
   , topLevelModuleName
     -- * Pattern tools
-  , patternHead, patternNames
+  , patternNames, patternQNames
     -- * Lenses
   , mapLhsOriginalPattern
     -- * Concrete instances
@@ -62,6 +61,8 @@ import Data.Typeable (Typeable)
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable)
 import Data.List
+import Data.Set (Set)
+
 import Agda.Syntax.Position
 import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
 import qualified Agda.Syntax.Common as Common
@@ -70,6 +71,7 @@ import Agda.Syntax.Notation
 import Agda.Syntax.Literal
 
 import Agda.Syntax.Concrete.Name
+import qualified Agda.Syntax.Abstract.Name as A
 
 import Agda.Utils.Lens
 
@@ -101,7 +103,14 @@ data Expr
   | Underscore !Range (Maybe String)           -- ^ ex: @_@ or @_A_5@
   | RawApp !Range [Expr]                       -- ^ before parsing operators
   | App !Range Expr (NamedArg Expr)            -- ^ ex: @e e@, @e {e}@, or @e {x = e}@
-  | OpApp !Range QName [NamedArg (OpApp Expr)] -- ^ ex: @e + e@
+  | OpApp !Range QName (Set A.Name)
+          [NamedArg (OpApp Expr)]              -- ^ ex: @e + e@
+                                               -- The 'QName' is
+                                               -- possibly ambiguous,
+                                               -- but it must
+                                               -- correspond to one of
+                                               -- the names in the
+                                               -- set.
   | WithApp !Range Expr [Expr]                 -- ^ ex: @e | e1 | .. | en@
   | HiddenArg !Range (Named_ Expr)             -- ^ ex: @{e}@ or @{x=e}@
   | InstanceArg !Range (Named_ Expr)           -- ^ ex: @{{e}}@ or @{{x=e}}@
@@ -131,7 +140,7 @@ data Expr
   | Equal !Range Expr Expr                     -- ^ ex: @a = b@, used internally in the parser
   deriving (Typeable)
 
-instance NFData Expr
+instance NFData Expr where rnf x = seq x ()
 
 -- | Concrete patterns. No literals in patterns at the moment.
 data Pattern
@@ -139,7 +148,12 @@ data Pattern
   | QuoteP !Range                           -- ^ @quote@
   | AppP Pattern (NamedArg Pattern)         -- ^ @p p'@ or @p {x = p'}@
   | RawAppP !Range [Pattern]                -- ^ @p1..pn@ before parsing operators
-  | OpAppP !Range QName [NamedArg Pattern]  -- ^ eg: @p => p'@ for operator @_=>_@
+  | OpAppP !Range QName (Set A.Name)
+           [NamedArg Pattern]               -- ^ eg: @p => p'@ for operator @_=>_@
+                                            -- The 'QName' is possibly
+                                            -- ambiguous, but it must
+                                            -- correspond to one of
+                                            -- the names in the set.
   | HiddenP !Range (Named_ Pattern)         -- ^ @{p}@ or @{x = p}@
   | InstanceP !Range (Named_ Pattern)       -- ^ @{{p}}@ or @{{x = p}}@
   | ParenP !Range Pattern                   -- ^ @(p)@
@@ -150,7 +164,7 @@ data Pattern
   | LitP Literal                            -- ^ @0@, @1@, etc.
   deriving (Typeable)
 
-instance NFData Pattern
+instance NFData Pattern where rnf x = seq x ()
 
 -- | A lambda binding is either domain free or typed.
 type LamBinding = LamBinding' TypedBindings
@@ -162,11 +176,14 @@ data LamBinding' a
 
 -- | A sequence of typed bindings with hiding information. Appears in dependent
 --   function spaces, typed lambdas, and telescopes.
+--
+--   If the individual binding contains hiding information as well, the
+--   'Hiding' in @TypedBindings@ must be the unit 'NotHidden'.
 
 type TypedBindings = TypedBindings' TypedBinding
 
 data TypedBindings' a = TypedBindings !Range (Arg a)
-     -- ^ . @(xs : e)@ or @{xs : e}@
+     -- ^ . @(xs : e)@ or @{xs : e}@ or something like @(x {y} _ : e)@.
   deriving (Typeable, Functor, Foldable, Traversable)
 
 data BoundName = BName
@@ -187,7 +204,7 @@ mkBoundName x f = BName x x f
 type TypedBinding = TypedBinding' Expr
 
 data TypedBinding' e
-  = TBind !Range [BoundName] e  -- ^ Binding @(x1 ... xn : A)@.
+  = TBind !Range [WithHiding BoundName] e  -- ^ Binding @(x1 ... xn : A)@.
   | TLet  !Range [Declaration]  -- ^ Let binding @(let Ds)@ or @(open M args)@.
   deriving (Typeable, Functor, Foldable, Traversable)
 
@@ -239,15 +256,7 @@ data LHSCore
              }
   deriving (Typeable)
 
-instance NFData LHSCore
-
-{- TRASH
-lhsCoreToPattern :: LHSCore -> Pattern
-lhsCoreToPattern (LHSHead f args) = OpAppP (fuseRange f args) (unqualify f) args
-lhsCoreToPattern (LHSProj d ps1 lhscore ps2) = OpAppP (fuseRange d ps) (unqualify) ps
-  where p = lhsCoreToPattern lhscore
-        ps = ps1 ++ p : ps2
--}
+instance NFData LHSCore where rnf x = seq x ()
 
 type RHS = RHS' Expr
 data RHS' e
@@ -432,49 +441,52 @@ appView e = AppView e []
     Patterns
  --------------------------------------------------------------------------}
 
--- | Get the leftmost symbol in a pattern.
-patternHead :: Pattern -> Maybe Name
-patternHead p =
-  case p of
-    IdentP x               -> return $ unqualify x
-    AppP p p'              -> patternHead p
-    RawAppP _ []           -> __IMPOSSIBLE__
-    RawAppP _ (p:_)        -> patternHead p
-    OpAppP _ name ps       -> return $ unqualify name
-    HiddenP _ (namedPat)   -> patternHead (namedThing namedPat)
-    ParenP _ p             -> patternHead p
-    WildP _                -> Nothing
-    AbsurdP _              -> Nothing
-    AsP _ x p              -> patternHead p
-    DotP{}                 -> Nothing
-    LitP (LitQName _ x)    -> Nothing -- return $ unqualify x -- does not compile
-    LitP _                 -> Nothing
-    QuoteP _               -> Nothing
-    InstanceP _ (namedPat) -> patternHead (namedThing namedPat)
-
-
 -- | Get all the identifiers in a pattern in left-to-right order.
-patternNames :: Pattern -> [Name]
-patternNames p =
+patternQNames :: Pattern -> [QName]
+patternQNames p =
   case p of
-    IdentP x               -> [unqualify x]
-    AppP p p'              -> concatMap patternNames [p, namedArg p']
-    RawAppP _ ps           -> concatMap patternNames  ps
-    OpAppP _ name ps       -> unqualify name : concatMap (patternNames . namedArg) ps
-    HiddenP _ (namedPat)   -> patternNames (namedThing namedPat)
-    ParenP _ p             -> patternNames p
+    IdentP x               -> [x]
+    AppP p p'              -> concatMap patternQNames [p, namedArg p']
+    RawAppP _ ps           -> concatMap patternQNames  ps
+    OpAppP _ x _ ps        -> x : concatMap (patternQNames . namedArg) ps
+    HiddenP _ (namedPat)   -> patternQNames (namedThing namedPat)
+    ParenP _ p             -> patternQNames p
     WildP _                -> []
     AbsurdP _              -> []
-    AsP _ x p              -> patternNames p
+    AsP _ x p              -> patternQNames p
     DotP{}                 -> []
     LitP _                 -> []
     QuoteP _               -> []
-    InstanceP _ (namedPat) -> patternNames (namedThing namedPat)
+    InstanceP _ (namedPat) -> patternQNames (namedThing namedPat)
+
+-- | Get all the identifiers in a pattern in left-to-right order.
+patternNames :: Pattern -> [Name]
+patternNames = map unqualify . patternQNames
 
 {--------------------------------------------------------------------------
     Instances
  --------------------------------------------------------------------------}
 
+-- Lenses
+------------------------------------------------------------------------
+
+instance LensRelevance TypedBindings where
+  getRelevance   (TypedBindings _ b) = getRelevance b
+  mapRelevance f (TypedBindings r b) = TypedBindings r $ mapRelevance f b
+
+instance LensHiding TypedBindings where
+  getHiding   (TypedBindings _ b) = getHiding b
+  mapHiding f (TypedBindings r b) = TypedBindings r $ mapHiding f b
+
+instance LensHiding LamBinding where
+  getHiding   (DomainFree ai _) = getHiding ai
+  getHiding   (DomainFull a)    = getHiding a
+  mapHiding f (DomainFree ai x) = DomainFree (mapHiding f ai) x
+  mapHiding f (DomainFull a)    = DomainFull $ mapHiding f a
+
+-- HasRange instances
+------------------------------------------------------------------------
+
 instance HasRange e => HasRange (OpApp e) where
   getRange e = case e of
     Ordinary e -> getRange e
@@ -489,7 +501,7 @@ instance HasRange Expr where
       Underscore r _     -> r
       App r _ _          -> r
       RawApp r _         -> r
-      OpApp r _ _        -> r
+      OpApp r _ _ _      -> r
       WithApp r _ _      -> r
       Lam r _ _          -> r
       AbsurdLam r _      -> r
@@ -617,7 +629,7 @@ instance HasRange AsName where
 instance HasRange Pattern where
   getRange (IdentP x)         = getRange x
   getRange (AppP p q)         = fuseRange p q
-  getRange (OpAppP r _ _)     = r
+  getRange (OpAppP r _ _ _)   = r
   getRange (RawAppP r _)      = r
   getRange (ParenP r _)       = r
   getRange (WildP r)          = r
@@ -629,6 +641,30 @@ instance HasRange Pattern where
   getRange (InstanceP r _)    = r
   getRange (DotP r _)         = r
 
+-- SetRange instances
+------------------------------------------------------------------------
+
+instance SetRange TypedBindings where
+  setRange r (TypedBindings _ b) = TypedBindings r b
+
+instance SetRange Pattern where
+  setRange r (IdentP x)         = IdentP (setRange r x)
+  setRange r (AppP p q)         = AppP (setRange r p) (setRange r q)
+  setRange r (OpAppP _ x ns ps) = OpAppP r x ns ps
+  setRange r (RawAppP _ ps)     = RawAppP r ps
+  setRange r (ParenP _ p)       = ParenP r p
+  setRange r (WildP _)          = WildP r
+  setRange r (AsP _ x p)        = AsP r (setRange r x) p
+  setRange r (AbsurdP _)        = AbsurdP r
+  setRange r (LitP l)           = LitP (setRange r l)
+  setRange r (QuoteP _)         = QuoteP r
+  setRange r (HiddenP _ p)      = HiddenP r p
+  setRange r (InstanceP _ p)    = InstanceP r p
+  setRange r (DotP _ e)         = DotP r e
+
+-- KillRange instances
+------------------------------------------------------------------------
+
 instance KillRange AsName where
   killRange (AsName n _) = killRange1 (flip AsName noRange) n
 
@@ -666,7 +702,7 @@ instance KillRange Expr where
   killRange (Underscore _ n)     = Underscore noRange n
   killRange (RawApp _ e)         = killRange1 (RawApp noRange) e
   killRange (App _ e a)          = killRange2 (App noRange) e a
-  killRange (OpApp _ n o)        = killRange2 (OpApp noRange) n o
+  killRange (OpApp _ n ns o)     = killRange3 (OpApp noRange) n ns o
   killRange (WithApp _ e es)     = killRange2 (WithApp noRange) e es
   killRange (HiddenArg _ n)      = killRange1 (HiddenArg noRange) n
   killRange (InstanceArg _ n)    = killRange1 (InstanceArg noRange) n
@@ -720,19 +756,19 @@ instance KillRange e => KillRange (OpApp e) where
   killRange (Ordinary e)                = killRange1 Ordinary e
 
 instance KillRange Pattern where
-  killRange (IdentP q)      = killRange1 IdentP q
-  killRange (AppP p n)      = killRange2 AppP p n
-  killRange (RawAppP _ p)   = killRange1 (RawAppP noRange) p
-  killRange (OpAppP _ n p)  = killRange2 (OpAppP noRange) n p
-  killRange (HiddenP _ n)   = killRange1 (HiddenP noRange) n
-  killRange (InstanceP _ n) = killRange1 (InstanceP noRange) n
-  killRange (ParenP _ p)    = killRange1 (ParenP noRange) p
-  killRange (WildP _)       = WildP noRange
-  killRange (AbsurdP _)     = AbsurdP noRange
-  killRange (AsP _ n p)     = killRange2 (AsP noRange) n p
-  killRange (DotP _ e)      = killRange1 (DotP noRange) e
-  killRange (LitP l)        = killRange1 LitP l
-  killRange (QuoteP _)      = QuoteP noRange
+  killRange (IdentP q)        = killRange1 IdentP q
+  killRange (AppP p n)        = killRange2 AppP p n
+  killRange (RawAppP _ p)     = killRange1 (RawAppP noRange) p
+  killRange (OpAppP _ n ns p) = killRange3 (OpAppP noRange) n ns p
+  killRange (HiddenP _ n)     = killRange1 (HiddenP noRange) n
+  killRange (InstanceP _ n)   = killRange1 (InstanceP noRange) n
+  killRange (ParenP _ p)      = killRange1 (ParenP noRange) p
+  killRange (WildP _)         = WildP noRange
+  killRange (AbsurdP _)       = AbsurdP noRange
+  killRange (AsP _ n p)       = killRange2 (AsP noRange) n p
+  killRange (DotP _ e)        = killRange1 (DotP noRange) e
+  killRange (LitP l)          = killRange1 LitP l
+  killRange (QuoteP _)        = QuoteP noRange
 
 instance KillRange Pragma where
   killRange (OptionsPragma _ s)           = OptionsPragma noRange s
