@@ -1,8 +1,9 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE DeriveDataTypeable   #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE StandaloneDeriving #-}
 
 {-| Names in the concrete syntax are just strings (or lists of strings for
     qualified names).
@@ -17,6 +18,8 @@ import qualified Data.ByteString.Char8 as ByteString
 import Data.List
 import Data.Typeable (Typeable)
 
+import GHC.Generics (Generic)
+
 import System.FilePath
 
 import Test.QuickCheck
@@ -42,7 +45,7 @@ data Name
   | NoName !Range NameId    -- ^ @_@.
   deriving (Typeable)
 
-instance NFData Name
+instance NFData Name where rnf x = seq x ()
 
 instance Underscore Name where
   underscore = NoName noRange __IMPOSSIBLE__
@@ -55,7 +58,7 @@ instance Underscore Name where
 data NamePart
   = Hole       -- ^ @_@ part.
   | Id RawName  -- ^ Identifier part.
-  deriving (Typeable)
+  deriving (Typeable, Generic)
 
 -- | Define equality on @Name@ to ignore range so same names in different
 --   locations are equal.
@@ -118,7 +121,7 @@ newtype TopLevelModuleName
 ------------------------------------------------------------------------
 
 nameToRawName :: Name -> RawName
-nameToRawName = show
+nameToRawName = prettyShow
 
 nameParts :: Name -> [NamePart]
 nameParts (Name _ ps)  = ps
@@ -181,7 +184,7 @@ qnameParts (QName x)  = [x]
 -- name is assumed to represent a top-level module name.
 
 toTopLevelModuleName :: QName -> TopLevelModuleName
-toTopLevelModuleName = TopLevelModuleName . map show . qnameParts
+toTopLevelModuleName = TopLevelModuleName . map prettyShow . qnameParts
 
 -- | Turns a top-level module name into a file name with the given
 -- suffix.
@@ -244,37 +247,38 @@ instance IsNoName QName where
 -- * Showing names
 ------------------------------------------------------------------------
 
+-- deriving instance Show Name
+-- deriving instance Show NamePart
+-- deriving instance Show QName
+
 -- TODO: 'Show' should output Haskell-parseable representations.
 -- The following instances are deprecated, and Pretty should be used
 -- instead.  Later, simply derive Show for these types:
 
 instance Show Name where
-    show (Name _ xs)  = concatMap show xs
-    show (NoName _ _) = "_"
+  show = prettyShow
 
 instance Show NamePart where
-    show Hole   = "_"
-    show (Id s) = rawNameToString s
+  show = prettyShow
 
 instance Show QName where
-    show (Qual m x) = show m ++ "." ++ show x
-    show (QName x)  = show x
+  show = prettyShow
 
 ------------------------------------------------------------------------
 -- * Printing names
 ------------------------------------------------------------------------
 
 instance Pretty Name where
-    pretty (Name _ xs)  = hcat $ map pretty xs
-    pretty (NoName _ _) = text $ "_"
+  pretty (Name _ xs)  = hcat $ map pretty xs
+  pretty (NoName _ _) = text $ "_"
 
 instance Pretty NamePart where
-    pretty Hole   = text $ "_"
-    pretty (Id s) = text $ rawNameToString s
+  pretty Hole   = text $ "_"
+  pretty (Id s) = text $ rawNameToString s
 
 instance Pretty QName where
-    pretty (Qual m x) = pretty m <> pretty "." <> pretty x
-    pretty (QName x)  = pretty x
+  pretty (Qual m x) = pretty m <> pretty "." <> pretty x
+  pretty (QName x)  = pretty x
 
 instance Pretty TopLevelModuleName where
   pretty (TopLevelModuleName ms) = text $ intercalate "." ms
@@ -289,6 +293,27 @@ instance Arbitrary TopLevelModuleName where
 instance CoArbitrary TopLevelModuleName where
   coarbitrary (TopLevelModuleName m) = coarbitrary m
 
+instance Arbitrary Name where
+  arbitrary = oneof
+    [ Name   <$> arbitrary <*> parts
+    , NoName <$> arbitrary <*> arbitrary
+    ]
+    where
+    parts = do
+      parts         <- listOf1 (elements ["x", "y", "z"])
+      startWithHole <- arbitrary
+      endWithHole   <- arbitrary
+      return $
+        (if startWithHole then (Hole :)    else id) $
+        (if endWithHole   then (++ [Hole]) else id) $
+        intersperse Hole (map Id parts)
+
+instance CoArbitrary NamePart
+
+instance CoArbitrary Name where
+  coarbitrary (Name _ ns)  = variant 0 . coarbitrary ns
+  coarbitrary (NoName _ i) = variant 1 . coarbitrary i
+
 ------------------------------------------------------------------------
 -- * Range instances
 ------------------------------------------------------------------------
