@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP                 #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 {-| The parser doesn't know about operators and parses everything as normal
@@ -45,57 +45,54 @@ import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
 
 import Agda.TypeChecking.Monad.Base (typeError, TypeError(..), LHSOrPatSyn(..))
-import Agda.TypeChecking.Monad.Benchmark (billSub)
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import Agda.TypeChecking.Monad.State (getScope)
 import Agda.TypeChecking.Monad.Options
 
 import Agda.Utils.Either
 import Agda.Utils.ReadP
+#if MIN_VERSION_base(4,8,0)
+import Agda.Utils.List hiding ( uncons )
+#else
 import Agda.Utils.List
+#endif
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+---------------------------------------------------------------------------
+-- * Billing
+---------------------------------------------------------------------------
+
+-- | Bills the operator parser.
+
+billToParser :: ScopeM a -> ScopeM a
+billToParser = Bench.billTo [Bench.Parsing, Bench.Operators]
+
 ---------------------------------------------------------------------------
 -- * Building the parser
 ---------------------------------------------------------------------------
 
-partsInScope :: FlatScope -> ScopeM (Set QName)
-partsInScope flat = do
-    (names, ops) <- localNames flat
-    let xs = concatMap parts names ++ concatMap notationNames ops
-    return $ Set.fromList xs
-    where
-        qual xs x = foldr Qual (QName x) xs
-        parts q = parts' (init $ qnameParts q) (unqualify q)
-        parts' ms (NoName _ _)   = []
-        parts' ms x@(Name _ [_]) = [qual ms x]
-                                   -- The first part should be qualified, but not the rest
-        parts' ms x@(Name _ xs)  = qual ms x : qual ms (Name noRange [first]) : [ QName $ Name noRange [i] | i <- iparts ]
-          where
-            first:iparts = [ i | i@(Id {}) <- xs ]
-
 type FlatScope = Map QName [AbstractName]
 
--- | Compute all unqualified defined names in scope and their fixities.
---   Note that overloaded names (constructors) can have several fixities.
---   Then we 'chooseFixity'. (See issue 1194.)
-getDefinedNames :: [KindOfName] -> FlatScope -> [(QName, Fixity')]
+-- | Compute all defined names in scope and their fixities/notations.
+-- Note that overloaded names (constructors) can have several
+-- fixities/notations. Then we 'mergeNotations'. (See issue 1194.)
+getDefinedNames :: [KindOfName] -> FlatScope -> [[NewNotation]]
 getDefinedNames kinds names =
-  [ (x, chooseFixity fixs)
+  [ mergeNotations $
+      map (\d -> namesToNotation x (A.qnameName $ anameName d)) ds
   | (x, ds) <- Map.toList names
   , any ((`elem` kinds) . anameKind) ds
-  , let fixs = map (A.nameFixity . A.qnameName . anameName) ds
-  , not (null fixs)
+  , not (null ds)
   -- Andreas, 2013-03-21 see Issue 822
   -- Names can have different kinds, i.e., 'defined' and 'constructor'.
   -- We need to consider all names that have *any* matching kind,
   -- not only those whose first appearing kind is matching.
- ]
+  ]
 
--- | Compute all names (first component) and operators (second component) in
---   scope.
+-- | Compute all names (first component) and operators/notations
+-- (second component) in scope.
 localNames :: FlatScope -> ScopeM ([QName], [NewNotation])
 localNames flat = do
   let defs = getDefinedNames allKindsOfNames flat
@@ -106,17 +103,16 @@ localNames flat = do
     , "defs  = " ++ show defs
     , "locals= " ++ show locals
     ]
-  return $ split $ uniqOn fst $ map localOp locals ++ defs
+  let localNots  = map localOp locals
+      localNames = Set.fromList $ map notaName localNots
+      otherNots  = filter (\n -> not (Set.member (notaName n) localNames))
+                          (concat defs)
+  return $ split $ localNots ++ otherNots
   where
-    localOp (x, y) = (QName x, A.nameFixity y)
-    split ops = partitionEithers $ concatMap opOrNot ops
-
-    opOrNot (q, Fixity' fx syn) = Left q : map Right (notaFromName ++ nota)
-      where
-        notaFromName = case unqualify q of
-          Name _ [_] -> []
-          x          -> [NewNotation q fx $ syntaxOf x]
-        nota = if null syn then [] else [NewNotation q fx syn]
+    localOp (x, y) = namesToNotation (QName x) y
+    split ops      = partitionEithers $ concatMap opOrNot ops
+    opOrNot n      = Left (notaName n) :
+                     if null (notation n) then [] else [Right n]
 
 -- | Data structure filled in by @buildParsers@.
 --   The top-level parser @pTop@ is of primary interest,
@@ -167,10 +163,10 @@ buildParsers r flat use = do
       [ "names = " ++ show names
       , "ops   = " ++ show ops
       , "cons  = " ++ show cons ]
-    let conparts   = Set.fromList $ concatMap notationNames $ map oldToNewNotation cons
-        opsparts   = Set.fromList $ concatMap notationNames $ ops
+    let conparts   = Set.fromList $ concatMap notationNames $ concat cons
+        opsparts   = Set.fromList $ concatMap notationNames ops
         allParts   = Set.union conparts opsparts
-        connames   = Set.fromList $ map fst cons
+        connames   = Set.fromList $ map (notaName . head) cons
         (non, fix) = partition nonfix ops
         set        = Set.fromList names
         isAtom   x = case use of
@@ -192,23 +188,16 @@ buildParsers r flat use = do
         level :: NewNotation -> Integer
         level = fixityLevel . notaFixity
 
-        isinfixl, isinfixr, isinfix, nonfix, isprefix, ispostfix :: NewNotation -> Bool
+        nonfix, isprefix, ispostfix :: NewNotation -> Bool
+        nonfix    = (== NonfixNotation)  . notationKind . notation
+        isprefix  = (== PrefixNotation)  . notationKind . notation
+        ispostfix = (== PostfixNotation) . notationKind . notation
 
-        isinfixl (NewNotation _ (LeftAssoc _ _) syn)  = isInfix syn
-        isinfixl _ = False
-
-        isinfixr (NewNotation _ (RightAssoc _ _) syn) = isInfix syn
-        isinfixr _ = False
-
-        isinfix (NewNotation _ (NonAssoc _ _) syn)    = isInfix syn
-        isinfix _ = False
-
-        nonfix    (NewNotation _ _ syn) = notationKind syn == NonfixNotation
-        isprefix  (NewNotation _ _ syn) = notationKind syn == PrefixNotation
-        ispostfix (NewNotation _ _ syn) = notationKind syn == PostfixNotation
-
-        isInfix :: Notation -> Bool
-        isInfix syn = notationKind syn == InfixNotation
+        isinfix :: Associativity -> NewNotation -> Bool
+        isinfix ass syn =
+          notationKind (notation syn) == InfixNotation
+            &&
+          fixityAssoc (notaFixity syn) == ass
 
         -- | Group operators by precedence level
         order :: [NewNotation] -> [[NewNotation]]
@@ -221,9 +210,9 @@ buildParsers r flat use = do
             []      -> [id]
             fs      -> fs
             where
-                inlfx   = fixP infixlP  isinfixl
-                inrfx   = fixP infixrP  isinfixr
-                infx    = fixP infixP   isinfix
+                inlfx   = fixP infixlP  (isinfix LeftAssoc)
+                inrfx   = fixP infixrP  (isinfix RightAssoc)
+                infx    = fixP infixP   (isinfix NonAssoc)
                 prefx   = fixP prefixP  isprefix
                 postfx  = fixP postfixP ispostfix
 
@@ -233,97 +222,21 @@ buildParsers r flat use = do
                         []  -> []
                         ops -> [ f $ choice $ map (opP p0) ops ]
 
-
----------------------------------------------------------------------------
--- * Expression instances
----------------------------------------------------------------------------
-
-instance IsExpr Expr where
-    exprView e = case e of
-        Ident x         -> LocalV x
-        App _ e1 e2     -> AppV e1 e2
-        OpApp r d es    -> OpAppV d es
-        HiddenArg _ e   -> HiddenArgV e
-        InstanceArg _ e -> InstanceArgV e
-        Paren _ e       -> ParenV e
-        Lam _ bs    e   -> LamV bs e
-        Underscore{}    -> WildV e
-        _               -> OtherV e
-    unExprView e = case e of
-        LocalV x      -> Ident x
-        AppV e1 e2    -> App (fuseRange e1 e2) e1 e2
-        OpAppV d es   -> OpApp (fuseRange d es) d es
-        HiddenArgV e  -> HiddenArg (getRange e) e
-        InstanceArgV e -> InstanceArg (getRange e) e
-        ParenV e      -> Paren (getRange e) e
-        LamV bs e     -> Lam (fuseRange bs e) bs e
-        WildV e       -> e
-        OtherV e      -> e
-
-
-instance IsExpr Pattern where
-    exprView e = case e of
-        IdentP x      -> LocalV x
-        AppP e1 e2    -> AppV e1 e2
-        OpAppP r d es -> OpAppV d ((map . fmap . fmap) Ordinary es)
-        HiddenP _ e   -> HiddenArgV e
-        InstanceP _ e -> InstanceArgV e
-        ParenP _ e    -> ParenV e
-        WildP{}       -> WildV e
-        _             -> OtherV e
-    unExprView e = case e of
-        LocalV x       -> IdentP x
-        AppV e1 e2     -> AppP e1 e2
-        OpAppV d es    -> let ess :: [NamedArg Pattern]
-                              ess = (map . fmap . fmap) (fromOrdinary __IMPOSSIBLE__) es
-                          in OpAppP (fuseRange d es) d ess
-        HiddenArgV e   -> HiddenP (getRange e) e
-        InstanceArgV e -> InstanceP (getRange e) e
-        ParenV e       -> ParenP (getRange e) e
-        LamV _ _       -> __IMPOSSIBLE__
-        WildV e        -> e
-        OtherV e       -> e
-
-{- TRASH
-instance IsExpr LHSCore where
-    exprView e = case e of
-        LHSHead f ps -> foldl AppV (LocalV f) $ map exprView ps
-        LHSProj d ps1 e ps2 -> foldl AppV (LocalV d) $
-          map exprView ps1 ++ exprView e : map exprView ps2
-    unExprView e = LHSHead f ps
-      where p :: Pattern
-            p = unExprView
-            (f, ps) = lhsArgs p
--}
-
 ---------------------------------------------------------------------------
 -- * Helpers for pattern and lhs parsing
 ---------------------------------------------------------------------------
 
--- Andreas, 2011-11-24 moved here from ConcreteToAbstract
-lhsArgs :: Pattern -> (Name, [NamedArg Pattern])
-lhsArgs p = case lhsArgs' p of
-              Just (x, args) -> (x, args)
-              Nothing        -> __IMPOSSIBLE__
-
--- | @lhsArgs' p@ splits a lhs @f ps@, given as a pattern @p@,
---   into @(f, ps)@.
-lhsArgs' :: Pattern -> Maybe (Name, [NamedArg Pattern])
-lhsArgs' p = case patternAppView p of
-    Common.Arg _ (Named _ (IdentP (QName x))) : ps -> Just (x, ps)
-    _                                              -> Nothing
-
 -- | View a pattern @p@ as a list @p0 .. pn@ where @p0@ is the identifier
 --   (in most cases a constructor).
 --
 --  Pattern needs to be parsed already (operators resolved).
 patternAppView :: Pattern -> [NamedArg Pattern]
 patternAppView p = case p of
-    AppP p arg    -> patternAppView p ++ [arg]
-    OpAppP _ x ps -> defaultNamedArg (IdentP x) : ps
-    ParenP _ p    -> patternAppView p
-    RawAppP _ _   -> __IMPOSSIBLE__
-    _             -> [ defaultNamedArg p ]
+    AppP p arg      -> patternAppView p ++ [arg]
+    OpAppP _ x _ ps -> defaultNamedArg (IdentP x) : ps
+    ParenP _ p      -> patternAppView p
+    RawAppP _ _     -> __IMPOSSIBLE__
+    _               -> [ defaultNamedArg p ]
 
 
 ---------------------------------------------------------------------------
@@ -336,7 +249,7 @@ parsePat prs p = case p of
     AppP p (Common.Arg info q) ->
         fullParen' <$> (AppP <$> parsePat prs p <*> (Common.Arg info <$> traverse (parsePat prs) q))
     RawAppP _ ps     -> fullParen' <$> (parsePat prs =<< parse prs ps)
-    OpAppP r d ps    -> fullParen' . OpAppP r d <$> (mapM . traverse . traverse) (parsePat prs) ps
+    OpAppP r d ns ps -> fullParen' . OpAppP r d ns <$> (mapM . traverse . traverse) (parsePat prs) ps
     HiddenP _ _      -> fail "bad hidden argument"
     InstanceP _ _    -> fail "bad instance argument"
     AsP r x p        -> AsP r x <$> parsePat prs p
@@ -402,7 +315,8 @@ parseLHS' lhsOrPatSyn top p = do
         rs        -> typeError $ AmbiguousParseForLHS lhsOrPatSyn p $
                        map (fullParen . fst) rs
     where
-        getNames kinds flat = map fst $ getDefinedNames kinds flat
+        getNames kinds flat =
+          map (notaName . head) $ getDefinedNames kinds flat
 
         -- validPattern returns an empty or singleton list (morally a Maybe)
         validPattern :: PatternCheckConfig -> Pattern -> [(Pattern, ParseLHS)]
@@ -455,24 +369,14 @@ classifyPattern conf p =
 
 
 -- | Parses a left-hand side, and makes sure that it defined the expected name.
---   TODO: check the arities of constructors. There is a possible ambiguity with
---   postfix constructors:
---      Assume _ * is a constructor. Then 'true *' can be parsed as either the
---      intended _* applied to true, or as true applied to a variable *. If we
---      check arities this problem won't appear.
 parseLHS :: Name -> Pattern -> ScopeM LHSCore
-parseLHS top p = do
+parseLHS top p = billToParser $ do
   res <- parseLHS' IsLHS (Just top) p
   case res of
     Right (f, lhs) -> return lhs
     _ -> typeError $ NoParseForLHS IsLHS p
 
 -- | Parses a pattern.
---   TODO: check the arities of constructors. There is a possible ambiguity with
---   postfix constructors:
---      Assume _ * is a constructor. Then 'true *' can be parsed as either the
---      intended _* applied to true, or as true applied to a variable *. If we
---      check arities this problem won't appear.
 parsePattern :: Pattern -> ScopeM Pattern
 parsePattern = parsePatternOrSyn IsLHS
 
@@ -480,7 +384,7 @@ parsePatternSyn :: Pattern -> ScopeM Pattern
 parsePatternSyn = parsePatternOrSyn IsPatSyn
 
 parsePatternOrSyn :: LHSOrPatSyn -> Pattern -> ScopeM Pattern
-parsePatternOrSyn lhsOrPatSyn p = do
+parsePatternOrSyn lhsOrPatSyn p = billToParser $ do
   res <- parseLHS' lhsOrPatSyn Nothing p
   case res of
     Left p -> return p
@@ -502,30 +406,13 @@ validConPattern cons p = case appView p of
 appView :: Pattern -> [Pattern]
 appView p = case p of
     AppP p a         -> appView p ++ [namedArg a]
-    OpAppP _ op ps   -> IdentP op : map namedArg ps
+    OpAppP _ op _ ps -> IdentP op : map namedArg ps
     ParenP _ p       -> appView p
     RawAppP _ _      -> __IMPOSSIBLE__
     HiddenP _ _      -> __IMPOSSIBLE__
     InstanceP _ _    -> __IMPOSSIBLE__
     _                -> [p]
 
--- | Collect all names in a pattern into a list of qualified names.
-patternQNames :: Pattern -> [QName]
-patternQNames p = case p of
-  RawAppP _ ps     -> concatMap patternQNames ps
-  IdentP q         -> [q]
-  ParenP _ p       -> patternQNames p
-  HiddenP _ p      -> patternQNames (namedThing p)
-  InstanceP _ p    -> patternQNames (namedThing p)
-  OpAppP r d ps    -> __IMPOSSIBLE__
-  AppP{}           -> __IMPOSSIBLE__
-  AsP r x p        -> patternQNames p
-  AbsurdP{}        -> []
-  WildP{}          -> []
-  DotP{}           -> []
-  LitP{}           -> []
-  QuoteP{}         -> []
-
 -- | Return all qualifiers occuring in a list of 'QName's.
 --   Each qualifier is returned as a list of names, e.g.
 --   for @Data.Nat._+_@ we return the list @[Data,Nat]@.
@@ -536,25 +423,16 @@ qualifierModules qs =
 -- | Parse a list of expressions into an application.
 parseApplication :: [Expr] -> ScopeM Expr
 parseApplication [e] = return e
-parseApplication es = do
+parseApplication es  = billToParser $ do
     -- Build the parser
     let ms = qualifierModules [ q | Ident q <- es ]
     flat <- flattenScope ms <$> getScope
-    -- Andreas, 2014-04-27 Time for building the parser is negligible
-    p <- -- billSub [Bench.Parsing, Bench.Operators, Bench.BuildParser] $
-      buildParsers (getRange es) flat UseBoundNames
+    p <- buildParsers (getRange es) flat UseBoundNames
 
     -- Parse
     case force $ parse (pTop p) es of
         [e] -> return e
-        [] -> do
-          -- When the parser fails and a name is not in scope, it is more
-          -- useful to say that to the user rather than just "failed".
-          inScope <- partsInScope flat
-          case [ x | Ident x <- es, not (Set.member x inScope) ] of
-              [] -> typeError $ NoParseForApplication es
-              xs -> typeError $ NotInScope xs
-
+        []  -> typeError $ NoParseForApplication es
         es' -> typeError $ AmbiguousParseForApplication es $ map fullParen es'
 
 parseModuleIdentifier :: Expr -> ScopeM QName
@@ -562,7 +440,7 @@ parseModuleIdentifier (Ident m) = return m
 parseModuleIdentifier e = typeError $ NotAModuleExpr e
 
 parseRawModuleApplication :: [Expr] -> ScopeM (QName, [NamedArg Expr])
-parseRawModuleApplication es = do
+parseRawModuleApplication es = billToParser $ do
     let e : es_args = es
     m <- parseModuleIdentifier e
 
@@ -574,12 +452,7 @@ parseRawModuleApplication es = do
     -- Parse
     case {-force $-} parse (pArgs p) es_args of -- TODO: not sure about forcing
         [as] -> return (m, as)
-        [] -> do
-          inScope <- partsInScope flat
-          case [ x | Ident x <- es_args, not (Set.member x inScope) ] of
-              [] -> typeError $ NoParseForApplication es
-              xs -> typeError $ NotInScope xs
-
+        []   -> typeError $ NoParseForApplication es
         ass -> do
           let f = fullParen . foldl (App noRange) (Ident m)
           typeError $ AmbiguousParseForApplication es
@@ -619,7 +492,7 @@ fullParen' e = case exprView e of
                 Hidden    -> e2
                 Instance  -> e2
                 NotHidden -> fullParen' <$> e2
-    OpAppV x es -> par $ unExprView $ OpAppV x $ (map . fmap . fmap . fmap) fullParen' es
+    OpAppV x ns es -> par $ unExprView $ OpAppV x ns $ (map . fmap . fmap . fmap) fullParen' es
     LamV bs e -> par $ unExprView $ LamV bs (fullParen e)
     where
         par = unExprView . ParenV
