@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP                 #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.Syntax.Concrete.Operators.Parser where
@@ -6,15 +6,20 @@ module Agda.Syntax.Concrete.Operators.Parser where
 import Control.Exception (throw)
 
 import Data.Maybe
+import Data.Set (Set)
 
 import Agda.Syntax.Position
+import qualified Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
 import Agda.Syntax.Concrete
+
 import Agda.TypeChecking.Monad.Base (TCErr(Exception))
-import Agda.Utils.ReadP
+
 import Agda.Utils.Monad
+import Agda.Utils.Pretty
+import Agda.Utils.ReadP
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -24,7 +29,9 @@ data ExprView e
     | WildV e
     | OtherV e
     | AppV e (NamedArg e)
-    | OpAppV QName [NamedArg (OpApp e)]
+    | OpAppV QName (Set A.Name) [NamedArg (OpApp e)]
+      -- ^ The 'QName' is possibly ambiguous, but it must correspond
+      -- to one of the names in the set.
     | HiddenArgV (Named_ e)
     | InstanceArgV (Named_ e)
     | LamV [LamBinding] e
@@ -38,6 +45,51 @@ class HasRange e => IsExpr e where
 instance IsExpr e => HasRange (ExprView e) where
   getRange = getRange . unExprView
 
+instance IsExpr Expr where
+    exprView e = case e of
+        Ident x         -> LocalV x
+        App _ e1 e2     -> AppV e1 e2
+        OpApp r d ns es -> OpAppV d ns es
+        HiddenArg _ e   -> HiddenArgV e
+        InstanceArg _ e -> InstanceArgV e
+        Paren _ e       -> ParenV e
+        Lam _ bs    e   -> LamV bs e
+        Underscore{}    -> WildV e
+        _               -> OtherV e
+    unExprView e = case e of
+        LocalV x       -> Ident x
+        AppV e1 e2     -> App (fuseRange e1 e2) e1 e2
+        OpAppV d ns es -> OpApp (fuseRange d es) d ns es
+        HiddenArgV e   -> HiddenArg (getRange e) e
+        InstanceArgV e -> InstanceArg (getRange e) e
+        ParenV e       -> Paren (getRange e) e
+        LamV bs e      -> Lam (fuseRange bs e) bs e
+        WildV e        -> e
+        OtherV e       -> e
+
+instance IsExpr Pattern where
+    exprView e = case e of
+        IdentP x         -> LocalV x
+        AppP e1 e2       -> AppV e1 e2
+        OpAppP r d ns es -> OpAppV d ns ((map . fmap . fmap) Ordinary es)
+        HiddenP _ e      -> HiddenArgV e
+        InstanceP _ e    -> InstanceArgV e
+        ParenP _ e       -> ParenV e
+        WildP{}          -> WildV e
+        _                -> OtherV e
+    unExprView e = case e of
+        LocalV x       -> IdentP x
+        AppV e1 e2     -> AppP e1 e2
+        OpAppV d ns es -> let ess :: [NamedArg Pattern]
+                              ess = (map . fmap . fmap) (fromOrdinary __IMPOSSIBLE__) es
+                          in OpAppP (fuseRange d es) d ns ess
+        HiddenArgV e   -> HiddenP (getRange e) e
+        InstanceArgV e -> InstanceP (getRange e) e
+        ParenV e       -> ParenP (getRange e) e
+        LamV _ _       -> __IMPOSSIBLE__
+        WildV e        -> e
+        OtherV e       -> e
+
 ---------------------------------------------------------------------------
 -- * Parser combinators
 ---------------------------------------------------------------------------
@@ -79,8 +131,9 @@ postop middleP = do
 -- Note: it would be better to take the decision of "postprocessing" at the same
 -- place as where the holes are discarded, however that would require a dependently
 -- typed function (or duplicated code)
-opP :: IsExpr e => ReadP e e -> NewNotation -> ReadP e (NewNotation,Range,[e])
-opP p nsyn@(NewNotation q _ syn) = do
+opP :: IsExpr e =>
+       ReadP e e -> NewNotation -> ReadP e (NewNotation,Range,[e])
+opP p nsyn@(NewNotation q _ _ syn) = do
   (range,es) <- worker (init $ qnameParts q) $ removeExternalHoles syn
   return (nsyn,range,es)
  where worker ms [IdPart x] = do r <- partP ms x; return (r,[])
@@ -97,7 +150,8 @@ opP p nsyn@(NewNotation q _ syn) = do
 -- | Given a name with a syntax spec, and a list of parsed expressions
 -- fitting it, rebuild the expression.
 rebuild :: forall e. IsExpr e => NewNotation -> Range -> [e] -> e
-rebuild (NewNotation name _ syn) r es = unExprView $ OpAppV (setRange r name) exprs
+rebuild (NewNotation name names _ syn) r es =
+  unExprView $ OpAppV (setRange r name) names exprs
   where
     exprs = map findExprFor [0..lastHole]
     filledHoles = zip es (filter isAHole syn)
@@ -118,7 +172,8 @@ rebuildBinding :: IsExpr e => ExprView e -> LamBinding
 rebuildBinding (LocalV (QName name)) = DomainFree defaultArgInfo $ mkBoundName_ name
 rebuildBinding (WildV e) =
   DomainFree defaultArgInfo $ mkBoundName_ $ Name noRange [Hole]
-rebuildBinding e = throw $ Exception (getRange e) "Expected variable name in binding position"
+rebuildBinding e = throw $ Exception (getRange e) $
+  text "Expected variable name in binding position"
 
 -- | Parse using the appropriate fixity, given a parser parsing the
 -- operator part, the name of the operator, and a parser of
