@@ -1,24 +1,32 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE DeriveFoldable     #-}
-{-# LANGUAGE DeriveFunctor      #-}
-{-# LANGUAGE DeriveTraversable  #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
 
 {-| Definitions for fixity, precedence levels, and declared syntax.
 -}
 module Agda.Syntax.Fixity where
 
 import Data.Foldable
+import Data.Function
 import Data.List as List
+import Data.Set (Set)
+import qualified Data.Set as Set
 import Data.Traversable
 import Data.Typeable (Typeable)
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
+import qualified Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Concrete.Name
 import Agda.Syntax.Notation
 
 import Agda.Utils.List
 
+#include "undefined.h"
+import Agda.Utils.Impossible
+
 -- * Notation coupled with 'Fixity'
 
 -- | The notation is handled as the fixity in the renamer.
@@ -36,27 +44,34 @@ data ThingWithFixity x = ThingWithFixity x Fixity'
 -- | All the notation information related to a name.
 data NewNotation = NewNotation
   { notaName   :: QName
-    -- ^ The concrete name the syntax or fixity belongs to.
+  , notaNames  :: Set A.Name
+    -- ^ The names the syntax and/or fixity belong to.
+    --
+    -- Invariant: The set is non-empty. Every name in the list matches
+    -- 'notaName'.
   , notaFixity :: Fixity
-    -- ^ Associativity and precedence (fixity) of the name.
+    -- ^ Associativity and precedence (fixity) of the names.
   , notation   :: Notation
-    -- ^ Syntax associated with the name.
+    -- ^ Syntax associated with the names.
   } deriving (Typeable, Show)
 
--- | If an operator has no specific notation, recover it from its name.
-oldToNewNotation :: (QName, Fixity') -> NewNotation
-oldToNewNotation (name, Fixity' f syn) = NewNotation
-  { notaName   = name
+-- | If an operator has no specific notation, then it is computed from
+-- its name.
+namesToNotation :: QName -> A.Name -> NewNotation
+namesToNotation q n = NewNotation
+  { notaName   = q
+  , notaNames  = Set.singleton n
   , notaFixity = f
-  , notation   = if null syn then syntaxOf $ unqualify name else syn
+  , notation   = if null syn then syntaxOf $ unqualify q else syn
   }
+  where Fixity' f syn = A.nameFixity n
 
 -- | Return the 'IdPart's of a notation, the first part qualified,
 --   the other parts unqualified.
 --   This allows for qualified use of operators, e.g.,
 --   @M.for x ∈ xs return e@, or @x ℕ.+ y@.
 notationNames :: NewNotation -> [QName]
-notationNames (NewNotation q _ parts) =
+notationNames (NewNotation q _ _ parts) =
   zipWith ($) (reQualify : repeat QName) [Name noRange [Id x] | IdPart x <- parts ]
   where
     -- The qualification of @q@.
@@ -84,44 +99,62 @@ syntaxOf (Name _ xs)  = mkSyn 0 xs
 defaultFixity' :: Fixity'
 defaultFixity' = Fixity' defaultFixity defaultNotation
 
--- | Removes copies of @defaultFixity'@ from a list of fixities.
---   Never returns an empty list, though, rather a singleton list
---   consisting of @defaultFixity'@.
-interestingFixities :: [Fixity'] -> [Fixity']
-interestingFixities fixs = if null fixs' then [defaultFixity'] else fixs'
-  where fixs' = filter (not . (== defaultFixity')) fixs
+noFixity' :: Fixity'
+noFixity' = Fixity' noFixity noNotation
 
--- | If different interesting fixities are available for the same symbol,
---   we take none of them.
-chooseFixity :: [Fixity'] -> Fixity'
-chooseFixity fixs = if allEqual fixs' then head fixs' else defaultFixity'
-  where fixs' = interestingFixities fixs
+-- | Merges all 'NewNotation's that have the same notation.
+--
+-- If all 'NewNotation's with a given notation have the same fixity,
+-- then this fixity is preserved, and otherwise it is replaced by
+-- 'defaultFixity'.
+--
+-- Precondition: No 'A.Name' may occur in more than one list element.
+-- Every 'NewNotation' must have the same 'notaName'.
+--
+-- Postcondition: No 'A.Name' occurs in more than one list element.
+mergeNotations :: [NewNotation] -> [NewNotation]
+mergeNotations = map (merge . fixFixities) . groupOn notation
+  where
+  fixFixities ns
+    | allEqual (map notaFixity ns) = ns
+    | otherwise                    =
+        map (\n -> n { notaFixity = defaultFixity }) ns
+
+  merge :: [NewNotation] -> NewNotation
+  merge []         = __IMPOSSIBLE__
+  merge ns@(n : _) = n { notaNames = Set.unions $ map notaNames ns }
 
 -- * Fixity
 
+-- | Associativity.
+
+data Associativity = NonAssoc | LeftAssoc | RightAssoc
+   deriving (Eq, Ord, Show, Typeable)
+
 -- | Fixity of operators.
 
-data Fixity
-  = LeftAssoc  { fixityRange :: Range, fixityLevel :: Integer }
-  | RightAssoc { fixityRange :: Range, fixityLevel :: Integer }
-  | NonAssoc   { fixityRange :: Range, fixityLevel :: Integer }
+data Fixity =
+  Fixity { fixityRange :: Range
+         , fixityLevel :: Integer
+         , fixityAssoc :: Associativity
+         }
   deriving (Typeable, Show)
 
 instance Eq Fixity where
-    LeftAssoc  _ n == LeftAssoc  _ m = n == m
-    RightAssoc _ n == RightAssoc _ m = n == m
-    NonAssoc   _ n == NonAssoc   _ m = n == m
-    _              == _              = False
+  f1 == f2 = compare f1 f2 == EQ
+
+instance Ord Fixity where
+  compare = compare `on` (\f -> (fixityLevel f, fixityAssoc f))
 
 -- For @instance Pretty Fixity@, see Agda.Syntax.Concrete.Pretty
 
 -- | The default fixity. Currently defined to be @'NonAssoc' 20@.
 defaultFixity :: Fixity
-defaultFixity = NonAssoc noRange 20
+defaultFixity = Fixity noRange 20 NonAssoc
 
 -- | Hack used for @syntax@ facility.
 noFixity :: Fixity
-noFixity = NonAssoc noRange (negate 666)
+noFixity = Fixity noRange (negate 666) NonAssoc
   -- Ts,ts,ts, why the number of the beast?  Revelation 13, 18
   --
   -- It's not the number of the beast, it's the negation of the
@@ -150,10 +183,10 @@ hiddenArgumentCtx Instance  = TopCtx
 -- | Do we need to bracket an operator application of the given fixity
 --   in a context with the given precedence.
 opBrackets :: Fixity -> Precedence -> Bool
-opBrackets (LeftAssoc _ n1)
-           (LeftOperandCtx   (LeftAssoc   _ n2)) | n1 >= n2       = False
-opBrackets (RightAssoc _ n1)
-           (RightOperandCtx  (RightAssoc  _ n2)) | n1 >= n2       = False
+opBrackets (Fixity _ n1 LeftAssoc)
+           (LeftOperandCtx   (Fixity _ n2 LeftAssoc))  | n1 >= n2       = False
+opBrackets (Fixity _ n1 RightAssoc)
+           (RightOperandCtx  (Fixity _ n2 RightAssoc)) | n1 >= n2       = False
 opBrackets f1
            (LeftOperandCtx  f2) | fixityLevel f1 > fixityLevel f2 = False
 opBrackets f1
@@ -204,9 +237,16 @@ instance HasRange Fixity where
   getRange = fixityRange
 
 instance KillRange Fixity where
-  killRange (LeftAssoc  _ n) = LeftAssoc  noRange n
-  killRange (RightAssoc _ n) = RightAssoc noRange n
-  killRange (NonAssoc   _ n) = NonAssoc   noRange n
+  killRange f = f { fixityRange = noRange }
 
 instance KillRange Fixity' where
-  killRange (Fixity' f n) = killRange1 (flip Fixity' n) f
+  killRange (Fixity' f n) = killRange2 Fixity' f n
+
+instance KillRange x => KillRange (ThingWithFixity x) where
+  killRange (ThingWithFixity c f) = ThingWithFixity (killRange c) f
+
+------------------------------------------------------------------------
+-- * Printing
+------------------------------------------------------------------------
+
+-- deriving instance Show Fixity'
