@@ -1,6 +1,7 @@
-{-# LANGUAGE DeriveDataTypeable   #-}
-{-# LANGUAGE FlexibleContexts     #-}
-{-# LANGUAGE StandaloneDeriving   #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE UndecidableInstances #-}
 
 {-| An info object contains additional information about a piece of abstract
@@ -11,6 +12,8 @@
 
 module Agda.Syntax.Info where
 
+import Prelude hiding (null)
+
 import Data.Typeable (Typeable)
 
 import qualified Agda.Syntax.Concrete.Name as C
@@ -20,6 +23,9 @@ import Agda.Syntax.Concrete
 import Agda.Syntax.Fixity
 import Agda.Syntax.Scope.Base (ScopeInfo, emptyScopeInfo)
 
+import Agda.Utils.Function
+import Agda.Utils.Null
+
 {--------------------------------------------------------------------------
     Meta information
  --------------------------------------------------------------------------}
@@ -27,7 +33,7 @@ import Agda.Syntax.Scope.Base (ScopeInfo, emptyScopeInfo)
 data MetaInfo = MetaInfo
   { metaRange          :: Range
   , metaScope          :: ScopeInfo
-  , metaNumber         :: Maybe Nat  -- ^ The 'MetaId', not the 'InteractionId'.
+  , metaNumber         :: Maybe MetaId
   , metaNameSuggestion :: String
   }
   deriving (Typeable, Show)
@@ -51,7 +57,7 @@ instance KillRange MetaInfo where
  --------------------------------------------------------------------------}
 
 newtype ExprInfo = ExprRange Range
-  deriving (Typeable, Show)
+  deriving (Typeable, Show, Null)
 
 exprNoRange :: ExprInfo
 exprNoRange = ExprRange noRange
@@ -95,7 +101,7 @@ instance KillRange ModuleInfo where
 ---------------------------------------------------------------------------
 
 newtype LetInfo = LetRange Range
-  deriving (Typeable, Show)
+  deriving (Typeable, Show, Null)
 
 instance HasRange LetInfo where
   getRange (LetRange r)   = r
@@ -172,7 +178,7 @@ instance KillRange MutualInfo where
  --------------------------------------------------------------------------}
 
 newtype LHSInfo = LHSRange Range
-  deriving (Typeable, Show)
+  deriving (Typeable, Show, Null)
 
 instance HasRange LHSInfo where
   getRange (LHSRange r) = r
@@ -184,26 +190,10 @@ instance KillRange LHSInfo where
     Pattern information
  --------------------------------------------------------------------------}
 
--- | For a general pattern we can either remember just the source code
---   position or the entire concrete pattern it came from.
-data PatInfo
+-- | For a general pattern we remember the source code position.
+newtype PatInfo
   = PatRange Range
-  | PatSource Range (Precedence -> Pattern)
-      -- ^ Even if we store the original pattern we have to know
-      --   whether to put parenthesis around it.
-  deriving (Typeable)
-
-instance Show PatInfo where
-  show (PatRange r) = "PatRange " ++ show r
-  show (PatSource r _) = "PatSource " ++ show r
-
-instance HasRange PatInfo where
-  getRange (PatRange r)    = r
-  getRange (PatSource r _) = r
-
-instance KillRange PatInfo where
-  killRange (PatRange r)    = PatRange noRange
-  killRange (PatSource r f) = PatSource noRange f
+  deriving (Typeable, Eq, Null, Show, HasRange, KillRange)
 
 -- | Empty range for patterns.
 patNoRange :: PatInfo
@@ -211,13 +201,15 @@ patNoRange = PatRange noRange
 
 -- | Constructor pattern info.
 data ConPatInfo = ConPatInfo
-  { patImplicit :: Bool
+  { patOrigin   :: ConPOrigin
     -- ^ Does this pattern come form the eta-expansion of an implicit pattern?
+    ---  Or from a user written constructor or record pattern?
   , patInfo     :: PatInfo
   }
+  deriving (Typeable)
 
 instance Show ConPatInfo where
-  show (ConPatInfo b i) = (if b then ("implicit " ++) else id) $ show i
+  show (ConPatInfo po i) = applyWhen (po == ConPImplicit) ("implicit " ++) $ show i
 
 instance HasRange ConPatInfo where
   getRange = getRange . patInfo
