@@ -1,3 +1,4 @@
+{-# LANGUAGE BangPatterns               #-}
 {-# LANGUAGE CPP                        #-}
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE DeriveFoldable             #-}
@@ -6,27 +7,35 @@
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE MultiParamTypeClasses      #-}
+{-# LANGUAGE StandaloneDeriving         #-}
 {-# LANGUAGE TemplateHaskell            #-}
-{-# LANGUAGE TypeSynonymInstances       #-}
 
 module Agda.Syntax.Internal
     ( module Agda.Syntax.Internal
     , module Agda.Syntax.Abstract.Name
     , module Agda.Utils.Pointer
+    , MetaId(..)
     ) where
 
 import Prelude hiding (foldr, mapM, null)
 
 import Control.Arrow ((***))
-import Control.Applicative
+import Control.Applicative hiding (empty)
 import Control.Monad.Identity hiding (mapM)
 import Control.Monad.State hiding (mapM)
 import Control.Parallel
 
-import Data.Foldable
+import Data.Foldable ( Foldable, foldMap )
 import Data.Function
 import qualified Data.List as List
 import Data.Maybe
+import Data.Monoid
+
+-- base-4.7 defines the Num instance for Sum
+#if !(MIN_VERSION_base(4,7,0))
+import Data.Orphans             ()
+#endif
+
 import Data.Traversable
 import Data.Typeable (Typeable)
 
@@ -34,17 +43,26 @@ import Agda.Syntax.Position
 import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
 import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Literal
+import Agda.Syntax.Abstract (IsProjP(..))
 import Agda.Syntax.Abstract.Name
 
 import Agda.Utils.Empty
+
+-- See Issue 1593.
+#if !MIN_VERSION_transformers(0,4,1)
+import Agda.Utils.Except ( Error(noMsg) )
+#endif
+
 import Agda.Utils.Functor
 import Agda.Utils.Geniplate
+import Agda.Utils.Lens
 import Agda.Utils.List
+import Agda.Utils.Maybe
 import Agda.Utils.Null
 import Agda.Utils.Permutation
 import Agda.Utils.Pointer
 import Agda.Utils.Size
-import Agda.Utils.Pretty
+import Agda.Utils.Pretty as P
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -157,9 +175,13 @@ data Abs a = Abs   { absName :: ArgName, unAbs :: a }
            | NoAbs { absName :: ArgName, unAbs :: a }
   deriving (Typeable, Functor, Foldable, Traversable)
 
+instance Decoration Abs where
+  traverseF f (Abs   x a) = Abs   x <$> f a
+  traverseF f (NoAbs x a) = NoAbs x <$> f a
+
 -- | Types are terms with a sort annotation.
 --
-data Type' a = El { getSort :: Sort, unEl :: a }
+data Type' a = El { _getSort :: Sort, unEl :: a }
   deriving (Typeable, Show, Functor, Foldable, Traversable)
 
 type Type = Type' Term
@@ -167,6 +189,22 @@ type Type = Type' Term
 instance Decoration Type' where
   traverseF f (El s a) = El s <$> f a
 
+class LensSort a where
+  lensSort ::  Lens' Sort a
+  getSort  :: a -> Sort
+  getSort a = a ^. lensSort
+
+instance LensSort (Type' a) where
+  lensSort f (El s a) = f s <&> \ s' -> El s' a
+
+-- General instance leads to overlapping instances.
+-- instance (Decoration f, LensSort a) => LensSort (f a) where
+instance LensSort a => LensSort (Common.Dom c a) where
+  lensSort = traverseF . lensSort
+
+instance LensSort a => LensSort (Abs a) where
+  lensSort = traverseF . lensSort
+
 -- | Sequence of types. An argument of the first type is bound in later types
 --   and so on.
 data Tele a = EmptyTel
@@ -175,14 +213,10 @@ data Tele a = EmptyTel
 
 type Telescope = Tele (Dom Type)
 
-instance Null (Tele a) where
-  null EmptyTel    = True
-  null ExtendTel{} = False
-  empty = EmptyTel
-
+-- | A traversal for the names in a telescope.
 mapAbsNamesM :: Applicative m => (ArgName -> m ArgName) -> Tele a -> m (Tele a)
 mapAbsNamesM f EmptyTel                  = pure EmptyTel
-mapAbsNamesM f (ExtendTel a (Abs x b))   = ExtendTel a <$> (Abs <$> f x <*> mapAbsNamesM f b)
+mapAbsNamesM f (ExtendTel a (  Abs x b)) = ExtendTel a <$> (  Abs <$> f x <*> mapAbsNamesM f b)
 mapAbsNamesM f (ExtendTel a (NoAbs x b)) = ExtendTel a <$> (NoAbs <$> f x <*> mapAbsNamesM f b)
   -- Ulf, 2013-11-06: Last case is really impossible but I'd rather find out we
   --                  violated that invariant somewhere other than here.
@@ -203,13 +237,16 @@ replaceEmptyName x = mapAbsNames $ \ y -> if null y then x else y
 
 -- | Sorts.
 --
-data Sort = Type Level
-          | Prop  -- ignore me
-          | Inf
-          | DLub Sort (Abs Sort)
-            -- ^ if the free variable occurs in the second sort
-            --   the whole thing should reduce to Inf, otherwise
-            --   it's the normal Lub
+data Sort
+  = Type Level  -- ^ @Set ℓ@.
+  | Prop        -- ^ Dummy sort.
+  | Inf         -- ^ @Setω@.
+  | SizeUniv    -- ^ @SizeUniv@, a sort inhabited by type @Size@.
+  | DLub Sort (Abs Sort)
+    -- ^ Dependent least upper bound.
+    --   If the free variable occurs in the second sort,
+    --   the whole thing should reduce to Inf,
+    --   otherwise it's the normal lub.
   deriving (Typeable, Show)
 
 -- | A level is a maximum expression of 0..n 'PlusLevel' expressions
@@ -219,33 +256,126 @@ data Sort = Type Level
 newtype Level = Max [PlusLevel]
   deriving (Show, Typeable)
 
-data PlusLevel = ClosedLevel Integer
-               | Plus Integer LevelAtom
+data PlusLevel
+  = ClosedLevel Integer     -- ^ @n@, to represent @Setₙ@.
+  | Plus Integer LevelAtom  -- ^ @n + ℓ@.
   deriving (Show, Typeable)
 
+-- | An atomic term of type @Level@.
 data LevelAtom
   = MetaLevel MetaId Elims
+    -- ^ A meta variable targeting @Level@ under some eliminations.
   | BlockedLevel MetaId Term
-  | NeutralLevel Term
+    -- ^ A term of type @Level@ whose reduction is blocked by a meta.
+  | NeutralLevel NotBlocked Term
+    -- ^ A neutral term of type @Level@.
   | UnreducedLevel Term
     -- ^ Introduced by 'instantiate', removed by 'reduce'.
   deriving (Show, Typeable)
 
--- | A meta variable identifier is just a natural number.
---
-newtype MetaId = MetaId { metaId :: Nat }
-    deriving (Eq, Ord, Num, Real, Enum, Integral, Typeable)
+-- | Even if we are not stuck on a meta during reduction
+--   we can fail to reduce a definition by pattern matching
+--   for another reason.
+data NotBlocked
+  = StuckOn Elim
+    -- ^ The 'Elim' is neutral and blocks a pattern match.
+  | Underapplied
+    -- ^ Not enough arguments were supplied to complete the matching.
+  | AbsurdMatch
+    -- ^ We matched an absurd clause, results in a neutral 'Def'.
+  | MissingClauses
+    -- ^ We ran out of clauses, all considered clauses
+    --   produced an actual mismatch.
+    --   This can happen when try to reduce a function application
+    --   but we are still missing some function clauses.
+    --   See "Agda.TypeChecking.Patterns.Match".
+  | ReallyNotBlocked
+    -- ^ Reduction was not blocked, we reached a whnf
+    --   which can be anything but a stuck @'Def'@.
+  deriving (Show, Typeable)
+
+-- | 'ReallyNotBlocked' is the unit.
+--   'MissingClauses' is dominant.
+--   @'StuckOn'{}@ should be propagated, if tied, we take the left.
+instance Monoid NotBlocked where
+  -- ReallyNotBlocked is neutral
+  mempty                       = ReallyNotBlocked
+  ReallyNotBlocked `mappend` b = b
+  -- MissingClauses is dominant (absorptive)
+  b@MissingClauses `mappend` _ = b
+  _ `mappend` b@MissingClauses = b
+  -- StuckOn is second strongest
+  b@StuckOn{}      `mappend` _ = b
+  _ `mappend` b@StuckOn{}      = b
+  b `mappend` _                = b
 
 -- | Something where a meta variable may block reduction.
-data Blocked t = Blocked MetaId t
-               | NotBlocked t
-    deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
+data Blocked t
+  = Blocked    { theBlockingMeta :: MetaId    , ignoreBlocking :: t }
+  | NotBlocked { blockingStatus  :: NotBlocked, ignoreBlocking :: t }
+  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  -- deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
 
+-- | Blocking by a meta is dominant.
 instance Applicative Blocked where
   pure = notBlocked
-  Blocked x f  <*> e = Blocked x $ f (ignoreBlocking e)
-  NotBlocked f <*> e = f <$> e
+  f <*> e = ((f $> ()) `mappend` (e $> ())) $> ignoreBlocking f (ignoreBlocking e)
 
+-- -- | Blocking by a meta is dominant.
+-- instance Applicative Blocked where
+--   pure = notBlocked
+--   Blocked x f     <*> e                = Blocked x $ f (ignoreBlocking e)
+--   NotBlocked nb f <*> Blocked    x   e = Blocked x $ f e
+--   NotBlocked nb f <*> NotBlocked nb' e = NotBlocked (nb `mappend` nb') $ f e
+
+-- | @'Blocked' t@ without the @t@.
+type Blocked_ = Blocked ()
+
+instance Monoid Blocked_ where
+  mempty = notBlocked ()
+  b@Blocked{}    `mappend` _              = b
+  _              `mappend` b@Blocked{}    = b
+  NotBlocked x _ `mappend` NotBlocked y _ = NotBlocked (x `mappend` y) ()
+
+-- See Issue 1573.
+#if !MIN_VERSION_transformers(0,4,1)
+instance Error Blocked_ where
+  noMsg = __IMPOSSIBLE__
+#endif
+
+-- | When trying to reduce @f es@, on match failed on one
+--   elimination @e ∈ es@ that came with info @r :: NotBlocked@.
+--   @stuckOn e r@ produces the new @NotBlocked@ info.
+--
+--   'MissingClauses' must be propagated, as this is blockage
+--   that can be lifted in the future (as more clauses are added).
+--
+--   @'StuckOn' e0@ is also propagated, since it provides more
+--   precise information as @StuckOn e@ (as @e0@ is the original
+--   reason why reduction got stuck and usually a subterm of @e@).
+--   An information like @StuckOn (Apply (Arg info (Var i [])))@
+--   (stuck on a variable) could be used by the lhs/coverage checker
+--   to trigger a split on that (pattern) variable.
+--
+--   In the remaining cases for @r@, we are terminally stuck
+--   due to @StuckOn e@.  Propagating @'AbsurdMatch'@ does not
+--   seem useful.
+--
+--   'Underapplied' must not be propagated, as this would mean
+--   that @f es@ is underapplied, which is not the case (it is stuck).
+--   Note that 'Underapplied' can only arise when projection patterns were
+--   missing to complete the original match (in @e@).
+--   (Missing ordinary pattern would mean the @e@ is of function type,
+--   but we cannot match against something of function type.)
+stuckOn :: Elim -> NotBlocked -> NotBlocked
+stuckOn e r =
+  case r of
+    MissingClauses   -> r
+    StuckOn{}        -> r
+    Underapplied     -> r'
+    AbsurdMatch      -> r'
+    ReallyNotBlocked -> r'
+  where r' = StuckOn e
 
 ---------------------------------------------------------------------------
 -- * Definitions
@@ -258,7 +388,7 @@ instance Applicative Blocked where
 --  the patterns to the order they occur in the telescope. The body
 --  binds the variables in the order they appear in the patterns.
 --
---  @clauseTel ~ permute clausePerm (patternVars clausPats)@
+--  @clauseTel ~ permute clausePerm (patternVars namedClausePats)@
 --
 --  Terms in dot patterns are valid in the clause telescope.
 --
@@ -266,12 +396,16 @@ instance Applicative Blocked where
 --  as variables. TODO: Change this!
 data Clause = Clause
     { clauseRange     :: Range
-    , clauseTel       :: Telescope     -- ^ The types of the pattern variables.
+    , clauseTel       :: Telescope
+      -- ^ @Δ@: The types of the pattern variables.
     , clausePerm      :: Permutation
+      -- ^ @π@ with @Γ ⊢ renamingR π : Δ@, which means @Δ ⊢ renaming π : Γ@.
     , namedClausePats :: [NamedArg Pattern]
+      -- ^ @let Γ = patternVars namedClausePats@
     , clauseBody      :: ClauseBody
+      -- ^ @λΓ.v@
     , clauseType      :: Maybe (Arg Type)
-      -- ^ The type of the rhs under @clauseTel@.
+      -- ^ @Δ ⊢ t@.  The type of the rhs under @clauseTel@.
       --   Used, e.g., by @TermCheck@.
       --   Can be 'Irrelevant' if we encountered an irrelevant projection
       --   pattern on the lhs.
@@ -281,25 +415,6 @@ data Clause = Clause
 clausePats :: Clause -> [Arg Pattern]
 clausePats = map (fmap namedThing) . namedClausePats
 
--- MOVED to Agda. Syntax.Internal.Patterns
--- -- | Translate the clause patterns to terms with free variables bound by the
--- --   clause telescope.
--- clauseArgs :: Clause -> Args
--- clauseArgs cl = evalState (argsToTerms $ namedClausePats cl) xs
---   where
---     perm = clausePerm cl
---     xs   = permute (invertP __IMPOSSIBLE__ perm) $ downFrom (size perm)
---
---     next = do x : xs <- get; put xs; return x
---
---     argsToTerms = traverse $ traverse $ patToTerm . namedThing
---     patToTerm p = case p of
---       VarP _      -> flip Var [] <$> next
---       DotP v      -> v <$ next   -- dot patterns count as variables
---       ConP c _ ps -> Con c <$> argsToTerms ps
---       LitP l      -> pure $ Lit l
---       ProjP{}     -> __IMPOSSIBLE__   -- TODO
-
 data ClauseBodyF a = Body a
                    | Bind (Abs (ClauseBodyF a))
                    | NoBody    -- ^ for absurd clauses.
@@ -307,6 +422,13 @@ data ClauseBodyF a = Body a
 
 type ClauseBody = ClauseBodyF Term
 
+imapClauseBody :: (Nat -> a -> b) -> ClauseBodyF a -> ClauseBodyF b
+imapClauseBody f b = go 0 b
+  where
+    go i  (Body x)  = Body (f i x)
+    go _   NoBody   = NoBody
+    go !i (Bind b)  = Bind $ go (i + 1) <$> b
+
 instance HasRange Clause where
   getRange = clauseRange
 
@@ -342,6 +464,8 @@ data Pattern' x
 type Pattern = Pattern' PatVarName
     -- ^ The @PatVarName@ is a name suggestion.
 
+-- | Type used when numbering pattern variables.
+type DeBruijnPattern = Pattern' (Int, PatVarName)
 
 namedVarP :: PatVarName -> Named (Ranged PatVarName) Pattern
 namedVarP x = Named named $ VarP x
@@ -354,7 +478,22 @@ namedVarP x = Named named $ VarP x
 --   The @Type@ is the type of the whole record pattern.
 --   The scope used for the type is given by any outer scope
 --   plus the clause's telescope ('clauseTel').
-type ConPatternInfo = Maybe (Bool, Arg Type)
+data ConPatternInfo = ConPatternInfo
+  { conPRecord :: Maybe ConPOrigin
+    -- ^ @Nothing@ if data constructor.
+    --   @Just@ if record constructor.
+  , conPType   :: Maybe (Arg Type)
+    -- ^ The type of the whole constructor pattern.
+    --   Should be present (@Just@) if constructor pattern is
+    --   is generated ordinarily by type-checking.
+    --   Could be absent (@Nothing@) if pattern comes from some
+    --   plugin (like Agsy).
+    --   Needed e.g. for with-clause stripping.
+  }
+  deriving (Typeable, Show)
+
+noConPatternInfo :: ConPatternInfo
+noConPatternInfo = ConPatternInfo Nothing Nothing
 
 -- | Extract pattern variables in left-to-right order.
 --   A 'DotP' is also treated as variable (see docu for 'Clause').
@@ -370,40 +509,70 @@ properlyMatching :: Pattern -> Bool
 properlyMatching VarP{} = False
 properlyMatching DotP{} = False
 properlyMatching LitP{} = True
-properlyMatching (ConP _ mt ps) = isNothing mt || -- not a record cons
+properlyMatching (ConP _ ci ps) = isNothing (conPRecord ci) || -- not a record cons
   List.any (properlyMatching . namedArg) ps  -- or one of subpatterns is a proper m
 properlyMatching ProjP{} = True
 
+instance IsProjP Pattern where
+  isProjP (ProjP d) = Just d
+  isProjP _         = Nothing
+
 -----------------------------------------------------------------------------
 -- * Explicit substitutions
 -----------------------------------------------------------------------------
 
 -- | Substitutions.
 
-infixr 4 :#
 data Substitution
 
-  = IdS                     -- Γ ⊢ IdS : Γ
+  = IdS
+    -- ^ Identity substitution.
+    --   @Γ ⊢ IdS : Γ@
 
-  | EmptyS                  -- Γ ⊢ EmptyS : ()
+  | EmptyS
+    -- ^ Empty substitution, lifts from the empty context.
+    --   Apply this to closed terms you want to use in a non-empty context.
+    --   @Γ ⊢ EmptyS : ()@
 
-                            --      Γ ⊢ ρ : Δ
-  | Wk !Int Substitution    -- -------------------
-                            -- Γ, Ψ ⊢ Wk |Ψ| ρ : Δ
+  | Term :# Substitution
+    -- ^ Substitution extension, ``cons''.
+    --   @
+    --     Γ ⊢ u : Aρ   Γ ⊢ ρ : Δ
+    --     ----------------------
+    --     Γ ⊢ u :# ρ : Δ, A
+    --   @
 
-                            -- Γ ⊢ u : Aρ  Γ ⊢ ρ : Δ
-  | Term :# Substitution    -- ---------------------
-                            --   Γ ⊢ u :# ρ : Δ, A
+  | Strengthen Empty Substitution
+    -- ^ Strengthening substitution.  First argument is @__IMPOSSIBLE__@.
+    --   Apply this to a term which does not contain variable 0
+    --   to lower all de Bruijn indices by one.
+    --   @
+    --             Γ ⊢ ρ : Δ
+    --     ---------------------------
+    --     Γ ⊢ Strengthen ρ : Δ, A
+    --   @
 
-    -- First argument is __IMPOSSIBLE__  --         Γ ⊢ ρ : Δ
-  | Strengthen Empty Substitution        -- ---------------------------
-                                         --   Γ ⊢ Strengthen ρ : Δ, A
+  | Wk !Int Substitution
+    -- ^ Weakning substitution, lifts to an extended context.
+    --   @
+    --         Γ ⊢ ρ : Δ
+    --     -------------------
+    --     Γ, Ψ ⊢ Wk |Ψ| ρ : Δ
+    --   @
 
-                            --        Γ ⊢ ρ : Δ
-  | Lift !Int Substitution  -- -------------------------
-                            -- Γ, Ψρ ⊢ Lift |Ψ| ρ : Δ, Ψ
+
+  | Lift !Int Substitution
+    -- ^ Lifting substitution.  Use this to go under a binder.
+    --   @Lift 1 ρ == var 0 :# Wk 1 ρ@.
+    --   @
+    --            Γ ⊢ ρ : Δ
+    --     -------------------------
+    --     Γ, Ψρ ⊢ Lift |Ψ| ρ : Δ, Ψ
+    --   @
   deriving (Show)
 
+infixr 4 :#
+
 ---------------------------------------------------------------------------
 -- * Absurd Lambda
 ---------------------------------------------------------------------------
@@ -511,19 +680,17 @@ typeDontCare = El Prop (Sort Prop)
 topSort :: Type
 topSort = El Inf (Sort Inf)
 
-prop :: Type
-prop = sort Prop
-
 sort :: Sort -> Type
 sort s = El (sSuc s) $ Sort s
 
 varSort :: Int -> Sort
-varSort n = Type $ Max [Plus 0 $ NeutralLevel $ Var n []]
+varSort n = Type $ Max [Plus 0 $ NeutralLevel mempty $ var n]
 
 -- | Get the next higher sort.
 sSuc :: Sort -> Sort
 sSuc Prop            = mkType 1
 sSuc Inf             = Inf
+sSuc SizeUniv        = SizeUniv
 sSuc (DLub a b)      = DLub (sSuc a) (fmap sSuc b)
 sSuc (Type l)        = Type $ levelSuc l
 
@@ -542,6 +709,7 @@ impossibleTerm file line = Lit $ LitString noRange $ unlines
   , "Location of the error: " ++ file ++ ":" ++ show line
   ]
 
+-- | Constructing a singleton telescope.
 class SgTel a where
   sgTel :: a -> Telescope
 
@@ -551,6 +719,9 @@ instance SgTel (ArgName, Dom Type) where
 instance SgTel (Dom (ArgName, Type)) where
   sgTel (Common.Dom ai (x, t)) = ExtendTel (Common.Dom ai t) $ Abs x EmptyTel
 
+instance SgTel (Dom Type) where
+  sgTel dom = sgTel (stringToArgName "_", dom)
+
 hackReifyToMeta :: Term
 hackReifyToMeta = DontCare $ Lit $ LitInt noRange (-42)
 
@@ -564,17 +735,13 @@ isHackReifyToMeta _ = False
 
 blockingMeta :: Blocked t -> Maybe MetaId
 blockingMeta (Blocked m _) = Just m
-blockingMeta (NotBlocked _) = Nothing
+blockingMeta NotBlocked{}  = Nothing
 
 blocked :: MetaId -> a -> Blocked a
 blocked x = Blocked x
 
 notBlocked :: a -> Blocked a
-notBlocked = NotBlocked
-
-ignoreBlocking :: Blocked a -> a
-ignoreBlocking (Blocked _ x) = x
-ignoreBlocking (NotBlocked x) = x
+notBlocked = NotBlocked ReallyNotBlocked
 
 ---------------------------------------------------------------------------
 -- * Simple operations on terms and types.
@@ -592,12 +759,9 @@ arity t = case ignoreSharing $ unEl t of
   Pi  _ b -> 1 + arity (unAbs b)
   _       -> 0
 
--- | Suggest a name for the first argument of a function of the given type.
-argName :: Type -> String
-argName = argN . ignoreSharing . unEl
-    where
-        argN (Pi _ b)  = "." ++ argNameToString (absName b)
-        argN _    = __IMPOSSIBLE__
+-- | Make a name that is not in scope.
+notInScopeName :: ArgName -> ArgName
+notInScopeName = stringToArgName . ("." ++) . argNameToString
 
 -- | Pick the better name suggestion, i.e., the one that is not just underscore.
 class Suggest a b where
@@ -610,6 +774,12 @@ instance Suggest String String where
 instance Suggest (Abs a) (Abs b) where
   suggest b1 b2 = suggest (absName b1) (absName b2)
 
+instance Suggest String (Abs b) where
+  suggest x b = suggest x (absName b)
+
+instance Suggest Name (Abs b) where
+  suggest n b = suggest (nameToArgName n) (absName b)
+
 ---------------------------------------------------------------------------
 -- * Eliminations.
 ---------------------------------------------------------------------------
@@ -702,6 +872,30 @@ swapElimArg (Apply (Common.Arg ai a)) = Common.Arg ai (Apply a)
 swapElimArg (Proj  d) = defaultArg (Proj  d)
 -}
 
+---------------------------------------------------------------------------
+-- * Null instances.
+---------------------------------------------------------------------------
+
+instance Null (Tele a) where
+  empty = EmptyTel
+  null EmptyTel    = True
+  null ExtendTel{} = False
+
+instance Null ClauseBody where
+  empty = NoBody
+  null NoBody = True
+  null _      = False
+
+-- | A 'null' clause is one with no patterns and no rhs.
+--   Should not exist in practice.
+instance Null Clause where
+  empty = Clause empty empty empty empty empty empty
+  null (Clause r tel perm pats body t)
+    =  null tel
+    && null pats
+    && null body
+
+
 ---------------------------------------------------------------------------
 -- * Show instances.
 ---------------------------------------------------------------------------
@@ -712,49 +906,16 @@ instance Show a => Show (Abs a) where
   showsPrec p (NoAbs x a) = showParen (p > 0) $
     showString "NoAbs " . shows x . showString " " . showsPrec 10 a
 
-instance Show MetaId where
-    show (MetaId n) = "_" ++ show n
-
-instance Show t => Show (Blocked t) where
-  showsPrec p (Blocked m x) = showParen (p > 0) $
-    showString "Blocked " . shows m . showString " " . showsPrec 10 x
-  showsPrec p (NotBlocked x) = showsPrec p x
+-- instance Show t => Show (Blocked t) where
+--   showsPrec p (Blocked m x) = showParen (p > 0) $
+--     showString "Blocked " . shows m . showString " " . showsPrec 10 x
+--   showsPrec p (NotBlocked x) = showsPrec p x
 
 ---------------------------------------------------------------------------
--- * Sized instances.
+-- * Sized instances and TermSize.
 ---------------------------------------------------------------------------
 
-instance Sized Term where
-  size v = case v of
-    Var _ vs    -> 1 + Prelude.sum (map size vs)
-    Def _ vs    -> 1 + Prelude.sum (map size vs)
-    Con _ vs    -> 1 + Prelude.sum (map size vs)
-    MetaV _ vs  -> 1 + Prelude.sum (map size vs)
-    Level l     -> size l
-    Lam _ f     -> 1 + size f
-    Lit _       -> 1
-    Pi a b      -> 1 + size a + size b
-    Sort s      -> 1
-    DontCare mv -> size mv
-    Shared p    -> size (derefPtr p)
-    ExtLam{}    -> __IMPOSSIBLE__
-
-instance Sized Type where
-  size = size . unEl
-
-instance Sized Level where
-  size (Max as) = 1 + Prelude.sum (map size as)
-
-instance Sized PlusLevel where
-  size (ClosedLevel _) = 1
-  size (Plus _ a)      = size a
-
-instance Sized LevelAtom where
-  size (MetaLevel _   vs) = 1 + Prelude.sum (map size vs)
-  size (BlockedLevel _ v) = size v
-  size (NeutralLevel   v) = size v
-  size (UnreducedLevel v) = size v
-
+-- | The size of a telescope is its length (as a list).
 instance Sized (Tele a) where
   size  EmptyTel         = 0
   size (ExtendTel _ tel) = 1 + size tel
@@ -762,9 +923,68 @@ instance Sized (Tele a) where
 instance Sized a => Sized (Abs a) where
   size = size . unAbs
 
-instance Sized a => Sized (Elim' a) where
-  size (Apply v) = size v
-  size  Proj{}   = 1
+-- | The size of a term is roughly the number of nodes in its
+--   syntax tree.  This number need not be precise for logical
+--   correctness of Agda, it is only used for reporting
+--   (and maybe decisions regarding performance).
+--
+--   Not counting towards the term size are:
+--
+--     * sort and color annotations,
+--     * projections.
+--
+class TermSize a where
+  termSize :: a -> Int
+  termSize = getSum . tsize
+
+  tsize :: a -> Sum Int
+
+instance (Foldable t, TermSize a) => TermSize (t a) where
+  tsize = foldMap tsize
+
+instance TermSize Term where
+  tsize v = case v of
+    Var _ vs    -> 1 + tsize vs
+    Def _ vs    -> 1 + tsize vs
+    Con _ vs    -> 1 + tsize vs
+    MetaV _ vs  -> 1 + tsize vs
+    Level l     -> tsize l
+    Lam _ f     -> 1 + tsize f
+    Lit _       -> 1
+    Pi a b      -> 1 + tsize a + tsize b
+    Sort s      -> tsize s
+    DontCare mv -> tsize mv
+    Shared p    -> tsize (derefPtr p)
+    ExtLam{}    -> __IMPOSSIBLE__
+
+instance TermSize Sort where
+  tsize s = case s of
+    Type l    -> 1 + tsize l
+    Prop      -> 1
+    Inf       -> 1
+    SizeUniv  -> 1
+    DLub s s' -> 1 + tsize s + tsize s'
+
+instance TermSize Level where
+  tsize (Max as) = 1 + tsize as
+
+instance TermSize PlusLevel where
+  tsize (ClosedLevel _) = 1
+  tsize (Plus _ a)      = tsize a
+
+instance TermSize LevelAtom where
+  tsize (MetaLevel _   vs) = 1 + tsize vs
+  tsize (BlockedLevel _ v) = tsize v
+  tsize (NeutralLevel _ v) = tsize v
+  tsize (UnreducedLevel v) = tsize v
+
+instance TermSize Substitution where
+  tsize IdS                = 1
+  tsize EmptyS             = 1
+  tsize (Wk _ rho)         = 1 + tsize rho
+  tsize (t :# rho)         = 1 + tsize t + tsize rho
+  tsize (Strengthen _ rho) = 1 + tsize rho
+  tsize (Lift _ rho)       = 1 + tsize rho
 
 ---------------------------------------------------------------------------
 -- * KillRange instances.
@@ -779,14 +999,14 @@ instance KillRange Term where
     Def c vs    -> killRange2 Def c vs
     Con c vs    -> killRange2 Con c vs
     MetaV m vs  -> killRange1 (MetaV m) vs
-    Lam i f     -> killRange1 Lam i f
+    Lam i f     -> killRange2 Lam i f
     Lit l       -> killRange1 Lit l
     Level l     -> killRange1 Level l
     Pi a b      -> killRange2 Pi a b
     Sort s      -> killRange1 Sort s
     DontCare mv -> killRange1 DontCare mv
     Shared p    -> Shared $ updatePtr killRange p
-    ExtLam{}    -> __IMPOSSIBLE__
+    ExtLam c vs -> killRange2 ExtLam c vs
 
 instance KillRange Level where
   killRange (Max as) = killRange1 Max as
@@ -798,19 +1018,31 @@ instance KillRange PlusLevel where
 instance KillRange LevelAtom where
   killRange (MetaLevel n as)   = killRange1 (MetaLevel n) as
   killRange (BlockedLevel m v) = killRange1 (BlockedLevel m) v
-  killRange (NeutralLevel v)   = killRange1 NeutralLevel v
+  killRange (NeutralLevel r v) = killRange1 (NeutralLevel r) v
   killRange (UnreducedLevel v) = killRange1 UnreducedLevel v
 
-instance KillRange Type where
+instance (KillRange a) => KillRange (Type' a) where
   killRange (El s v) = killRange2 El s v
 
 instance KillRange Sort where
   killRange s = case s of
     Prop       -> Prop
     Inf        -> Inf
+    SizeUniv   -> SizeUniv
     Type a     -> killRange1 Type a
     DLub s1 s2 -> killRange2 DLub s1 s2
 
+instance KillRange Substitution where
+  killRange IdS                  = IdS
+  killRange EmptyS               = EmptyS
+  killRange (Wk n rho)           = killRange1 (Wk n) rho
+  killRange (t :# rho)           = killRange2 (:#) t rho
+  killRange (Strengthen err rho) = killRange1 (Strengthen err) rho
+  killRange (Lift n rho)         = killRange1 (Lift n) rho
+
+instance KillRange ConPatternInfo where
+  killRange (ConPatternInfo mr mt) = killRange1 (ConPatternInfo mr) mt
+
 instance KillRange Pattern where
   killRange p =
     case p of
@@ -820,9 +1052,6 @@ instance KillRange Pattern where
       LitP l           -> killRange1 LitP l
       ProjP q          -> killRange1 ProjP q
 
-instance KillRange Permutation where
-  killRange = id
-
 instance KillRange Clause where
   killRange (Clause r tel perm ps body t) = killRange6 Clause r tel perm ps body t
 
@@ -857,8 +1086,16 @@ instanceUniverseBiT' [] [t| ([Term], Term)                |]
 -- * Simple pretty printing
 -----------------------------------------------------------------------------
 
-showTerm :: Term -> String
-showTerm = show . pretty
+instance Pretty Substitution where
+  prettyPrec p rho = brackets $ pr rho
+    where
+    pr rho = case rho of
+      IdS              -> text "idS"
+      EmptyS           -> text "ε"
+      t :# rho         -> prettyPrec 1 t <+> text ":#" <+> pr rho
+      Strengthen _ rho -> text "↓" <+> pr rho
+      Wk n rho         -> text ("↑" ++ show n) <+> pr rho
+      Lift n rho       -> text ("⇑" ++ show n) <+> pr rho
 
 instance Pretty Term where
   prettyPrec p v =
@@ -879,7 +1116,7 @@ instance Pretty Term where
             , nest 2 $ pretty (unAbs b) ]
       Sort s      -> pretty s
       Level l     -> pretty l
-      MetaV x els -> text (show x) `pApp` els
+      MetaV x els -> pretty x `pApp` els
       DontCare v  -> pretty v
       Shared{}    -> __IMPOSSIBLE__
       ExtLam{}    -> __IMPOSSIBLE__
@@ -914,7 +1151,7 @@ instance Pretty LevelAtom where
     case a of
       MetaLevel x els  -> prettyPrec p (MetaV x els)
       BlockedLevel _ v -> prettyPrec p v
-      NeutralLevel v   -> prettyPrec p v
+      NeutralLevel _ v -> prettyPrec p v
       UnreducedLevel v -> prettyPrec p v
 
 instance Pretty Sort where
@@ -925,6 +1162,7 @@ instance Pretty Sort where
       Type l -> mparens (p > 9) $ text "Set" <+> prettyPrec 10 l
       Prop -> text "Prop"
       Inf -> text "Setω"
+      SizeUniv -> text "SizeUniv"
       DLub s b -> mparens (p > 9) $
         text "dlub" <+> prettyPrec 10 s
                     <+> parens (sep [ text ("λ " ++ show (absName b) ++ " ->")
@@ -944,3 +1182,24 @@ instance Pretty a => Pretty (Arg a) where
       NotHidden -> prettyPrec p
       Hidden    -> braces . pretty
       Instance  -> braces . braces . pretty
+
+instance Pretty a => Pretty (Pattern' a) where
+  prettyPrec n (VarP x)      = prettyPrec n x
+  prettyPrec _ (DotP t)      = text "." P.<> prettyPrec 10 t
+  prettyPrec n (ConP c i ps) = mparens (n > 0) $
+    text (show $ conName c) <+> fsep (map (pretty . namedArg) ps)
+  -- -- Version with printing record type:
+  -- prettyPrec _ (ConP c i ps) = (if b then braces else parens) $ prTy $
+  --   text (show $ conName c) <+> fsep (map (pretty . namedArg) ps)
+  --   where
+  --     b = maybe False (== ConPImplicit) $ conPRecord i
+  --     prTy d = caseMaybe (conPType i) d $ \ t -> d  <+> text ":" <+> pretty t
+  prettyPrec _ (LitP l)      = text (show l)
+  prettyPrec _ (ProjP q)     = text (show q)
+
+instance Pretty a => Pretty (ClauseBodyF a) where
+  pretty b = case b of
+    Bind (NoAbs _ b) -> pretty b
+    Bind (Abs   x b) -> text (show x ++ ".") <+> pretty b
+    Body t           -> pretty t
+    NoBody           -> text "()"
