@@ -3,11 +3,13 @@
 {-# LANGUAGE FlexibleInstances      #-}
 {-# LANGUAGE FunctionalDependencies #-}
 {-# LANGUAGE MultiParamTypeClasses  #-}
-{-# LANGUAGE OverlappingInstances   #-}
 {-# LANGUAGE TupleSections          #-}
-{-# LANGUAGE TypeSynonymInstances   #-}
 {-# LANGUAGE UndecidableInstances   #-}  -- because of func. deps.
 
+#if __GLASGOW_HASKELL__ <= 708
+{-# LANGUAGE OverlappingInstances #-}
+#endif
+
 module Agda.Syntax.Internal.Pattern where
 
 import Control.Applicative
@@ -49,7 +51,12 @@ class FunArity a where
   funArity :: a -> Int
 
 -- | Get the number of initial 'Apply' patterns.
+
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPABLE #-} IsProjP p => FunArity [p] where
+#else
 instance IsProjP p => FunArity [p] where
+#endif
   funArity = length . takeWhile (isNothing . isProjP)
 
 -- | Get the number of initial 'Apply' patterns in a clause.
@@ -57,7 +64,11 @@ instance FunArity Clause where
   funArity = funArity . clausePats
 
 -- | Get the number of common initial 'Apply' patterns in a list of clauses.
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} FunArity [Clause] where
+#else
 instance FunArity [Clause] where
+#endif
   funArity []  = 0
   funArity cls = minimum $ map funArity cls
 
@@ -90,14 +101,25 @@ instance LabelPatVars (Pattern' x) (Pattern' (i,x)) i where
     where next = do (x:xs) <- get; put xs; return x
 
 -- | Augment pattern variables with their de Bruijn index.
-{-# SPECIALIZE numberPatVars :: Permutation -> [NamedArg (Pattern' x)] -> [(NamedArg (Pattern' (Int, x)))] #-}
+{-# SPECIALIZE numberPatVars :: Permutation -> [NamedArg (Pattern' x)] -> [NamedArg (Pattern' (Int, x))] #-}
+{-# SPECIALIZE numberPatVars :: Permutation -> [NamedArg Pattern] -> [NamedArg DeBruijnPattern] #-}
+--
+--  Example:
+--  @
+--    f : (A : Set) (n : Nat) (v : Vec A n) -> ...
+--    f A .(suc n) (cons n x xs)
+--
+--    clauseTel = (A : Set) (n : Nat) (x : A) (xs : Vec A n)
+--    perm      = Perm 5 [0,2,3,4]
+--    invertP __IMPOSSIBLE__ perm = Perm 4 [0,__IMPOSSIBLE__,1,2,3]
+--    flipP ... = Perm 4 [3,__IMPOSSIBLE__,2,1,0]
+--    pats      = A .(suc 2) (cons n x xs)
+--    dBpats    = 3 .(suc 2) (cons 2 1 0 )
+--  @
+--
 numberPatVars :: LabelPatVars a b Int => Permutation -> a -> b
 numberPatVars perm ps = evalState (labelPatVars ps) $
-  permute (invertP __IMPOSSIBLE__ perm) $ downFrom $ size perm
-
-instance IsProjP Pattern where
-  isProjP (ProjP d) = Just d
-  isProjP _         = Nothing
+  permPicks $ flipP $ invertP __IMPOSSIBLE__ perm
 
 patternsToElims :: Permutation -> [I.NamedArg Pattern] -> [Elim]
 patternsToElims perm ps = map build' $ numberPatVars perm ps
