@@ -87,4 +87,4 @@ instance KillRange Literal where
   killRange (LitFloat  r x) = LitFloat  (killRange r) x
   killRange (LitString r x) = LitString (killRange r) x
   killRange (LitChar   r x) = LitChar   (killRange r) x
-  killRange (LitQName  r x) = LitQName  (killRange r) x
+  killRange (LitQName  r x) = killRange2 LitQName r x
