@@ -23,6 +23,7 @@ import Data.Maybe
 import Data.Typeable (Typeable)
 
 import Agda.Syntax.Common
+import Agda.Syntax.Position
 
 import Agda.Utils.Except ( MonadError(throwError) )
 import Agda.Utils.List
@@ -54,7 +55,13 @@ data GenPart
   | NormalHole (NamedArg () Int)
     -- ^ Argument is where the expression should go.
   | IdPart RawName
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Show, Eq, Ord)
+
+instance KillRange GenPart where
+  killRange p = case p of
+    IdPart x     -> IdPart x
+    BindHole i   -> BindHole i
+    NormalHole x -> NormalHole $ killRange x
 
 -- | Get a flat list of identifier parts of a notation.
 stringParts :: Notation -> [RawName]
@@ -83,7 +90,7 @@ data NotationKind
   | PostfixNotation -- ^ Ex: @bla_blub_@.
   | NonfixNotation  -- ^ Ex: @bla_blub@.
   | NoNotation
-   deriving (Eq)
+   deriving (Eq, Show)
 
 -- | Classify a notation by presence of leading and/or trailing hole.
 notationKind :: Notation -> NotationKind
