@@ -1,10 +1,10 @@
-
 {-| This module defines the things required by Alex and some other
     Alex related things.
 -}
 module Agda.Syntax.Parser.Alex
     ( -- * Alex requirements
       AlexInput(..)
+    , lensLexInput
     , alexInputPrevChar
     , alexGetChar, alexGetByte
       -- * Lex actions
@@ -23,14 +23,19 @@ import Data.Word
 import Agda.Syntax.Position
 import Agda.Syntax.Parser.Monad
 
+import Agda.Utils.Lens
 import Agda.Utils.Monad
 
 -- | This is what the lexer manipulates.
 data AlexInput = AlexInput
-		    { lexPos	    :: !Position    -- ^ current position
-		    , lexInput	    :: String	    -- ^ current input
-		    , lexPrevChar   :: !Char	    -- ^ previously read character
-		    }
+  { lexPos      :: !Position    -- ^ current position
+  , lexInput    :: String       -- ^ current input
+  , lexPrevChar :: !Char        -- ^ previously read character
+  }
+
+-- | A lens for 'lexInput'.
+lensLexInput :: Lens' String AlexInput
+lensLexInput f r = f (lexInput r) <&> \ s -> r { lexInput = s }
 
 -- | Get the previously lexed character. Same as 'lexPrevChar'. Alex needs this
 --   to be defined to handle \"patterns with a left-context\".
@@ -44,11 +49,11 @@ alexGetChar :: AlexInput -> Maybe (Char, AlexInput)
 alexGetChar (AlexInput { lexInput = []  }) = Nothing
 alexGetChar (AlexInput { lexInput = c:s, lexPos = p }) =
     Just (c, AlexInput
-		 { lexInput	= s
-		 , lexPos	= movePos p c
-		 , lexPrevChar	= c
-		 }
-	 )
+                 { lexInput     = s
+                 , lexPos       = movePos p c
+                 , lexPrevChar  = c
+                 }
+         )
 
 -- | A variant of 'alexGetChar'.
 --
@@ -66,19 +71,19 @@ alexGetByte ai =
 getLexInput :: Parser AlexInput
 getLexInput = getInp <$> get
     where
-	getInp s = AlexInput
-		    { lexPos	    = parsePos s
-		    , lexInput	    = parseInp s
-		    , lexPrevChar   = parsePrevChar s
-		    }
+        getInp s = AlexInput
+                    { lexPos        = parsePos s
+                    , lexInput      = parseInp s
+                    , lexPrevChar   = parsePrevChar s
+                    }
 
 setLexInput :: AlexInput -> Parser ()
 setLexInput inp = modify upd
     where
-	upd s = s { parsePos	    = lexPos inp
-		  , parseInp	    = lexInput inp
-		  , parsePrevChar   = lexPrevChar inp
-		  }
+        upd s = s { parsePos        = lexPos inp
+                  , parseInp        = lexInput inp
+                  , parsePrevChar   = lexPrevChar inp
+                  }
 
 {--------------------------------------------------------------------------
     Lex actions
