@@ -6,15 +6,13 @@
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE NoMonomorphismRestriction  #-}
+{-# LANGUAGE ScopedTypeVariables        #-}
+{-# LANGUAGE TemplateHaskell            #-}
 
-#if __GLASGOW_HASKELL__ <= 700
+#if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE OverlappingInstances #-}
 #endif
 
-{-# LANGUAGE ScopedTypeVariables  #-}
-{-# LANGUAGE TemplateHaskell      #-}
-{-# LANGUAGE TypeSynonymInstances #-}
-
 {-| Position information for syntax. Crucial for giving good error messages.
 -}
 
@@ -47,10 +45,12 @@ module Agda.Syntax.Position
   , rangeToInterval
   , continuous
   , continuousPerLine
+  , PrintRange(..)
   , HasRange(..)
   , SetRange(..)
   , KillRange(..)
   , KillRangeT
+  , killRangeMap
   , killRange1, killRange2, killRange3, killRange4, killRange5, killRange6, killRange7
   , killRange8, killRange9, killRange10, killRange11, killRange12, killRange13, killRange14
   , killRange15, killRange16, killRange17, killRange18, killRange19
@@ -66,13 +66,15 @@ module Agda.Syntax.Position
 
 import Prelude hiding (null)
 
-import Control.Applicative
+import Control.Applicative hiding (empty)
 import Control.Monad
 
 import Data.Foldable (Foldable)
 import Data.Function
 import Data.Int
 import Data.List hiding (null)
+import Data.Map (Map)
+import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable (Traversable)
@@ -81,9 +83,10 @@ import Data.Typeable (Typeable)
 import Test.QuickCheck.All
 
 import Agda.Utils.FileName hiding (tests)
+import Agda.Utils.List hiding (tests)
 import Agda.Utils.Maybe
 import Agda.Utils.Null
-import Agda.Utils.Pretty ( (<>), Pretty(pretty) )
+import Agda.Utils.Pretty
 import Agda.Utils.TestHelpers
 import Agda.Utils.QuickCheck
 
@@ -171,6 +174,10 @@ rightMargin r@(Range is) =
   if null is then r else
   let i = last is in Range [ i { iStart = iEnd i } ]
 
+-- | Wrapper to indicate that range should be printed.
+newtype PrintRange a = PrintRange a
+  deriving (Eq, Ord, HasRange, SetRange, KillRange)
+
 -- | Things that have a range are instances of this class.
 class HasRange t where
     getRange :: t -> Range
@@ -215,12 +222,19 @@ class HasRange t => SetRange t where
 instance SetRange Range where
   setRange = const
 
+instance SetRange a => SetRange [a] where
+  setRange r = fmap $ setRange r
+
 -- | Killing the range of an object sets all range information to 'noRange'.
 class KillRange a where
   killRange :: KillRangeT a
 
 type KillRangeT a = a -> a
 
+-- | Remove ranges in keys and values of a map.
+killRangeMap :: (KillRange k, KillRange v) => KillRangeT (Map k v)
+killRangeMap = Map.mapKeysMonotonic killRange . Map.map killRange
+
 killRange1 :: KillRange a => (a -> b) -> a -> b
 
 killRange2 :: (KillRange a, KillRange b) => (a -> b -> c) -> a -> b -> c
@@ -356,9 +370,38 @@ instance KillRange Bool where
 instance KillRange Int where
   killRange = id
 
+instance KillRange Integer where
+  killRange = id
+
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPABLE #-} KillRange a => KillRange [a] where
+#else
 instance KillRange a => KillRange [a] where
+#endif
   killRange = map killRange
 
+-- | Overlaps with @KillRange [a]@.
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} KillRange String where
+#else
+instance KillRange String where
+#endif
+  killRange = id
+
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPABLE #-} KillRange a => KillRange (Map k a) where
+#else
+instance KillRange a => KillRange (Map k a) where
+#endif
+  killRange = fmap killRange
+
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPABLE #-} (Ord a, KillRange a) => KillRange (Set a) where
+#else
+instance (Ord a, KillRange a) => KillRange (Set a) where
+#endif
+  killRange = Set.map killRange
+
 instance (KillRange a, KillRange b) => KillRange (a, b) where
   killRange (x, y) = (killRange x, killRange y)
 
@@ -366,6 +409,10 @@ instance (KillRange a, KillRange b, KillRange c) =>
          KillRange (a, b, c) where
   killRange (x, y, z) = killRange3 (,,) x y z
 
+instance (KillRange a, KillRange b, KillRange c, KillRange d) =>
+         KillRange (a, b, c, d) where
+  killRange (x, y, z, u) = killRange4 (,,,) x y z u
+
 instance KillRange a => KillRange (Maybe a) where
   killRange = fmap killRange
 
@@ -379,27 +426,34 @@ instance (KillRange a, KillRange b) => KillRange (Either a b) where
 
 -- TODO: 'Show' should output Haskell-parseable representations.
 -- The following instances are deprecated, and Pretty should be used
--- instead.  Later, simply derive Show for these types:
+-- instead.  Later, simply derive Show for these types.
 
+-- ASR (02 December 2014). This instance is not used anymore (module
+-- the test suite) when reporting errors. See Issue 1293.
 instance Show a => Show (Position' (Maybe a)) where
-    show (Pn Nothing  _ l c) = show l ++ "," ++ show c
-    show (Pn (Just f) _ l c) = show f ++ ":" ++ show l ++ "," ++ show c
+  show (Pn Nothing  _ l c) = show l ++ "," ++ show c
+  show (Pn (Just f) _ l c) = show f ++ ":" ++ show l ++ "," ++ show c
 
 instance Show a => Show (Interval' (Maybe a)) where
-    show (Interval s e) = file ++ start ++ "-" ++ end
-        where
-            f   = srcFile s
-            sl  = posLine s
-            el  = posLine e
-            sc  = posCol s
-            ec  = posCol e
-            file = case f of
-                     Nothing -> ""
-                     Just f  -> show f ++ ":"
-            start = show sl ++ "," ++ show sc
-            end
-                | sl == el  = show ec
-                | otherwise = show el ++ "," ++ show ec
+  show (Interval s e) = file ++ start ++ "-" ++ end
+    where
+      f  = srcFile s
+      sl = posLine s
+      el = posLine e
+      sc = posCol s
+      ec = posCol e
+
+      file :: String
+      file = case f of
+               Nothing -> ""
+               Just f  -> show f ++ ":"
+
+      start :: String
+      start = show sl ++ "," ++ show sc
+
+      end :: String
+      end | sl == el  = show ec
+          | otherwise = show el ++ "," ++ show ec
 
 instance Show a => Show (Range' (Maybe a)) where
   show r = case rangeToInterval r of
@@ -411,9 +465,38 @@ instance Show a => Show (Range' (Maybe a)) where
 ------------------------------------------------------------------------
 
 instance Pretty a => Pretty (Position' (Maybe a)) where
-    pretty (Pn Nothing  _ l c) = pretty l <> pretty "," <> pretty c
-    pretty (Pn (Just f) _ l c) =
-      pretty f <> pretty ":" <> pretty l <> pretty "," <> pretty c
+  pretty (Pn Nothing  _ l c) = pretty l <> pretty "," <> pretty c
+  pretty (Pn (Just f) _ l c) =
+    pretty f <> pretty ":" <> pretty l <> pretty "," <> pretty c
+
+instance Pretty a => Pretty (Interval' (Maybe a)) where
+  pretty (Interval s e) = file <> start <> pretty "-" <> end
+    where
+      f  = srcFile s
+      sl = posLine s
+      el = posLine e
+      sc = posCol s
+      ec = posCol e
+
+      file :: Doc
+      file = case f of
+               Nothing -> empty
+               Just f  -> pretty f <> colon
+
+      start :: Doc
+      start = pretty sl <> comma <> pretty sc
+
+      end :: Doc
+        | sl == el  = pretty ec
+        | otherwise = pretty el <> comma <> pretty ec
+
+instance Pretty a => Pretty (Range' (Maybe a)) where
+  pretty r = case rangeToInterval r of
+    Nothing -> empty
+    Just i  -> pretty i
+
+instance (Pretty a, HasRange a) => Pretty (PrintRange a) where
+  pretty (PrintRange a) = pretty a <+> parens (text "at" <+> pretty (getRange a))
 
 {--------------------------------------------------------------------------
     Functions on postitions and ranges
@@ -599,6 +682,22 @@ prop_continuous r =
   rPositions cr == makeInterval (rPositions r)
   where cr = continuous r
 
+prop_continuousPerLine :: Range -> Bool
+prop_continuousPerLine r =
+  rangeInvariant r'
+    &&
+  distinct lineNumbers
+  where
+  r'@(Range is') = continuousPerLine r
+
+  lineNumbers = concatMap lines is'
+    where
+    lines i | s == e    = [s]
+            | otherwise = [s, e]
+      where
+      s = posLine (iStart i)
+      e = posLine (iEnd   i)
+
 prop_fuseIntervals :: Interval' Integer -> Property
 prop_fuseIntervals i1 =
   forAll (intervalInSameFileAs i1) $ \i2 ->
