@@ -1,10 +1,9 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE DeriveDataTypeable  #-}
-{-# LANGUAGE GADTs               #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE GADTs #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE RankNTypes          #-}
-{-# LANGUAGE TupleSections       #-}
-
+{-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE TupleSections #-}
 
 {-| This module defines the notion of a scope and operations on scopes.
 -}
@@ -23,6 +22,8 @@ import Data.Typeable (Typeable)
 
 -- import Debug.Trace (trace)
 
+import Agda.Benchmarking
+
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Fixity
@@ -113,7 +114,7 @@ data LocalVar
     --   (List not empty).
   deriving (Typeable)
 
-instance NFData LocalVar
+instance NFData LocalVar where rnf x = seq x ()
 
 instance Eq LocalVar where
   (==) = (==) `on` localVar
@@ -448,7 +449,11 @@ setScopeAccess a s = (`updateScopeNameSpaces` s) $ AssocList.mapWithKey $ const
 
 -- | Update a particular name space.
 setNameSpace :: NameSpaceId -> NameSpace -> Scope -> Scope
-setNameSpace nsid ns = updateScopeNameSpaces $ AssocList.update nsid ns
+setNameSpace nsid ns = modifyNameSpace nsid $ const ns
+
+-- | Modify a particular name space.
+modifyNameSpace :: NameSpaceId -> (NameSpace -> NameSpace) -> Scope -> Scope
+modifyNameSpace nsid f = updateScopeNameSpaces $ AssocList.updateAt nsid f
 
 -- | Add names to a scope.
 addNamesToScope :: NameSpaceId -> C.Name -> [AbstractName] -> Scope -> Scope
@@ -522,7 +527,7 @@ renameCanonicalNames renD renM = mapScope_ renameD renameM
 -- | Remove private name space of a scope.
 --
 --   Should be a right identity for 'exportedNamesInScope'.
---   @exportedNamesInScope . restrictPrivate == exportedNamesInscope@.
+--   @exportedNamesInScope . restrictPrivate == exportedNamesInScope@.
 restrictPrivate :: Scope -> Scope
 restrictPrivate s
   = setNameSpace PrivateNS  emptyNameSpace
@@ -556,11 +561,10 @@ publicModules scope = Map.filterWithKey (\ m _ -> reachable m) allMods
 
 everythingInScope :: ScopeInfo -> NameSpace
 everythingInScope scope = allThingsInScope $ mergeScopes $
-    [ s | (m, s) <- Map.toList (scopeModules scope), m `elem` current ]
+    (s0 :) $ map look $ scopeParents s0
   where
-    this    = scopeCurrent scope
-    parents = maybe __IMPOSSIBLE__ scopeParents $ Map.lookup this $ scopeModules scope
-    current = this : parents
+    look m = fromMaybe __IMPOSSIBLE__ $ Map.lookup m $ scopeModules scope
+    s0     = look $ scopeCurrent scope
 
 -- | Compute a flattened scope. Only include unqualified names or names
 -- qualified by modules in the first argument.
@@ -576,8 +580,8 @@ flattenScope ms scope =
     imported = Map.unionsWith (++)
                [ qual c (build ms' exportedNamesInScope $ moduleScope a)
                | (c, a) <- Map.toList $ scopeImports root
-               , let m   = C.qnameParts c
-                     ms' = map (drop (length m)) $ filter (m `isPrefixOf`) ms
+               , let -- get the suffixes of c in ms
+                     ms' = mapMaybe (maybePrefixMatch $ C.qnameParts c) ms
                , not $ null ms' ]
     qual c = Map.mapKeys (q c)
       where
@@ -585,11 +589,10 @@ flattenScope ms scope =
         q (C.Qual m x) = C.Qual m . q x
 
     build :: [[C.Name]] -> (forall a. InScope a => Scope -> ThingsInScope a) -> Scope -> Map C.QName [AbstractName]
-    build ms getNames s =
-      Map.unionWith (++)
-        (Map.mapKeys (\x -> C.QName x) (getNames s))
-        $ Map.unionsWith (++) $
-          [ Map.mapKeys (\y -> C.Qual x y) $ build ms' exportedNamesInScope $ moduleScope m
+    build ms getNames s = Map.unionsWith (++) $
+        (Map.mapKeysMonotonic C.QName $ getNames s) :
+          [ Map.mapKeysMonotonic (\ y -> C.Qual x y) $
+              build ms' exportedNamesInScope $ moduleScope m
           | (x, mods) <- Map.toList (getNames s)
           , let ms' = [ tl | hd:tl <- ms, hd == x ]
           , not $ null ms'
@@ -603,7 +606,7 @@ scopeLookup :: InScope a => C.QName -> ScopeInfo -> [a]
 scopeLookup q scope = map fst $ scopeLookup' q scope
 
 scopeLookup' :: forall a. InScope a => C.QName -> ScopeInfo -> [(a, Access)]
-scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ topImports ++ imports
+scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ maybeToList topImports ++ imports
   where
 
     -- 1. Finding a name in the current scope and its parents.
@@ -639,11 +642,11 @@ scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ topImports ++
         -- trace ("mods ++ defs = " ++ show (mods ++ defs)) $ do
         m <- nub $ mods ++ defs -- record types will appear both as a mod and a def
         -- Get the scope of module m, if any, and remove its private definitions.
-        let ss  = maybeToList $ Map.lookup m $ scopeModules scope
+        let ss  = Map.lookup m $ scopeModules scope
             ss' = restrictPrivate <$> ss
         -- trace ("ss  = " ++ show ss ) $ do
         -- trace ("ss' = " ++ show ss') $ do
-        s' <- ss'
+        s' <- maybeToList ss'
         findName q s'
       where
         lookupName :: forall a. InScope a => C.Name -> Scope -> [(a, Access)]
@@ -651,21 +654,21 @@ scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ topImports ++
 
     -- 2. Finding a name in the top imports.
 
-    topImports :: [(a, Access)]
+    topImports :: Maybe (a, Access)
     topImports = case (inScopeTag :: InScopeTag a) of
-      NameTag   -> []
-      ModuleTag -> map (first (`AbsModule` Defined)) (imported q)
+      NameTag   -> Nothing
+      ModuleTag -> first (`AbsModule` Defined) <$> imported q
 
-    imported :: C.QName -> [(A.ModuleName, Access)]
-    imported q = map (,PublicAccess) $ maybeToList $ Map.lookup q $ scopeImports root
+    imported :: C.QName -> Maybe (A.ModuleName, Access)
+    imported q = fmap (,PublicAccess) $ Map.lookup q $ scopeImports root
 
     -- 3. Finding a name in the imports belonging to an initial part of the qualifier.
 
     imports :: [(a, Access)]
     imports = do
       (m, x) <- splitName q
-      m <- fst <$> imported m
-      findName x (restrictPrivate $ moduleScope m)
+      m <- maybeToList $ fst <$> imported m
+      findName x $ restrictPrivate $ moduleScope m
 
     -- return all possible splittings, e.g.
     -- splitName X.Y.Z = [(X, Y.Z), (X.Y, Z)]
@@ -680,15 +683,19 @@ scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ topImports ++
 data AllowAmbiguousConstructors = AllowAmbiguousConstructors | NoAmbiguousConstructors
   deriving (Eq)
 
--- | Find the shortest concrete name that maps (uniquely) to a given abstract
---   name.
-inverseScopeLookup :: Either A.ModuleName A.QName -> ScopeInfo -> Maybe C.QName
+-- | Find the concrete names that map (uniquely) to a given abstract name.
+--   Sort by length, shortest first.
+
+inverseScopeLookup :: Either A.ModuleName A.QName -> ScopeInfo -> [C.QName]
 inverseScopeLookup = inverseScopeLookup' AllowAmbiguousConstructors
 
-inverseScopeLookup' :: AllowAmbiguousConstructors -> Either A.ModuleName A.QName -> ScopeInfo -> Maybe C.QName
-inverseScopeLookup' ambCon name scope = case name of
-  Left  m -> best $ filter unambiguousModule $ findModule m
-  Right q -> best $ filter unambiguousName   $ findName nameMap q
+inverseScopeLookup' :: AllowAmbiguousConstructors -> Either A.ModuleName A.QName -> ScopeInfo -> [C.QName]
+inverseScopeLookup' ambCon name scope = billToPure [ Scoping , InverseScopeLookup ] $
+  -- trace ("importMap = " ++ show importMap) $
+  -- trace ("moduleMap = " ++ show moduleMap) $
+  case name of
+    Left  m -> best $ filter unambiguousModule $ findModule m
+    Right q -> best $ filter unambiguousName   $ findName nameMap q
   where
     this = scopeCurrent scope
     current = this : scopeParents (moduleScope this)
@@ -704,9 +711,16 @@ inverseScopeLookup' ambCon name scope = case name of
     len (C.QName _)  = 1
     len (C.Qual _ x) = 1 + len x
 
-    best :: [C.QName] -> Maybe C.QName
-    best xs = headMaybe $ sortBy (compare `on` len) xs
+    best :: [C.QName] -> [C.QName]
+    best = sortBy (compare `on` len) . filter (not . internalName)
 
+    internalName :: C.QName -> Bool
+    internalName C.QName{} = False
+    internalName (C.Qual m n) = intern m || internalName n
+      where
+      -- Recognize fresh names created Parser.y
+      intern (C.Name _ [ C.Id ('.' : '#' : _) ]) = True
+      intern _ = False
     unique :: forall a . [a] -> Bool
     unique []      = __IMPOSSIBLE__
     unique [_]     = True
@@ -729,32 +743,34 @@ inverseScopeLookup' ambCon name scope = case name of
     findModule q = findName moduleMap q ++
                    fromMaybe [] (Map.lookup q importMap)
 
-    importMap = Map.unionsWith (++) $ do
+    importMap = Map.fromListWith (++) $ do
       (m, s) <- scopes
       (x, y) <- Map.toList $ scopeImports s
-      return $ Map.singleton y [x]
+      return (y, [x])
 
-    moduleMap = Map.unionsWith (++) $ do
+    moduleMap = Map.fromListWith (++) $ do
       (m, s)  <- scopes
       (x, ms) <- Map.toList (allNamesInScope s)
       q       <- amodName <$> ms
-      return $ Map.singleton q [(m, x)]
+      return (q, [(m, x)])
 
-    nameMap = Map.unionsWith (++) $ do
+    nameMap = Map.fromListWith (++) $ do
       (m, s)  <- scopes
       (x, ms) <- Map.toList (allNamesInScope s)
       q       <- anameName <$> ms
-      return $ Map.singleton q [(m, x)]
+      return (q, [(m, x)])
 
--- | Takes the first component of 'inverseScopeLookup'.
-inverseScopeLookupName :: A.QName -> ScopeInfo -> Maybe C.QName
+-- | Find the concrete names that map (uniquely) to a given abstract qualified name.
+--   Sort by length, shortest first.
+inverseScopeLookupName :: A.QName -> ScopeInfo -> [C.QName]
 inverseScopeLookupName x = inverseScopeLookup (Right x)
 
-inverseScopeLookupName' :: AllowAmbiguousConstructors -> A.QName -> ScopeInfo -> Maybe C.QName
+inverseScopeLookupName' :: AllowAmbiguousConstructors -> A.QName -> ScopeInfo -> [C.QName]
 inverseScopeLookupName' ambCon x = inverseScopeLookup' ambCon (Right x)
 
--- | Takes the second component of 'inverseScopeLookup'.
-inverseScopeLookupModule :: A.ModuleName -> ScopeInfo -> Maybe C.QName
+-- | Find the concrete names that map (uniquely) to a given abstract module name.
+--   Sort by length, shortest first.
+inverseScopeLookupModule :: A.ModuleName -> ScopeInfo -> [C.QName]
 inverseScopeLookupModule x = inverseScopeLookup (Left x)
 
 ------------------------------------------------------------------------
@@ -786,7 +802,7 @@ instance Show NameSpace where
 instance Show Scope where
   show (scope@Scope{ scopeName = name, scopeParents = parents, scopeImports = imps }) =
     unlines $
-      [ "* scope " ++ show name ] ++ ind (
+      [ "scope " ++ show name ] ++ ind (
         concat [ blockOfLines (show nsid) (lines $ show $ scopeNameSpace nsid scope)
                | nsid <- [minBound..maxBound] ]
       ++ blockOfLines "imports"  (case Map.keys imps of
