@@ -1,4 +1,9 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE NondecreasingIndentation #-}
+
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
 
 {-| The scope monad with operations.
 -}
@@ -16,6 +21,8 @@ import Data.List as List
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Maybe
+import Data.Set (Set)
+import qualified Data.Set as Set
 import Data.Traversable
 
 import Agda.Syntax.Common
@@ -36,6 +43,7 @@ import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Null (unlessNull)
 import Agda.Utils.Size
+import Agda.Utils.Pretty
 import Agda.Utils.Tuple
 
 #include "undefined.h"
@@ -93,11 +101,14 @@ getCurrentScope = getNamedScope =<< getCurrentModule
 -- | Create a new module with an empty scope (Bool is True if it is a datatype module)
 createModule :: Bool -> A.ModuleName -> ScopeM ()
 createModule b m = do
+  reportSLn "scope.createModule" 10 $ "createModule " ++ prettyShow m
   s <- getCurrentScope
   let parents = scopeName s : scopeParents s
-  modifyScopes $ Map.insert m emptyScope { scopeName           = m
-                                         , scopeParents        = parents
-                                         , scopeDatatypeModule = b }
+      sm = emptyScope { scopeName           = m
+                      , scopeParents        = parents
+                      , scopeDatatypeModule = b }
+  -- Andreas, 2015-07-02: internall error if module is not new.
+  modifyScopes $ Map.insertWith __IMPOSSIBLE__ m sm
 
 -- | Apply a function to the scope info.
 modifyScopeInfo :: (ScopeInfo -> ScopeInfo) -> ScopeM ()
@@ -197,42 +208,64 @@ data ResolvedName = VarName A.Name
                   | ConstructorName [AbstractName]
                   | PatternSynResName AbstractName
                   | UnknownName
-  deriving (Show)
+  deriving (Show, Eq)
 
 -- | Look up the abstract name referred to by a given concrete name.
 resolveName :: C.QName -> ScopeM ResolvedName
-resolveName = resolveName' allKindsOfNames
+resolveName = resolveName' allKindsOfNames Nothing
 
 -- | Look up the abstract name corresponding to a concrete name of
---   a certain kind.
+--   a certain kind and/or from a given set of names.
 --   Sometimes we know already that we are dealing with a constructor
 --   or pattern synonym (e.g. when we have parsed a pattern).
 --   Then, we can ignore conflicting definitions of that name
 --   of a different kind. (See issue 822.)
-resolveName' :: [KindOfName] -> C.QName -> ScopeM ResolvedName
-resolveName' kinds x = do
+resolveName' ::
+  [KindOfName] -> Maybe (Set A.Name) -> C.QName -> ScopeM ResolvedName
+resolveName' kinds names x = do
   scope <- getScope
-  let vars = AssocList.mapKeysMonotonic C.QName $ scopeLocals scope
+  let vars     = AssocList.mapKeysMonotonic C.QName $ scopeLocals scope
+      retVar y = return $ VarName $ y { nameConcrete = unqualify x }
+      aName    = A.qnameName . anameName
   case lookup x vars of
     -- Case: we have a local variable x.
-    Just (LocalVar y)  -> return $ VarName $ y { nameConcrete = unqualify x }
-    -- Case: ... but is shadowed by some imports.
-    Just (ShadowedVar y ys) -> typeError $ AmbiguousName x $ A.qualify_ y : map anameName ys
+    Just (LocalVar y)  -> retVar y
+    -- Case: ... but is (perhaps) shadowed by some imports.
+    Just (ShadowedVar y ys) -> case names of
+      Nothing -> shadowed ys
+      Just ns -> case filter (\y -> aName y `Set.member` ns) ys of
+        [] -> retVar y
+        ys -> shadowed ys
+      where
+      shadowed ys =
+        typeError $ AmbiguousName x $ A.qualify_ y : map anameName ys
     -- Case: we do not have a local variable x.
-    Nothing -> case filter ((`elem` kinds) . anameKind . fst) $ scopeLookup' x scope of
-      [] -> return UnknownName
-      ds | all ((==ConName) . anameKind . fst) ds ->
-        return $ ConstructorName
-               $ map (\ (d, _) -> updateConcreteName d $ unqualify x) ds
-      [(d, a)] | anameKind d == FldName -> return $ FieldName $ updateConcreteName d (unqualify x)
-      [(d, a)] | anameKind d == PatternSynName ->
-                  return $ PatternSynResName (updateConcreteName d $ unqualify x)
-      [(d, a)] -> return $ DefinedName a $ updateConcreteName d (unqualify x)
-      ds  -> typeError $ AmbiguousName x (map (anameName . fst) ds)
+    Nothing -> do
+      -- Consider only names of one of the given kinds
+      let filtKind = filter $ \ y -> anameKind (fst y) `elem` kinds
+      -- Consider only names in the given set of names
+          filtName = filter $ \ y -> maybe True (Set.member (aName (fst y))) names
+      case filtKind $ filtName $ scopeLookup' x scope of
+        [] -> return UnknownName
+
+        ds       | all ((==ConName) . anameKind . fst) ds ->
+          return $ ConstructorName $ map (upd . fst) ds
+
+        [(d, a)] | anameKind d == FldName ->
+          return $ FieldName $ upd d
+
+        [(d, a)] | anameKind d == PatternSynName ->
+          return $ PatternSynResName $ upd d
+
+        [(d, a)] ->
+          return $ DefinedName a $ upd d
+
+        ds -> typeError $ AmbiguousName x (map (anameName . fst) ds)
   where
+  upd d = updateConcreteName d $ unqualify x
   updateConcreteName :: AbstractName -> C.Name -> AbstractName
-  updateConcreteName d@(AbsName { anameName = an@(A.QName { qnameName = qn }) }) x =
-    d { anameName = an { qnameName = qn { nameConcrete = x } } }
+  updateConcreteName d@(AbsName { anameName = A.QName qm qn }) x =
+    d { anameName = A.QName (setRange (getRange x) qm) (qn { nameConcrete = x }) }
 
 -- | Look up a module in the scope.
 resolveModule :: C.QName -> ScopeM AbstractModule
@@ -243,19 +276,25 @@ resolveModule x = do
     []                -> typeError $ NoSuchModule x
     ms                -> typeError $ AmbiguousModule x (map amodName ms)
 
--- | Get the fixity of a name. The name is assumed to be in scope.
-getFixity :: C.QName -> ScopeM Fixity'
-getFixity x = do
-  r <- resolveName x
+-- | Get the notation of a name. The name is assumed to be in scope.
+getNotation
+  :: C.QName
+  -> Set A.Name
+     -- ^ The name must correspond to one of the names in this set.
+  -> ScopeM NewNotation
+getNotation x ns = do
+  r <- resolveName' allKindsOfNames (Just ns) x
   case r of
-    VarName y           -> return $ nameFixity y
-    DefinedName _ d     -> return $ aFixity d
-    FieldName d         -> return $ aFixity d
-    ConstructorName ds  -> return $ chooseFixity $ map aFixity ds
-    PatternSynResName n -> return $ aFixity n
+    VarName y           -> return $ namesToNotation x y
+    DefinedName _ d     -> return $ notation d
+    FieldName d         -> return $ notation d
+    ConstructorName ds  -> case mergeNotations $ map notation ds of
+                             [n] -> return n
+                             _   -> __IMPOSSIBLE__
+    PatternSynResName n -> return $ notation n
     UnknownName         -> __IMPOSSIBLE__
   where
-    aFixity = nameFixity . qnameName . anameName
+    notation = namesToNotation x . qnameName . anameName
 
 -- * Binding names
 
@@ -319,7 +358,7 @@ type WSM = StateT Out ScopeM
 --   alternative to qualification by their module).
 --   (See Issue 836).
 copyScope :: C.QName -> A.ModuleName -> Scope -> ScopeM (Scope, (A.Ren A.ModuleName, A.Ren A.QName))
-copyScope oldc new s = first (inScopeBecause $ Applied oldc) <$> runStateT (copy new s) (Map.empty, Map.empty)
+copyScope oldc new s = first (inScopeBecause $ Applied oldc) <$> runStateT (copy new s) ([], [])
   where
     -- | A memoizing algorithm, the renamings serving as memo structure.
     copy :: A.ModuleName -> Scope -> StateT (A.Ren A.ModuleName, A.Ren A.QName) ScopeM Scope
@@ -334,6 +373,7 @@ copyScope oldc new s = first (inScopeBecause $ Applied oldc) <$> runStateT (copy
                   , scopeParents = scopeParents s0
                   }
       where
+        rnew = getRange new
         new' = killRange new
         newL = A.mnameToList new'
         old  = scopeName s
@@ -351,42 +391,57 @@ copyScope oldc new s = first (inScopeBecause $ Applied oldc) <$> runStateT (copy
             _ -> lensAnameName f d
 
         -- Adding to memo structure.
-        addName x y = modify $ second $ Map.insert x y
-        addMod  x y = modify $ first  $ Map.insert x y
+        addName x y = modify $ second $ ((x, y):)
+        addMod  x y = modify $ first  $ ((x, y):)
 
         -- Querying the memo structure.
-        findName x = Map.lookup x <$> gets snd
-        findMod  x = Map.lookup x <$> gets fst
+        findName x = lookup x <$> gets snd
+        findMod  x = lookup x <$> gets fst
 
         -- Change a binding M.x -> old.M'.y to M.x -> new.M'.y
         renName :: A.QName -> WSM A.QName
         renName x = do
-          lift $ reportSLn "scope.copy" 50 $ "  Copying " ++ show x
-          -- If we've seen it already, just return its copy.
-          (`fromMaybeM` findName x) $ do
-          -- We have not processed this name @x@, so copy it to some @y@.
           -- Check whether we have already seen a module of the same name.
           -- If yes, use its copy as @y@.
           y <- ifJustM (findMod $ qnameToMName x) (return . mnameToQName) $ {- else -} do
             -- First time, generate a fresh name for it.
             i <- lift fresh
             return $ A.qualify new' $ (qnameName x) { nameId = i }
+          lift $ reportSLn "scope.copy" 50 $ "  Copying " ++ show x ++ " to " ++ show y
+          -- Andreas, 2015-08-11 Issue 1619:
+          -- Names copied by a module macro should get the module macro's
+          -- range as declaration range
+          -- (maybe rather the one of the open statement).
+          -- For now, we just set their range
+          -- to the new module name's one, which fixes issue 1619.
+          y <- return $ setRange rnew y
           addName x y
           return y
 
         -- Change a binding M.x -> old.M'.y to M.x -> new.M'.y
         renMod :: A.ModuleName -> WSM A.ModuleName
         renMod x = do
-          -- If we've seen it already, just return its copy.
-          (`fromMaybeM` findMod x) $ do
-          -- We have not processed this name @x@, so copy it to some @y@.
+          -- Andreas, issue 1607:
+          -- If we have already copied this module, return the copy.
+          ifJustM (findMod x) return $ {- else -} do
+
           -- Check whether we have seen it already, yet as  name.
           -- If yes, use its copy as @y@.
           y <- ifJustM (findName $ mnameToQName x) (return . qnameToMName) $ {- else -} do
-             return $ A.mnameFromList $ (newL ++) $ drop (size old) $ A.mnameToList x
+             -- Andreas, Jesper, 2015-07-02: Issue 1597
+             -- Don't blindly drop a prefix of length of the old qualifier.
+             -- If things are imported by open public they do not have the old qualifier
+             -- as prefix.  Those need just to be linked, not copied.
+             -- return $ A.mnameFromList $ (newL ++) $ drop (size old) $ A.mnameToList x
+             caseMaybe (maybePrefixMatch (A.mnameToList old) (A.mnameToList x)) (return x) $ \ suffix -> do
+               return $ A.mnameFromList $ newL ++ suffix
+          -- Andreas, Jesper, 2015-07-02: Issue 1597
+          -- Don't copy a module over itself, it will just be emptied of its contents.
+          if (x == y) then return x else do
           addMod x y
           -- We need to copy the contents of included modules recursively
-          s0 <- lift $ createModule False y >> getNamedScope x
+          lift $ createModule False y
+          s0 <- lift $ getNamedScope x
           s  <- withCurrentModule' y $ copy y s0
           lift $ modifyNamedScope y (const s)
           return y
@@ -426,7 +481,9 @@ openModule_ :: C.QName -> ImportDirective -> ScopeM ()
 openModule_ cm dir = do
   current <- getCurrentModule
   m <- amodName <$> resolveModule cm
-  let acc = namespace current m
+  let acc | not (publicOpen dir)      = PrivateNS
+          | m `isSubModuleOf` current = PublicNS
+          | otherwise                 = ImportedNS
   -- Get the scope exported by module to be opened.
   s <- setScopeAccess acc <$>
         (applyImportDirectiveM cm dir . inScopeBecause (Opened cm) . removeOnlyQualified . restrictPrivate =<< getNamedScope m)
@@ -444,11 +501,6 @@ openModule_ cm dir = do
       Nothing -> x
       Just ys -> shadowLocal ys x
   where
-    namespace m0 m1
-      | not (publicOpen dir)  = PrivateNS
-      | m1 `isSubModuleOf` m0 = PublicNS
-      | otherwise             = ImportedNS
-
     -- Only checks for clashes that would lead to the same
     -- name being exported twice from the module.
     checkForClashes new = when (publicOpen dir) $ do
