@@ -1,9 +1,10 @@
-{-# LANGUAGE CPP                        #-}
-{-# LANGUAGE DeriveFunctor              #-}
-{-# LANGUAGE FlexibleInstances          #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE MultiParamTypeClasses      #-}
-{-# LANGUAGE TypeSynonymInstances       #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 
 -- | The monad for the termination checker.
 --
@@ -13,17 +14,21 @@
 
 module Agda.Termination.Monad where
 
-import Control.Applicative
+import Prelude hiding (null)
+
+import Control.Applicative hiding (empty)
 import Control.Monad.Reader
 import Control.Monad.Writer
 import Control.Monad.State
 
+import Data.Foldable (Foldable)
 import Data.Functor ((<$>))
 import qualified Data.List as List
+import Data.Traversable (Traversable)
 
 import Agda.Interaction.Options
 
-import Agda.Syntax.Abstract (QName,IsProjP(..))
+import Agda.Syntax.Abstract (QName, IsProjP(..), AllNames)
 import Agda.Syntax.Common   (Delayed(..), Induction(..), Dom(..))
 import Agda.Syntax.Internal
 import Agda.Syntax.Literal
@@ -34,15 +39,21 @@ import Agda.Termination.Order (Order,le,unknown)
 import Agda.Termination.RecCheck (anyDefs)
 
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad.Benchmark
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records
+import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
+import Agda.Utils.Function
+import Agda.Utils.Functor
+import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
-import Agda.Utils.Pretty (Pretty)
+import Agda.Utils.Null
+import Agda.Utils.Pretty (Pretty, prettyShow)
 import qualified Agda.Utils.Pretty as P
 import Agda.Utils.VarSet (VarSet)
 import qualified Agda.Utils.VarSet as VarSet
@@ -102,10 +113,14 @@ data TerEnv = TerEnv
   , terDelayed :: Delayed
     -- ^ Are we checking a delayed definition?
   , terMaskArgs :: [Bool]
-    -- ^ Only consider the 'True' arguments for establishing termination.
+    -- ^ Only consider the 'notMasked' 'False' arguments for establishing termination.
   , terMaskResult :: Bool
-    -- ^ Only consider guardedness if 'True'.
-  , terPatterns :: [DeBruijnPat]
+    -- ^ Only consider guardedness if 'False' (not masked).
+  , _terSizeDepth :: Int  -- lazy by intention!
+    -- ^ How many @SIZELT@ relations do we have in the context
+    --   (= clause telescope).  Used to approximate termination
+    --   for metas in call args.
+  , terPatterns :: MaskedDeBruijnPats
     -- ^ The patterns of the clause we are checking.
   , terPatternsRaise :: !Int
     -- ^ Number of additional binders we have gone under
@@ -149,8 +164,9 @@ defaultTerEnv = TerEnv
   , terCurrent                  = __IMPOSSIBLE__ -- needs to be set!
   , terTarget                   = Nothing
   , terDelayed                  = NotDelayed
-  , terMaskArgs                 = repeat True    -- use all arguments
-  , terMaskResult               = True           -- use result
+  , terMaskArgs                 = repeat False   -- use all arguments (mask none)
+  , terMaskResult               = False          -- use result (do not mask)
+  , _terSizeDepth               = __IMPOSSIBLE__ -- needs to be set!
   , terPatterns                 = __IMPOSSIBLE__ -- needs to be set!
   , terPatternsRaise            = 0
   , terGuarded                  = le -- not initially guarded
@@ -170,7 +186,7 @@ class (Functor m, Monad m) => MonadTer m where
 -- | Termination monad.
 
 newtype TerM a = TerM { terM :: ReaderT TerEnv TCM a }
-  deriving (Functor, Applicative, Monad)
+  deriving (Functor, Applicative, Monad, MonadBench Phase)
 
 instance MonadTer TerM where
   terAsk     = TerM $ ask
@@ -292,10 +308,13 @@ terGetMaskResult = terAsks terMaskResult
 terSetMaskResult :: Bool -> TerM a -> TerM a
 terSetMaskResult b = terLocal $ \ e -> e { terMaskResult = b }
 
-terGetPatterns :: TerM DeBruijnPats
-terGetPatterns = raiseDBP <$> terAsks terPatternsRaise <*> terAsks terPatterns
+terGetPatterns :: TerM (MaskedDeBruijnPats)
+terGetPatterns = do
+  n   <- terAsks terPatternsRaise
+  mps <- terAsks terPatterns
+  return $ if n == 0 then mps else map (fmap (fmap (n +))) mps
 
-terSetPatterns :: DeBruijnPats -> TerM a -> TerM a
+terSetPatterns :: MaskedDeBruijnPats -> TerM a -> TerM a
 terSetPatterns ps = terLocal $ \ e -> e { terPatterns = ps }
 
 terRaise :: TerM a -> TerM a
@@ -317,6 +336,11 @@ terUnguarded = terSetGuarded unknown
 terPiGuarded :: TerM a -> TerM a
 terPiGuarded m = ifM terGetGuardingTypeConstructors m $ terUnguarded m
 
+-- | Lens for '_terSizeDepth'.
+
+terSizeDepth :: Lens' Int TerEnv
+terSizeDepth f e = f (_terSizeDepth e) <&> \ i -> e { _terSizeDepth = i }
+
 -- | Lens for 'terUsableVars'.
 
 terGetUsableVars :: TerM VarSet
@@ -343,6 +367,7 @@ terSetUseSizeLt = terModifyUseSizeLt . const
 withUsableVars :: UsableSizeVars a => a -> TerM b -> TerM b
 withUsableVars pats m = do
   vars <- usableSizeVars pats
+  reportSLn "term.size" 20 $ "usableSizeVars = " ++ show vars
   terSetUsableVars vars $ m
 
 -- | Set 'terUseSizeLt' when going under constructor @c@.
@@ -353,12 +378,14 @@ conUseSizeLt c m = do
     (const $ terSetUseSizeLt False m)
 
 -- | Set 'terUseSizeLt' for arguments following projection @q@.
+--   We disregard j<i after a non-coinductive projection.
+--   However, the projection need not be recursive (Issue 1470).
 projUseSizeLt :: QName -> TerM a -> TerM a
-projUseSizeLt q m = isCoinductiveProjection q >>= (`terSetUseSizeLt` m)
--- projUseSizeLt q m = do
---   ifM (liftTCM $ isProjectionButNotCoinductive q)
---     (terSetUseSizeLt False m)
---     (terSetUseSizeLt True  m)
+projUseSizeLt q m = do
+  co <- isCoinductiveProjection False q
+  reportSLn "term.size" 20 $ applyUnless co ("not " ++) $
+    "using SIZELT vars after projection " ++ prettyShow q
+  terSetUseSizeLt co m
 
 -- | For termination checking purposes flat should not be considered a
 --   projection. That is, it flat doesn't preserve either structural order
@@ -393,29 +420,44 @@ isProjectionButNotCoinductive qn = liftTCM $ do
 --
 --      isCoinductiveProjection (Stream.tail) = return True
 --   @
-isCoinductiveProjection :: MonadTCM tcm => QName -> tcm Bool
-isCoinductiveProjection q = liftTCM $ do
+isCoinductiveProjection :: MonadTCM tcm => Bool -> QName -> tcm Bool
+isCoinductiveProjection mustBeRecursive q = liftTCM $ do
+  reportSLn "term.guardedness" 40 $ "checking isCoinductiveProjection " ++ prettyShow q
   flat <- fmap nameOfFlat <$> coinductionKit
   -- yes for ♭
   if Just q == flat then return True else do
-  pdef <- getConstInfo q
-  case isProjection_ (theDef pdef) of
-    Just Projection{ projProper = Just{}, projFromType = r, projIndex = n }
-      -> caseMaybeM (isRecord r) __IMPOSSIBLE__ $ \ rdef -> do
-           -- no for inductive or non-recursive record
-           if recInduction rdef /= Just CoInductive then return False else do
-           if not (recRecursive rdef) then return False else do
-           -- TODO: the following test for recursiveness of a projection should be cached.
-           -- E.g., it could be stored in the @Projection@ component.
-           -- Now check if type of field mentions mutually recursive symbol.
-           -- Get the type of the field by dropping record parameters and record argument.
-           let TelV tel core = telView' (defType pdef)
-               tel' = drop n $ telToList tel
-           -- Check if any recursive symbols appear in the record type.
-           -- Q (2014-07-01): Should we normalize the type?
-           names <- anyDefs (r : recMutual rdef) (map (snd . unDom) tel', core)
-           return $ not $ null names
-    _ -> return False
+    pdef <- getConstInfo q
+    case isProjection_ (theDef pdef) of
+      Just Projection{ projProper = Just{}, projFromType = r, projIndex = n } ->
+        caseMaybeM (isRecord r) __IMPOSSIBLE__ $ \ rdef -> do
+          -- no for inductive or non-recursive record
+          if recInduction rdef /= Just CoInductive then return False else do
+            reportSLn "term.guardedness" 40 $ prettyShow q ++ " is coinductive"
+            if not mustBeRecursive then return True else do
+              reportSLn "term.guardedness" 40 $ prettyShow q ++ " must be recursive"
+              if not (recRecursive rdef) then return False else do
+                reportSLn "term.guardedness" 40 $ prettyShow q ++ " has been declared recursive, doing actual check now..."
+                -- TODO: the following test for recursiveness of a projection should be cached.
+                -- E.g., it could be stored in the @Projection@ component.
+                -- Now check if type of field mentions mutually recursive symbol.
+                -- Get the type of the field by dropping record parameters and record argument.
+                let TelV tel core = telView' (defType pdef)
+                    tel' = drop n $ telToList tel
+                -- Check if any recursive symbols appear in the record type.
+                -- Q (2014-07-01): Should we normalize the type?
+                reportSDoc "term.guardedness" 40 $ sep
+                  [ text "looking for recursive occurrences in"
+                  , prettyTCM (telFromList tel')
+                  , text "and"
+                  , prettyTCM core
+                  ]
+                names <- anyDefs (r : recMutual rdef) (map (snd . unDom) tel', core)
+                reportSDoc "term.guardedness" 40 $
+                  text "found" <+> sep (map prettyTCM names)
+                return $ not $ null names
+      _ -> do
+        reportSLn "term.guardedness" 40 $ prettyShow q ++ " is not a proper projection"
+        return False
 
 
 -- * De Bruijn patterns.
@@ -485,21 +527,61 @@ class UsableSizeVars a where
   usableSizeVars :: a -> TerM VarSet
 
 instance UsableSizeVars DeBruijnPat where
-  usableSizeVars p =
+  usableSizeVars p = do
+    let none = return mempty
     case p of
-      VarDBP i    -> ifM terGetUseSizeLt (return $ VarSet.singleton i) (return $ mempty)
+      VarDBP i    -> ifM terGetUseSizeLt (return $ VarSet.singleton i) {- else -} none
       ConDBP c ps -> conUseSizeLt c $ usableSizeVars ps
-      LitDBP{}    -> return mempty
-      TermDBP{}   -> return mempty
-      ProjDBP{}   -> return mempty
+      LitDBP{}    -> none
+      TermDBP{}   -> none
+      ProjDBP{}   -> none
 
-instance UsableSizeVars [DeBruijnPat] where
+instance UsableSizeVars DeBruijnPats where
   usableSizeVars ps =
     case ps of
       []               -> return mempty
       (ProjDBP q : ps) -> projUseSizeLt q $ usableSizeVars ps
       (p         : ps) -> mappend <$> usableSizeVars p <*> usableSizeVars ps
 
+instance UsableSizeVars (Masked DeBruijnPat) where
+  usableSizeVars (Masked m p) = do
+    let none = return mempty
+    case p of
+      VarDBP i    -> ifM terGetUseSizeLt (return $ VarSet.singleton i) {- else -} none
+      ConDBP c ps -> if m then none else conUseSizeLt c $ usableSizeVars ps
+      LitDBP{}    -> none
+      TermDBP{}   -> none
+      ProjDBP{}   -> none
+
+instance UsableSizeVars MaskedDeBruijnPats where
+  usableSizeVars ps =
+    case ps of
+      []                          -> return mempty
+      (Masked _ (ProjDBP q) : ps) -> projUseSizeLt q $ usableSizeVars ps
+      (p                    : ps) -> mappend <$> usableSizeVars p <*> usableSizeVars ps
+
+-- * Masked patterns (which are not eligible for structural descent, only for size descent)
+
+type MaskedDeBruijnPats = [Masked DeBruijnPat]
+
+data Masked a = Masked
+  { getMask   :: Bool  -- ^ True if thing not eligible for structural descent.
+  , getMasked :: a     -- ^ Thing.
+  } deriving (Eq, Ord, Show, Functor, Foldable, Traversable)
+
+masked :: a -> Masked a
+masked = Masked True
+
+notMasked :: a -> Masked a
+notMasked = Masked False
+
+instance Decoration Masked where
+  traverseF f (Masked m a) = Masked m <$> f a
+
+-- | Print masked things in double parentheses.
+instance PrettyTCM a => PrettyTCM (Masked a) where
+  prettyTCM (Masked m a) = applyWhen m (parens . parens) $ prettyTCM a
+
 -- * Call pathes
 
 -- | The call information is stored as free monoid
@@ -513,12 +595,31 @@ instance UsableSizeVars [DeBruijnPat] where
 --   Performance-wise, I could not see a difference between Set and list.
 
 newtype CallPath = CallPath { callInfos :: [CallInfo] }
-  deriving (Show, Monoid)
+  deriving (Show, Monoid, AllNames)
 
 -- | Only show intermediate nodes.  (Drop last 'CallInfo').
 instance Pretty CallPath where
-  pretty (CallPath cis0) = if List.null cis then P.empty else
+  pretty (CallPath cis0) = if null cis then empty else
     P.hsep (map (\ ci -> arrow P.<+> P.pretty ci) cis) P.<+> arrow
     where
       cis   = init cis0
       arrow = P.text "-->"
+
+-- * Size depth estimation
+
+-- | A very crude way of estimating the @SIZELT@ chains
+--   @i > j > k@ in context.  Returns 3 in this case.
+--   Overapproximates.
+
+-- TODO: more precise analysis, constructing a tree
+-- of relations between size variables.
+terSetSizeDepth :: Telescope -> TerM a -> TerM a
+terSetSizeDepth tel cont = do
+  n <- liftTCM $ sum <$> do
+    forM (telToList tel) $ \ dom -> do
+      a <- reduce $ snd $ unDom dom
+      ifM (isJust <$> isSizeType a) (return 1) {- else -} $
+        case ignoreSharing $ unEl a of
+          MetaV{} -> return 1
+          _       -> return 0
+  terLocal (set terSizeDepth n) cont
