@@ -1,7 +1,6 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE ImplicitParams       #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE ImplicitParams #-}
 
 -- | An Abstract domain of relative sizes, i.e., differences
 --   between size of formal function parameter and function argument
@@ -31,7 +30,7 @@ import Agda.Termination.Semiring (HasZero(..), Semiring)
 import qualified Agda.Termination.Semiring as Semiring
 
 import Agda.Utils.PartialOrd hiding (tests)
-import Agda.Utils.Pretty hiding (empty)
+import Agda.Utils.Pretty
 import Agda.Utils.QuickCheck
 import Agda.Utils.TestHelpers
 
@@ -93,25 +92,6 @@ instance PartialOrd Order where
     (Mat{}  , _      ) -> __IMPOSSIBLE__
     (_      , Mat{}  ) -> __IMPOSSIBLE__
 
--- | Pointwise comparison.
---   Only matrices with the same dimension are comparable.
-instance (Ord i, PartialOrd a) => PartialOrd (Matrix i a) where
-  comparable m n
-    | size m /= size n = POAny
-    | otherwise        = Fold.fold $
-                           zipMatrices onlym onlyn both trivial m n
-    where
-      -- If an element is only in @m@, then its 'Unknown' in @n@
-      -- so it gotten better at best, in any case, not worse.
-      onlym o = POGT
-      -- If an element is only in @n@, then its 'Unknown' in @m@
-      -- so we have strictly less information.
-      onlyn o = POLT
-      both    = comparable
-      -- The zero element of the result sparse matrix is the
-      -- neutral element of the monoid.
-      trivial = (==mempty)
-
 -- | A partial order, aimed at deciding whether a call graph gets
 --   worse during the completion.
 --
