@@ -1,13 +1,18 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE TupleSections #-}
 
 module Agda.TypeChecking.CompiledClause.Compile where
 
+import Prelude hiding (null)
+
 import Data.Maybe
 import Data.Monoid
 import qualified Data.Map as Map
 import Data.List (genericReplicate, nubBy, findIndex)
 import Data.Function
 
+import Debug.Trace
+
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
 import Agda.TypeChecking.CompiledClause
@@ -16,10 +21,14 @@ import Agda.TypeChecking.Coverage.SplitTree
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Pretty (prettyTCM, nest, sep, text)
 
-import Agda.Utils.Functor (($>))
+import Agda.Utils.Functor
+import Agda.Utils.Maybe
+import Agda.Utils.Null
 import Agda.Utils.List
+import Agda.Utils.Pretty (Pretty(..), prettyShow)
+import qualified Agda.Utils.Pretty as P
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -35,7 +44,7 @@ compileClauses ::
   Maybe (QName, Type) -- ^ Translate record patterns and coverage check with given type?
   -> [Clause] -> TCM CompiledClauses
 compileClauses mt cs = do
-  let cls = [(clausePats c, clauseBody c) | c <- cs]
+  let cls = [ Cl (clausePats c) (clauseBody c) | c <- cs ]
   case mt of
     Nothing -> return $ compile cls
     Just (q, t)  -> do
@@ -43,7 +52,7 @@ compileClauses mt cs = do
 
       reportSDoc "tc.cc" 30 $ sep $ do
         (text "clauses patterns  before compilation") : do
-          map (prettyTCM . map unArg . fst) cls
+          map (prettyTCM . map unArg . clPats) cls
       reportSDoc "tc.cc" 50 $ do
         sep [ text "clauses before compilation"
             , (nest 2 . text . show) cs
@@ -56,16 +65,21 @@ compileClauses mt cs = do
       cc <- translateCompiledClauses cc
       return cc
 
-type Cl  = ([I.Arg Pattern], ClauseBody)
+-- | Stripped-down version of 'Agda.Syntax.Internal.Clause'
+--   used in clause compiler.
+data Cl = Cl
+  { clPats :: [I.Arg Pattern]
+  , clBody :: ClauseBody
+  } deriving (Show)
+
+instance Pretty Cl where
+  pretty (Cl ps b) = P.prettyList ps P.<+> P.text "->" P.<+> pretty b
+
 type Cls = [Cl]
 
 compileWithSplitTree :: SplitTree -> Cls -> CompiledClauses
 compileWithSplitTree t cs = case t of
-  SplitAt i ts ->
-    -- the coverage checker does not count dot patterns as variables
-    -- in case trees however, they count as variable patterns
-    let n = i -- countInDotPatterns i cs
-    in  Case n $ compiles ts $ splitOn (length ts == 1) n cs
+  SplitAt i ts -> Case i $ compiles ts $ splitOn (length ts == 1) i cs
         -- if there is just one case, we force expansion of catch-alls
         -- this is needed to generate a sound tree on which we can
         -- collapse record pattern splits
@@ -74,33 +88,28 @@ compileWithSplitTree t cs = case t of
 
   where
     compiles :: SplitTrees -> Case Cls -> Case CompiledClauses
-    compiles ts br@Branches{ conBranches = cons
+    compiles ts br@Branches{ projPatterns = cop
+                           , conBranches = cons
                            , litBranches = lits
-                           , catchAllBranch = Nothing }
-      | Map.null lits = emptyBranches { conBranches = updCons cons }
+                           , catchAllBranch = catchAll }
+      = Branches
+          { projPatterns   = cop
+          , conBranches    = updCons cons
+          , litBranches    = compile <$> lits
+          , catchAllBranch = compile <$> catchAll
+          }
       where
-        updCons = Map.mapWithKey $ \ c cl -> case lookup c ts of
-                    Nothing -> __IMPOSSIBLE__
-                    Just t  -> fmap (compileWithSplitTree t) cl
-    compiles ts br    = fmap compile br
-
-    -- increase split index by number of dot patterns we have skipped
-    countInDotPatterns :: Int -> [Cl] -> Int
-    countInDotPatterns i [] = __IMPOSSIBLE__
-    countInDotPatterns i ((ps, _) : _) = i + loop i (map unArg ps) where
-      loop 0 ps            = 0
-      loop i []            = __IMPOSSIBLE__
-      loop i (DotP{} : ps) = 1 + loop i ps
-      loop i (_      : ps) = loop (i - 1) ps
-
+        updCons = Map.mapWithKey $ \ c cl ->
+         caseMaybe (lookup c ts) compile compileWithSplitTree <$> cl
+         -- When the split tree is finished, we continue with @compile@.
 
 compile :: Cls -> CompiledClauses
 compile cs = case nextSplit cs of
-  Just n  -> Case n $ fmap compile $ splitOn False n cs
-  Nothing -> case map (getBody . snd) cs of
+  Just (isRecP, n)-> Case n $ fmap compile $ splitOn isRecP n cs
+  Nothing -> case map (getBody . clBody) cs of
     -- It's possible to get more than one clause here due to
     -- catch-all expansion.
-    Just t : _  -> Done (map (fmap name) $ fst $ head cs) (shared t)
+    Just t : _  -> Done (map (fmap name) $ clPats $ head cs) (shared t)
     Nothing : _ -> Fail
     []          -> __IMPOSSIBLE__
   where
@@ -112,11 +121,23 @@ compile cs = case nextSplit cs of
 
 -- | Get the index of the next argument we need to split on.
 --   This the number of the first pattern that does a match in the first clause.
-nextSplit :: Cls -> Maybe Int
-nextSplit []          = __IMPOSSIBLE__
-nextSplit ((ps, _):_) = findIndex (not . isVar . unArg) ps
+nextSplit :: Cls -> Maybe (Bool, Int)
+nextSplit []            = __IMPOSSIBLE__
+nextSplit (Cl ps _ : _) = headMaybe $ catMaybes $
+  zipWith (\ p n -> (,n) <$> properSplit (unArg p)) ps [0..]
+
+-- | Is is not a variable pattern?
+--   And if yes, is it a record pattern?
+properSplit :: Pattern -> Maybe Bool
+properSplit (ConP _ cpi _) = Just $ isJust $ conPRecord cpi
+properSplit LitP{}  = Just False
+properSplit ProjP{} = Just False
+properSplit VarP{}  = Nothing
+properSplit DotP{}  = Nothing
 
 -- | Is this a variable pattern?
+--
+--   Maintain invariant: @isVar = isNothing . properSplit@!
 isVar :: Pattern -> Bool
 isVar VarP{}  = True
 isVar DotP{}  = True
@@ -127,15 +148,18 @@ isVar ProjP{} = False
 -- | @splitOn single n cs@ will force expansion of catch-alls
 --   if @single@.
 splitOn :: Bool -> Int -> Cls -> Case Cls
-splitOn single n cs = mconcat $ map (fmap (:[]) . splitC n) $ expandCatchAlls single n cs
+splitOn single n cs = mconcat $ map (fmap (:[]) . splitC n) $
+  -- (\ cs -> trace ("splitting on " ++ show n ++ " after expandCatchAlls " ++ show single ++ ": " ++ prettyShow (P.prettyList cs)) cs) $
+    expandCatchAlls single n cs
 
 splitC :: Int -> Cl -> Case Cl
-splitC n (ps, b) = case unArg p of
-  ProjP d     -> conCase d $ WithArity 0 (ps0 ++ ps1, b)
-  ConP c _ qs -> conCase (conName c) $ WithArity (length qs) (ps0 ++ map (fmap namedThing) qs ++ ps1, b)
-  LitP l      -> litCase l (ps0 ++ ps1, b)
-  VarP{}      -> catchAll (ps, b)
-  DotP{}      -> catchAll (ps, b)
+splitC n (Cl ps b) = case unArg p of
+  ProjP d     -> projCase d $ Cl (ps0 ++ ps1) b
+  ConP c _ qs -> conCase (conName c) $ WithArity (length qs) $
+                   Cl (ps0 ++ map (fmap namedThing) qs ++ ps1) b
+  LitP l      -> litCase l $ Cl (ps0 ++ ps1) b
+  VarP{}      -> catchAll $ Cl ps b
+  DotP{}      -> catchAll $ Cl ps b
   where
     (ps0, p, ps1) = extractNthElement' n ps
 
@@ -155,6 +179,27 @@ splitC n (ps, b) = case unArg p of
 -- they come before/after a record pattern, otherwise we get into
 -- trouble when we want to eliminate splits on records later.
 --
+-- Another example (see Issue 1650):
+-- @
+--   f (x, (y, z)) true  = a
+--   f _           false = b
+-- @
+-- Split tree:
+-- @
+--   0 (first argument of f)
+--    \- 1 (second component of the pair)
+--        \- 3 (last argument of f)
+--            \-- true  -> a
+--             \- false -> b
+-- @
+-- We would like to get the following case tree:
+-- @
+--   case 0 of
+--   _,_ -> case 1 of
+--          _,_ -> case 3 of true  -> a; false -> b
+--          _   -> case 3 of true  -> a; false -> b
+--   _          -> case 3 of true  -> a; false -> b
+-- @
 expandCatchAlls :: Bool -> Int -> Cls -> Cls
 expandCatchAlls single n cs =
   -- Andreas, 2013-03-22
@@ -162,16 +207,16 @@ expandCatchAlls single n cs =
   -- we force expansion
   if single then doExpand =<< cs else
   case cs of
-  _            | all (isCatchAllNth . fst) cs -> cs
-  (ps, b) : cs | not (isCatchAllNth ps) -> (ps, b) : expandCatchAlls False n cs
-               | otherwise -> map (expand ps b) expansions ++ (ps, b) : expandCatchAlls False n cs
+  _            | all (isCatchAllNth . clPats) cs -> cs
+  Cl ps b : cs | not (isCatchAllNth ps) -> Cl ps b : expandCatchAlls False n cs
+               | otherwise -> map (expand ps b) expansions ++ Cl ps b : expandCatchAlls False n cs
   _ -> __IMPOSSIBLE__
   where
     -- In case there is only one branch in the split tree, we expand all
     -- catch-alls for this position
     -- The @expansions@ are collected from all the clauses @cs@ then.
     -- Note: @expansions@ could be empty, so we keep the orignal clause.
-    doExpand c@(ps, b)
+    doExpand c@(Cl ps b)
       | isVar $ unArg $ nth ps = map (expand ps b) expansions ++ [c]
       | otherwise              = [c]
 
@@ -188,19 +233,19 @@ expandCatchAlls single n cs =
     -- These are the cases the wildcard needs to be expanded into.
     expansions = nubBy ((==) `on` (classify . unArg))
                . filter (not . isVar . unArg)
-               . map (nth . fst)
+               . map (nth . clPats)
                $ cs
 
     expand ps b q =
       case unArg q of
-        ConP c mt qs' -> (ps0 ++ [q $> ConP c mt conPArgs] ++ ps1,
-                         substBody n' m (Con c conArgs) b)
+        ConP c mt qs' -> Cl (ps0 ++ [q $> ConP c mt conPArgs] ++ ps1)
+                            (substBody n' m (Con c conArgs) b)
           where
             m        = length qs'
             -- replace all direct subpatterns of q by _
             conPArgs = map (fmap ($> VarP underscore)) qs'
             conArgs  = zipWith (\ q n -> q $> var n) qs' $ downFrom m
-        LitP l -> (ps0 ++ [q $> LitP l] ++ ps1, substBody n' 0 (Lit l) b)
+        LitP l -> Cl (ps0 ++ [q $> LitP l] ++ ps1) (substBody n' 0 (Lit l) b)
         _ -> __IMPOSSIBLE__
       where
         (ps0, rest) = splitAt n ps
@@ -216,7 +261,7 @@ expandCatchAlls single n cs =
 substBody :: Int -> Int -> Term -> ClauseBody -> ClauseBody
 substBody _ _ _ NoBody = NoBody
 substBody 0 m v b = case b of
-  Bind   b -> foldr (.) id (replicate m (Bind . Abs underscore)) $ subst v (absBody $ raise m b)
+  Bind   b -> foldr (.) id (replicate m (Bind . Abs underscore)) $ subst 0 v (absBody $ raise m b)
   _        -> __IMPOSSIBLE__
 substBody n m v b = case b of
   Bind b   -> Bind $ fmap (substBody (n - 1) m v) b
