@@ -15,16 +15,17 @@ import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.LevelConstraints
+import Agda.TypeChecking.SizedTypes
 
-import {-# SOURCE #-} Agda.TypeChecking.Rules.Term (checkExpr, checkArguments')
+import {-# SOURCE #-} Agda.TypeChecking.Rules.Term
 import {-# SOURCE #-} Agda.TypeChecking.Conversion
 import {-# SOURCE #-} Agda.TypeChecking.MetaVars
 import {-# SOURCE #-} Agda.TypeChecking.Empty
--- import {-# SOURCE #-} Agda.TypeChecking.UniversePolymorphism -- RETIRED
 
 import Agda.Utils.Except ( MonadError(throwError) )
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Pretty (prettyShow)
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -166,6 +167,7 @@ solveConstraint_ c0@(Guarded c pid)         = do
   ifM (isProblemSolved pid) (solveConstraint_ c)
                             (addConstraint c0)
 solveConstraint_ (IsEmpty r t)              = isEmptyType r t
+solveConstraint_ (CheckSizeLtSat t)         = checkSizeLtSat t
 solveConstraint_ (UnBlock m)                =
   ifM (isFrozen m) (addConstraint $ UnBlock m) $ do
     inst <- mvInstantiation <$> lookupMeta m
@@ -173,7 +175,7 @@ solveConstraint_ (UnBlock m)                =
     case inst of
       BlockedConst t -> do
         reportSDoc "tc.constr.blocked" 15 $
-          text ("blocked const " ++ show m ++ " :=") <+> prettyTCM t
+          text ("blocked const " ++ prettyShow m ++ " :=") <+> prettyTCM t
         assignTerm m [] t
       PostponedTypeCheckingProblem cl unblock -> enterClosure cl $ \prob -> do
         ifNotM unblock (addConstraint $ UnBlock m) $ do
@@ -196,3 +198,4 @@ checkTypeCheckingProblem :: TypeCheckingProblem -> TCM Term
 checkTypeCheckingProblem p = case p of
   CheckExpr e t                  -> checkExpr e t
   CheckArgs eh ei r args t0 t1 k -> checkArguments' eh ei r args t0 t1 k
+  CheckLambda args body target   -> checkPostponedLambda args body target
