@@ -20,7 +20,7 @@ import Agda.Syntax.Abstract.Name
 
 import Agda.Utils.Monad
 
-#include "../../undefined.h"
+#include "undefined.h"
 import Agda.Utils.Impossible
 
 type SplitTree  = SplitTree'  QName
@@ -93,4 +93,5 @@ instance Arbitrary CName where
   arbitrary = CName <$> elements
     [ "zero", "suc", "nil", "cons", "left", "right", "just", "nothing" ]
 
+testSplitTreePrinting :: IO ()
 testSplitTreePrinting = sample (arbitrary :: Gen (SplitTree' CName))
