@@ -7,6 +7,7 @@ import Control.Applicative ((<$>))
 import Data.Maybe (fromMaybe)
 
 import Agda.Syntax.Common
+import Agda.Syntax.Position
 import Agda.Syntax.Internal as I
 
 import Agda.TypeChecking.Monad
@@ -16,7 +17,7 @@ import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Size
 
-#include "../undefined.h"
+#include "undefined.h"
 import Agda.Utils.Impossible
 
 ---------------------------------------------------------------------------
@@ -125,10 +126,12 @@ isDataOrRecordType d = do
     Record{}   -> return $ Just IsRecord
     _          -> return $ Nothing
 
+-- | Precodition: 'Term' is 'reduce'd.
 isDataOrRecord :: Term -> TCM (Maybe QName)
-isDataOrRecord (Def d _)  = fmap (const d) <$> isDataOrRecordType d
-isDataOrRecord (Shared p) = isDataOrRecord (derefPtr p)
-isDataOrRecord _          = return Nothing
+isDataOrRecord v = do
+  case ignoreSharing v of
+    Def d _ -> fmap (const d) <$> isDataOrRecordType d
+    _       -> return Nothing
 
 getNumberOfParameters :: QName -> TCM (Maybe Nat)
 getNumberOfParameters d = do
