@@ -1,6 +1,7 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+
+-- | Tools for 'DisplayTerm' and 'DisplayForm'.
 
 module Agda.TypeChecking.DisplayForm where
 
@@ -12,6 +13,7 @@ import Data.Traversable (traverse)
 
 import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
 import Agda.Syntax.Internal
+import Agda.Syntax.Scope.Base (inverseScopeLookupName)
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Substitute
@@ -23,6 +25,19 @@ import Agda.Utils.Maybe
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+-- | Convert a 'DisplayTerm' into a 'Term'.
+dtermToTerm :: DisplayTerm -> Term
+dtermToTerm dt = case dt of
+  DWithApp d ds vs -> dtermToTerm d `apply` (map (defaultArg . dtermToTerm) ds ++ vs)
+  DCon c args      -> Con c $ map (fmap dtermToTerm) args
+  DDef f es        -> Def f $ map (fmap dtermToTerm) es
+  DDot v           -> v
+  DTerm v          -> v
+
+-- | Get the arities of all display forms for a name.
+displayFormArities :: QName -> TCM [Int]
+displayFormArities q = map (length . dfPats . openThing) . defDisplay <$> getConstInfo q
+
 -- | Find a matching display form for @q vs@.
 --   In essence this tries to reqwrite @q vs@ with any
 --   display form @q ps --> dt@ and returns the instantiated
@@ -51,7 +66,7 @@ displayForm q vs = do
       , "displayForms: " ++ show dfs
       , "arguments   : " ++ show vs
       , "matches     : " ++ show ms
-      , "result      : " ++ show (foldr (const . Just) Nothing ms)
+      , "result      : " ++ show (headMaybe ms)
       ]
     -- Return the first display form that matches.
     return $ headMaybe ms
@@ -62,16 +77,20 @@ displayForm q vs = do
 --  `catchError` \_ -> return Nothing
 
   where
-    inScope _ _ = True  -- TODO: distinguish between with display forms and other display forms
---     inScope scope d = case hd d of
---       Just h  -> isJust $ inverseScopeLookupName h scope
---       Nothing -> __IMPOSSIBLE__ -- TODO: currently all display forms have heads
     -- 'hd' is only used in the commented-out code for 'inScope' above.
-    -- hd (DTerm (Def x _))    = Just x
-    -- hd (DTerm (Con x _))    = Just $ conName x
-    -- hd (DTerm (Shared p))   = hd (DTerm $ derefPtr p)
-    -- hd (DWithApp d _ _) = hd d
-    -- hd _                 = Nothing
+    inScope _ d | isWithDisplay d = True
+    inScope scope d = case hd d of
+      Just h  -> not . null $ inverseScopeLookupName h scope
+      Nothing -> __IMPOSSIBLE__ -- TODO: currently all display forms have heads
+
+    isWithDisplay DWithApp{} = True
+    isWithDisplay _          = False
+
+    hd (DTerm (Def x _))    = Just x
+    hd (DTerm (Con x _))    = Just $ conName x
+    hd (DTerm (Shared p))   = hd (DTerm $ derefPtr p)
+    hd (DWithApp d _ _) = hd d
+    hd _                 = Nothing
 
 -- | Match a 'DisplayForm' @q ps = v@ against @q vs@.
 --   Return the 'DisplayTerm' @v[us]@ if the match was successful,
