@@ -1,6 +1,8 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ConstraintKinds #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 
 -- | Computing the free variables of a term.
 --
@@ -21,18 +23,15 @@
 
 module Agda.TypeChecking.Free
     ( FreeVars(..)
-    , Free
+    , Free, FreeV, FreeVS
     , IgnoreSorts(..)
-    , freeVars
-    , freeVarsIgnore
-    , allVars
-    , relevantVars
-    , rigidVars
-    , freeIn, isBinderUsed
-    , freeInIgnoringSorts, freeInIgnoringSortAnn
+    , allFreeVars, allRelevantVars, allRelevantVarsIgnoring
+    , freeIn, freeInIgnoringSorts, isBinderUsed
     , relevantIn, relevantInIgnoringSortAnn
     , Occurrence(..)
     , occurrence
+    , closed
+    , freeVars -- only for testing
     ) where
 
 import Control.Applicative hiding (empty)
@@ -40,19 +39,30 @@ import Control.Monad.Reader
 
 import Data.Foldable (foldMap)
 import Data.Monoid
+import Data.IntSet (IntSet)
+import qualified Data.IntSet as Set
+
+import qualified Agda.Benchmarking as Bench
 
 import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
 import Agda.Syntax.Internal
 
+import Agda.TypeChecking.Free.Lazy
+  ( Free'(..) , FreeEnv(..), initFreeEnv
+  , VarOcc(..), IgnoreSorts(..), Variable, SingleVar
+  )
+import qualified Agda.TypeChecking.Free.Lazy as Free
+
 import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Monad
-import Agda.Utils.VarSet (VarSet)
-import qualified Agda.Utils.VarSet as Set
+import Agda.Utils.Singleton
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+type VarSet = IntSet
+
 -- | Free variables of a term, (disjointly) partitioned into strongly and
 --   and weakly rigid variables, flexible variables and irrelevant variables.
 data FreeVars = FV
@@ -74,7 +84,16 @@ data FreeVars = FV
     --   in irrelevant positions.
   , unusedVars        :: VarSet
     -- ^ Variables in 'UnusedArg'uments.
-  }
+  } deriving (Eq, Show)
+
+mapSRV, mapUGV, mapWRV, mapFXV, mapIRV, mapUUV
+  :: (VarSet -> VarSet) -> FreeVars -> FreeVars
+mapSRV f fv = fv { stronglyRigidVars = f $ stronglyRigidVars fv }
+mapUGV f fv = fv { unguardedVars     = f $ unguardedVars     fv }
+mapWRV f fv = fv { weaklyRigidVars   = f $ weaklyRigidVars   fv }
+mapFXV f fv = fv { flexibleVars      = f $ flexibleVars      fv }
+mapIRV f fv = fv { irrelevantVars    = f $ irrelevantVars    fv }
+mapUUV f fv = fv { unusedVars        = f $ unusedVars        fv }
 
 -- | Rigid variables: either strongly rigid, unguarded, or weakly rigid.
 rigidVars :: FreeVars -> VarSet
@@ -102,11 +121,13 @@ data Occurrence
   | Unused
   deriving (Eq,Show)
 
-{- NO LONGER
--- | @occurrence x fv@ ignores irrelevant variables in @fv@
--}
-occurrence :: Nat -> FreeVars -> Occurrence
-occurrence x fv
+-- | Compute an occurrence of a single variable in a piece of internal syntax.
+occurrence :: FreeV a => Nat -> a -> Occurrence
+occurrence x v = occurrenceFV x $ freeVars v
+
+-- | Extract occurrence of a single variable from computed free variables.
+occurrenceFV :: Nat -> FreeVars -> Occurrence
+occurrenceFV x fv
   | x `Set.member` stronglyRigidVars fv = StronglyRigid
   | x `Set.member` unguardedVars     fv = Unguarded
   | x `Set.member` weaklyRigidVars   fv = WeaklyRigid
@@ -124,7 +145,7 @@ flexible fv =
        , flexibleVars      = relevantVars fv
        }
 
--- | Mark rigid variables as non-strongly.  Useful when traversion arguments of variables.
+-- | Mark rigid variables as non-strongly.  Useful when traversing arguments of variables.
 weakly :: FreeVars -> FreeVars
 weakly fv = fv
   { stronglyRigidVars = Set.empty
@@ -132,7 +153,7 @@ weakly fv = fv
   , weaklyRigidVars   = rigidVars fv
   }
 
--- | Mark unguarded variables as strongly rigid.  Useful when traversion arguments of inductive constructors.
+-- | Mark unguarded variables as strongly rigid.  Useful when traversing arguments of inductive constructors.
 strongly :: FreeVars -> FreeVars
 strongly fv = fv
   { stronglyRigidVars = stronglyRigidVars fv `Set.union` unguardedVars fv
@@ -184,9 +205,99 @@ instance Monoid FreeVars where
 delete :: Nat -> FreeVars -> FreeVars
 delete n (FV sv gv rv fv iv uv) = FV (Set.delete n sv) (Set.delete n gv) (Set.delete n rv) (Set.delete n fv) (Set.delete n iv) (Set.delete n uv)
 
--- | @subtractFV n fv@ subtracts $n$ from each free variable in @fv@.
-subtractFV :: Nat -> FreeVars -> FreeVars
-subtractFV n (FV sv gv rv fv iv uv) = FV (Set.subtract n sv) (Set.subtract n gv) (Set.subtract n rv) (Set.subtract n fv) (Set.subtract n iv) (Set.subtract n uv)
+instance Singleton Variable FreeVars where
+  singleton (i, VarOcc o r) = mod (Set.insert i) empty where
+    mod :: (VarSet -> VarSet) -> FreeVars -> FreeVars
+    mod = case (o, r) of
+      (_, Irrelevant) -> mapIRV
+      (_, UnusedArg ) -> mapUUV
+      (Free.StronglyRigid, _) -> mapSRV
+      (Free.Unguarded    , _) -> mapUGV
+      (Free.WeaklyRigid  , _) -> mapWRV
+      (Free.Flexible     , _) -> mapFXV
+
+-- * Collecting free variables.
+
+bench :: a -> a
+bench = Bench.billToPure [ Bench.Typing , Bench.Free ]
+
+type Free a = Free' a Any
+type FreeV a = Free' a FreeVars
+type FreeVS a = Free' a VarSet
+
+-- | Doesn't go inside solved metas, but collects the variables from a
+-- metavariable application @X ts@ as @flexibleVars@.
+{-# SPECIALIZE freeVars :: FreeV a => a -> FreeVars #-}
+freeVars :: (Monoid c, Singleton Variable c, Free' a c) => a -> c
+freeVars = freeVarsIgnore IgnoreNot
+
+{-# SPECIALIZE freeVarsIgnore :: FreeV a => IgnoreSorts -> a -> FreeVars #-}
+freeVarsIgnore :: (Monoid c, Singleton Variable c, Free' a c) => IgnoreSorts -> a -> c
+freeVarsIgnore = runFree singleton
+
+-- Specialization to typical monoids
+{-# SPECIALIZE runFree :: Free' a Any      => SingleVar Any      -> IgnoreSorts -> a -> Any #-}
+{-# SPECIALIZE runFree :: Free' a All      => SingleVar All      -> IgnoreSorts -> a -> All #-}
+{-# SPECIALIZE runFree :: Free' a VarSet   => SingleVar VarSet   -> IgnoreSorts -> a -> VarSet #-}
+{-# SPECIALIZE runFree :: Free' a FreeVars => SingleVar FreeVars -> IgnoreSorts -> a -> FreeVars #-}
+-- Specialization to Term
+{-# SPECIALIZE runFree :: SingleVar Any      -> IgnoreSorts -> Term -> Any #-}
+{-# SPECIALIZE runFree :: SingleVar All      -> IgnoreSorts -> Term -> All #-}
+{-# SPECIALIZE runFree :: SingleVar VarSet   -> IgnoreSorts -> Term -> VarSet #-}
+{-# SPECIALIZE runFree :: SingleVar FreeVars -> IgnoreSorts -> Term -> FreeVars #-}
+runFree :: (Monoid c, Free' a c) => SingleVar c -> IgnoreSorts -> a -> c
+runFree singleton i t = -- bench $  -- Benchmarking is expensive (4% on std-lib)
+  freeVars' t `runReader` (initFreeEnv singleton) { feIgnoreSorts = i }
+
+freeIn'' :: Free a => (VarOcc -> Bool) -> IgnoreSorts -> Nat -> a -> Bool
+freeIn'' test ig x t =
+  getAny $ runFree (\ (y, o) -> Any $ x == y && test o) ig t
+
+-- | @freeIn' = freeIn'' (const True)@
+freeIn' :: Free a => IgnoreSorts -> Nat -> a -> Bool
+freeIn' ig x t =
+  getAny $ runFree (Any . (x ==) . fst) ig t
+
+{-# SPECIALIZE freeIn :: Nat -> Term -> Bool #-}
+freeIn :: Free a => Nat -> a -> Bool
+freeIn = freeIn' IgnoreNot
+
+freeInIgnoringSorts :: Free a => Nat -> a -> Bool
+freeInIgnoringSorts = freeIn' IgnoreAll
+
+freeInIgnoringSortAnn :: Free a => Nat -> a -> Bool
+freeInIgnoringSortAnn = freeIn' IgnoreInAnnotations
+
+relevantInIgnoringSortAnn :: Free a => Nat -> a -> Bool
+relevantInIgnoringSortAnn = freeIn'' (not . irrelevantOrUnused . varRelevance) IgnoreInAnnotations
+
+relevantIn :: Free a => Nat -> a -> Bool
+relevantIn = freeIn'' (not . irrelevantOrUnused . varRelevance) IgnoreAll
+
+-- | Is the variable bound by the abstraction actually used?
+isBinderUsed :: Free a => Abs a -> Bool
+isBinderUsed NoAbs{}   = False
+isBinderUsed (Abs _ x) = 0 `freeIn` x
+
+-- | Is the term entirely closed (no free variables)?
+closed :: Free' a All => a -> Bool
+closed t = getAll $ runFree (const $ All False) IgnoreNot t
+
+-- | Collect all free variables.
+allFreeVars :: Free' a VarSet => a -> VarSet
+allFreeVars = runFree (Set.singleton . fst) IgnoreNot
+
+-- | Collect all relevant free variables, possibly ignoring sorts.
+allRelevantVarsIgnoring :: Free' a VarSet => IgnoreSorts -> a -> VarSet
+allRelevantVarsIgnoring = runFree sg
+  where sg (i, VarOcc _ r) = if irrelevantOrUnused r then Set.empty else Set.singleton i
+
+-- | Collect all relevant free variables.
+allRelevantVars :: Free' a VarSet => a -> VarSet
+allRelevantVars = allRelevantVarsIgnoring IgnoreNot
+
+
+{- OLD
 
 -- | A single unguarded variable.
 singleton :: Nat -> FreeVars
@@ -214,14 +325,6 @@ initFreeConf = FreeConf
   , fcContext     = 0
   }
 
--- | Doesn't go inside solved metas, but collects the variables from a
--- metavariable application @X ts@ as @flexibleVars@.
-freeVars :: Free a => a -> FreeVars
-freeVars t = freeVars' t `runReader` initFreeConf
-
-freeVarsIgnore :: Free a => IgnoreSorts -> a -> FreeVars
-freeVarsIgnore i t = freeVars' t `runReader` initFreeConf{ fcIgnoreSorts = i }
-
 -- | Return type of fold over syntax.
 type FreeT = Reader FreeConf FreeVars
 
@@ -280,6 +383,7 @@ instance Free Sort where
       Type a     -> freeVars' a
       Prop       -> mempty
       Inf        -> mempty
+      SizeUniv   -> mempty
       DLub s1 s2 -> weakly <$> freeVars' (s1, s2)
 
 instance Free Level where
@@ -292,7 +396,7 @@ instance Free PlusLevel where
 instance Free LevelAtom where
   freeVars' l = case l of
     MetaLevel _ vs   -> flexible <$> freeVars' vs
-    NeutralLevel v   -> freeVars' v
+    NeutralLevel _ v -> freeVars' v
     BlockedLevel _ v -> freeVars' v
     UnreducedLevel v -> freeVars' v
 
@@ -336,25 +440,44 @@ instance Free ClauseBody where
 instance Free Clause where
   freeVars' = freeVars' . clauseBody
 
+bench :: a -> a
+bench = Bench.billToPure [ Bench.Typing , Bench.Free ]
+
+-- | Doesn't go inside solved metas, but collects the variables from a
+-- metavariable application @X ts@ as @flexibleVars@.
+freeVars :: Free a => a -> FreeVars
+freeVars t =
+  freeVars' t `runReader` initFreeConf
+
+freeVarsIgnore :: Free a => IgnoreSorts -> a -> FreeVars
+freeVarsIgnore i t =
+  freeVars' t `runReader` initFreeConf{ fcIgnoreSorts = i }
+
 freeIn :: Free a => Nat -> a -> Bool
-freeIn v t = v `Set.member` allVars (freeVars t)
+freeIn v t = bench $
+  v `Set.member` allVars (freeVars t)
 
 freeInIgnoringSorts :: Free a => Nat -> a -> Bool
-freeInIgnoringSorts v t =
+freeInIgnoringSorts v t = bench $
   v `Set.member` allVars (freeVarsIgnore IgnoreAll t)
 
 freeInIgnoringSortAnn :: Free a => Nat -> a -> Bool
-freeInIgnoringSortAnn v t =
+freeInIgnoringSortAnn v t = bench $
   v `Set.member` allVars (freeVarsIgnore IgnoreInAnnotations t)
 
 relevantInIgnoringSortAnn :: Free a => Nat -> a -> Bool
-relevantInIgnoringSortAnn v t =
+relevantInIgnoringSortAnn v t = bench $
   v `Set.member` relevantVars (freeVarsIgnore IgnoreInAnnotations t)
 
 relevantIn :: Free a => Nat -> a -> Bool
-relevantIn v t = v `Set.member` relevantVars (freeVarsIgnore IgnoreAll t)
+relevantIn v t = bench $
+  v `Set.member` relevantVars (freeVarsIgnore IgnoreAll t)
 
 -- | Is the variable bound by the abstraction actually used?
 isBinderUsed :: Free a => Abs a -> Bool
 isBinderUsed NoAbs{}   = False
 isBinderUsed (Abs _ x) = 0 `freeIn` x
+
+-- -}
+-- -}
+-- -}
