@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP           #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE TupleSections #-}
 
 module Agda.TypeChecking.Injectivity where
@@ -193,9 +193,9 @@ useInjectivity cmp a u v = do
             , text "  ps   =" <+> prettyList (map (text . show) ps)
             ]
           -- and this is the order the variables occur in the patterns
-          let ms' = permute (invertP __IMPOSSIBLE__ $ compactP perm) ms
-          let sub = parallelS (reverse ms)
-          margs <- runReaderT (evalStateT (mapM metaElim ps) ms') sub
+          let msAux = permute (invertP __IMPOSSIBLE__ $ compactP perm) ms
+          let sub   = parallelS (reverse ms)
+          margs <- runReaderT (evalStateT (mapM metaElim ps) msAux) sub
           reportSDoc "tc.inj.invert" 20 $ vcat
             [ text "inversion"
             , nest 2 $ vcat
