@@ -1,5 +1,9 @@
 {-# LANGUAGE CPP #-}
 
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
+
 module Agda.TypeChecking.InstanceArguments where
 
 import Control.Applicative
@@ -13,11 +17,13 @@ import Agda.Syntax.Position
 import Agda.Syntax.Scope.Base
 import Agda.Syntax.Internal as I
 
+import Agda.TypeChecking.Errors ()
+import Agda.TypeChecking.Implicit (implicitArgs)
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty
-import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 
 import {-# SOURCE #-} Agda.TypeChecking.Constraints
@@ -29,6 +35,7 @@ import Agda.Utils.Except ( MonadError(catchError, throwError), runExceptT )
 import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Pretty (prettyShow)
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -120,92 +127,92 @@ initializeIFSMeta s t = do
 --   its type again.
 findInScope :: MetaId -> Maybe Candidates -> TCM ()
 findInScope m Nothing = do
-  reportSLn "tc.instance" 20 $ "The type of the FindInScope constraint isn't known, trying to find it again."
-  t <- getMetaType m
-  cands <- initialIFSCandidates t
-  case cands of
-    Nothing -> addConstraint $ FindInScope m Nothing
-    Just {} -> findInScope m cands
+  -- Andreas, 2015-02-07: New metas should be created with range of the
+  -- current instance meta, thus, we set the range.
+  mv <- lookupMeta m
+  setCurrentRange mv $ do
+    reportSLn "tc.instance" 20 $ "The type of the FindInScope constraint isn't known, trying to find it again."
+    t <- getMetaType m
+    cands <- initialIFSCandidates t
+    case cands of
+      Nothing -> addConstraint $ FindInScope m Nothing
+      Just {} -> findInScope m cands
 findInScope m (Just cands) = whenJustM (findInScope' m cands) $ addConstraint . FindInScope m . Just
 
 -- | Result says whether we need to add constraint, and if so, the set of
 --   remaining candidates.
 findInScope' :: MetaId -> Candidates -> TCM (Maybe Candidates)
 findInScope' m cands = ifM (isFrozen m) (return (Just cands)) $ do
-    -- Andreas, 2013-12-28 issue 1003:
-    -- If instance meta is already solved, simply discard the constraint.
-    ifM (isInstantiatedMeta m) (return Nothing) $ do
-    reportSLn "tc.instance" 15 $
-      "findInScope 2: constraint: " ++ show m ++ "; candidates left: " ++ show (length cands)
-    t <- normalise =<< getMetaTypeInContext m
-    reportSDoc "tc.instance" 15 $ text "findInScope 3: t =" <+> prettyTCM t
-    reportSLn "tc.instance" 70 $ "findInScope 3: t: " ++ show t
+  -- Andreas, 2013-12-28 issue 1003:
+  -- If instance meta is already solved, simply discard the constraint.
+  ifM (isInstantiatedMeta m) (return Nothing) $ do
+    -- Andreas, 2015-02-07: New metas should be created with range of the
+    -- current instance meta, thus, we set the range.
     mv <- lookupMeta m
-    -- If there are recursive instances, it's not safe to instantiate
-    -- metavariables in the goal, so we freeze them before checking candidates.
-    -- Metas that are rigidly constrained need not be frozen.
-    isRec <- orM $ map (isRecursive . unEl . snd) cands
-    let shouldFreeze rigid m
-          | elem m rigid = return False
-          | otherwise    = not <$> isFrozen m
-    metas <- if not isRec then return [] else do
-      rigid <- rigidlyConstrainedMetas
-      filterM (shouldFreeze rigid) (allMetas t)
-    forM_ metas $ \ m -> updateMetaVar m $ \ mv -> mv { mvFrozen = Frozen }
-    cands <- checkCandidates m t cands
-    reportSLn "tc.instance" 15 $
-      "findInScope 4: cands left: " ++ show (length cands)
-    unfreezeMeta metas
-    case cands of
+    setCurrentRange mv $ do
+      reportSLn "tc.instance" 15 $
+        "findInScope 2: constraint: " ++ prettyShow m ++ "; candidates left: " ++ show (length cands)
+      t <- normalise =<< getMetaTypeInContext m
+      reportSDoc "tc.instance" 15 $ text "findInScope 3: t =" <+> prettyTCM t
+      reportSLn "tc.instance" 70 $ "findInScope 3: t: " ++ show t
+      -- If there are recursive instances, it's not safe to instantiate
+      -- metavariables in the goal, so we freeze them before checking candidates.
+      -- Metas that are rigidly constrained need not be frozen.
+      isRec <- orM $ map (isRecursive . unEl . snd) cands
+      let shouldFreeze rigid m
+            | elem m rigid = return False
+            | otherwise    = not <$> isFrozen m
+      metas <- if not isRec then return [] else do
+        rigid <- rigidlyConstrainedMetas
+        filterM (shouldFreeze rigid) (allMetas t)
+      forM_ metas $ \ m -> updateMetaVar m $ \ mv -> mv { mvFrozen = Frozen }
+      cands <- checkCandidates m t cands
+      reportSLn "tc.instance" 15 $
+        "findInScope 4: cands left: " ++ show (length cands)
+      unfreezeMeta metas
+      case cands of
 
-      [] -> do
-        reportSDoc "tc.instance" 15 $
-          text "findInScope 5: not a single candidate found..."
-        typeError $ IFSNoCandidateInScope t
+        [] -> do
+          reportSDoc "tc.instance" 15 $
+            text "findInScope 5: not a single candidate found..."
+          typeError $ IFSNoCandidateInScope t
 
-      [(term, t')] -> do
-        reportSDoc "tc.instance" 15 $ vcat
-          [ text "findInScope 5: found one candidate"
-          , nest 2 $ prettyTCM term
-          , text "of type " <+> prettyTCM t'
-          , text "for type" <+> prettyTCM t
-          ]
+        [(term, t')] -> do
+          reportSDoc "tc.instance" 15 $ vcat
+            [ text "findInScope 5: found one candidate"
+            , nest 2 $ prettyTCM term
+            , text "of type " <+> prettyTCM t'
+            , text "for type" <+> prettyTCM t
+            ]
 
-        -- if t' takes initial hidden arguments, apply them
-        ca <- liftTCM $ runExceptT $ checkArguments ExpandLast ExpandInstanceArguments (getRange mv) [] t' t
-        case ca of
-          Left _ -> __IMPOSSIBLE__
-          Right (args, t'') -> do
-            -- @args@ are the hidden arguments @t'@ takes, @t''@ is @t' `apply` args@
-{- TODO
-        (args, t'') <- implicitArgs (...) t'
-        do
--}
-            leqType t'' t
-            ctxArgs <- getContextArgs
-            v <- (`applyDroppingParameters` args) =<< reduce term
-            assignV DirEq m ctxArgs v
-            reportSDoc "tc.instance" 10 $ vcat
-              [ text "solved by instance search:"
-              , prettyTCM m <+> text ":=" <+> prettyTCM v
-              ]
-            return Nothing
+          -- If t' takes initial hidden and instance arguments, apply them.
+          -- Taking also instance arguments facilitates recursive instance search.
+          (args, t'') <- implicitArgs (-1) notVisible t'
+          leqType t'' t
+          ctxArgs <- getContextArgs
+          v <- (`applyDroppingParameters` args) =<< reduce term
+          assignV DirEq m ctxArgs v
+          reportSDoc "tc.instance" 10 $ vcat
+            [ text "solved by instance search:"
+            , prettyTCM m <+> text ":=" <+> prettyTCM v
+            ]
+          return Nothing
 
-      cs -> do
-        reportSDoc "tc.instance" 15 $
-          text ("findInScope 5: more than one candidate found: ") <+>
-          prettyTCM (List.map fst cs)
-        return (Just cs)
-    where
-      -- | Check whether a type is a function type with an instance domain.
-      isRecursive :: Term -> TCM Bool
-      isRecursive v = do
-        v <- reduce v
-        case ignoreSharing v of
-          Pi (Dom info _) t ->
-            if getHiding info == Instance then return True else
-              isRecursive $ unEl $ unAbs t
-          _ -> return False
+        cs -> do
+          reportSDoc "tc.instance" 15 $
+            text ("findInScope 5: more than one candidate found: ") <+>
+            prettyTCM (List.map fst cs)
+          return (Just cs)
+      where
+        -- | Check whether a type is a function type with an instance domain.
+        isRecursive :: Term -> TCM Bool
+        isRecursive v = do
+          v <- reduce v
+          case ignoreSharing v of
+            Pi (Dom info _) t ->
+              if getHiding info == Instance then return True else
+                isRecursive $ unEl $ unAbs t
+            _ -> return False
 
 -- | A meta _M is rigidly constrained if there is a constraint _M us == D vs,
 -- for inert D. Such metas can safely be instantiated by recursive instance
@@ -243,6 +250,7 @@ rigidlyConstrainedMetas = do
         UnBlock{}     -> return Nothing
         Guarded{}     -> return Nothing  -- don't look inside Guarded, since the inner constraint might not fire
         IsEmpty{}     -> return Nothing
+        CheckSizeLtSat{} -> return Nothing
         FindInScope{} -> return Nothing
 
 -- | Given a meta @m@ of type @t@ and a list of candidates @cands@,
@@ -257,48 +265,53 @@ checkCandidates m t cands = localTCState $ disableDestructiveUpdate $ do
   dropSameCandidates cands
   where
     checkCandidateForMeta :: MetaId -> Type -> Term -> Type -> TCM Bool
-    checkCandidateForMeta m t term t' =
-      verboseBracket "tc.instance" 20 ("checkCandidateForMeta " ++ show m) $ do
-      liftTCM $ flip catchError handle $ do
-        reportSLn "tc.instance" 70 $ "  t: " ++ show t ++ "\n  t':" ++ show t' ++ "\n  term: " ++ show term ++ "."
-        reportSDoc "tc.instance" 20 $ vcat
-          [ text "checkCandidateForMeta"
-          , text "t    =" <+> prettyTCM t
-          , text "t'   =" <+> prettyTCM t'
-          , text "term =" <+> prettyTCM term
-          ]
-        localTCState $ do
-           -- domi: we assume that nothing below performs direct IO (except
-           -- for logging and such, I guess)
-          ca <- runExceptT $ checkArguments ExpandLast ExpandInstanceArguments noRange [] t' t
-          case ca of
-            Left _ -> return False
-            Right (args, t'') -> do
+    checkCandidateForMeta m t term t' = do
+      -- Andreas, 2015-02-07: New metas should be created with range of the
+      -- current instance meta, thus, we set the range.
+      mv <- lookupMeta m
+      setCurrentRange mv $ do
+        verboseBracket "tc.instance" 20 ("checkCandidateForMeta " ++ prettyShow m) $ do
+          liftTCM $ flip catchError handle $ do
+            reportSLn "tc.instance" 70 $ "  t: " ++ show t ++ "\n  t':" ++ show t' ++ "\n  term: " ++ show term ++ "."
+            reportSDoc "tc.instance" 20 $ vcat
+              [ text "checkCandidateForMeta"
+              , text "t    =" <+> prettyTCM t
+              , text "t'   =" <+> prettyTCM t'
+              , text "term =" <+> prettyTCM term
+              ]
+            -- domi: we assume that nothing below performs direct IO
+            -- (except for logging and such, I guess)
+            localTCState $ do
+
+              -- Apply hidden and instance arguments (recursive inst. search!).
+              (args, t'') <- implicitArgs (-1) notVisible t'
+
               reportSDoc "tc.instance" 20 $
                 text "instance search: checking" <+> prettyTCM t''
                 <+> text "<=" <+> prettyTCM t
               -- if constraints remain, we abort, but keep the candidate
               flip (ifNoConstraints_ $ leqType t'' t) (const $ return True) $ do
-              --tel <- getContextTelescope
-              ctxArgs <- getContextArgs
-              v <- (`applyDroppingParameters` args) =<< reduce term
-              reportSDoc "tc.instance" 15 $ vcat
-                [ text "instance search: attempting"
-                , nest 2 $ prettyTCM m <+> text ":=" <+> prettyTCM v
-                ]
-              assign DirEq m ctxArgs v
---              assign m ctxArgs (term `apply` args)
-              -- make a pass over constraints, to detect cases where some are made
-              -- unsolvable by the assignment, but don't do this for FindInScope's
-              -- to prevent loops. We currently also ignore UnBlock constraints
-              -- to be on the safe side.
-              solveAwakeConstraints' True
-              return True
-      where
-        handle err = do
-          reportSDoc "tc.instance" 50 $
-            text "assignment failed:" <+> prettyTCM err
-          return False
+                ctxArgs <- getContextArgs
+                v <- (`applyDroppingParameters` args) =<< reduce term
+                reportSDoc "tc.instance" 15 $ vcat
+                  [ text "instance search: attempting"
+                  , nest 2 $ prettyTCM m <+> text ":=" <+> prettyTCM v
+                  ]
+                assign DirEq m ctxArgs v
+      --          assign m ctxArgs (term `apply` args)
+                -- make a pass over constraints, to detect cases where some are made
+                -- unsolvable by the assignment, but don't do this for FindInScope's
+                -- to prevent loops. We currently also ignore UnBlock constraints
+                -- to be on the safe side.
+                solveAwakeConstraints' True
+                return True
+          where
+            handle :: TCErr -> TCM Bool
+            handle err = do
+              reportSDoc "tc.instance" 50 $
+                text "assignment failed:" <+> prettyTCM err
+              return False
+
     isIFSConstraint :: Constraint -> Bool
     isIFSConstraint FindInScope{} = True
     isIFSConstraint UnBlock{}     = True -- otherwise test/fail/Issue723 loops
