@@ -1,9 +1,8 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE RelaxedPolyRec       #-}
-{-# LANGUAGE TupleSections        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE RelaxedPolyRec #-}
+{-# LANGUAGE TupleSections #-}
 
 module Agda.TypeChecking.MetaVars where
 
@@ -42,8 +41,10 @@ import Agda.TypeChecking.SizedTypes (boundedSizeMetaHook, isSizeProblem)
 import Agda.TypeChecking.MetaVars.Occurs
 
 import Agda.Utils.Except
-  ( Error(noMsg)
-  , ExceptT
+  ( ExceptT
+#if !MIN_VERSION_transformers(0,4,1)
+  , Error(noMsg)
+#endif
   , MonadError(throwError)
   , runExceptT
   )
@@ -54,6 +55,7 @@ import Agda.Utils.Monad
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty (prettyShow)
 import qualified Agda.Utils.VarSet as Set
 
 #include "undefined.h"
@@ -70,7 +72,7 @@ findIdx vs v = findIndex (==v) (reverse vs)
 -- | Check whether a meta variable is a place holder for a blocked term.
 isBlockedTerm :: MetaId -> TCM Bool
 isBlockedTerm x = do
-    reportSLn "tc.meta.blocked" 12 $ "is " ++ show x ++ " a blocked term? "
+    reportSLn "tc.meta.blocked" 12 $ "is " ++ prettyShow x ++ " a blocked term? "
     i <- mvInstantiation <$> lookupMeta x
     let r = case i of
             BlockedConst{}                 -> True
@@ -110,7 +112,7 @@ assignTerm x tel v = do
 -- | Skip frozen check.  Used for eta expanding frozen metas.
 assignTerm' :: MetaId -> [I.Arg ArgName] -> Term -> TCM ()
 assignTerm' x tel v = do
-    reportSLn "tc.meta.assign" 70 $ show x ++ " := " ++ show v ++ "\n  in " ++ show tel
+    reportSLn "tc.meta.assign" 70 $ prettyShow x ++ " := " ++ show v ++ "\n  in " ++ show tel
      -- verify (new) invariants
     whenM (not <$> asks envAssignMetas) __IMPOSSIBLE__
 
@@ -133,7 +135,7 @@ assignTerm' x tel v = do
     modifyMetaStore $ ins x i
     etaExpandListeners x
     wakeupConstraints x
-    reportSLn "tc.meta.assign" 20 $ "completed assignment of " ++ show x
+    reportSLn "tc.meta.assign" 20 $ "completed assignment of " ++ prettyShow x
   where
     metaInstance tel v = InstV tel v
     ins x i store = Map.adjust (inst i) x store
@@ -201,7 +203,7 @@ newIFSMetaCtx s t vs cands = do
   let perm = idP (size tel)
   x <- newMeta' OpenIFS i normalMetaPriority perm (HasType () t)
   reportSDoc "tc.meta.new" 50 $ fsep
-    [ nest 2 $ text (show x) <+> text ":" <+> prettyTCM t
+    [ nest 2 $ pretty x <+> text ":" <+> prettyTCM t
     ]
   addConstraint $ FindInScope x cands
   return $ MetaV x $ map Apply vs
@@ -241,7 +243,7 @@ newValueMetaCtx' b t vs = do
   reportSDoc "tc.meta.new" 50 $ fsep
     [ text "new meta:"
     , nest 2 $ prettyTCM vs <+> text "|-"
-    , nest 2 $ text (show x) <+> text ":" <+> prettyTCM t
+    , nest 2 $ pretty x <+> text ":" <+> prettyTCM t
     ]
   etaExpandMetaSafe x
   -- Andreas, 2012-09-24: for Metas X : Size< u add constraint X+1 <= u
@@ -250,7 +252,7 @@ newValueMetaCtx' b t vs = do
   return u
 
 newTelMeta :: Telescope -> TCM Args
-newTelMeta tel = newArgsMeta (abstract tel $ El Prop $ Sort Prop)
+newTelMeta tel = newArgsMeta (abstract tel $ typeDontCare)
 
 type Condition = I.Dom Type -> Abs Type -> Bool
 
@@ -339,6 +341,7 @@ blockTermOnProblem t v pid =
         -- that we compare against the blocked term once it's unblocked. This way
         -- blocked terms can be instantiated before they are unblocked, thus making
         -- constraint solving a bit more robust against instantiation order.
+        -- Andreas, 2015-05-22: DontRunMetaOccursCheck to avoid Issue585-17.
         v   <- newValueMeta DontRunMetaOccursCheck t
         i   <- liftTCM fresh
         -- This constraint is woken up when unblocking, so it doesn't need a problem id.
@@ -362,7 +365,8 @@ postponeTypeCheckingProblem_ p = do
   postponeTypeCheckingProblem p (unblock p)
   where
     unblock (CheckExpr _ t)           = unblockedTester t
-    unblock (CheckArgs _ _ _ _ t _ _) = unblockedTester t
+    unblock (CheckArgs _ _ _ _ t _ _) = unblockedTester t  -- The type of the head of the application.
+    unblock (CheckLambda _ _ t)       = unblockedTester t
 
 -- | Create a postponed type checking problem @e : t@ that waits for conditon
 --   @unblock@.  A new meta is created in the current context that has as
@@ -393,9 +397,11 @@ postponeTypeCheckingProblem p unblock = do
   addConstraint (UnBlock m)
   return v
 
+-- | Type of the term that is produced by solving the 'TypeCheckingProblem'.
 problemType :: TypeCheckingProblem -> Type
-problemType (CheckExpr _ t) = t
-problemType (CheckArgs _ _ _ _ _ t _) = t
+problemType (CheckExpr _ t           ) = t
+problemType (CheckArgs _ _ _ _ _ t _ ) = t  -- The target type of the application.
+problemType (CheckLambda _ _ t       ) = t
 
 -- | Eta expand metavariables listening on the current meta.
 etaExpandListeners :: MetaId -> TCM ()
@@ -437,59 +443,62 @@ allMetaKinds = [minBound .. maxBound]
 --   Don't do anything if the metavariable is a blocked term.
 etaExpandMeta :: [MetaKind] -> MetaId -> TCM ()
 etaExpandMeta kinds m = whenM (isEtaExpandable m) $ do
-  verboseBracket "tc.meta.eta" 20 ("etaExpandMeta " ++ show m) $ do
-  let waitFor x = do
-        reportSDoc "tc.meta.eta" 20 $ do
-          text "postponing eta-expansion of meta variable" <+>
-            prettyTCM m <+>
-            text "which is blocked by" <+> prettyTCM x
-        listenToMeta (EtaExpand m) x
-      dontExpand = do
-        reportSDoc "tc.meta.eta" 20 $ do
-          text "we do not expand meta variable" <+> prettyTCM m <+>
-            text ("(requested was expansion of " ++ show kinds ++ ")")
-  meta           <- lookupMeta m
-  let HasType _ a = mvJudgement meta
-  TelV tel b     <- telView a
-  -- if the target type @b@ of @m@ is a meta variable @x@ itself
-  -- (@NonBlocked (MetaV{})@),
-  -- or it is blocked by a meta-variable @x@ (@Blocked@), we cannot
-  -- eta expand now, we have to postpone this.  Once @x@ is
-  -- instantiated, we can continue eta-expanding m.  This is realized
-  -- by adding @m@ to the listeners of @x@.
-  ifBlocked (unEl b) (\ x _ -> waitFor x) $ \ t -> case ignoreSharing t of
-    lvl@(Def r es) ->
-      ifM (isEtaRecord r) {- then -} (do
-        let ps = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-        let expand = do
-              u <- withMetaInfo' meta $ newRecordMetaCtx r ps tel $ teleArgs tel
-              inTopContext $ do
-                verboseS "tc.meta.eta" 15 $ do
-                  du <- prettyTCM u
-                  reportSLn "tc.meta.eta" 15 $ "eta expanding: " ++ show m ++ " --> " ++ show du
-                -- Andreas, 2012-03-29: No need for occurrence check etc.
-                -- we directly assign the solution for the meta
-                -- 2012-05-23: We also bypass the check for frozen.
-                noConstraints $ assignTerm' m (telToArgs tel) u  -- should never produce any constraints
-        if Records `elem` kinds then
-          expand
-         else if (SingletonRecords `elem` kinds) then do
-           singleton <- isSingletonRecord r ps
-           case singleton of
-             Left x      -> waitFor x
-             Right False -> dontExpand
-             Right True  -> expand
-          else dontExpand
-      ) $ {- else -} ifM (andM [ return $ Levels `elem` kinds
-                      , typeInType
-                      , (Just lvl ==) <$> getBuiltin' builtinLevel
-                      ]) (do
-        reportSLn "tc.meta.eta" 20 $ "Expanding level meta to 0 (type-in-type)"
-        -- Andreas, 2012-03-30: No need for occurrence check etc.
-        -- we directly assign the solution for the meta
-        noConstraints $ assignTerm m (telToArgs tel) (Level $ Max [])
-     ) $ {- else -} dontExpand
-    _ -> dontExpand
+  verboseBracket "tc.meta.eta" 20 ("etaExpandMeta " ++ prettyShow m) $ do
+    let waitFor x = do
+          reportSDoc "tc.meta.eta" 20 $ do
+            text "postponing eta-expansion of meta variable" <+>
+              prettyTCM m <+>
+              text "which is blocked by" <+> prettyTCM x
+          listenToMeta (EtaExpand m) x
+        dontExpand = do
+          reportSDoc "tc.meta.eta" 20 $ do
+            text "we do not expand meta variable" <+> prettyTCM m <+>
+              text ("(requested was expansion of " ++ show kinds ++ ")")
+    meta           <- lookupMeta m
+    let HasType _ a = mvJudgement meta
+    TelV tel b     <- telView a
+    -- if the target type @b@ of @m@ is a meta variable @x@ itself
+    -- (@NonBlocked (MetaV{})@),
+    -- or it is blocked by a meta-variable @x@ (@Blocked@), we cannot
+    -- eta expand now, we have to postpone this.  Once @x@ is
+    -- instantiated, we can continue eta-expanding m.  This is realized
+    -- by adding @m@ to the listeners of @x@.
+    ifBlocked (unEl b) (\ x _ -> waitFor x) $ \ t -> case ignoreSharing t of
+      lvl@(Def r es) ->
+        ifM (isEtaRecord r) {- then -} (do
+          let ps = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+          let expand = do
+                u <- withMetaInfo' meta $ newRecordMetaCtx r ps tel $ teleArgs tel
+                inTopContext $ do
+                  verboseS "tc.meta.eta" 15 $ do
+                    du <- prettyTCM u
+                    reportSDoc "tc.meta.eta" 15 $ sep
+                      [ text "eta expanding: " <+> pretty m <+> text " --> "
+                      , nest 2 $ prettyTCM u
+                      ]
+                  -- Andreas, 2012-03-29: No need for occurrence check etc.
+                  -- we directly assign the solution for the meta
+                  -- 2012-05-23: We also bypass the check for frozen.
+                  noConstraints $ assignTerm' m (telToArgs tel) u  -- should never produce any constraints
+          if Records `elem` kinds then
+            expand
+           else if (SingletonRecords `elem` kinds) then do
+             singleton <- isSingletonRecord r ps
+             case singleton of
+               Left x      -> waitFor x
+               Right False -> dontExpand
+               Right True  -> expand
+            else dontExpand
+        ) $ {- else -} ifM (andM [ return $ Levels `elem` kinds
+                        , typeInType
+                        , (Just lvl ==) <$> getBuiltin' builtinLevel
+                        ]) (do
+          reportSLn "tc.meta.eta" 20 $ "Expanding level meta to 0 (type-in-type)"
+          -- Andreas, 2012-03-30: No need for occurrence check etc.
+          -- we directly assign the solution for the meta
+          noConstraints $ assignTerm m (telToArgs tel) (Level $ Max [])
+       ) $ {- else -} dontExpand
+      _ -> dontExpand
 
 -- | Eta expand blocking metavariables of record type, and reduce the
 -- blocked thing.
@@ -544,161 +553,161 @@ assignWrapper dir x es v doAssign = do
 assign :: CompareDirection -> MetaId -> Args -> Term -> TCM ()
 assign dir x args v = do
 
-        mvar <- lookupMeta x  -- information associated with meta x
-        let t = jMetaType $ mvJudgement mvar
+  mvar <- lookupMeta x  -- information associated with meta x
+  let t = jMetaType $ mvJudgement mvar
 
-        -- Andreas, 2011-05-20 TODO!
-        -- full normalization  (which also happens during occurs check)
-        -- is too expensive! (see Issue 415)
-        -- need to do something cheaper, especially if
-        -- we are dealing with a Miller pattern that can be solved
-        -- immediately!
-        -- Ulf, 2011-08-25 DONE!
-        -- Just instantiating the top-level meta, which is cheaper. The occurs
-        -- check will first try without unfolding any definitions (treating
-        -- arguments to definitions as flexible), if that fails it tries again
-        -- with full unfolding.
-        v <- instantiate v
-        reportSLn "tc.meta.assign" 50 $ "MetaVars.assign: assigning to " ++ show v
+  -- Andreas, 2011-05-20 TODO!
+  -- full normalization  (which also happens during occurs check)
+  -- is too expensive! (see Issue 415)
+  -- need to do something cheaper, especially if
+  -- we are dealing with a Miller pattern that can be solved
+  -- immediately!
+  -- Ulf, 2011-08-25 DONE!
+  -- Just instantiating the top-level meta, which is cheaper. The occurs
+  -- check will first try without unfolding any definitions (treating
+  -- arguments to definitions as flexible), if that fails it tries again
+  -- with full unfolding.
+  v <- instantiate v
+  reportSLn "tc.meta.assign" 50 $ "MetaVars.assign: assigning to " ++ show v
 
-        case (ignoreSharing v, mvJudgement mvar) of
-            (Sort Inf, HasType{}) -> typeError SetOmegaNotValidType
-            _                     -> return ()
+  case (ignoreSharing v, mvJudgement mvar) of
+      (Sort Inf, HasType{}) -> typeError SetOmegaNotValidType
+      _                     -> return ()
 
-        -- We don't instantiate frozen mvars
-        when (mvFrozen mvar == Frozen) $ do
-          reportSLn "tc.meta.assign" 25 $ "aborting: meta is frozen!"
-          patternViolation
+  -- We don't instantiate frozen mvars
+  when (mvFrozen mvar == Frozen) $ do
+    reportSLn "tc.meta.assign" 25 $ "aborting: meta is frozen!"
+    patternViolation
 
-        -- We never get blocked terms here anymore. TODO: we actually do. why?
-        whenM (isBlockedTerm x) patternViolation
+  -- We never get blocked terms here anymore. TODO: we actually do. why?
+  whenM (isBlockedTerm x) patternViolation
 
-        -- Andreas, 2010-10-15 I want to see whether rhs is blocked
-        reportSLn "tc.meta.assign" 50 $ "MetaVars.assign: I want to see whether rhs is blocked"
-        reportSDoc "tc.meta.assign" 25 $ do
-          v0 <- reduceB v
-          case v0 of
-            Blocked m0 _ -> text "r.h.s. blocked on:" <+> prettyTCM m0
-            NotBlocked{} -> text "r.h.s. not blocked"
+  -- Andreas, 2010-10-15 I want to see whether rhs is blocked
+  reportSLn "tc.meta.assign" 50 $ "MetaVars.assign: I want to see whether rhs is blocked"
+  reportSDoc "tc.meta.assign" 25 $ do
+    v0 <- reduceB v
+    case v0 of
+      Blocked m0 _ -> text "r.h.s. blocked on:" <+> prettyTCM m0
+      NotBlocked{} -> text "r.h.s. not blocked"
 
-        -- Turn the assignment problem @_X args >= SizeLt u@ into
-        -- @_X args = SizeLt (_Y args@ and constraint
-        -- @_Y args >= u@.
-        subtypingForSizeLt dir x mvar t args v $ \ v -> do
+  -- Turn the assignment problem @_X args >= SizeLt u@ into
+  -- @_X args = SizeLt (_Y args@ and constraint
+  -- @_Y args >= u@.
+  subtypingForSizeLt dir x mvar t args v $ \ v -> do
 
-        -- Normalise and eta contract the arguments to the meta. These are
-        -- usually small, and simplifying might let us instantiate more metas.
+    -- Normalise and eta contract the arguments to the meta. These are
+    -- usually small, and simplifying might let us instantiate more metas.
 
-        -- MOVED TO expandProjectedVars:
-        -- args <- etaContract =<< normalise args
+    -- MOVED TO expandProjectedVars:
+    -- args <- etaContract =<< normalise args
 
-        -- Also, try to expand away projected vars in meta args.
-        expandProjectedVars args v $ \ args v -> do
+    -- Also, try to expand away projected vars in meta args.
+    expandProjectedVars args v $ \ args v -> do
 
-        -- If we had the type here we could save the work we put
-        -- into expanding projected variables.
-        -- catchConstraint (ValueCmp CmpEq ? (MetaV m $ map Apply args) v) $ do
+      -- If we had the type here we could save the work we put
+      -- into expanding projected variables.
+      -- catchConstraint (ValueCmp CmpEq ? (MetaV m $ map Apply args) v) $ do
 
-        -- Andreas, 2011-04-21 do the occurs check first
-        -- e.g. _1 x (suc x) = suc (_2 x y)
-        -- even though the lhs is not a pattern, we can prune the y from _2
-        let varsL = freeVars args
-        let relVL = Set.toList $ relevantVars varsL
+      -- Andreas, 2011-04-21 do the occurs check first
+      -- e.g. _1 x (suc x) = suc (_2 x y)
+      -- even though the lhs is not a pattern, we can prune the y from _2
+      let relVL = Set.toList $ allRelevantVars args
 {- Andreas, 2012-04-02: DontCare no longer present
-        -- take away top-level DontCare constructors
-        args <- return $ map (fmap stripDontCare) args
+      -- take away top-level DontCare constructors
+      args <- return $ map (fmap stripDontCare) args
 -}
-        -- Andreas, 2011-10-06 only irrelevant vars that are direct
-        -- arguments to the meta, hence, can be abstracted over, may
-        -- appear on the rhs.  (test/fail/Issue483b)
-        -- Update 2011-03-27: Also irr. vars under record constructors.
-        let fromIrrVar (Var i [])   = return [i]
-            fromIrrVar (Con c vs)   =
-              ifM (isNothing <$> isRecordConstructor (conName c)) (return []) $
-                concat <$> mapM (fromIrrVar . {- stripDontCare .-} unArg) vs
-            fromIrrVar (Shared p)   = fromIrrVar (derefPtr p)
-            fromIrrVar _ = return []
-        irrVL <- concat <$> mapM fromIrrVar
-                   [ v | Arg info v <- args, irrelevantOrUnused (getRelevance info) ]
-        reportSDoc "tc.meta.assign" 20 $
-            let pr (Var n []) = text (show n)
-                pr (Def c []) = prettyTCM c
-                pr _          = text ".."
-            in vcat
-                 [ text "mvar args:" <+> sep (map (pr . unArg) args)
-                 , text "fvars lhs (rel):" <+> sep (map (text . show) relVL)
-                 , text "fvars lhs (irr):" <+> sep (map (text . show) irrVL)
-                 ]
+      -- Andreas, 2011-10-06 only irrelevant vars that are direct
+      -- arguments to the meta, hence, can be abstracted over, may
+      -- appear on the rhs.  (test/fail/Issue483b)
+      -- Update 2011-03-27: Also irr. vars under record constructors.
+      let fromIrrVar (Var i [])   = return [i]
+          fromIrrVar (Con c vs)   =
+            ifM (isNothing <$> isRecordConstructor (conName c)) (return []) $
+              concat <$> mapM (fromIrrVar . {- stripDontCare .-} unArg) vs
+          fromIrrVar (Shared p)   = fromIrrVar (derefPtr p)
+          fromIrrVar _ = return []
+      irrVL <- concat <$> mapM fromIrrVar
+                 [ v | Arg info v <- args, irrelevantOrUnused (getRelevance info) ]
+      reportSDoc "tc.meta.assign" 20 $
+          let pr (Var n []) = text (show n)
+              pr (Def c []) = prettyTCM c
+              pr _          = text ".."
+          in vcat
+               [ text "mvar args:" <+> sep (map (pr . unArg) args)
+               , text "fvars lhs (rel):" <+> sep (map (text . show) relVL)
+               , text "fvars lhs (irr):" <+> sep (map (text . show) irrVL)
+               ]
 
-        -- Check that the x doesn't occur in the right hand side.
-        -- Prune mvars on rhs such that they can only depend on varsL.
-        -- Herein, distinguish relevant and irrelevant vars,
-        -- since when abstracting irrelevant lhs vars, they may only occur
-        -- irrelevantly on rhs.
-        v <- liftTCM $ occursCheck x (relVL, irrVL) v
+      -- Check that the x doesn't occur in the right hand side.
+      -- Prune mvars on rhs such that they can only depend on lhs vars.
+      -- Herein, distinguish relevant and irrelevant vars,
+      -- since when abstracting irrelevant lhs vars, they may only occur
+      -- irrelevantly on rhs.
+      v <- liftTCM $ occursCheck x (relVL, irrVL) v
 
-        reportSLn "tc.meta.assign" 15 "passed occursCheck"
-        verboseS "tc.meta.assign" 30 $ do
-          let n = size v
-          when (n > 200) $ reportSDoc "tc.meta.assign" 30 $
-            sep [ text "size" <+> text (show n)
---                , nest 2 $ text "type" <+> prettyTCM t
-                , nest 2 $ text "term" <+> prettyTCM v
-                ]
+      reportSLn "tc.meta.assign" 15 "passed occursCheck"
+      verboseS "tc.meta.assign" 30 $ do
+        let n = termSize v
+        when (n > 200) $ reportSDoc "tc.meta.assign" 30 $
+          sep [ text "size" <+> text (show n)
+--              , nest 2 $ text "type" <+> prettyTCM t
+              , nest 2 $ text "term" <+> prettyTCM v
+              ]
 
-        -- Check linearity of @ids@
-        -- Andreas, 2010-09-24: Herein, ignore the variables which are not
-        -- free in v
-        -- Ulf, 2011-09-22: we need to respect irrelevant vars as well, otherwise
-        -- we'll build solutions where the irrelevant terms are not valid
-        let fvs = allVars $ freeVars v
-        reportSDoc "tc.meta.assign" 20 $
-          text "fvars rhs:" <+> sep (map (text . show) $ Set.toList fvs)
+      -- Check linearity of @ids@
+      -- Andreas, 2010-09-24: Herein, ignore the variables which are not
+      -- free in v
+      -- Ulf, 2011-09-22: we need to respect irrelevant vars as well, otherwise
+      -- we'll build solutions where the irrelevant terms are not valid
+      let fvs = allFreeVars v
+      reportSDoc "tc.meta.assign" 20 $
+        text "fvars rhs:" <+> sep (map (text . show) $ Set.toList fvs)
 
-        -- Check that the arguments are variables
-        mids <- do
-          res <- runExceptT $ inverseSubst args
-          case res of
-            -- all args are variables
-            Right ids -> do
-              reportSDoc "tc.meta.assign" 50 $
-                text "inverseSubst returns:" <+> sep (map prettyTCM ids)
-              return $ Just ids
-            -- we have proper values as arguments which could be cased on
-            -- here, we cannot prune, since offending vars could be eliminated
-            Left CantInvert  -> return Nothing
-            -- we have non-variables, but these are not eliminateable
-            Left NeutralArg  -> Just <$> attemptPruning x args fvs
-            -- we have a projected variable which could not be eta-expanded away:
-            -- same as neutral
-            Left (ProjectedVar i qs) -> Just <$> attemptPruning x args fvs
+      -- Check that the arguments are variables
+      mids <- do
+        res <- runExceptT $ inverseSubst args
+        case res of
+          -- all args are variables
+          Right ids -> do
+            reportSDoc "tc.meta.assign" 50 $
+              text "inverseSubst returns:" <+> sep (map prettyTCM ids)
+            return $ Just ids
+          -- we have proper values as arguments which could be cased on
+          -- here, we cannot prune, since offending vars could be eliminated
+          Left CantInvert  -> return Nothing
+          -- we have non-variables, but these are not eliminateable
+          Left NeutralArg  -> Just <$> attemptPruning x args fvs
+          -- we have a projected variable which could not be eta-expanded away:
+          -- same as neutral
+          Left (ProjectedVar i qs) -> Just <$> attemptPruning x args fvs
 
-        case mids of
-          Nothing  -> patternViolation -- Ulf 2014-07-13: actually not needed after all: attemptInertRHSImprovement x args v
-          Just ids -> do
-            -- Check linearity
-            ids <- do
-              res <- runExceptT $ checkLinearity {- (`Set.member` fvs) -} ids
-              case res of
-                -- case: linear
-                Right ids -> return ids
-                -- case: non-linear variables that could possibly be pruned
-                Left ()   -> attemptPruning x args fvs
+      case mids of
+        Nothing  -> patternViolation -- Ulf 2014-07-13: actually not needed after all: attemptInertRHSImprovement x args v
+        Just ids -> do
+          -- Check linearity
+          ids <- do
+            res <- runExceptT $ checkLinearity {- (`Set.member` fvs) -} ids
+            case res of
+              -- case: linear
+              Right ids -> return ids
+              -- case: non-linear variables that could possibly be pruned
+              Left ()   -> attemptPruning x args fvs
 
-            -- Solve.
-            m <- getContextSize
-            assignMeta' m x t (length args) ids v
-    where
-        attemptPruning x args fvs = do
-          -- non-linear lhs: we cannot solve, but prune
-          killResult <- prune x args $ Set.toList fvs
-          reportSDoc "tc.meta.assign" 10 $
-            text "pruning" <+> prettyTCM x <+> do
-            text $
-              if killResult `elem` [PrunedSomething,PrunedEverything] then "succeeded"
-               else "failed"
-          patternViolation
+          -- Solve.
+          m <- getContextSize
+          assignMeta' m x t (length args) ids v
+  where
+    attemptPruning x args fvs = do
+      -- non-linear lhs: we cannot solve, but prune
+      killResult <- prune x args $ Set.toList fvs
+      reportSDoc "tc.meta.assign" 10 $
+        text "pruning" <+> prettyTCM x <+> do
+        text $
+          if killResult `elem` [PrunedSomething,PrunedEverything] then "succeeded"
+           else "failed"
+      patternViolation
 
+{- UNUSED
 -- | When faced with @_X us == D vs@ for an inert D we can solve this by
 --   @_X xs := D _Ys@ with new constraints @_Yi us == vi@. This is important
 --   for instance arguments, where knowing the head D might enable progress.
@@ -786,8 +795,8 @@ attemptInertRHSImprovement m args v = do
               patternViolation
             | otherwise  = return ()
       case fmap ignoreSharing b of
-        Blocked{}    -> notNeutral v
-        NotBlocked v ->
+        Blocked{}      -> notNeutral v
+        NotBlocked r v ->                      -- Andrea(s) 2014-12-06 can r be useful?
           case v of
             Var x _    -> checkRHS (Var x [])
             Def f _    -> checkRHS (Def f [])
@@ -801,7 +810,7 @@ attemptInertRHSImprovement m args v = do
             Lam{}      -> notNeutral v
             ExtLam{}   -> __IMPOSSIBLE__
             Shared{}   -> __IMPOSSIBLE__
-
+-- END UNUSED -}
 
 -- | @assignMeta m x t ids u@ solves @x ids = u@ for meta @x@ of type @t@,
 --   where term @u@ lives in a context of length @m@.
@@ -871,54 +880,68 @@ assignMeta' m x t n ids v = do
 -- | Turn the assignment problem @_X args <= SizeLt u@ into
 -- @_X args = SizeLt (_Y args)@ and constraint
 -- @_Y args <= u@.
-subtypingForSizeLt :: CompareDirection -> MetaId -> MetaVariable -> Type -> Args -> Term -> (Term -> TCM ()) -> TCM ()
+subtypingForSizeLt
+  :: CompareDirection -- ^ @dir@
+  -> MetaId           -- ^ The meta variable @x@.
+  -> MetaVariable     -- ^ Its associated information @mvar <- lookupMeta x@.
+  -> Type             -- ^ Its type @t = jMetaType $ mvJudgement mvar@
+  -> Args             -- ^ Its arguments.
+  -> Term             -- ^ Its to-be-assigned value @v@, such that @x args `dir` v@.
+  -> (Term -> TCM ()) -- ^ Continuation taking its possibly assigned value.
+  -> TCM ()
 subtypingForSizeLt DirEq x mvar t args v cont = cont v
 subtypingForSizeLt dir   x mvar t args v cont = do
   let fallback = cont v
   -- Check whether we have built-ins SIZE and SIZELT
   (mSize, mSizeLt) <- getBuiltinSize
   caseMaybe mSize   fallback $ \ qSize   -> do
-  caseMaybe mSizeLt fallback $ \ qSizeLt -> do
-  -- Check whether v is a SIZELT
-  v <- reduce v
-  case ignoreSharing v of
-    Def q [Apply (Arg ai u)] | q == qSizeLt -> do
-      -- Clone the meta into a new size meta @y@.
-      -- To this end, we swap the target of t for Size.
-      TelV tel _ <- telView t
-      let size = sizeType_ qSize
-          t'   = telePi tel size
-      y <- newMeta (mvInfo mvar) (mvPriority mvar) (mvPermutation mvar)
-                   (HasType __IMPOSSIBLE__ t')
-      -- Note: no eta-expansion of new meta possible/necessary.
-      -- Add the size constraint @y args `dir` u@.
-      let yArgs = MetaV y $ map Apply args
-      addConstraint $ dirToCmp (`ValueCmp` size) dir yArgs u
-      -- We continue with the new assignment problem, and install
-      -- an exception handler, since we created a meta and a constraint,
-      -- so we cannot fall back to the original handler.
-      let xArgs = MetaV x $ map Apply args
-          v'    = Def qSizeLt [Apply $ Arg ai yArgs]
-          c     = dirToCmp (`ValueCmp` sizeUniv) dir xArgs v'
-      catchConstraint c $ cont v'
-    _ -> fallback
+    caseMaybe mSizeLt fallback $ \ qSizeLt -> do
+      -- Check whether v is a SIZELT
+      v <- reduce v
+      case ignoreSharing v of
+        Def q [Apply (Arg ai u)] | q == qSizeLt -> do
+          -- Clone the meta into a new size meta @y@.
+          -- To this end, we swap the target of t for Size.
+          TelV tel _ <- telView t
+          let size = sizeType_ qSize
+              t'   = telePi tel size
+          y <- newMeta (mvInfo mvar) (mvPriority mvar) (mvPermutation mvar)
+                       (HasType __IMPOSSIBLE__ t')
+          -- Note: no eta-expansion of new meta possible/necessary.
+          -- Add the size constraint @y args `dir` u@.
+          let yArgs = MetaV y $ map Apply args
+          addConstraint $ dirToCmp (`ValueCmp` size) dir yArgs u
+          -- We continue with the new assignment problem, and install
+          -- an exception handler, since we created a meta and a constraint,
+          -- so we cannot fall back to the original handler.
+          let xArgs = MetaV x $ map Apply args
+              v'    = Def qSizeLt [Apply $ Arg ai yArgs]
+              c     = dirToCmp (`ValueCmp` sizeUniv) dir xArgs v'
+          catchConstraint c $ cont v'
+        _ -> fallback
 
 -- | Eta-expand bound variables like @z@ in @X (fst z)@.
-expandProjectedVars :: (Normalise a, TermLike a, PrettyTCM a, NoProjectedVar a, Subst a, PrettyTCM b, Subst b) =>
+expandProjectedVars :: (Normalise a, TermLike a, Show a, PrettyTCM a, NoProjectedVar a, Subst a, PrettyTCM b, Subst b) =>
   a -> b -> (a -> b -> TCM c) -> TCM c
 expandProjectedVars args v ret = loop (args, v) where
   loop (args, v) = do
-    reportSDoc "tc.meta.assign.proj" 40 $ text "meta args: " <+> prettyTCM args
+    reportSDoc "tc.meta.assign.proj" 45 $ text "meta args: " <+> prettyTCM args
     args <- etaContract =<< normalise args
-    reportSDoc "tc.meta.assign.proj" 40 $ text "norm args: " <+> prettyTCM args
+    reportSDoc "tc.meta.assign.proj" 45 $ text "norm args: " <+> prettyTCM args
+    reportSDoc "tc.meta.assign.proj" 85 $ text "norm args: " <+> text (show args)
     let done = ret args v
     case noProjectedVar args of
-      Right ()              -> done
+      Right ()              -> do
+        reportSDoc "tc.meta.assign.proj" 40 $
+          text "no projected var found in args: " <+> prettyTCM args
+        done
       Left (ProjVarExc i _) -> etaExpandProjectedVar i (args, v) done loop
 
 -- | Eta-expand a de Bruijn index of record type in context and passed term(s).
 etaExpandProjectedVar :: (PrettyTCM a, Subst a) => Int -> a -> TCM c -> (a -> TCM c) -> TCM c
 etaExpandProjectedVar i v fail succeed = do
+  reportSDoc "tc.meta.assign.proj" 40 $
+    text "trying to expand projected variable" <+> prettyTCM (var i)
   caseMaybeM (etaExpandBoundVar i) fail $ \ (delta, sigma, tau) -> do
     reportSDoc "tc.meta.assign.proj" 25 $
       text "eta-expanding var " <+> prettyTCM (var i) <+>
@@ -932,13 +955,19 @@ class NoProjectedVar a where
 
 data ProjVarExc = ProjVarExc Int [QName]
 
-instance Error ProjVarExc where
-  noMsg = __IMPOSSIBLE__
+-- ASR (17 June 2015). Unused Error instance.
+-- instance Error ProjVarExc where
+--   noMsg = __IMPOSSIBLE__
 
 instance NoProjectedVar Term where
-  noProjectedVar (Var i es)
-    | qs@(_:_) <- takeWhileJust id $ map isProjElim es = Left $ ProjVarExc i qs
-  noProjectedVar _ = return ()
+  noProjectedVar t =
+    case ignoreSharing t of
+      Var i es
+        | qs@(_:_) <- takeWhileJust id $ map isProjElim es -> Left $ ProjVarExc i qs
+      -- Andreas, 2015-09-12 Issue 1316:
+      -- Also look in inductive record constructors
+      Con (ConHead _ Inductive (_:_)) vs -> noProjectedVar vs
+      _ -> return ()
 
 instance NoProjectedVar a => NoProjectedVar (I.Arg a) where
   noProjectedVar = Fold.mapM_ noProjectedVar
@@ -1020,42 +1049,19 @@ checkLinearity ids0 = do
         (return [p])
         (throwError ())
 
-{- UNNECESSARILY COMPLICATED
--- | Turn non-det substitution into proper substitution, if possible.
---   Writes a list of non-linear variables that need to be pruned.
---   If a non-linear variable is @elemFVs@, hence, not prunable,
---   the error is thrown.
-checkLinearity :: (Nat -> Bool) -> SubstCand -> ErrorT () (WriterT [Nat] TCM) SubstCand
-checkLinearity elemFVs ids0 = do
-  let ids = sortBy (compare `on` fst) ids0
-  let grps = groupOn fst ids
-  concat <$> mapM makeLinear grps
-  where
-    -- | Non-determinism can be healed if type is singleton. [Issue 593]
-    --   (Same as for irrelevance.)
-    makeLinear :: SubstCand -> ErrorT () TCM SubstCand
-    makeLinear []                = __IMPOSSIBLE__
-    makeLinear grp@[_]           = return grp
-    makeLinear grp@(p@(i,t) : _) = do
-      ifM ((Right True ==) <$> do isSingletonTypeModuloRelevance =<< typeOfBV i)
-        {- then -} (return [p])
-        {- else -} $ do
-        ifM (elemFVs i)
-          {- then -} (throwError ())          -- non-prunable non-linear var
-          {- else -} (tell [i] >> return grp) -- possibly prunable non-lin var
--}
-
 -- Intermediate result in the following function
 type Res = [(I.Arg Nat, Term)]
 
 -- | Exceptions raised when substitution cannot be inverted.
 data InvertExcept
   = CantInvert                -- ^ Cannot recover.
-  | NeutralArg                -- ^ A neutral arg: can't invert, but maybe prune.
+  | NeutralArg                -- ^ A potentially neutral arg: can't invert, but can try pruning.
   | ProjectedVar Int [QName]  -- ^ Try to eta-expand var to remove projs.
 
+#if !MIN_VERSION_transformers(0,4,1)
 instance Error InvertExcept where
   noMsg = CantInvert
+#endif
 
 -- | Check that arguments @args@ to a metavar are in pattern fragment.
 --   Assumes all arguments already in whnf and eta-reduced.
@@ -1100,20 +1106,13 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
           case isRC of
             Just (_, Record{ recFields = fs })
               | length fs == length vs -> do
-                let aux (Arg _ v) (Arg info' f) =
-                      (Arg (ArgInfo { argInfoColors = argInfoColors info -- TODO guilhem
-                                    , argInfoHiding = min (argInfoHiding info)
-                                                          (argInfoHiding info')
-                                    , argInfoRelevance = max (argInfoRelevance info)
-                                                             (argInfoRelevance info')
-                                    })
-                           v,) -- OLD: (stripDontCare v),
-                       $ t `applyE` [Proj f]
+                let aux (Arg _ v) (Arg info' f) = (Arg ai v,) $ t `applyE` [Proj f] where
+                     ai = ArgInfo
+                       { argInfoColors    = argInfoColors info -- TODO guilhem
+                       , argInfoHiding    = min (getHiding info) (getHiding info')
+                       , argInfoRelevance = max (getRelevance info) (getRelevance info')
+                       }
                 res <- loop $ zipWith aux vs fs
--- Andreas, 2013-09-22, applyDef not needed after all
--- since f (because taken from recFields) is the original record projection.
---                        <$> do liftTCM $ applyDef f (defaultArg t)
---                res <- loop =<< zipWithM aux vs fs
                 return $ res `append` vars
               | otherwise -> fallback
             Just _  -> __IMPOSSIBLE__
@@ -1127,9 +1126,9 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
         Arg _ DontCare{}                                    -> return vars
 
         -- Distinguish args that can be eliminated (Con,Lit,Lam,unsure) ==> failure
-        -- from those that can only put somewhere as a whole ==> return Nothing
+        -- from those that can only put somewhere as a whole ==> neutralArg
         Arg _ Var{}      -> neutralArg
-        Arg _ Def{}      -> failure
+        Arg _ Def{}      -> neutralArg  -- Note that this Def{} is in normal form and might be prunable.
         Arg _ Lam{}      -> failure
         Arg _ Lit{}      -> failure
         Arg _ MetaV{}    -> failure
