@@ -1,6 +1,5 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
 
 module Agda.TypeChecking.MetaVars.Mention where
 
@@ -43,7 +42,7 @@ instance MentionsMeta LevelAtom where
     MetaLevel m vs   -> x == m || mentionsMeta x vs
     BlockedLevel m _ -> x == m    -- if it's blocked on a different meta it doesn't matter if it mentions the meta somewhere else
     UnreducedLevel l -> mentionsMeta x l
-    NeutralLevel l   -> mentionsMeta x l
+    NeutralLevel _ l -> mentionsMeta x l
 
 instance MentionsMeta Type where
     mentionsMeta x (El s t) = mentionsMeta x (s, t)
@@ -53,6 +52,7 @@ instance MentionsMeta Sort where
     Type l     -> mentionsMeta x l
     Prop       -> False
     Inf        -> False
+    SizeUniv   -> False
     DLub s1 s2 -> mentionsMeta x (s1, s2)
 
 instance MentionsMeta t => MentionsMeta (Abs t) where
@@ -106,6 +106,7 @@ instance MentionsMeta Constraint where
                                   -- what metas it depends on
     FindInScope{}       -> True   -- this needs to be woken up for any meta
     IsEmpty r t         -> mm t
+    CheckSizeLtSat t    -> mm t
     where
       mm v = mentionsMeta x v
 
