@@ -1,6 +1,18 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE NoMonomorphismRestriction #-}
+
+{- | The occurs check for unification.  Does pruning on the fly.
+
+  When hitting a meta variable:
+
+  - Compute flex/rigid for its arguments.
+  - Compare to allowed variables.
+  - Mark arguments with rigid occurrences of disallowed variables for deletion.
+  - Attempt to delete marked arguments.
+  - We don't need to check for success, we can just continue occurs checking.
+-}
 
 module Agda.TypeChecking.MetaVars.Occurs where
 
@@ -9,14 +21,20 @@ import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State
 
+import Data.Foldable (foldMap)
 import Data.List
 import Data.Maybe
+import Data.Monoid
 import Data.Set (Set)
 import qualified Data.Set as Set
 
+import qualified Agda.Benchmarking as Bench
+
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
+
 import Agda.TypeChecking.Monad
+import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Free hiding (Occurrence(..))
@@ -38,6 +56,7 @@ import Agda.Utils.List (takeWhileJust)
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Size
 import qualified Agda.Utils.VarSet as VarSet
 
@@ -87,8 +106,23 @@ modifyOccursCheckDefs f = stOccursCheckDefs %= f
 initOccursCheck :: MetaVariable -> TCM ()
 initOccursCheck mv = modifyOccursCheckDefs . const =<<
   if (miMetaOccursCheck (mvInfo mv) == DontRunMetaOccursCheck)
-   then return Set.empty
-   else maybe (return Set.empty) lookupMutualBlock =<< asks envMutualBlock
+   then do
+     reportSLn "tc.meta.occurs" 20 $
+       "initOccursCheck: we do not look into definitions"
+     return Set.empty
+   else do
+     reportSLn "tc.meta.occurs" 20 $
+       "initOccursCheck: we look into the following definitions:"
+     mb <- asks envMutualBlock
+     case mb of
+       Nothing -> do
+         reportSLn "tc.meta.occurs" 20 $ "(none)"
+         return Set.empty
+       Just b  -> do
+         ds <- lookupMutualBlock b
+         reportSDoc "tc.meta.occurs" 20 $ sep $ map prettyTCM $ Set.toList ds
+         return ds
+
 
 -- | Is a def in the list of stuff to be checked?
 defNeedsChecking :: QName -> TCM Bool
@@ -114,7 +148,7 @@ defArgs NoUnfold  _   = Flex
 defArgs YesUnfold ctx = weakly ctx
 
 unfold :: UnfoldStrategy -> Term -> TCM (Blocked Term)
-unfold NoUnfold  v = NotBlocked <$> instantiate v
+unfold NoUnfold  v = notBlocked <$> instantiate v
 unfold YesUnfold v = reduceB v
 
 -- | Leave the top position.
@@ -166,41 +200,43 @@ class Occurs t where
 
 -- | When assigning @m xs := v@, check that @m@ does not occur in @v@
 --   and that the free variables of @v@ are contained in @xs@.
-occursCheck :: MetaId -> Vars -> Term -> TCM Term
-occursCheck m xs v = liftTCM $ do
+occursCheck
+  :: (Occurs a, InstantiateFull a, PrettyTCM a)
+  => MetaId -> Vars -> a -> TCM a
+occursCheck m xs v = Bench.billTo [ Bench.Typing, Bench.OccursCheck ] $ do
   mv <- lookupMeta m
   initOccursCheck mv
       -- TODO: Can we do this in a better way?
   let redo m = m -- disableDestructiveUpdate m >> m
   -- First try without normalising the term
   redo (occurs NoUnfold  Top m xs v) `catchError` \_ -> do
-  initOccursCheck mv
-  redo (occurs YesUnfold Top m xs v) `catchError` \err -> case err of
-                          -- Produce nicer error messages
-    TypeError _ cl -> case clValue cl of
-      MetaOccursInItself{} ->
-        typeError . GenericError . show =<<
-          fsep [ text ("Refuse to construct infinite term by instantiating " ++ show m ++ " to")
-               , prettyTCM =<< instantiateFull v
-               ]
-      MetaCannotDependOn _ _ i ->
-        ifM (isSortMeta m `and2M` (not <$> hasUniversePolymorphism))
-        ( typeError . GenericError . show =<<
-          fsep [ text ("Cannot instantiate the metavariable " ++ show m ++ " to")
-               , prettyTCM v
-               , text "since universe polymorphism is disabled"
-               ]
-        ) {- else -}
-        ( typeError . GenericError . show =<<
-            fsep [ text ("Cannot instantiate the metavariable " ++ show m ++ " to solution")
-                 , prettyTCM v
-                 , text "since it contains the variable"
-                 , enterClosure cl $ \_ -> prettyTCM (Var i [])
-                 , text $ "which is not in scope of the metavariable or irrelevant in the metavariable but relevant in the solution"
+    initOccursCheck mv
+    redo (occurs YesUnfold Top m xs v) `catchError` \err -> case err of
+                            -- Produce nicer error messages
+      TypeError _ cl -> case clValue cl of
+        MetaOccursInItself{} ->
+          typeError . GenericError . show =<<
+            fsep [ text ("Refuse to construct infinite term by instantiating " ++ prettyShow m ++ " to")
+                 , prettyTCM =<< instantiateFull v
                  ]
-          )
+        MetaCannotDependOn _ _ i ->
+          ifM (isSortMeta m `and2M` (not <$> hasUniversePolymorphism))
+          ( typeError . GenericError . show =<<
+            fsep [ text ("Cannot instantiate the metavariable " ++ prettyShow m ++ " to")
+                 , prettyTCM v
+                 , text "since universe polymorphism is disabled"
+                 ]
+          ) {- else -}
+          ( typeError . GenericError . show =<<
+              fsep [ text ("Cannot instantiate the metavariable " ++ prettyShow m ++ " to solution")
+                   , prettyTCM v
+                   , text "since it contains the variable"
+                   , enterClosure cl $ \_ -> prettyTCM (Var i [])
+                   , text $ "which is not in scope of the metavariable or irrelevant in the metavariable but relevant in the solution"
+                   ]
+            )
+        _ -> throwError err
       _ -> throwError err
-    _ -> throwError err
 
 instance Occurs Term where
   occurs red ctx m xs v = do
@@ -208,90 +244,88 @@ instance Occurs Term where
     -- occurs' ctx $ ignoreBlocking v  -- fails test/succeed/DontPruneBlocked
     case v of
       -- Don't fail on blocked terms or metas
-      NotBlocked v        -> occurs' ctx v
+      NotBlocked _ v      -> occurs' ctx v
       -- Blocked _ v@MetaV{} -> occurs' ctx v  -- does not help with issue 856
       Blocked _ v         -> occurs' Flex v
     where
       occurs' ctx v = do
-      reportSDoc "tc.meta.occurs" 45 $
-        text ("occursCheck " ++ show m ++ " (" ++ show ctx ++ ") of ") <+> prettyTCM v
-      reportSDoc "tc.meta.occurs" 70 $
-        nest 2 $ text $ show v
-      case v of
-        Var i es   -> do
-          if (i `allowedVar` xs) then Var i <$> occ (weakly ctx) es else do
-            -- if the offending variable is of singleton type,
-            -- eta-expand it away
-            isST <- isSingletonType =<< typeOfBV i
-            case isST of
-              -- cannot decide, blocked by meta-var
-              Left mid -> patternViolation' 70 $ "Disallowed var " ++ show i ++ " not obviously singleton"
-              -- not a singleton type
-              Right Nothing -> -- abort Rigid turns this error into PatternErr
-                abort (strongly ctx) $ MetaCannotDependOn m (takeRelevant xs) i
-              -- is a singleton type with unique inhabitant sv
-              Right (Just sv) -> return $ sv `applyE` es
-        Lam h f     -> Lam h <$> occ (leaveTop ctx) f
-        Level l     -> Level <$> occ ctx l  -- stay in Top
-        Lit l       -> return v
-        DontCare v  -> dontCare <$> occurs red Irrel m (goIrrelevant xs) v
-        Def d es    -> Def d <$> occDef d (leaveTop ctx) es
-        Con c vs    -> Con c <$> occ (leaveTop ctx) vs  -- if strongly rigid, remain so
-        Pi a b      -> uncurry Pi <$> occ (leaveTop ctx) (a,b)
-        Sort s      -> Sort <$> occ (leaveTop ctx) s
-        v@Shared{}  -> updateSharedTerm (occ ctx) v
-        ExtLam{}    -> __IMPOSSIBLE__
-        MetaV m' es -> do
-            -- Check for loop
-            --   don't fail hard on this, since we might still be on the top-level
-            --   after some killing (Issue 442)
-            --
-            -- Andreas, 2013-02-18  Issue 795 demonstrates that a recursive
-            -- occurrence of a meta could be solved by the identity.
-            --   ? (Q A) = Q (? A)
-            -- So, do not throw an error.
-            -- I guess the error was there from times when occurrence check
-            -- was done after the "lhs=linear variables" check, but now
-            -- occurrence check comes first.
-            {-
-            when (m == m') $ if ctx == Top then patternViolation else
-              abort ctx $ MetaOccursInItself m'
-            -}
-            when (m == m') $ patternViolation' 50 $ "occursCheck failed: Found " ++ show m
+        reportSDoc "tc.meta.occurs" 45 $
+          text ("occursCheck " ++ prettyShow m ++ " (" ++ show ctx ++ ") of ") <+> prettyTCM v
+        reportSDoc "tc.meta.occurs" 70 $
+          nest 2 $ text $ show v
+        case v of
+          Var i es   -> do
+            if (i `allowedVar` xs) then Var i <$> occ (weakly ctx) es else do
+              -- if the offending variable is of singleton type,
+              -- eta-expand it away
+              isST <- isSingletonType =<< typeOfBV i
+              case isST of
+                -- cannot decide, blocked by meta-var
+                Left mid -> patternViolation' 70 $ "Disallowed var " ++ show i ++ " not obviously singleton"
+                -- not a singleton type
+                Right Nothing -> -- abort Rigid turns this error into PatternErr
+                  abort (strongly ctx) $ MetaCannotDependOn m (takeRelevant xs) i
+                -- is a singleton type with unique inhabitant sv
+                Right (Just sv) -> return $ sv `applyE` es
+          Lam h f     -> Lam h <$> occ (leaveTop ctx) f
+          Level l     -> Level <$> occ ctx l  -- stay in Top
+          Lit l       -> return v
+          DontCare v  -> dontCare <$> occurs red Irrel m (goIrrelevant xs) v
+          Def d es    -> Def d <$> occDef d (leaveTop ctx) es
+          Con c vs    -> Con c <$> occ (leaveTop ctx) vs  -- if strongly rigid, remain so
+          Pi a b      -> uncurry Pi <$> occ (leaveTop ctx) (a,b)
+          Sort s      -> Sort <$> occ (leaveTop ctx) s
+          v@Shared{}  -> updateSharedTerm (occ ctx) v
+          ExtLam{}    -> __IMPOSSIBLE__
+          MetaV m' es -> do
+              -- Check for loop
+              --   don't fail hard on this, since we might still be on the top-level
+              --   after some killing (Issue 442)
+              --
+              -- Andreas, 2013-02-18  Issue 795 demonstrates that a recursive
+              -- occurrence of a meta could be solved by the identity.
+              --   ? (Q A) = Q (? A)
+              -- So, do not throw an error.
+              -- I guess the error was there from times when occurrence check
+              -- was done after the "lhs=linear variables" check, but now
+              -- occurrence check comes first.
+              -- WAS:
+              -- when (m == m') $ if ctx == Top then patternViolation else
+              --   abort ctx $ MetaOccursInItself m'
+              when (m == m') $ patternViolation' 50 $ "occursCheck failed: Found " ++ prettyShow m
 
-            -- The arguments of a meta are in a flexible position
-            (MetaV m' <$> occurs red Flex m xs es) `catchError` \err -> do
-              reportSDoc "tc.meta.kill" 25 $ vcat
-                [ text $ "error during flexible occurs check, we are " ++ show ctx
-                , text $ show err
-                ]
-              case err of
-                -- On pattern violations try to remove offending
-                -- flexible occurrences (if not already in a flexible context)
-                PatternErr{} | ctx /= Flex -> do
-                  reportSLn "tc.meta.kill" 20 $
-                    "oops, pattern violation for " ++ show m'
-                  -- Andreas, 2014-03-02, see issue 1070:
-                  -- Do not prune when meta is projected!
-                  -- WAS: let vs = takeWhileJust isApplyElim es
-                  caseMaybe (allApplyElims es) (throwError err) $ \ vs -> do
-                  killResult <- prune m' vs (takeRelevant xs)
-                  if (killResult == PrunedEverything)
-                    -- after successful pruning, restart occurs check
-                    then occurs red ctx m xs =<< instantiate (MetaV m' es)
-                    else throwError err
-                _ -> throwError err
-        where
-          occ ctx v = occurs red ctx m xs v
-          -- a data or record type constructor propagates strong occurrences
-          -- since e.g. x = List x is unsolvable
-          occDef d ctx vs = do
-            def <- theDef <$> getConstInfo d
-            whenM (defNeedsChecking d) $ do
-              tallyDef d
-              reportSLn "tc.meta.occurs" 30 $ "Checking for occurrences in " ++ show d
-              metaOccurs m def
-            if (defIsDataOrRecord def) then (occ ctx vs) else (occ (defArgs red ctx) vs)
+              -- The arguments of a meta are in a flexible position
+              (MetaV m' <$> occurs red Flex m xs es) `catchError` \err -> do
+                reportSDoc "tc.meta.kill" 25 $ vcat
+                  [ text $ "error during flexible occurs check, we are " ++ show ctx
+                  , text $ show err
+                  ]
+                case err of
+                  -- On pattern violations try to remove offending
+                  -- flexible occurrences (if not already in a flexible context)
+                  PatternErr{} | ctx /= Flex -> do
+                    reportSLn "tc.meta.kill" 20 $
+                      "oops, pattern violation for " ++ prettyShow m'
+                    -- Andreas, 2014-03-02, see issue 1070:
+                    -- Do not prune when meta is projected!
+                    caseMaybe (allApplyElims es) (throwError err) $ \ vs -> do
+                      killResult <- prune m' vs (takeRelevant xs)
+                      if (killResult == PrunedEverything)
+                        -- after successful pruning, restart occurs check
+                        then occurs red ctx m xs =<< instantiate (MetaV m' es)
+                        else throwError err
+                  _ -> throwError err
+          where
+            occ ctx v = occurs red ctx m xs v
+            -- a data or record type constructor propagates strong occurrences
+            -- since e.g. x = List x is unsolvable
+            occDef d ctx vs = do
+              def <- theDef <$> getConstInfo d
+              whenM (defNeedsChecking d) $ do
+                tallyDef d
+                reportSLn "tc.meta.occurs" 30 $ "Checking for occurrences in " ++ show d
+                metaOccurs m def
+              if (defIsDataOrRecord def) then (occ ctx vs) else (occ (defArgs red ctx) vs)
 
   metaOccurs m v = do
     v <- instantiate v
@@ -307,7 +341,7 @@ instance Occurs Term where
       Sort s     -> metaOccurs m s
       Shared p   -> metaOccurs m $ derefPtr p
       ExtLam{}   -> __IMPOSSIBLE__
-      MetaV m' vs | m == m' -> patternViolation' 50 $ "Found occurrence of " ++ show m
+      MetaV m' vs | m == m' -> patternViolation' 50 $ "Found occurrence of " ++ prettyShow m
                   | otherwise -> metaOccurs m vs
 
 instance Occurs QName where
@@ -361,16 +395,16 @@ instance Occurs LevelAtom where
       MetaLevel m' args -> do
         MetaV m' args <- ignoreSharing <$> occurs red ctx m xs (MetaV m' args)
         return $ MetaLevel m' args
-      NeutralLevel v    -> NeutralLevel    <$> occurs red ctx m xs v
+      NeutralLevel r v  -> NeutralLevel r  <$> occurs red ctx m xs v
       BlockedLevel m' v -> BlockedLevel m' <$> occurs red Flex m xs v
       UnreducedLevel v  -> UnreducedLevel  <$> occurs red ctx m xs v
 
   metaOccurs m l = do
     l <- instantiate l
     case l of
-      MetaLevel m' args -> metaOccurs m (MetaV m' args)
-      NeutralLevel v    -> metaOccurs m v
-      BlockedLevel m v  -> metaOccurs m v
+      MetaLevel m' args -> metaOccurs m $ MetaV m' args
+      NeutralLevel _ v  -> metaOccurs m v
+      BlockedLevel _ v  -> metaOccurs m v
       UnreducedLevel v  -> metaOccurs m v
 
 
@@ -389,6 +423,7 @@ instance Occurs Sort where
       Type a     -> Type <$> occurs red ctx m xs a
       Prop       -> return s'
       Inf        -> return s'
+      SizeUniv   -> return s'
 
   metaOccurs m s = do
     s <- instantiate s
@@ -397,6 +432,7 @@ instance Occurs Sort where
       Type a     -> metaOccurs m a
       Prop       -> return ()
       Inf        -> return ()
+      SizeUniv   -> return ()
 
 instance Occurs a => Occurs (Elim' a) where
   occurs red ctx m xs e@Proj{}  = return e
@@ -449,24 +485,17 @@ prune :: MetaId -> Args -> [Nat] -> TCM PruneResult
 prune m' vs xs = do
   caseEitherM (runExceptT $ mapM (hasBadRigid xs) $ map unArg vs)
     (const $ return PrunedNothing) $ \ kills -> do
-  reportSDoc "tc.meta.kill" 10 $ vcat
-    [ text "attempting kills"
-    , nest 2 $ vcat
-      [ text "m'    =" <+> text (show m')
-      , text "xs    =" <+> text (show xs)
-      , text "vs    =" <+> prettyList (map prettyTCM vs)
-      , text "kills =" <+> text (show kills)
+    reportSDoc "tc.meta.kill" 10 $ vcat
+      [ text "attempting kills"
+      , nest 2 $ vcat
+        [ text "m'    =" <+> pretty m'
+        -- , text "xs    =" <+> text (show xs)
+        , text "xs    =" <+> prettyList (map (prettyTCM . var) xs)
+        , text "vs    =" <+> prettyList (map prettyTCM vs)
+        , text "kills =" <+> text (show kills)
+        ]
       ]
-    ]
-{- Andreas, 2011-05-11 REDUNDANT CODE
-  reportSLn "tc.meta.kill" 20 $
-    "attempting to prune meta " ++ show m' ++ "\n" ++
-    "  kills: " ++ show kills
-  if not (or kills)
-    then return False -- nothing to kill
-    else do
--}
-  killArgs kills m'
+    killArgs kills m'
 
 -- | @hasBadRigid xs v = Just True@ iff one of the rigid variables in @v@ is not in @xs@.
 --   Actually we can only prune if a bad variable is in the head. See issue 458.
@@ -479,7 +508,8 @@ hasBadRigid :: [Nat] -> Term -> ExceptT () TCM Bool
 hasBadRigid xs t = do
   -- We fail if we encounter a matchable argument.
   let failure = throwError ()
-  t <- liftTCM $ reduce t
+  tb <- liftTCM $ reduceB t
+  let t = ignoreBlocking tb
   case ignoreSharing t of
     Var x _      -> return $ notElem x xs
     -- Issue 1153: A lambda has to be considered matchable.
@@ -489,13 +519,13 @@ hasBadRigid xs t = do
     -- The following types of arguments cannot be eliminated by a pattern
     -- match: data, record, Pi, levels, sorts
     -- Thus, their offending rigid variables are bad.
-    v@(Def f es) -> ifNotM (isNeutral f es) failure $ {- else -} do
-      return $ es `rigidVarsNotContainedIn` xs
+    v@(Def f es) -> ifNotM (isNeutral tb f es) failure $ {- else -} do
+      es `rigidVarsNotContainedIn` xs
     -- Andreas, 2012-05-03: There is room for further improvement.
     -- We could also consider a defined f which is not blocked by a meta.
-    Pi a b       -> return $ (a,b) `rigidVarsNotContainedIn` xs
-    Level v      -> return $ v `rigidVarsNotContainedIn` xs
-    Sort s       -> return $ s `rigidVarsNotContainedIn` xs
+    Pi a b       -> (a,b) `rigidVarsNotContainedIn` xs
+    Level v      -> v `rigidVarsNotContainedIn` xs
+    Sort s       -> s `rigidVarsNotContainedIn` xs
     -- Since constructors can be eliminated by pattern-matching,
     -- offending variables under a constructor could be removed by
     -- the right instantiation of the meta variable.
@@ -513,24 +543,124 @@ hasBadRigid xs t = do
 
 -- | Check whether a term @Def f es@ is finally stuck.
 --   Currently, we give only a crude approximation.
-isNeutral :: MonadTCM tcm => QName -> Elims -> tcm Bool
-isNeutral f es = liftTCM $ do
+isNeutral :: MonadTCM tcm => Blocked t -> QName -> Elims -> tcm Bool
+isNeutral b f es = liftTCM $ do
   let yes = return True
+      no  = return False
   def <- getConstInfo f
   case theDef def of
     Axiom{}    -> yes
     Datatype{} -> yes
     Record{}   -> yes
-    _          -> return False
-      -- TODO: more precise analysis
-      -- We need to check whether a function is stuck on a variable
-      -- (not meta variable), but the API does not help us...
+    Function{} -> case b of
+      NotBlocked StuckOn{}   _ -> yes
+      NotBlocked AbsurdMatch _ -> yes
+      _                        -> no
+    _          -> no
 
--- This could be optimized, by not computing the whole variable set
--- at once, but allow early failure
-rigidVarsNotContainedIn :: Free a => a -> [Nat] -> Bool
-rigidVarsNotContainedIn v xs =
-  not $ rigidVars (freeVars v) `VarSet.isSubsetOf` VarSet.fromList xs
+-- | Check whether any of the variables (given as de Bruijn indices)
+--   occurs *definitely* in the term in a rigid position.
+--   Reduces the term successively to remove variables in dead subterms.
+--   This fixes issue 1386.
+rigidVarsNotContainedIn :: (MonadTCM tcm, FoldRigid a) => a -> [Nat] -> tcm Bool
+rigidVarsNotContainedIn v is = liftTCM $ do
+  n0 <- getContextSize
+  let -- allowed variables as de Bruijn levels
+      levels = Set.fromList $ map (n0-1 -) is
+      -- test if index is forbidden by converting it to level
+      test i = do
+        n <- getContextSize
+        -- get de Bruijn level for i
+        let l = n-1 - i
+            -- If l >= n0 then it is a bound variable and can be
+            -- ignored.  Otherwise, it has to be in the allowed levels.
+            forbidden = l < n0 && not (l `Set.member` levels)
+        when forbidden $
+          reportSLn "tc.meta.kill" 20 $
+            "found forbidden de Bruijn level " ++ show l
+        return $ Any forbidden
+  getAny <$> foldRigid id test v
+
+-- | Collect the *definitely* rigid variables in a monoid.
+--   We need to successively reduce the expression to do this.
+
+class FoldRigid a where
+--  foldRigid :: (MonadTCM tcm, Monoid (tcm m)) => (tcm m -> tcm m) -> (Nat -> tcm m) -> a -> tcm m
+  foldRigid :: (Monoid (TCM m)) => (TCM m -> TCM m) -> (Nat -> TCM m) -> a -> TCM m
+
+instance FoldRigid Term where
+  foldRigid abs f t = do
+    b <- liftTCM $ reduceB t
+    case ignoreSharing $ ignoreBlocking b of
+      Var i es   -> f i `mappend` fold es
+      Lam _ t    -> fold t
+      Lit{}      -> mempty
+      Def f es   -> case b of
+        Blocked{}                   -> mempty
+        NotBlocked MissingClauses _ -> mempty
+        _        -> fold es
+      Con _ ts   -> fold ts
+      Pi a b     -> fold (a,b)
+      Sort s     -> fold s
+      Level l    -> fold l
+      MetaV{}    -> mempty
+      DontCare{} -> mempty
+      Shared{}   -> __IMPOSSIBLE__
+      ExtLam{}   -> __IMPOSSIBLE__
+    where fold = foldRigid abs f
+
+instance FoldRigid Type where
+  foldRigid abs f (El s t) = foldRigid abs f (s,t)
+
+instance FoldRigid Sort where
+  foldRigid abs f s =
+    case s of
+      Type l     -> fold l
+      Prop       -> mempty
+      Inf        -> mempty
+      SizeUniv   -> mempty
+      DLub s1 s2 -> fold (s1, s2)
+    where fold = foldRigid abs f
+
+instance FoldRigid Level where
+  foldRigid abs f (Max ls) = foldRigid abs f ls
+
+instance FoldRigid PlusLevel where
+  foldRigid abs f ClosedLevel{} = mempty
+  foldRigid abs f (Plus _ l)    = foldRigid abs f l
+
+instance FoldRigid LevelAtom where
+  foldRigid abs f l =
+    case l of
+      MetaLevel{} -> mempty
+      NeutralLevel MissingClauses _ -> mempty
+      NeutralLevel _              l -> fold l
+      BlockedLevel _              l -> fold l
+      UnreducedLevel              l -> fold l
+    where fold = foldRigid abs f
+
+instance (Subst a, FoldRigid a) => FoldRigid (Abs a) where
+  foldRigid abs f b = underAbstraction_ b $ foldRigid abs f
+
+instance FoldRigid a => FoldRigid (I.Arg a) where
+  foldRigid abs f a =
+    case getRelevance a of
+      Irrelevant -> mempty
+      UnusedArg  -> mempty
+      _          -> foldRigid abs f $ unArg a
+
+instance FoldRigid a => FoldRigid (I.Dom a) where
+  foldRigid abs f dom = foldRigid abs f $ unDom dom
+
+instance FoldRigid a => FoldRigid (Elim' a) where
+  foldRigid abs f (Apply a) = foldRigid abs f a
+  foldRigid abs f Proj{}    = mempty
+
+instance FoldRigid a => FoldRigid [a] where
+  foldRigid abs f = foldMap $ foldRigid abs f
+
+instance (FoldRigid a, FoldRigid b) => FoldRigid (a,b) where
+  foldRigid abs f (a,b) = foldRigid abs f a `mappend` foldRigid abs f b
 
 
 data PruneResult
@@ -550,11 +680,7 @@ killArgs kills m = do
   mv <- lookupMeta m
   allowAssign <- asks envAssignMetas
   if mvFrozen mv == Frozen || not allowAssign then return PrunedNothing else do
-{- Andreas 2011-04-26, allow pruning in MetaS
-  case mvJudgement mv of
-    IsSort _    -> return False
-    HasType _ a -> do
--}
+      -- Andreas 2011-04-26, we allow pruning in MetaV and MetaS
       let a = jMetaType $ mvJudgement mv
       TelV tel b <- telView' <$> instantiateFull a
       let args         = zip (telToList tel) (kills ++ repeat False)
@@ -566,7 +692,6 @@ killArgs kills m = do
         -- Only successful if all occurrences were killed
         -- Andreas, 2011-05-09 more precisely, check that at least
         -- the in 'kills' prescribed kills were carried out
-        -- OLD CODE: return (map unArg kills' == kills)
         return $ if (and $ zipWith implies kills $ map unArg kills')
                    then PrunedEverything
                    else PrunedSomething
@@ -594,7 +719,7 @@ killArgs kills m = do
 killedType :: [(I.Dom (ArgName, Type), Bool)] -> Type -> ([I.Arg Bool], Type)
 killedType [] b = ([], b)
 killedType ((arg@(Dom info _), kill) : kills) b
-  | dontKill  = (Arg info False : args, mkPi arg b') -- OLD: telePi (telFromList [arg]) b')
+  | dontKill  = (Arg info False : args, mkPi arg b')
   | otherwise = (Arg info True  : args, strengthen __IMPOSSIBLE__ b')
   where
     (args, b') = killedType kills b
@@ -615,32 +740,14 @@ performKill kills m a = do
       lam b a = Lam (argInfo a) (Abs "v" b)
       tel     = map ("v" <$) (reverse kills)
       u       = MetaV m' $ map Apply vars
-{- OLD CODE
-      hs   = reverse [ argHiding a | a <- kills ]
-      lam h b = Lam h (Abs "v" b)
-      u       = foldr lam (MetaV m' vars) hs
--}
   dbg m' u
   assignTerm m tel u
   where
     dbg m' u = reportSDoc "tc.meta.kill" 10 $ vcat
       [ text "actual killing"
       , nest 2 $ vcat
-        [ text "new meta:" <+> text (show m')
+        [ text "new meta:" <+> pretty m'
         , text "kills   :" <+> text (show kills)
-        , text "inst    :" <+> text (show m) <+> text ":=" <+> prettyTCM u
+        , text "inst    :" <+> pretty m <+> text ":=" <+> prettyTCM u
         ]
       ]
-
-{-
-
-  When hitting a meta variable:
-
-  - Compute flex/rigid for its arguments
-  - Compare to allowed variables
-  - Mark arguments with rigid occurrences of disallowed
-    variables for deletion
-  - Attempt to delete marked arguments
-  - We don't need to check for success, we can just
-    continue occurs checking.
--}
