@@ -29,14 +29,14 @@ class (Functor m, Applicative m, Monad m) => HasBuiltins m where
 
 litType :: Literal -> TCM Type
 litType l = case l of
-    LitInt _ n    -> do
-      primZero
-      when_ (n > 0) $ primSuc
-      el <$> primNat
-    LitFloat _ _  -> el <$> primFloat
-    LitChar _ _   -> el <$> primChar
-    LitString _ _ -> el <$> primString
-    LitQName _ _  -> el <$> primQName
+  LitInt _ n    -> do
+    _ <- primZero
+    when_ (n > 0) $ primSuc
+    el <$> primNat
+  LitFloat _ _  -> el <$> primFloat
+  LitChar _ _   -> el <$> primChar
+  LitString _ _ -> el <$> primString
+  LitQName _ _  -> el <$> primQName
   where
     el t = El (mkType 0) t
 
@@ -86,20 +86,13 @@ constructorForm :: Term -> TCM Term
 constructorForm v = constructorForm' primZero primSuc v
 
 constructorForm' :: Applicative m => m Term -> m Term -> Term -> m Term
-constructorForm' pZero pSuc v = case ignoreSharing v of
-{- 2012-04-02 changed semantics of DontCare
--- Andreas, 2011-10-03, the following line restores IrrelevantLevel
-    DontCare v                  -> constructorForm v
--}
-    Lit (LitInt r n)            -> cons (Lit . LitInt r) n
---     Level (Max [])              -> primLevelZero
---     Level (Max [ClosedLevel n]) -> cons primLevelZero primLevelSuc (Level . Max . (:[]) . ClosedLevel) n
-    _                           -> pure v
-  where
-    cons lit n
-      | n == 0    = pZero
-      | n > 0     = (`apply` [defaultArg $ lit $ n - 1]) <$> pSuc
-      | otherwise = pure v
+constructorForm' pZero pSuc v =
+  case ignoreSharing v of
+    Lit (LitInt r n)
+      | n == 0    -> pZero
+      | n > 0     -> (`apply` [defaultArg $ Lit $ LitInt r $ n - 1]) <$> pSuc
+      | otherwise -> pure v
+    _ -> pure v
 
 ---------------------------------------------------------------------------
 -- * The names of built-in things
@@ -109,12 +102,12 @@ primInteger, primFloat, primChar, primString, primBool, primTrue, primFalse,
     primList, primNil, primCons, primIO, primNat, primSuc, primZero,
     primNatPlus, primNatMinus, primNatTimes, primNatDivSucAux, primNatModSucAux,
     primNatEquality, primNatLess,
-    primSize, primSizeLt, primSizeSuc, primSizeInf,
+    primSizeUniv, primSize, primSizeLt, primSizeSuc, primSizeInf, primSizeMax,
     primInf, primSharp, primFlat,
     primEquality, primRefl,
     primRewrite, -- Name of rewrite relation
     primLevel, primLevelZero, primLevelSuc, primLevelMax,
-    primIrrAxiom, primSizeMax,
+    primIrrAxiom,
     -- builtins for reflection:
     primQName, primArgInfo, primArgArgInfo, primArg, primArgArg, primAgdaTerm, primAgdaTermVar,
     primAgdaTermLam, primAgdaTermExtLam, primAgdaTermDef, primAgdaTermCon, primAgdaTermPi,
@@ -153,6 +146,7 @@ primNatDivSucAux = getBuiltin builtinNatDivSucAux
 primNatModSucAux = getBuiltin builtinNatModSucAux
 primNatEquality  = getBuiltin builtinNatEquals
 primNatLess      = getBuiltin builtinNatLess
+primSizeUniv     = getBuiltin builtinSizeUniv
 primSize         = getBuiltin builtinSize
 primSizeLt       = getBuiltin builtinSizeLt
 primSizeSuc      = getBuiltin builtinSizeSuc
@@ -228,9 +222,11 @@ primAgdaDefinition                = getBuiltin builtinAgdaDefinition
 builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinNatTimes, builtinNatDivSucAux, builtinNatModSucAux, builtinNatEquals,
   builtinNatLess, builtinInteger, builtinFloat, builtinChar, builtinString,
-  builtinBool, builtinTrue, builtinFalse, builtinList, builtinNil,
-  builtinCons, builtinIO, builtinSize, builtinSizeLt, builtinSizeSuc,
-  builtinSizeInf, builtinSizeMax, builtinInf, builtinSharp, builtinFlat,
+  builtinBool, builtinTrue, builtinFalse,
+  builtinList, builtinNil, builtinCons, builtinIO,
+  builtinSizeUniv, builtinSize, builtinSizeLt,
+  builtinSizeSuc, builtinSizeInf, builtinSizeMax,
+  builtinInf, builtinSharp, builtinFlat,
   builtinEquality, builtinRefl, builtinRewrite, builtinLevelMax,
   builtinLevel, builtinLevelZero, builtinLevelSuc, builtinIrrAxiom,
   builtinQName, builtinAgdaSort, builtinAgdaSortSet, builtinAgdaSortLit,
@@ -274,6 +270,7 @@ builtinList                          = "LIST"
 builtinNil                           = "NIL"
 builtinCons                          = "CONS"
 builtinIO                            = "IO"
+builtinSizeUniv                      = "SIZEUNIV"
 builtinSize                          = "SIZE"
 builtinSizeLt                        = "SIZELT"
 builtinSizeSuc                       = "SIZESUC"
@@ -346,6 +343,27 @@ builtinAgdaDefinitionPostulate       = "AGDADEFINITIONPOSTULATE"
 builtinAgdaDefinitionPrimitive       = "AGDADEFINITIONPRIMITIVE"
 builtinAgdaDefinition                = "AGDADEFINITION"
 
+-- | Builtins that come without a definition in Agda syntax.
+--   These are giving names to Agda internal concepts which
+--   cannot be assigned an Agda type.
+--
+--   An example would be a user-defined name for @Set@.
+--
+--     {-# BUILTIN TYPE Type #-}
+--
+--   The type of @Type@ would be @Type : Level → Setω@
+--   which is not valid Agda.
+
+builtinsNoDef :: [String]
+builtinsNoDef =
+  [ builtinSizeUniv
+  , builtinSize
+  , builtinSizeLt
+  , builtinSizeSuc
+  , builtinSizeInf
+  , builtinSizeMax
+  ]
+
 -- | The coinductive primitives.
 
 data CoinductionKit = CoinductionKit
@@ -356,17 +374,21 @@ data CoinductionKit = CoinductionKit
 
 -- | Tries to build a 'CoinductionKit'.
 
-coinductionKit :: TCM (Maybe CoinductionKit)
-coinductionKit = (do
+coinductionKit' :: TCM CoinductionKit
+coinductionKit' = do
   Def inf   _ <- ignoreSharing <$> primInf
   Def sharp _ <- ignoreSharing <$> primSharp
   Def flat  _ <- ignoreSharing <$> primFlat
-  return $ Just $ CoinductionKit
+  return $ CoinductionKit
     { nameOfInf   = inf
     , nameOfSharp = sharp
     , nameOfFlat  = flat
-    })
-    `catchError` \_ -> return Nothing
+    }
+
+coinductionKit :: TCM (Maybe CoinductionKit)
+coinductionKit =
+  (Just <$> coinductionKit')
+  `catchError` \_ -> return Nothing
 
 ------------------------------------------------------------------------
 -- * Builtin equality
