@@ -1,4 +1,5 @@
 {-# LANGUAGE FlexibleContexts #-}
+
 module Agda.TypeChecking.Monad.Env where
 
 import Control.Monad.Reader
@@ -70,18 +71,29 @@ performedSimplification' simpl = local $ \ e -> e { envSimplification = simpl `m
 getSimplification :: MonadReader TCEnv m => m Simplification
 getSimplification = asks envSimplification
 
+-- * Controlling reduction.
+
+-- | Lens for 'AllowedReductions'.
+updateAllowedReductions :: (AllowedReductions -> AllowedReductions) -> TCEnv -> TCEnv
+updateAllowedReductions f e = e { envAllowedReductions = f (envAllowedReductions e) }
+
+modifyAllowedReductions :: (AllowedReductions -> AllowedReductions) -> TCM a -> TCM a
+modifyAllowedReductions = local . updateAllowedReductions
+
+putAllowedReductions :: AllowedReductions -> TCM a -> TCM a
+putAllowedReductions = modifyAllowedReductions . const
+
 -- | Reduce @Def f vs@ only if @f@ is a projection.
 onlyReduceProjections :: TCM a -> TCM a
-onlyReduceProjections = local $ \ e -> e { envAllowedReductions = [ProjectionReductions] }
-
-dontReduceProjections :: TCM a -> TCM a
-dontReduceProjections = local $ \ e -> e { envAllowedReductions = allReductions \\ [ProjectionReductions] }
-
-dontReduceLevels :: TCM a -> TCM a
-dontReduceLevels = local $ \ e -> e { envAllowedReductions = allReductions \\ [LevelReductions] }
+onlyReduceProjections = putAllowedReductions [ProjectionReductions]
 
+-- | Allow all reductions except for non-terminating functions (default).
 allowAllReductions :: TCM a -> TCM a
-allowAllReductions = local $ \ e -> e { envAllowedReductions = allReductions }
+allowAllReductions = putAllowedReductions allReductions
+
+-- | Allow all reductions including non-terminating functions.
+allowNonTerminatingReductions :: TCM a -> TCM a
+allowNonTerminatingReductions = putAllowedReductions $ [NonTerminatingReductions] ++ allReductions
 
 insideDotPattern :: TCM a -> TCM a
 insideDotPattern = local $ \e -> e { envInsideDotPattern = True }
@@ -89,3 +101,9 @@ insideDotPattern = local $ \e -> e { envInsideDotPattern = True }
 isInsideDotPattern :: TCM Bool
 isInsideDotPattern = asks envInsideDotPattern
 
+isReifyingUnquoted :: TCM Bool
+isReifyingUnquoted = asks envReifyUnquoted
+
+nowReifyingUnquoted :: TCM a -> TCM a
+nowReifyingUnquoted = local $ \e -> e { envReifyUnquoted = True }
+
