@@ -1,4 +1,3 @@
-
 module Agda.TypeChecking.Monad.Imports where
 
 import Control.Monad.State
@@ -11,20 +10,21 @@ import qualified Data.Set as Set
 import Agda.Syntax.Abstract.Name
 import qualified Agda.Syntax.Concrete.Name as C
 import Agda.TypeChecking.Monad.Base
+import Agda.Utils.Lens
 import Agda.Utils.Monad
 import Agda.Utils.Time
 import Agda.Utils.Hash
 
 addImport :: ModuleName -> TCM ()
 addImport m =
-    modify $ \s -> s { stImportedModules = Set.insert m $ stImportedModules s }
+    stImportedModules %= Set.insert m
 
 addImportCycleCheck :: C.TopLevelModuleName -> TCM a -> TCM a
 addImportCycleCheck m =
     local $ \e -> e { envImportPath = m : envImportPath e }
 
 getImports :: TCM (Set ModuleName)
-getImports = gets stImportedModules
+getImports = use stImportedModules
 
 isImported :: ModuleName -> TCM Bool
 isImported m = Set.member m <$> getImports
@@ -33,49 +33,48 @@ getImportPath :: TCM [C.TopLevelModuleName]
 getImportPath = asks envImportPath
 
 visitModule :: ModuleInfo -> TCM ()
-visitModule mi = modify $ \s ->
-  s { stVisitedModules =
-        Map.insert (toTopLevelModuleName $ iModuleName $ miInterface mi)
-                   mi $
-          stVisitedModules s }
+visitModule mi =
+  stVisitedModules %=
+    Map.insert (toTopLevelModuleName $ iModuleName $ miInterface mi) mi
 
 setVisitedModules :: VisitedModules -> TCM ()
-setVisitedModules ms = modify $ \s -> s { stVisitedModules = ms }
+setVisitedModules ms = stVisitedModules .= ms
 
 getVisitedModules :: TCM VisitedModules
-getVisitedModules = gets stVisitedModules
+getVisitedModules = use stVisitedModules
 
 isVisited :: C.TopLevelModuleName -> TCM Bool
-isVisited x = gets $ Map.member x . stVisitedModules
+isVisited x = Map.member x <$> use stVisitedModules
 
 getVisitedModule :: C.TopLevelModuleName
                  -> TCM (Maybe ModuleInfo)
-getVisitedModule x = gets $ Map.lookup x . stVisitedModules
+getVisitedModule x = Map.lookup x <$> use stVisitedModules
 
 getDecodedModules :: TCM DecodedModules
-getDecodedModules = stDecodedModules . stPersistent <$> get
+getDecodedModules = stDecodedModules . stPersistentState <$> get
 
 setDecodedModules :: DecodedModules -> TCM ()
 setDecodedModules ms = modify $ \s ->
-  s { stPersistent = (stPersistent s) { stDecodedModules = ms } }
+  s { stPersistentState = (stPersistentState s) { stDecodedModules = ms } }
 
 getDecodedModule :: C.TopLevelModuleName -> TCM (Maybe Interface)
-getDecodedModule x = Map.lookup x . stDecodedModules . stPersistent <$> get
+getDecodedModule x = Map.lookup x . stDecodedModules . stPersistentState <$> get
 
 storeDecodedModule :: Interface -> TCM ()
 storeDecodedModule i = modify $ \s ->
-  s { stPersistent =
-        (stPersistent s) { stDecodedModules =
+  s { stPersistentState =
+        (stPersistentState s) { stDecodedModules =
           Map.insert (toTopLevelModuleName $ iModuleName i) i $
-            (stDecodedModules $ stPersistent s)
+            (stDecodedModules $ stPersistentState s)
         }
   }
 
 dropDecodedModule :: C.TopLevelModuleName -> TCM ()
 dropDecodedModule x = modify $ \s ->
-  s { stPersistent = (stPersistent s) { stDecodedModules =
-                       Map.delete x $ stDecodedModules $ stPersistent s
-                     }
+  s { stPersistentState =
+        (stPersistentState s) { stDecodedModules =
+                                  Map.delete x $ stDecodedModules $ stPersistentState s
+                              }
   }
 
 withImportPath :: [C.TopLevelModuleName] -> TCM a -> TCM a
@@ -87,4 +86,4 @@ checkForImportCycle :: TCM ()
 checkForImportCycle = do
     m:ms <- getImportPath
     when (m `elem` ms) $ typeError $ CyclicModuleDependency
-				   $ dropWhile (/= m) $ reverse (m:ms)
+                                   $ dropWhile (/= m) $ reverse (m:ms)
