@@ -1,13 +1,13 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TupleSections        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP               #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TupleSections     #-}
 
 module Agda.TypeChecking.Monad.MetaVars where
 
 import Control.Applicative
 import Control.Monad.State
 import Control.Monad.Reader
+import Control.Monad.Writer
 
 import qualified Data.Map as Map
 import qualified Data.Set as Set
@@ -31,6 +31,7 @@ import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Tuple
 import Agda.Utils.Size
 
@@ -51,7 +52,7 @@ modifyMetaStore f = stMetaStore %= f
 -- | Lookup a meta variable
 lookupMeta :: MetaId -> TCM MetaVariable
 lookupMeta m = fromMaybeM failure $ Map.lookup m <$> getMetaStore
-  where failure = fail $ "no such meta variable " ++ show m
+  where failure = fail $ "no such meta variable " ++ prettyShow m
 
 updateMetaVar :: MetaId -> (MetaVariable -> MetaVariable) -> TCM ()
 updateMetaVar m f = modifyMetaStore $ Map.adjust f m
@@ -169,7 +170,7 @@ getMetaNameSuggestion mi = miNameSuggestion . mvInfo <$> lookupMeta mi
 setMetaNameSuggestion :: MetaId -> MetaNameSuggestion -> TCM ()
 setMetaNameSuggestion mi s = do
   reportSLn "tc.meta.name" 20 $
-    "setting name of meta " ++ show mi ++ " to " ++ s
+    "setting name of meta " ++ prettyShow mi ++ " to " ++ s
   updateMetaVar mi $ \ mvar ->
     mvar { mvInfo = (mvInfo mvar) { miNameSuggestion = s }}
 
@@ -251,16 +252,16 @@ lookupInteractionId ii = fromMaybeM err2 $ ipMeta <$> lookupInteractionPoint ii
     err2 = typeError $ GenericError $ "No type nor action available for hole " ++ show ii
 
 -- | Generate new meta variable.
-newMeta :: MetaInfo -> MetaPriority -> Permutation -> Judgement Type a -> TCM MetaId
+newMeta :: MetaInfo -> MetaPriority -> Permutation -> Judgement a -> TCM MetaId
 newMeta = newMeta' Open
 
 -- | Generate a new meta variable with some instantiation given.
 --   For instance, the instantiation could be a 'PostponedTypeCheckingProblem'.
 newMeta' :: MetaInstantiation -> MetaInfo -> MetaPriority -> Permutation ->
-            Judgement Type a -> TCM MetaId
+            Judgement a -> TCM MetaId
 newMeta' inst mi p perm j = do
   x <- fresh
-  let j' = fmap (const x) j  -- fill the identifier part of the judgement
+  let j' = j { jMetaId = x }  -- fill the identifier part of the judgement
       mv = MetaVar mi p perm j' inst Set.empty Instantiable
   -- printing not available (import cycle)
   -- reportSDoc "tc.meta.new" 50 $ text "new meta" <+> prettyTCM j'
@@ -332,17 +333,36 @@ clearMetaListeners m =
 -- * Freezing and unfreezing metas.
 ---------------------------------------------------------------------------
 
--- | Freeze all meta variables.
-freezeMetas :: TCM ()
-freezeMetas = modifyMetaStore $ Map.map freeze where
-  freeze :: MetaVariable -> MetaVariable
-  freeze mvar = mvar { mvFrozen = Frozen }
+-- | Freeze all so far unfrozen metas for the duration of the given computation.
+withFreezeMetas :: TCM a -> TCM a
+withFreezeMetas cont = do
+  ms <- Set.fromList <$> freezeMetas
+  x  <- cont
+  unfreezeMetas' (`Set.member` ms)
+  return x
+
+-- | Freeze all meta variables and return the list of metas that got frozen.
+freezeMetas :: TCM [MetaId]
+freezeMetas = execWriterT $ stMetaStore %== Map.traverseWithKey freeze
+  where
+  freeze :: Monad m => MetaId -> MetaVariable -> WriterT [MetaId] m MetaVariable
+  freeze m mvar
+    | mvFrozen mvar == Frozen = return mvar
+    | otherwise = do
+        tell [m]
+        return $ mvar { mvFrozen = Frozen }
 
 -- | Thaw all meta variables.
 unfreezeMetas :: TCM ()
-unfreezeMetas = modifyMetaStore $ Map.map unfreeze where
-  unfreeze :: MetaVariable -> MetaVariable
-  unfreeze mvar = mvar { mvFrozen = Instantiable }
+unfreezeMetas = unfreezeMetas' $ const True
+
+-- | Thaw some metas, as indicated by the passed condition.
+unfreezeMetas' :: (MetaId -> Bool) -> TCM ()
+unfreezeMetas' cond = modifyMetaStore $ Map.mapWithKey unfreeze where
+  unfreeze :: MetaId -> MetaVariable -> MetaVariable
+  unfreeze m mvar
+    | cond m    = mvar { mvFrozen = Instantiable }
+    | otherwise = mvar
 
 isFrozen :: MetaId -> TCM Bool
 isFrozen x = do
@@ -385,7 +405,7 @@ instance UnFreezeMeta PlusLevel where
 instance UnFreezeMeta LevelAtom where
   unfreezeMeta (MetaLevel x _)    = unfreezeMeta x
   unfreezeMeta (BlockedLevel _ t) = unfreezeMeta t
-  unfreezeMeta (NeutralLevel t)   = unfreezeMeta t
+  unfreezeMeta (NeutralLevel _ t) = unfreezeMeta t
   unfreezeMeta (UnreducedLevel t) = unfreezeMeta t
 
 instance UnFreezeMeta a => UnFreezeMeta [a] where
@@ -393,4 +413,3 @@ instance UnFreezeMeta a => UnFreezeMeta [a] where
 
 instance UnFreezeMeta a => UnFreezeMeta (Abs a) where
   unfreezeMeta = Fold.mapM_ unfreezeMeta
-
