@@ -1,15 +1,14 @@
 -- {-# LANGUAGE CPP #-}
 
 module Agda.TypeChecking.Monad.Open
-	( makeOpen
-	, makeClosed
-	, getOpen
-	, tryOpen
-	) where
+        ( makeOpen
+        , makeClosed
+        , getOpen
+        , tryOpen
+        ) where
 
 import Control.Applicative
 import Control.Monad
-import Control.Monad.Error
 import Data.List
 
 import Agda.TypeChecking.Substitute
@@ -17,6 +16,8 @@ import Agda.TypeChecking.Monad.Base
 
 import {-# SOURCE #-} Agda.TypeChecking.Monad.Context
 
+import Agda.Utils.Except ( MonadError(catchError) )
+
 -- | Create an open term in the current context.
 makeOpen :: a -> TCM (Open a)
 makeOpen x = do
