@@ -1,23 +1,26 @@
-{-# LANGUAGE CPP              #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
 
 module Agda.TypeChecking.Monad.Options where
 
+import Prelude hiding (mapM)
+
 import Control.Applicative
-import Control.Monad.Reader
-import Control.Monad.State
+import Control.Monad.Reader hiding (mapM)
+import Control.Monad.State  hiding (mapM)
+
 import Data.Maybe
+import Data.Traversable
+
 import Text.PrettyPrint
 import System.Directory
 import System.FilePath
 
-import Paths_Agda (getDataFileName)
--- NB: find Paths_Agda.hs in dist/build/autogen/
-
 import Agda.Syntax.Concrete
 import {-# SOURCE #-} Agda.TypeChecking.Errors
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.State
+import Agda.TypeChecking.Monad.Benchmark
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
 import qualified Agda.Interaction.Options.Lenses as Lens
@@ -25,6 +28,7 @@ import Agda.Interaction.Response
 
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.FileName
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Lens
 import Agda.Utils.List
@@ -45,6 +49,7 @@ setPragmaOptions opts = do
     Left err   -> __IMPOSSIBLE__
     Right opts -> do
       stPragmaOptions .= optPragmaOptions opts
+      updateBenchmarkingStatus
 
 -- | Sets the command line options (both persistent and pragma options
 -- are updated).
@@ -71,6 +76,7 @@ setCommandLineOptions opts =
           getIncludeDirs
       modify $ Lens.setCommandLineOptions opts{ optIncludeDirs = Right incs }
              . Lens.setPragmaOptions (optPragmaOptions opts)
+      updateBenchmarkingStatus
 
 class (Functor m, Applicative m, Monad m) => HasOptions m where
   -- | Returns the pragma options which are currently in effect.
@@ -182,7 +188,7 @@ setIncludeDirs incs relativeTo = do
   incs <- return $  map (mkAbsolute . (filePath root </>)) incs
 
   -- Andreas, 2013-10-30  Add default include dir
-  libdir <- liftIO $ getDataFileName ("lib")
+  libdir <- liftIO $ defaultLibDir
       -- NB: This is an absolute file name, but
       -- Agda.Utils.FilePath wants to check absoluteness anyway.
   let primdir = mkAbsolute $ libdir </> "prim"
@@ -216,11 +222,13 @@ setInputFile file =
 
 -- | Should only be run if 'hasInputFile'.
 getInputFile :: TCM AbsolutePath
-getInputFile =
-    do  mf <- optInputFile <$> commandLineOptions
-        case mf of
-            Just file -> liftIO $ absolute file
-            Nothing   -> __IMPOSSIBLE__
+getInputFile = fromMaybeM __IMPOSSIBLE__ $
+  getInputFile'
+
+-- | Return the 'optInputFile' as 'AbsolutePath', if any.
+getInputFile' :: TCM (Maybe AbsolutePath)
+getInputFile' = mapM (liftIO . absolute) =<< do
+  optInputFile <$> commandLineOptions
 
 hasInputFile :: TCM Bool
 hasInputFile = isJust <$> optInputFile <$> commandLineOptions
