@@ -1,18 +1,23 @@
-{-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleContexts  #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DoAndIfThenElse #-}
+{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards     #-}
+{-# LANGUAGE PatternGuards #-}
 
 module Agda.TypeChecking.Monad.Signature where
 
-import Control.Applicative
+import Prelude hiding (null)
+
+import Control.Arrow (first, second, (***))
+import Control.Applicative hiding (empty)
 import Control.Monad.State
 import Control.Monad.Reader
 
-import Data.List
+import Data.List hiding (null)
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Maybe
+import Data.Monoid
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Abstract (Ren)
@@ -30,20 +35,23 @@ import Agda.TypeChecking.Monad.Exception ( ExceptionT )
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Open
 import Agda.TypeChecking.Monad.State
+import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import {-# SOURCE #-} Agda.TypeChecking.Polarity
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike
 
+import Agda.Utils.Except ( Error )
+import Agda.Utils.Functor
+import Agda.Utils.Lens
+import Agda.Utils.List
 import Agda.Utils.Map as Map
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
-import Agda.Utils.Size
+import Agda.Utils.Null
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty
-import Agda.Utils.Lens
-import Agda.Utils.List
-import Agda.Utils.Except ( Error )
+import Agda.Utils.Size
 import qualified Agda.Utils.HashMap as HMap
 
 #include "undefined.h"
@@ -59,8 +67,7 @@ addConstant q d = do
               _             -> tel
   let d' = abstract tel' $ d { defName = q }
   reportSLn "tc.signature" 30 $ "lambda-lifted definition = " ++ show d'
-  modifySignature $ \sig -> sig
-    { sigDefinitions = HMap.insertWith (+++) q d' $ sigDefinitions sig }
+  modifySignature $ updateDefinitions $ HMap.insertWith (+++) q d'
   i <- currentOrFreshMutualBlock
   setMutualBlock i q
   where
@@ -129,43 +136,58 @@ markStatic q = modifySignature $ updateDefinition q $ mark
 unionSignatures :: [Signature] -> Signature
 unionSignatures ss = foldr unionSignature emptySignature ss
   where
-    unionSignature (Sig a b) (Sig c d) = Sig (Map.union a c) (HMap.union b d)
+    unionSignature (Sig a b c) (Sig a' b' c') =
+      Sig (Map.union a a')
+          (HMap.union b b')              -- definitions are unique (in at most one module)
+          (HMap.unionWith mappend c c')  -- rewrite rules are accumulated
 
 -- | Add a section to the signature.
-addSection :: ModuleName -> Nat -> TCM ()
-addSection m fv = do
+--
+--   The current context will be stored as the cumulative module parameters
+--   for this section.
+addSection :: ModuleName -> TCM ()
+addSection m = do
   tel <- getContextTelescope
-  let sec = Section tel fv
-  modifySignature $ \sig -> sig { sigSections = Map.insert m sec $ sigSections sig }
+  let sec = Section tel
+  -- Make sure we do not overwrite an existing section!
+  whenJustM (getSection m) $ \ sec' -> do
+    -- At least not with different content!
+    if (sec == sec') then do
+      -- Andreas, 2015-12-02: test/Succeed/Issue1701II.agda
+      -- reports a "redundantly adding existing section".
+      reportSLn "tc.section" 10 $ "warning: redundantly adding existing section " ++ show m
+      reportSLn "tc.section" 60 $ "with content " ++ show sec
+    else do
+      reportSLn "impossible" 10 $ "overwriting existing section " ++ show m
+      reportSLn "impossible" 60 $ "of content   " ++ show sec'
+      reportSLn "impossible" 60 $ "with content " ++ show sec
+      __IMPOSSIBLE__
+  -- Add the new section.
+  modifySignature $ over sigSections $ Map.insert m sec
 
 -- | Lookup a section. If it doesn't exist that just means that the module
 --   wasn't parameterised.
 lookupSection :: ModuleName -> TCM Telescope
 lookupSection m = do
-  sig  <- sigSections <$> getSignature
-  isig <- sigSections <$> getImportedSignature
-  return $ maybe EmptyTel secTelescope $ Map.lookup m sig `mplus` Map.lookup m isig
+  sig  <- use $ stSignature . sigSections
+  isig <- use $ stImports   . sigSections
+  return $ maybe EmptyTel (^. secTelescope) $ Map.lookup m sig `mplus` Map.lookup m isig
 
 -- Add display forms to all names @xn@ such that @x = x1 es1@, ... @xn-1 = xn esn@.
 addDisplayForms :: QName -> TCM ()
 addDisplayForms x = do
   def  <- getConstInfo x
   args <- getContextArgs
-{- OLD
-  n    <- do
-    proj <- isProjection x
-    return $ case proj of
-      Just (_, n) -> n
-      Nothing     -> 0
--}
   add (drop (projectionArgs $ theDef def) args) x x []
   where
     add args top x vs0 = do
       def <- getConstInfo x
       let cs = defClauses def
+          isCopy = defCopy def
       case cs of
         [ Clause{ namedClausePats = pats, clauseBody = b } ]
-          | all (isVar . namedArg) pats
+          | isCopy
+          , all (isVar . namedArg) pats
           , Just (m, Def y es) <- strip (b `apply` vs0)
           , Just vs <- mapM isApplyElim es -> do
               let ps = raise 1 $ map unArg vs
@@ -175,7 +197,8 @@ addDisplayForms x = do
               addDisplayForm y df
               add args top y vs
         _ -> do
-              let reason = case cs of
+          let reason = if not isCopy then "not a copy" else
+                  case cs of
                     []    -> "no clauses"
                     _:_:_ -> "many clauses"
                     [ Clause{ clauseBody = b } ] -> case strip b of
@@ -185,8 +208,9 @@ addDisplayForms x = do
                         | m > length args -> "too many args"
                         | otherwise       -> "args=" ++ show args ++ " es=" ++ show es
                       Just (m, v) -> "not a def body"
-              reportSLn "tc.display.section" 30 $ "no display form from " ++ show x ++ " because " ++ reason
-              return ()
+          reportSLn "tc.display.section" 30 $
+            "no display form from " ++ show x ++ " because " ++ reason
+
     strip (Body v)   = return (0, unSpine v)
     strip  NoBody    = Nothing
     strip (Bind b)   = do
@@ -206,6 +230,10 @@ applySection
   -> Ren ModuleName -- ^ Imported modules (given as renaming).
   -> TCM ()
 applySection new ptel old ts rd rm = do
+  applySection' new ptel old ts rd rm
+
+applySection' :: ModuleName -> Telescope -> ModuleName -> Args -> Ren QName -> Ren ModuleName -> TCM ()
+applySection' new ptel old ts rd rm = do
   reportSLn "tc.mod.apply" 10 $ render $ vcat
     [ text "applySection"
     , text "new  =" <+> text (show new)
@@ -216,9 +244,9 @@ applySection new ptel old ts rd rm = do
   reportSLn "tc.mod.apply" 80 $ render $ vcat
     [ text "arguments:  " <+> text (show ts)
     ]
-  mapM_ (copyDef ts) $ Map.toList rd
-  mapM_ (copySec ts) $ Map.toList rm
-  mapM_ computePolarity (Map.elems rd)
+  mapM_ (copyDef ts) rd
+  mapM_ (copySec ts) rm
+  mapM_ computePolarity (map snd rd)
   where
     -- Andreas, 2013-10-29
     -- Here, if the name x is not imported, it persists as
@@ -228,15 +256,12 @@ applySection new ptel old ts rd rm = do
     -- I guess it would make sense to mark non-imported names
     -- as such (out-of-scope) and let splitting fail if it would
     -- produce out-of-scope constructors.
-    copyName x = Map.findWithDefault x x rd
+    copyName x = fromMaybe x $ lookup x rd
 
-    argsToUse new = do
-      let m = mnameFromList $ commonPrefix (mnameToList old) (mnameToList new)
+    argsToUse x = do
+      let m = commonParentModule old x
       reportSLn "tc.mod.apply" 80 $ "Common prefix: " ++ show m
-      let ms = tail . map mnameFromList . inits . mnameToList $ m
-      ps <- sequence [ maybe 0 secFreeVars <$> getSection m | m <- ms ]
-      reportSLn "tc.mod.apply" 80 $ "  params: " ++ show (zip ms ps)
-      return $ sum ps
+      size <$> lookupSection m
 
     copyDef :: Args -> (QName, QName) -> TCM ()
     copyDef ts (x, y) = do
@@ -245,60 +270,55 @@ applySection new ptel old ts rd rm = do
       copyDef' np def
       where
         copyDef' np d = do
-          reportSLn "tc.mod.apply" 80 $ "making new def for " ++ show y ++ " from " ++ show x ++ " with " ++ show np ++ " args"
-          reportSLn "tc.mod.apply" 80 $ "args = " ++ show ts' ++ "\n" ++
-                                        "old type = " ++ showTerm (unEl $ defType d) ++ "\n" ++
-                                        "new type = " ++ showTerm (unEl t)
+          reportSLn "tc.mod.apply" 60 $ "making new def for " ++ show y ++ " from " ++ show x ++ " with " ++ show np ++ " args " ++ show abstr
+          reportSLn "tc.mod.apply" 80 $
+            "args = " ++ show ts' ++ "\n" ++
+            "old type = " ++ prettyShow (defType d) ++ "\n" ++
+            "new type = " ++ prettyShow t
           addConstant y =<< nd y
           makeProjection y
+          -- Issue1238: the copied def should be an 'instance' if the original
+          -- def is one. Skip constructors since the original constructor will
+          -- still work as an instance.
+          unless isCon $ whenJust inst $ \ c -> addNamedInstance y c
           -- Set display form for the old name if it's not a constructor.
 {- BREAKS fail/Issue478
           -- Andreas, 2012-10-20 and if we are not an anonymous module
           -- unless (isAnonymousModuleName new || isCon || size ptel > 0) $ do
 -}
-          -- Issue1238: the copied def should be an 'instance' if the original
-          -- def is one. Skip constructors since the original constructor will
-          -- still work as an instance.
-          unless isCon $ flip (maybe (return ())) inst $ \c -> addNamedInstance y c
-
-          unless (isCon || size ptel > 0) $ do
+          -- BREAKS fail/Issue1643a
+          -- -- Andreas, 2015-09-09 Issue 1643:
+          -- -- Do not add a display form for a bare module alias.
+          -- when (not isCon && size ptel == 0 && not (null ts)) $ do
+          when (not isCon && size ptel == 0) $ do
             addDisplayForms y
           where
             ts' = take np ts
             t   = defType d `apply` ts'
             pol = defPolarity d `apply` ts'
             occ = defArgOccurrences d `apply` ts'
-            rew = defRewriteRules d `apply` ts'
             inst = defInstance d
+            abstr = defAbstract d
             -- the name is set by the addConstant function
             nd :: QName -> TCM Definition
-            nd y = Defn (defArgInfo d) y t pol occ [] (-1) noCompiledRep rew inst <$> def  -- TODO: mutual block?
+            nd y = Defn (defArgInfo d) y t pol occ [] (-1) noCompiledRep inst <$> def  -- TODO: mutual block?
             oldDef = theDef d
             isCon  = case oldDef of { Constructor{} -> True ; _ -> False }
             mutual = case oldDef of { Function{funMutual = m} -> m              ; _ -> [] }
             extlam = case oldDef of { Function{funExtLam = e} -> e              ; _ -> Nothing }
             with   = case oldDef of { Function{funWith = w}   -> copyName <$> w ; _ -> Nothing }
-{- THIS BREAKS A LOT OF THINGS:
-            -- Andreas, 2013-10-21:
-            -- Even if we apply the record argument, we stay a projection.
+            -- Andreas, 2015-05-11, to fix issue 1413:
+            -- Even if we apply the record argument (must be @var 0@), we stay a projection.
             -- This is because we may abstract the record argument later again.
             -- See succeed/ProjectionNotNormalized.agda
+            isVar0 t = case ignoreSharing $ unArg t of Var 0 [] -> True; _ -> False
             proj   = case oldDef of
               Function{funProjection = Just p@Projection{projIndex = n}}
+                | size ts < n || (size ts == n && isVar0 (last ts))
                 -> Just $ p { projIndex    = n - size ts
                             , projDropPars = projDropPars p `apply` ts
                             }
               _ -> Nothing
--}
-            -- NB (Andreas, 2013-10-19):
-            -- If we apply the record argument, we are no longer a projection!
-            proj   = case oldDef of
-              Function{funProjection = Just p@Projection{projIndex = n}} | size ts < n
-                -> Just $ p { projIndex    = n - size ts
-                            , projDropPars = projDropPars p `apply` ts
-                            }
-              _ -> Nothing
-
             def =
               case oldDef of
                 Constructor{ conPars = np, conData = d } -> return $
@@ -313,8 +333,8 @@ applySection new ptel old ts rd rm = do
                 Record{ recPars = np, recConType = t, recTel = tel } -> return $
                   oldDef { recPars    = np - size ts'
                          , recClause  = Just cl
-                         , recConType = apply t ts
-                         , recTel     = apply tel ts
+                         , recConType = apply t ts'
+                         , recTel     = apply tel ts'
                          }
                 _ -> do
                   cc <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
@@ -324,7 +344,7 @@ applySection new ptel old ts rd rm = do
                         , funDelayed        = NotDelayed
                         , funInv            = NotInjective
                         , funMutual         = mutual
-                        , funAbstr          = ConcreteDef
+                        , funAbstr          = ConcreteDef -- OR: abstr -- ?!
                         , funProjection     = proj
                         , funStatic         = False
                         , funCopy           = True
@@ -348,28 +368,55 @@ applySection new ptel old ts rd rm = do
                         , clauseType      = Just $ defaultArg t
                         }
 
+    {- Example
+
+    module Top Θ where
+      module A Γ where
+        module M Φ where
+      module B Δ where
+        module N Ψ where
+          module O Ψ' where
+        open A public     -- introduces only M --> A.M into the *scope*
+    module C Ξ = Top.B ts
+
+    new section C
+      tel = Ξ.(Θ.Δ)[ts]
+
+    calls
+      1. copySec ts (Top.A.M, C.M)
+      2. copySec ts (Top.B.N, C.N)
+      3. copySec ts (Top.B.N.O, C.N.O)
+    with
+      old = Top.B
+
+    For 1.
+      Common prefix is: Top
+      totalArgs = |Θ|   (section Top)
+      tel       = Θ.Γ.Φ (section Top.A.M)
+      ts'       = take totalArgs ts
+      Θ₂        = drop totalArgs Θ
+      new section C.M
+        tel =  Θ₂.Γ.Φ[ts']
+    -}
     copySec :: Args -> (ModuleName, ModuleName) -> TCM ()
     copySec ts (x, y) = do
-      np  <- argsToUse x
-      tel <- lookupSection x
-      let fv = size tel - np
+      totalArgs <- argsToUse x
+      tel       <- lookupSection x
+      let sectionTel =  apply tel $ take totalArgs ts
       reportSLn "tc.mod.apply" 80 $ "Copying section " ++ show x ++ " to " ++ show y
-      reportSLn "tc.mod.apply" 80 $ "  free variables: " ++ show fv
-      reportSLn "tc.mod.apply" 80 $ "  ts  = " ++ show ts
-      reportSLn "tc.mod.apply" 80 $ "  tel = " ++ show tel
-      reportSLn "tc.mod.apply" 80 $ "  np  = " ++ show np
-      addCtxTel (apply tel $ take np ts) $ addSection y fv
+      reportSLn "tc.mod.apply" 80 $ "  ts           = " ++ intercalate "; " (map prettyShow ts)
+      reportSLn "tc.mod.apply" 80 $ "  totalArgs    = " ++ show totalArgs
+      reportSLn "tc.mod.apply" 80 $ "  tel          = " ++ intercalate " " (map (fst . unDom) $ telToList tel)  -- only names
+      reportSLn "tc.mod.apply" 80 $ "  sectionTel   = " ++ intercalate " " (map (fst . unDom) $ telToList ptel) -- only names
+      addCtxTel sectionTel $ addSection y
 
+-- | Add a display form to a definition (could be in this or imported signature).
 addDisplayForm :: QName -> DisplayForm -> TCM ()
 addDisplayForm x df = do
   d <- makeOpen df
-  modifyImportedSignature (add d)
-  modifySignature (add d)
-  where
-    add df sig = sig { sigDefinitions = HMap.adjust addDf x defs }
-      where
-        addDf def = def { defDisplay = df : defDisplay def }
-        defs      = sigDefinitions sig
+  let add = updateDefinition x $ \ def -> def{ defDisplay = d : defDisplay def }
+  modifyImportedSignature add
+  modifySignature add
 
 canonicalName :: QName -> TCM QName
 canonicalName x = do
@@ -424,13 +471,24 @@ class (Functor m, Applicative m, Monad m) => HasConstInfo m where
   -- | Lookup the definition of a name. The result is a closed thing, all free
   --   variables have been abstracted over.
   getConstInfo :: QName -> m Definition
+  -- | Lookup the rewrite rules with the given head symbol.
+  getRewriteRulesFor :: QName -> m RewriteRules
 
 {-# SPECIALIZE getConstInfo :: QName -> TCM Definition #-}
 
+defaultGetRewriteRulesFor :: (Monad m) => m TCState -> QName -> m RewriteRules
+defaultGetRewriteRulesFor getTCState q = do
+  st <- getTCState
+  let sig = st^.stSignature
+      imp = st^.stImports
+      look s = HMap.lookup q $ s ^. sigRewriteRules
+  return $ mconcat $ catMaybes [look sig, look imp]
+
 instance HasConstInfo (TCMT IO) where
+  getRewriteRulesFor = defaultGetRewriteRulesFor get
   getConstInfo q = join $ pureTCM $ \st env ->
-    let defs  = sigDefinitions $ st^.stSignature
-        idefs = sigDefinitions $ st^.stImports
+    let defs  = st^.(stSignature . sigDefinitions)
+        idefs = st^.(stImports . sigDefinitions)
     in case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
         []  -> fail $ "Unbound name: " ++ show q ++ " " ++ showQNameId q
         [d] -> mkAbs env d
@@ -452,16 +510,13 @@ instance HasConstInfo (TCMT IO) where
             _             -> q
 
           dropLastModule q@QName{ qnameModule = m } =
-            q{ qnameModule = mnameFromList $ init' $ mnameToList m }
-
-          init' [] = {-'-} __IMPOSSIBLE__
-          init' xs = init xs
+            q{ qnameModule = mnameFromList $ ifNull (mnameToList m) __IMPOSSIBLE__ init }
 
 instance (HasConstInfo m, Error err) => HasConstInfo (ExceptionT err m) where
   getConstInfo = lift . getConstInfo
+  getRewriteRulesFor = lift . getRewriteRulesFor
 
 {-# INLINE getConInfo #-}
-{-# SPECIALIZE getConstInfo :: QName -> TCM Definition #-}
 getConInfo :: MonadTCM tcm => ConHead -> tcm Definition
 getConInfo = liftTCM . getConstInfo . conName
 
@@ -479,50 +534,20 @@ getPolarity' CmpLeq q = getPolarity q -- composition with Covariant is identity
 setPolarity :: QName -> [Polarity] -> TCM ()
 setPolarity q pol = modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
 
--- | Return a finite list of argument occurrences.
-getArgOccurrences :: QName -> TCM [Occurrence]
-getArgOccurrences d = defArgOccurrences <$> getConstInfo d
-
-{- OLD
--- | Return a finite list of argument occurrences.
-getArgOccurrences :: QName -> TCM [Occurrence]
-getArgOccurrences d = do
-  def <- theDef <$> getConstInfo d
-  return $ getArgOccurrences_ def
-
-getArgOccurrences_ :: Defn -> [Occurrence]
-getArgOccurrences_ def = case def of
-    Function { funArgOccurrences  = os } -> os
-    Datatype { dataArgOccurrences = os } -> os
-    Record   { recArgOccurrences  = os } -> os
-    Constructor{}                        -> [] -- repeat StrictPos
-    _                                    -> [] -- repeat Mixed
--}
-
+-- | Get argument occurrence info for argument @i@ of definition @d@ (never fails).
 getArgOccurrence :: QName -> Nat -> TCM Occurrence
 getArgOccurrence d i = do
   def <- getConstInfo d
   return $ case theDef def of
     Constructor{} -> StrictPos
-    _             -> (defArgOccurrences def ++ repeat Mixed) !! i
+    _             -> fromMaybe Mixed $ defArgOccurrences def !!! i
 
 setArgOccurrences :: QName -> [Occurrence] -> TCM ()
-setArgOccurrences d os =
-  modifySignature $ updateDefinition d $ updateDefArgOccurrences $ const os
+setArgOccurrences d os = modifyArgOccurrences d $ const os
 
-{- OLD
-getArgOccurrence :: QName -> Nat -> TCM Occurrence
-getArgOccurrence d i = do
-  def <- theDef <$> getConstInfo d
-  return $ case def of
-    Function { funArgOccurrences  = os } -> look i os
-    Datatype { dataArgOccurrences = os } -> look i os
-    Record   { recArgOccurrences  = os } -> look i os
-    Constructor{}                        -> StrictPos
-    _                                    -> Mixed
-  where
-    look i os = (os ++ repeat Mixed) !! fromIntegral i
--}
+modifyArgOccurrences :: QName -> ([Occurrence] -> [Occurrence]) -> TCM ()
+modifyArgOccurrences d f =
+  modifySignature $ updateDefinition d $ updateDefArgOccurrences f
 
 -- | Get the mutually recursive identifiers.
 getMutual :: QName -> TCM [QName]
@@ -552,30 +577,22 @@ mutuallyRecursive d d' = (d `elem`) <$> getMutual d'
 --   A.B.C say, A and A.B do not exist.
 getSection :: ModuleName -> TCM (Maybe Section)
 getSection m = do
-  sig  <- sigSections <$> getSignature
-  isig <- sigSections <$> getImportedSignature
+  sig  <- use $ stSignature . sigSections
+  isig <- use $ stImports   . sigSections
   return $ Map.lookup m sig <|> Map.lookup m isig
 
--- | Look up the number of free variables of a section. This is equal to the
---   number of parameters if we're currently inside the section and 0 otherwise.
-getSecFreeVars :: ModuleName -> TCM Nat
-getSecFreeVars m = do
-  top <- currentModule
-  case top `isSubModuleOf` m || top == m of
-    True  -> maybe 0 secFreeVars <$> getSection m
-    False -> return 0
-
--- | Compute the number of free variables of a module. This is the sum of
---   the free variables of its sections.
-getModuleFreeVars :: ModuleName -> TCM Nat
-getModuleFreeVars m = sum <$> ((:) <$> getAnonymousVariables m <*> mapM getSecFreeVars ms)
-  where
-    ms = map mnameFromList . inits . mnameToList $ m
+-- | Get the number of parameters to the current module.
+getCurrentModuleFreeVars :: TCM Nat
+getCurrentModuleFreeVars = size <$> (lookupSection =<< currentModule)
 
 -- | Compute the number of free variables of a defined name. This is the sum of
---   the free variables of the sections it's contained in.
+--   number of parameters shared with the current module and the number of
+--   anonymous variables (if the name comes from a let-bound module).
 getDefFreeVars :: QName -> TCM Nat
-getDefFreeVars q = getModuleFreeVars (qnameModule q)
+getDefFreeVars q = do
+  let m = qnameModule q
+  m0   <- commonParentModule m <$> currentModule
+  (+) <$> getAnonymousVariables m <*> (size <$> lookupSection m0)
 
 -- | Compute the context variables to apply a definition to.
 freeVarsToApply :: QName -> TCM Args
@@ -632,20 +649,43 @@ ignoreAbstractMode = local $ \e -> e { envAbstractMode = IgnoreAbstractMode,
                                        -- Allowing destructive updates when ignoring
                                        -- abstract may break the abstraction.
 
+-- | Enter concrete or abstract mode depending on whether the given identifier
+--   is concrete or abstract.
+inConcreteOrAbstractMode :: QName -> TCM a -> TCM a
+inConcreteOrAbstractMode q cont = do
+  -- Andreas, 2015-07-01: If we do not ignoreAbstractMode here,
+  -- we will get ConcreteDef for abstract things, as they are turned into axioms.
+  a <- ignoreAbstractMode $ defAbstract <$> getConstInfo q
+  case a of
+    AbstractDef -> inAbstractMode cont
+    ConcreteDef -> inConcreteMode cont
+
 -- | Check whether a name might have to be treated abstractly (either if we're
 --   'inAbstractMode' or it's not a local name). Returns true for things not
 --   declared abstract as well, but for those 'makeAbstract' will have no effect.
 treatAbstractly :: MonadReader TCEnv m => QName -> m Bool
 treatAbstractly q = asks $ treatAbstractly' q
 
+-- | Andreas, 2015-07-01:
+--   If the @current@ module is a weak suffix of the identifier module,
+--   we can see through its abstract definition if we are abstract.
+--   (Then @treatAbstractly'@ returns @False@).
+--
+--   If I am not mistaken, then we cannot see definitions in the @where@
+--   block of an abstract function from the perspective of the function,
+--   because then the current module is a strict prefix of the module
+--   of the local identifier.
+--   This problem is fixed by removing trailing anonymous module name parts
+--   (underscores) from both names.
 treatAbstractly' :: QName -> TCEnv -> Bool
 treatAbstractly' q env = case envAbstractMode env of
   ConcreteMode       -> True
   IgnoreAbstractMode -> False
   AbstractMode       -> not $ current == m || current `isSubModuleOf` m
   where
-    current = envCurrentModule env
-    m       = qnameModule q
+    current = dropAnon $ envCurrentModule env
+    m       = dropAnon $ qnameModule q
+    dropAnon (MName ms) = MName $ reverse $ dropWhile isNoName $ reverse ms
 
 -- | Get type of a constant, instantiated to the current context.
 typeOfConst :: QName -> TCM Type
@@ -667,8 +707,30 @@ sortOfConst q =
             Datatype{dataSort = s} -> return s
             _                      -> fail $ "Expected " ++ show q ++ " to be a datatype."
 
+-- | The number of parameters of a definition.
+defPars :: Definition -> Int
+defPars d = case theDef d of
+    Axiom{}                  -> 0
+    def@Function{}           -> projectionArgs def
+    Datatype  {dataPars = n} -> n
+    Record     {recPars = n} -> n
+    Constructor{conPars = n} -> n
+    Primitive{}              -> 0
+
+-- | The number of dropped parameters for a definition.
+--   0 except for projection(-like) functions and constructors.
+droppedPars :: Definition -> Int
+droppedPars d = case theDef d of
+    Axiom{}                  -> 0
+    def@Function{}           -> projectionArgs def
+    Datatype  {dataPars = _} -> 0  -- not dropped
+    Record     {recPars = _} -> 0  -- not dropped
+    Constructor{conPars = n} -> n
+    Primitive{}              -> 0
+
 -- | Is it the name of a record projection?
-isProjection :: QName -> TCM (Maybe Projection)
+{-# SPECIALIZE isProjection :: QName -> TCM (Maybe Projection) #-}
+isProjection :: HasConstInfo m => QName -> m (Maybe Projection)
 isProjection qn = isProjection_ . theDef <$> getConstInfo qn
 
 isProjection_ :: Defn -> Maybe Projection
@@ -677,23 +739,35 @@ isProjection_ def =
     Function { funProjection = result } -> result
     _                                   -> Nothing
 
+-- | Returns @True@ if we are dealing with a proper projection,
+--   i.e., not a projection-like function nor a record field value
+--   (projection applied to argument).
 isProperProjection :: Defn -> Bool
-isProperProjection = isJust . (projProper <=< isProjection_)
--- isProperProjection = maybe False projProper . isProjection_
+isProperProjection d = caseMaybe (isProjection_ d) False $ \ isP ->
+  if projIndex isP <= 0 then False else isJust $ projProper isP
 
--- | Number of dropped initial arguments.
+-- | Number of dropped initial arguments of a projection(-like) function.
 projectionArgs :: Defn -> Int
-projectionArgs = maybe 0 (pred . projIndex) . isProjection_
+projectionArgs = maybe 0 (max 0 . pred . projIndex) . isProjection_
+
+-- | Check whether a definition uses copatterns.
+usesCopatterns :: QName -> TCM Bool
+usesCopatterns q = do
+  d <- theDef <$> getConstInfo q
+  return $ case d of
+    Function{ funCopatternLHS = b } -> b
+    _ -> False
 
 -- | Apply a function @f@ to its first argument, producing the proper
 --   postfix projection if @f@ is a projection.
 applyDef :: QName -> I.Arg Term -> TCM Term
 applyDef f a = do
-  -- get the original projection, if existing
-  res <- (projProper =<<) <$> isProjection f
-  case res of
-    Nothing -> return $ Def f [Apply a]
-    Just f' -> return $ unArg a `applyE` [Proj f']
+  let fallback = return $ Def f [Apply a]
+  caseMaybeM (isProjection f) fallback $ \ isP -> do
+    if projIndex isP <= 0 then fallback else do
+      -- Get the original projection, if existing.
+      caseMaybe (projProper isP) fallback $ \ f' -> do
+        return $ unArg a `applyE` [Proj f']
 
 -- | @getDefType f t@ computes the type of (possibly projection-(like))
 --   function @t@ whose first argument has type @t@.
@@ -709,8 +783,9 @@ getDefType f t = do
   def <- getConstInfo f
   let a = defType def
   -- if @f@ is not a projection (like) function, @a@ is the correct type
-  caseMaybe (isProjection_ $ theDef def) (return $ Just a) $
-    \ (Projection{ projIndex = n }) -> do
+      fallback = return $ Just a
+  caseMaybe (isProjection_ $ theDef def) fallback $
+    \ (Projection{ projIndex = n }) -> if n <= 0 then fallback else do
       -- otherwise, we have to instantiate @a@ to the "parameters" of @f@
       let npars | n == 0    = __IMPOSSIBLE__
                 | otherwise = n - 1
