@@ -1,10 +1,10 @@
-{-# LANGUAGE CPP           #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE TupleSections #-}
 
 -- | Stuff for sized types that does not require modules
---   'Agda.TypeChecking.Reduce' or 'Agda.TypeChecking.Constraints'
---   (which import 'Agda.TypeChecking.Monad').
+--   "Agda.TypeChecking.Reduce" or "Agda.TypeChecking.Constraints"
+--   (which import "Agda.TypeChecking.Monad").
 
 module Agda.TypeChecking.Monad.SizedTypes where
 
@@ -20,9 +20,11 @@ import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Signature
 import Agda.TypeChecking.Monad.State
+import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute ()
 
 import Agda.Utils.Except ( MonadError(catchError) )
+import Agda.Utils.List
 import Agda.Utils.Monad
 
 #include "undefined.h"
@@ -85,8 +87,8 @@ haveSizedTypes = do
   `catchError` \_ -> return False
 
 -- | Add polarity info to a SIZE builtin.
-builtinSizeHook :: String -> QName -> Term -> Type -> TCM ()
-builtinSizeHook s q e' t = do
+builtinSizeHook :: String -> QName -> Type -> TCM ()
+builtinSizeHook s q t = do
   when (s `elem` [builtinSizeLt, builtinSizeSuc]) $ do
     modifySignature $ updateDefinition q
       $ updateDefPolarity       (const [Covariant])
@@ -135,7 +137,9 @@ sizeSuc n v | n < 0     = __IMPOSSIBLE__
             | n == 0    = return v
             | otherwise = do
   Def suc [] <- ignoreSharing <$> primSizeSuc
-  return $ iterate (sizeSuc_ suc) v !! n
+  return $ case iterate (sizeSuc_ suc) v !!! n of
+             Nothing -> __IMPOSSIBLE__
+             Just t  -> t
 
 sizeSuc_ :: QName -> Term -> Term
 sizeSuc_ suc v = Def suc [Apply $ defaultArg v]
@@ -174,6 +178,7 @@ data DeepSizeView
   | DSizeVar Nat Offset
   | DSizeMeta MetaId Elims Offset
   | DOtherSize Term
+  deriving (Show)
 
 data SizeViewComparable a
   = NotComparable
