@@ -1,7 +1,6 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE ScopedTypeVariables  #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 -- | Tools to manipulate patterns in abstract syntax
 --   in the TCM (type checking monad).
@@ -42,7 +41,7 @@ expandLitPattern p = traverse (traverse expand) p
           let zero  = A.ConP cinfo (A.AmbQ [setRange r $ conName z]) []
               suc p = A.ConP cinfo (A.AmbQ [setRange r $ conName s]) [defaultNamedArg p]
               info  = A.PatRange r
-              cinfo = A.ConPatInfo False info
+              cinfo = A.ConPatInfo ConPCon info
               p'    = foldr ($) zero $ genericReplicate n suc
           return $ foldr (A.AsP info) p' xs
       _ -> return p
@@ -74,13 +73,12 @@ instance ExpandPatternSynonyms A.Pattern where
     A.VarP{}             -> return p
     A.WildP{}            -> return p
     A.DotP{}             -> return p
-    A.ImplicitP{}        -> return p
     A.LitP{}             -> return p
     A.AbsurdP{}          -> return p
     A.ConP i ds as       -> A.ConP i ds <$> expandPatternSynonyms as
     A.DefP i q as        -> A.DefP i q <$> expandPatternSynonyms as
     A.AsP i x p          -> A.AsP i x <$> expandPatternSynonyms p
-    A.PatternSynP i x as -> setCurrentRange (getRange i) $ do
+    A.PatternSynP i x as -> setCurrentRange i $ do
       p <- killRange <$> lookupPatternSyn x
         -- Must expand arguments before instantiating otherwise pattern
         -- synonyms could get into dot patterns (which is __IMPOSSIBLE__)
@@ -90,7 +88,7 @@ instance ExpandPatternSynonyms A.Pattern where
         instPatternSyn :: A.PatternSynDefn -> [A.NamedArg A.Pattern] -> TCM A.Pattern
         instPatternSyn (ns, p) as = do
           p <- expandPatternSynonyms p
-          case A.insertImplicitPatSynArgs (A.ImplicitP . PatRange) (getRange x) ns as of
+          case A.insertImplicitPatSynArgs (A.WildP . PatRange) (getRange x) ns as of
             Nothing       -> typeError $ BadArgumentsToPatternSynonym x
             Just (_, _:_) -> typeError $ TooFewArgumentsToPatternSynonym x
             Just (s, [])  -> return $ setRange (getRange i) $ A.substPattern s p
