@@ -2,8 +2,13 @@
 {-# LANGUAGE DeriveFunctor       #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
+-- | Pattern matcher used in the reducer for clauses that
+--   have not been compiled to case trees yet.
+
 module Agda.TypeChecking.Patterns.Match where
 
+import Prelude hiding (null)
+
 import Data.Monoid
 import Data.Traversable (traverse)
 
@@ -18,6 +23,7 @@ import Agda.TypeChecking.Pretty
 
 import Agda.Utils.Functor (for, ($>))
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 
@@ -28,24 +34,32 @@ import Agda.Utils.Impossible
 --   it is due to a particular meta variable.
 data Match a = Yes Simplification [a]
              | No
-             | DontKnow (Maybe MetaId)
+             | DontKnow (Blocked ())
   deriving Functor
 
-instance Monoid (Match a) where
-    mempty = Yes mempty []
+instance Null (Match a) where
+  empty = Yes empty empty
+  null (Yes simpl as) = null simpl && null as
+  null _              = False
 
-    Yes s us   `mappend` Yes s' vs        = Yes (s `mappend` s') (us ++ vs)
-    Yes _ _    `mappend` No               = No
-    Yes _ _    `mappend` DontKnow m       = DontKnow m
-    No         `mappend` _                = No
+-- 'mappend' is UNUSED.
+--
+-- instance Monoid (Match a) where
+--     mempty = Yes mempty []
 
-    -- Nothing means blocked by a variable.  In this case no instantiation of
-    -- meta-variables will make progress.
-    DontKnow _ `mappend` DontKnow Nothing = DontKnow Nothing
+--     Yes s us   `mappend` Yes s' vs        = Yes (s `mappend` s') (us ++ vs)
+--     Yes _ _    `mappend` No               = No
+--     Yes _ _    `mappend` DontKnow m       = DontKnow m
+--     No         `mappend` _                = No
 
-    -- One could imagine DontKnow _ `mappend` No = No, but would break the
-    -- equivalence to case-trees.
-    DontKnow m `mappend` _                = DontKnow m
+--     -- @NotBlocked (StuckOn e)@ means blocked by a variable.
+--     -- In this case, no instantiation of
+--     -- meta-variables will make progress.
+--     DontKnow b `mappend` DontKnow b'      = DontKnow $ b `mappend` b'
+
+--     -- One could imagine DontKnow _ `mappend` No = No, but would break the
+--     -- equivalence to case-trees.
+--     DontKnow m `mappend` _                = DontKnow m
 
 -- | Instead of 'zipWithM', we need to use this lazy version
 --   of combining pattern matching computations.
@@ -66,26 +80,26 @@ instance Monoid (Match a) where
 -- upon failure, no further matching is performed.
 
 foldMatch
-  :: forall a b . (a -> b -> ReduceM (Match Term, b))
-  -> [a] -> [b] -> ReduceM (Match Term, [b])
+  :: forall p v . (p -> v -> ReduceM (Match Term, v))
+  -> [p] -> [v] -> ReduceM (Match Term, [v])
 foldMatch match = loop where
-  loop :: [a] -> [b] -> ReduceM (Match Term, [b])
+  loop :: [p] -> [v] -> ReduceM (Match Term, [v])
   loop ps0 vs0 = do
-  case (ps0, vs0) of
-    ([], []) -> return (mempty, [])
-    (p : ps, v : vs) -> do
-      (r, v') <- match p v
-      case r of
-        No         -> return (No        , v' : vs)
-        DontKnow m -> return (DontKnow m, v' : vs)
-        Yes s us   -> do
-          (r', vs') <- loop ps vs
-          let vs1 = v' : vs'
-          case r' of
-            Yes s' us' -> return (Yes (s `mappend` s') (us ++ us'), vs1)
-            No         -> return (No                              , vs1)
-            DontKnow m -> return (DontKnow m                      , vs1)
-    _ -> __IMPOSSIBLE__
+    case (ps0, vs0) of
+      ([], []) -> return (empty, [])
+      (p : ps, v : vs) -> do
+        (r, v') <- match p v
+        case r of
+          No         -> return (No        , v' : vs)
+          DontKnow m -> return (DontKnow m, v' : vs)
+          Yes s us   -> do
+            (r', vs') <- loop ps vs
+            let vs1 = v' : vs'
+            case r' of
+              Yes s' us' -> return (Yes (s `mappend` s') (us ++ us'), vs1)
+              No         -> return (No                              , vs1)
+              DontKnow m -> return (DontKnow m                      , vs1)
+      _ -> __IMPOSSIBLE__
 
 -- | @matchCopatterns ps es@ matches spine @es@ against copattern spine @ps@.
 --
@@ -102,36 +116,35 @@ foldMatch match = loop where
 --   to come to a decision).
 matchCopatterns :: [I.NamedArg Pattern] -> [Elim] -> ReduceM (Match Term, [Elim])
 matchCopatterns ps vs = do
-    traceSDoc "tc.match" 50
-     (vcat [ text "matchCopatterns"
-           , nest 2 $ text "ps =" <+> fsep (punctuate comma $ map (prettyTCM . namedArg) ps)
-           , nest 2 $ text "vs =" <+> fsep (punctuate comma $ map prettyTCM vs)
-           ]) $ do
-    -- Buggy, see issue 1124:
-    -- mapFst mconcat . unzip <$> zipWithM' (matchCopattern . namedArg) ps vs
-    foldMatch (matchCopattern . namedArg) ps vs
+  traceSDoc "tc.match" 50
+    (vcat [ text "matchCopatterns"
+          , nest 2 $ text "ps =" <+> fsep (punctuate comma $ map (prettyTCM . namedArg) ps)
+          , nest 2 $ text "vs =" <+> fsep (punctuate comma $ map prettyTCM vs)
+          ]) $ do
+     -- Buggy, see issue 1124:
+     -- mapFst mconcat . unzip <$> zipWithM' (matchCopattern . namedArg) ps vs
+     foldMatch (matchCopattern . namedArg) ps vs
 
 -- | Match a single copattern.
 matchCopattern :: Pattern -> Elim -> ReduceM (Match Term, Elim)
 matchCopattern (ProjP p) elim@(Proj q)
   | p == q    = return (Yes YesSimplification [], elim)
   | otherwise = return (No                      , elim)
-matchCopattern (ProjP p) elim@Apply{}
-              = return (DontKnow Nothing, elim)
-matchCopattern _ elim@Proj{} = return (DontKnow Nothing, elim)
-matchCopattern p (Apply v)   = mapSnd Apply <$> matchPattern p v
+matchCopattern ProjP{} Apply{}   = __IMPOSSIBLE__
+matchCopattern _       Proj{}    = __IMPOSSIBLE__
+matchCopattern p       (Apply v) = mapSnd Apply <$> matchPattern p v
 
 matchPatterns :: [I.NamedArg Pattern] -> [I.Arg Term] -> ReduceM (Match Term, [I.Arg Term])
 matchPatterns ps vs = do
-    traceSDoc "tc.match" 50
-     (vcat [ text "matchPatterns"
-           , nest 2 $ text "ps =" <+> fsep (punctuate comma $ map (text . show) ps)
-           , nest 2 $ text "vs =" <+> fsep (punctuate comma $ map prettyTCM vs)
-           ]) $ do
-    -- Buggy, see issue 1124:
-    -- (ms,vs) <- unzip <$> zipWithM' (matchPattern . namedArg) ps vs
-    -- return (mconcat ms, vs)
-    foldMatch (matchPattern . namedArg) ps vs
+  traceSDoc "tc.match" 50
+    (vcat [ text "matchPatterns"
+          , nest 2 $ text "ps =" <+> fsep (punctuate comma $ map (text . show) ps)
+          , nest 2 $ text "vs =" <+> fsep (punctuate comma $ map prettyTCM vs)
+          ]) $ do
+     -- Buggy, see issue 1124:
+     -- (ms,vs) <- unzip <$> zipWithM' (matchPattern . namedArg) ps vs
+     -- return (mconcat ms, vs)
+     foldMatch (matchPattern . namedArg) ps vs
 
 -- | Match a single pattern.
 matchPattern :: Pattern -> I.Arg Term -> ReduceM (Match Term, I.Arg Term)
@@ -143,27 +156,18 @@ matchPattern p u = case (p, u) of
     w <- reduceB' v
     let arg' = arg $> ignoreBlocking w
     case ignoreSharing <$> w of
-        NotBlocked (Lit l')
-            | l == l'          -> return (Yes YesSimplification [] , arg')
-            | otherwise        -> return (No                       , arg')
-        NotBlocked (MetaV x _) -> return (DontKnow $ Just x        , arg')
-        Blocked x _            -> return (DontKnow $ Just x        , arg')
-        _                      -> return (DontKnow Nothing         , arg')
-
-{- Andreas, 2012-04-02 NO LONGER UP-TO-DATE
-matchPattern (Arg h' r' (ConP c _ ps))     (Arg h Irrelevant v) = do
-          -- Andreas, 2010-09-07 matching a record constructor against
-          -- something irrelevant will just continue matching against
-          -- irrelevant stuff
-                (m, vs) <- matchPatterns ps $
-                  repeat $ Arg NotHidden Irrelevant $ DontCare __IMPOSSIBLE__
-                return (m, Arg h Irrelevant $ Con c vs)
--}
+      NotBlocked _ (Lit l')
+          | l == l'            -> return (Yes YesSimplification []    , arg')
+          | otherwise          -> return (No                          , arg')
+      NotBlocked _ (MetaV x _) -> return (DontKnow $ Blocked x ()     , arg')
+      Blocked x _              -> return (DontKnow $ Blocked x ()     , arg')
+      NotBlocked r t           -> return (DontKnow $ NotBlocked r' () , arg')
+        where r' = stuckOn (Apply arg') r
 
   -- Case record pattern: always succeed!
   -- This case is necessary if we want to use the clauses before
   -- record pattern translation (e.g., in type-checking definitions by copatterns).
-  (ConP con@(ConHead c _ ds) Just{} ps, arg@(Arg info v))
+  (ConP con@(ConHead c _ ds) ConPatternInfo{conPRecord = Just{}} ps, arg@(Arg info v))
      -- precondition: con actually comes with the record fields
      | size ds == size ps -> mapSnd (Arg info . Con con) <$> do
          matchPatterns ps $ for ds $ \ d -> Arg info $ v `applyE` [Proj d]
@@ -179,36 +183,30 @@ matchPattern (Arg h' r' (ConP c _ ps))     (Arg h Irrelevant v) = do
         -- 2) whatInduction c sometimes crashes because c may point to
         --    an axiom at this stage (if we are checking the
         --    projection functions for a record type).
+{-
         w <- case ignoreSharing <$> w of
-               NotBlocked (Def f es) ->
+               NotBlocked r (Def f es) ->   -- Andreas, 2014-06-12 TODO: r == ReallyNotBlocked sufficient?
                  unfoldDefinitionE True reduceB' (Def f []) f es
                    -- reduceB is used here because some constructors
                    -- are actually definitions which need to be
                    -- unfolded (due to open public).
                _ -> return w
-        let v = ignoreBlocking w
-        case ignoreSharing <$> w of
-
-{- Andreas, 2013-10-27 the following considered HARMFUL:
-          -- Andreas, 2010-09-07 matching a record constructor against
-          -- something irrelevant will just continue matching against
-          -- irrelevant stuff
-          -- NotBlocked (Sort Prop)
-          _  | isIrrelevant info -> do
-                (m, vs) <- matchPatterns ps $
-                  repeat $ setRelevance Irrelevant $ defaultArg $ Sort Prop
-                    -- repeat looks very bad here (non-termination!)
-                return (m, Arg info $ Con c vs)
 -}
-          NotBlocked (Con c' vs)
-            | c == c'            -> do
+        w <- case w of
+               NotBlocked r u -> unfoldCorecursion u  -- Andreas, 2014-06-12 TODO: r == ReallyNotBlocked sufficient?
+               _ -> return w
+        let v = ignoreBlocking w
+            arg = Arg info v  -- the reduced argument
+        case ignoreSharing <$> w of
+          NotBlocked _ (Con c' vs)
+            | c == c'               -> do
                 (m, vs) <- yesSimplification <$> matchPatterns ps vs
                 return (m, Arg info $ Con c' vs)
-            | otherwise           -> return (No, Arg info v) -- NOTE: v the reduced thing(shadowing!). Andreas, 2013-07-03
-          NotBlocked (MetaV x vs) -> return (DontKnow $ Just x, Arg info v)
-          Blocked x _             -> return (DontKnow $ Just x, Arg info v)
-          _                       -> return (DontKnow Nothing, Arg info v)
-
+            | otherwise             -> return (No                          , arg)
+          NotBlocked _ (MetaV x vs) -> return (DontKnow $ Blocked x ()     , arg)
+          Blocked x _               -> return (DontKnow $ Blocked x ()     , arg)
+          NotBlocked r _            -> return (DontKnow $ NotBlocked r' () , arg)
+            where r' = stuckOn (Apply arg) r
 -- ASR (08 November 2014). The type of the function could be
 --
 -- @(Match Term, [I.Arg Term]) -> (Match Term, [I.Arg Term])@.
