@@ -1,7 +1,6 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
 
 module Agda.TypeChecking.Polarity where
 
@@ -12,6 +11,7 @@ import Data.List
 import Data.Maybe
 import Data.Traversable (traverse)
 
+import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
 
@@ -21,10 +21,12 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Free hiding (Occurrence(..))
+import Agda.TypeChecking.Positivity.Occurrence
 
 import Agda.Interaction.Options
 
 import Agda.Utils.List
+import Agda.Utils.Maybe ( whenNothingM )
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
 import Agda.Utils.Size
@@ -78,9 +80,10 @@ nextPolarity :: [Polarity] -> (Polarity, [Polarity])
 nextPolarity []       = (Invariant, [])
 nextPolarity (p : ps) = (p, ps)
 
--- | Replace 'Nonvariant' by 'Invariant'.
+-- | Replace 'Nonvariant' by 'Covariant'.
+--   (Arbitrary bias, but better than 'Invariant', see issue 1596).
 purgeNonvariant :: [Polarity] -> [Polarity]
-purgeNonvariant = map (\ p -> if p == Nonvariant then Invariant else p)
+purgeNonvariant = map (\ p -> if p == Nonvariant then Covariant else p)
 
 ------------------------------------------------------------------------
 -- * Computing the polarity of a symbol.
@@ -88,13 +91,13 @@ purgeNonvariant = map (\ p -> if p == Nonvariant then Invariant else p)
 
 -- | Main function of this module.
 computePolarity :: QName -> TCM ()
-computePolarity x = do
+computePolarity x = inConcreteOrAbstractMode x $ do
   reportSLn "tc.polarity.set" 25 $ "Computing polarity of " ++ show x
 
   -- get basic polarity from positivity analysis
   def      <- getConstInfo x
-  let pol0 = map polFromOcc $ defArgOccurrences def
---  let pol0 = map polFromOcc $ getArgOccurrences_ $ theDef def
+  let npars = droppedPars def
+  let pol0 = replicate npars Nonvariant ++ map polFromOcc (defArgOccurrences def)
   reportSLn "tc.polarity.set" 15 $ "Polarity of " ++ show x ++ " from positivity: " ++ show pol0
 
 {-
@@ -119,25 +122,31 @@ computePolarity x = do
   -- t <- instantiateFull t -- Andreas, 2014-04-11 Issue 1099: needed for
   --                        -- variable occurrence test in  dependentPolarity.
   reportSDoc "tc.polarity.set" 15 $ text "Refining polarity with type " <+> prettyTCM t
-  pol <- enablePhantomTypes (theDef def) <$> dependentPolarity t pol1
+  pol <- dependentPolarity t (enablePhantomTypes (theDef def) pol1) pol1
   reportSLn "tc.polarity.set" 10 $ "Polarity of " ++ show x ++ ": " ++ show pol
 
   -- set the polarity in the signature
-  setPolarity x $ pol -- purgeNonvariant pol -- temporarily disable non-variance
+  setPolarity x $ drop npars pol -- purgeNonvariant pol -- temporarily disable non-variance
 
   -- make 'Nonvariant' args 'UnusedArg' in type and clause telescope
   -- Andreas 2012-11-18: skip this for abstract definitions (fixing issue 755).
   -- This means that the most precise type for abstract definitions
   -- is not available, even to other abstract definitions.
   -- A proper fix would be to introduce a second type for use within abstract.
-  t <- if (defAbstract def == AbstractDef) then return t else
-         nonvariantToUnusedArg pol t
+  --
+  -- Andreas, 2015-07-01: I thought one should do this for
+  -- abstract local definitions in @where@ blocks to fix Issue 1366b,
+  --but it is not necessary.
+  -- t <- if (defAbstract def == AbstractDef) && not (isAnonymousModuleName $ qnameModule x)
+  t <- if (defAbstract def == AbstractDef)
+         then return t
+         else nonvariantToUnusedArg pol t
   modifySignature $ updateDefinition x $
    updateTheDef (nonvariantToUnusedArgInDef pol) . updateDefType (const t)
 
 -- | Data and record parameters are used as phantom arguments all over
 --   the test suite (and possibly in user developments).
---   @enablePhantomTypes@ turns 'Nonvariant' parameters to 'Invariant'
+--   @enablePhantomTypes@ turns 'Nonvariant' parameters to 'Covariant'
 --   to enable phantoms.
 enablePhantomTypes :: Defn -> [Polarity] -> [Polarity]
 enablePhantomTypes def pol = case def of
@@ -170,14 +179,20 @@ usagePolarity def = case def of
 
 -- | Make arguments 'Invariant' if the type of a not-'Nonvariant'
 --   later argument depends on it.
-dependentPolarity :: Type -> [Polarity] -> TCM [Polarity]
-dependentPolarity t []          = return []  -- all remaining are 'Invariant'
-dependentPolarity t pols@(p:ps) = do
+--   Also, enable phantom types by turning 'Nonvariant' into something
+--   else if it is a data/record parameter but not a size argument. [See issue 1596]
+--
+--   Precondition: the "phantom" polarity list has the same length as the polarity list.
+dependentPolarity :: Type -> [Polarity] -> [Polarity] -> TCM [Polarity]
+dependentPolarity t _      []          = return []  -- all remaining are 'Invariant'
+dependentPolarity t []     (_ : _)     = __IMPOSSIBLE__
+dependentPolarity t (q:qs) pols@(p:ps) = do
   t <- reduce $ unEl t
   case ignoreSharing t of
-    Pi a b -> do
+    Pi dom b -> do
       let c = absBody b
-      ps <- dependentPolarity c ps
+      ps <- dependentPolarity c qs ps
+      let mp = ifM (isJust <$> isSizeType (unDom dom)) (return p) (return q)
       p  <- case b of
               Abs{} | p /= Invariant  ->
                 -- Andreas, 2014-04-11 see Issue 1099
@@ -185,8 +200,8 @@ dependentPolarity t pols@(p:ps) = do
                 -- hence metas must have been instantiated before!
                 ifM (relevantInIgnoringNonvariant 0 c ps)
                   (return Invariant)
-                  (return p)
-              _ -> return p
+                  mp
+              _ -> mp
       return $ p : ps
     _ -> return pols
 
@@ -254,59 +269,66 @@ nonvariantToUnusedArgInClause pol cl@Clause{clauseTel = tel, clausePerm = perm,
 ------------------------------------------------------------------------
 
 -- | Hack for polarity of size indices.
+--   As a side effect, this sets the positivity of the size index.
+--   See test/succeed/PolaritySizeSucData.agda for a case where this is needed.
 sizePolarity :: QName -> [Polarity] -> TCM [Polarity]
 sizePolarity d pol0 = do
   let exit = return pol0
   ifM (not . optSizedTypes <$> pragmaOptions) exit $ do
-  def <- getConstInfo d
-  case theDef def of
-    Datatype{ dataPars = np, dataCons = cons } -> do
-      let TelV tel _      = telView' $ defType def
-          (parTel, ixTel) = genericSplitAt np $ telToList tel
-      case ixTel of
-        []                 -> exit  -- No size index
-        Dom _ (_, a) : _ -> ifM ((/= Just BoundedNo) <$> isSizeType a) exit $ do
-          -- we assume the size index to be 'Covariant' ...
-          let pol   = genericTake np pol0
-              polCo = pol ++ [Covariant]
-              polIn = pol ++ [Invariant]
-          setPolarity d $ polCo
-          -- and seek confirm it by looking at the constructor types
-          let check c = do
-                t <- defType <$> getConstInfo c
-                addCtxTel (telFromList parTel) $ do
---OLD:                  let pars = reverse [ defaultArg $ var i | i <- [0..np - 1] ]
-                  let pars = map (defaultArg . var) $ downFrom np
-                  TelV conTel target <- telView =<< (t `piApplyM` pars)
-                  case conTel of
-                    EmptyTel  -> return False  -- no size argument
-                    ExtendTel arg  tel ->
-                      ifM ((/= Just BoundedNo) <$> isSizeType (unDom arg)) (return False) $ do -- also no size argument
-                        -- First constructor argument has type Size
+    def <- getConstInfo d
+    case theDef def of
+      Datatype{ dataPars = np, dataCons = cons } -> do
+        let TelV tel _      = telView' $ defType def
+            (parTel, ixTel) = genericSplitAt np $ telToList tel
+        case ixTel of
+          []                 -> exit  -- No size index
+          Dom _ (_, a) : _ -> ifM ((/= Just BoundedNo) <$> isSizeType a) exit $ do
+            -- we assume the size index to be 'Covariant' ...
+            let pol   = genericTake np pol0
+                polCo = pol ++ [Covariant]
+                polIn = pol ++ [Invariant]
+            setPolarity d $ polCo
+            -- and seek confirm it by looking at the constructor types
+            let check c = do
+                  t <- defType <$> getConstInfo c
+                  addCtxTel (telFromList parTel) $ do
+  --OLD:                  let pars = reverse [ defaultArg $ var i | i <- [0..np - 1] ]
+                    let pars = map (defaultArg . var) $ downFrom np
+                    TelV conTel target <- telView =<< (t `piApplyM` pars)
+                    case conTel of
+                      EmptyTel  -> return False  -- no size argument
+                      ExtendTel arg  tel ->
+                        ifM ((/= Just BoundedNo) <$> isSizeType (unDom arg)) (return False) $ do -- also no size argument
+                          -- First constructor argument has type Size
 
-                        -- check that only positive occurences in tel
-                        let isPos = underAbstraction arg tel $ \ tel -> do
-                              pols <- zipWithM polarity [0..] $ map (snd . unDom) $ telToList tel
-                              reportSDoc "tc.polarity.size" 25 $
-                                text $ "to pass size polarity check, the following polarities need all to be covariant: " ++ show pols
-                              return $ all (`elem` [Nonvariant, Covariant]) pols
+                          -- check that only positive occurences in tel
+                          let isPos = underAbstraction arg tel $ \ tel -> do
+                                pols <- zipWithM polarity [0..] $ map (snd . unDom) $ telToList tel
+                                reportSDoc "tc.polarity.size" 25 $
+                                  text $ "to pass size polarity check, the following polarities need all to be covariant: " ++ show pols
+                                return $ all (`elem` [Nonvariant, Covariant]) pols
 
-                        -- check that the size argument appears in the
-                        -- right spot in the target type
-                        let sizeArg = size tel
-                            isLin = addContext conTel $ checkSizeIndex d np sizeArg target
+                          -- check that the size argument appears in the
+                          -- right spot in the target type
+                          let sizeArg = size tel
+                              isLin = addContext conTel $ checkSizeIndex d np sizeArg target
 
-                        ok <- isPos `and2M` isLin
-                        reportSDoc "tc.polarity.size" 15 $
-                          text "constructor" <+> prettyTCM c <+>
-                          text (if ok then "passes" else "fails") <+>
-                          text "size polarity check"
-                        return ok
+                          ok <- isPos `and2M` isLin
+                          reportSDoc "tc.polarity.size" 15 $
+                            text "constructor" <+> prettyTCM c <+>
+                            text (if ok then "passes" else "fails") <+>
+                            text "size polarity check"
+                          return ok
 
-          ifM (andM $ map check cons)
-              (return polCo) -- yes, we have a sized type here
-              (return polIn) -- no, does not conform to the rules of sized types
-    _ -> exit
+            ifNotM (andM $ map check cons)
+                (return polIn) -- no, does not conform to the rules of sized types
+              $ do  -- yes, we have a sized type here
+                -- Andreas, 2015-07-01
+                -- As a side effect, mark the size also covariant for subsequent
+                -- positivity checking (which feeds back into polarity analysis).
+                modifyArgOccurrences d $ \ occ -> take np occ ++ [JustPos]
+                return polCo
+      _ -> exit
 
 -- | @checkSizeIndex d np i a@ checks that constructor target type @a@
 --   has form @d ps (↑ i) idxs@ where @|ps| = np@.
@@ -320,7 +342,7 @@ checkSizeIndex d np i a = do
     text "has size index successor of " <+> prettyTCM (var i)
   case ignoreSharing $ unEl a of
     Def d0 es -> do
-      unlessM (isJust <$> sameDef d d0) __IMPOSSIBLE__
+      whenNothingM (sameDef d d0) __IMPOSSIBLE__
       s <- sizeView $ unArg ix
       case s of
         SizeSuc v | Var j [] <- ignoreSharing v, i == j
@@ -401,5 +423,5 @@ instance HasPolarity LevelAtom where
   polarities i l = case l of
     MetaLevel _ vs   -> map (const Invariant) <$> polarities i vs
     BlockedLevel _ v -> polarities i v
-    NeutralLevel v   -> polarities i v
+    NeutralLevel _ v -> polarities i v
     UnreducedLevel v -> polarities i v
