@@ -1,22 +1,30 @@
 {-# LANGUAGE CPP                  #-}
 {-# LANGUAGE FlexibleContexts     #-}
 {-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE TupleSections        #-}
 {-# LANGUAGE UndecidableInstances #-}
 
 -- | Check that a datatype is strictly positive.
 module Agda.TypeChecking.Positivity where
 
+import Prelude hiding (null)
+
 import Control.Applicative hiding (empty)
 import Control.DeepSeq
 import Control.Monad.Reader
 
-import Data.Set (Set)
-import qualified Data.Set as Set
+import Data.Either
+import Data.Graph (SCC(..), flattenSCC)
+import Data.List as List hiding (null)
+import Data.Maybe (mapMaybe, fromMaybe)
 import Data.Map (Map)
 import qualified Data.Map as Map
-import Data.List as List
-import Data.Maybe (mapMaybe, fromMaybe)
+import Data.Set (Set)
+import qualified Data.Set as Set
+
+import Debug.Trace
+
+import Test.QuickCheck
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -26,99 +34,135 @@ import Agda.TypeChecking.Records (unguardedRecord, recursiveRecord)
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin (primInf, CoinductionKit(..), coinductionKit)
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 
-import Agda.Utils.Size
+import qualified Agda.Utils.Graph.AdjacencyMap.Unidirectional as Graph
+import Agda.Utils.Functor
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
+import qualified Agda.Utils.Permutation as Perm
 import Agda.Utils.SemiRing
-import qualified Agda.Utils.Graph.AdjacencyMap as Graph
-import Agda.Utils.Graph.AdjacencyMap (Graph)
+import Agda.Utils.Size
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+type Graph n e = Graph.Graph n n e
+
 -- | Check that the datatypes in the mutual block containing the given
 --   declarations are strictly positive.
 --
 --   Also add information about positivity and recursivity of records
 --   to the signature.
 checkStrictlyPositive :: Set QName -> TCM ()
-checkStrictlyPositive qs = disableDestructiveUpdate $ do
+checkStrictlyPositive qset = disableDestructiveUpdate $ do
   -- compute the occurrence graph for qs
-  reportSDoc "tc.pos.tick" 100 $ text "positivity of" <+> prettyTCM (Set.toList qs)
-  g <- Graph.filterEdges (\ (Edge o _) -> o /= Unused) <$> buildOccurrenceGraph qs
-  let gstar = Graph.transitiveClosure $ fmap occ g
+  let qs = Set.toList qset
+  reportSDoc "tc.pos.tick" 100 $ text "positivity of" <+> prettyTCM qs
+  -- remove @Unused@ edges
+  g <- Graph.clean <$> buildOccurrenceGraph qset
+  let gstar = Graph.gaussJordanFloydWarshallMcNaughtonYamada $ fmap occ g
   reportSDoc "tc.pos.tick" 100 $ text "constructed graph"
   reportSLn "tc.pos.graph" 5 $ "Positivity graph: N=" ++ show (size $ Graph.nodes g) ++
                                " E=" ++ show (length $ Graph.edges g)
   reportSDoc "tc.pos.graph" 10 $ vcat
-    [ text "positivity graph for" <+> prettyTCM (Set.toList qs)
+    [ text "positivity graph for" <+> (fsep $ map prettyTCM qs)
     , nest 2 $ prettyTCM g
     ]
+  reportSLn "tc.pos.graph" 5 $
+    "Positivity graph (completed): E=" ++ show (length $ Graph.edges gstar)
   reportSDoc "tc.pos.graph" 50 $ vcat
     [ text "transitive closure of positivity graph for" <+>
-      prettyTCM (Set.toList qs)
+      prettyTCM qs
     , nest 2 $ prettyTCM gstar
     ]
 
   -- remember argument occurrences for qs in the signature
-  mapM_ (setArgs gstar) $ Set.toList qs
+  setArgOccs qset qs gstar
   reportSDoc "tc.pos.tick" 100 $ text "set args"
 
   -- check positivity for all strongly connected components of the graph for qs
-  let sccs = Graph.sccs gstar
+  let sccs' = Graph.sccs' gstar
+      sccs  = map flattenSCC sccs'
+  reportSDoc "tc.pos.graph.sccs" 10 $ do
+    let (triv, others) = partitionEithers $ for sccs' $ \ scc -> case scc of
+          AcyclicSCC v -> Left v
+          CyclicSCC vs -> Right vs
+    sep [ text $ show (length triv) ++ " trivial sccs"
+        , text $ show (length others) ++ " non-trivial sccs with lengths " ++
+            show (map length others)
+        ]
   reportSDoc "tc.pos.graph.sccs" 15 $ text $ "  sccs = " ++ show sccs
   forM_ sccs $ \ scc -> setMut [ q | DefNode q <- scc ]
-  mapM_ (checkPos g) $ Set.toList qs
+  mapM_ (checkPos g gstar) $ qs
   reportSDoc "tc.pos.tick" 100 $ text "checked positivity"
 
   where
-    checkPos g q = do
+    checkPos :: Graph Node Edge ->
+                Graph Node Occurrence ->
+                QName -> TCM ()
+    checkPos g gstar q = inConcreteOrAbstractMode q $ do
       -- we check positivity only for data or record definitions
       whenJustM (isDatatype q) $ \ dr -> do
-      reportSDoc "tc.pos.check" 10 $ text "Checking positivity of" <+> prettyTCM q
-      -- get all pathes from q to q that exhibit a non-strictly occurrence
-      -- or, in case of records, any recursive occurrence
-      let critical IsData   = \ (Edge o _) -> o <= JustPos
-          critical IsRecord = \ (Edge o _) -> o /= Unused
-          loops      = filter (critical dr) $ Graph.allPaths (critical dr) (DefNode q) (DefNode q) g
+        reportSDoc "tc.pos.check" 10 $ text "Checking positivity of" <+> prettyTCM q
 
-      -- if we have a negative loop, raise error
-      whenM positivityCheckEnabled $ do
-        forM_ [ how | Edge o how <- loops, o <= JustPos ] $ \ how -> do
-          err <- fsep $
-            [prettyTCM q] ++ pwords "is not strictly positive, because it occurs" ++
-            [prettyTCM how]
-          setCurrentRange (getRange q) $ typeError $ GenericDocError err
+        let loop      = Graph.lookup (DefNode q) (DefNode q) gstar
+            how msg p =
+              fsep $ [prettyTCM q] ++ pwords "is" ++
+                case filter (p . occ) $
+                     -- Graph.allTrails (DefNode q) (DefNode q) g of -- exponential, see Issue 1612
+                     Graph.allPaths (p . occ) (DefNode q) (DefNode q) g of
+                  Edge _ how : _ -> pwords (msg ++ ", because it occurs") ++
+                                    [prettyTCM how]
+                  _              -> pwords $ msg ++ "."
 
-      -- if we find an unguarded record, mark it as such
-      when (dr == IsRecord) $ do
-       case headMaybe [ how | Edge o how <- loops, o <= StrictPos ] of
-        Just how -> do
-          reportSDoc "tc.pos.record" 5 $ sep
-            [ prettyTCM q <+> text "is not guarded, because it occurs"
-            , prettyTCM how
-            ]
-          unguardedRecord q
-          checkInduction q
-        -- otherwise, if the record is recursive, mark it as well
-        Nothing -> forM_ (take 1 [ how | Edge GuardPos how <- loops ]) $ \ how -> do
-          reportSDoc "tc.pos.record" 5 $ sep
-            [ prettyTCM q <+> text "is recursive, because it occurs"
-            , prettyTCM how
-            ]
-          recursiveRecord q
-          checkInduction q
+                  -- For an example of code that exercises the latter,
+                  -- uninformative clause above, see
+                  -- test/fail/BadInductionRecursion5.agda.
+
+                  -- If a suitable StarSemiRing instance can be
+                  -- defined for Edge, then
+                  -- gaussJordanFloydWarshallMcNaughtonYamada can be
+                  -- used instead of allTrails, thus avoiding the
+                  -- uninformative clause.
+
+        -- if we have a negative loop, raise error
+        whenM positivityCheckEnabled $
+          case loop of
+            Just o | p o -> do
+              err <- how "not strictly positive" p
+              setCurrentRange q $ typeError $ GenericDocError err
+              where p = (<= JustPos)
+            _ -> return ()
+
+        -- if we find an unguarded record, mark it as such
+        when (dr == IsRecord) $
+          case loop of
+            Just o | p o -> do
+              reportSDoc "tc.pos.record" 5 $ how "not guarded" p
+              unguardedRecord q
+              checkInduction q
+              where p = (<= StrictPos)
+            _ ->
+              -- otherwise, if the record is recursive, mark it as well
+              case loop of
+                Just o | p o -> do
+                  reportSDoc "tc.pos.record" 5 $ how "recursive" p
+                  recursiveRecord q
+                  checkInduction q
+                  where p = (== GuardPos)
+                _ -> return ()
 
     checkInduction q = whenM positivityCheckEnabled $ do
       -- Check whether the recursive record has been declared as
       -- 'Inductive' or 'Coinductive'.  Otherwise, error.
-      unlessM (isJust . recInduction . theDef <$> getConstInfo q) $ do
-        traceCall (SetRange $ nameBindingSite $ qnameName q) $ do
+      unlessM (isJust . recInduction . theDef <$> getConstInfo q) $
+        setCurrentRange (nameBindingSite $ qnameName q) $
           typeError . GenericDocError =<<
             text "Recursive record" <+> prettyTCM q <+>
             text "needs to be declared as either inductive or coinductive"
@@ -137,28 +181,33 @@ checkStrictlyPositive qs = disableDestructiveUpdate $ do
     setMut [q] = return ()  -- no mutual recursion
     setMut qs  = forM_ qs $ \ q -> setMutual q (delete q qs)
 
-    -- Set the polarity of the arguments to a definition
-    setArgs g q = do
-      reportSDoc "tc.pos.args" 10 $ text "checking args of" <+> prettyTCM q
-      n <- getDefArity =<< getConstInfo q
-      let nArgs = maximum $ n :
-                    [ i + 1 | (ArgNode q1 i) <- Set.toList $ Graph.nodes g
-                    , q1 == q ]
-          findOcc i = fromMaybe Unused $ Graph.lookup (ArgNode q i) (DefNode q) g
-          args = map findOcc [0..nArgs - 1]
-      reportSDoc "tc.pos.args" 10 $ sep
-        [ text "args of" <+> prettyTCM q <+> text "="
-        , nest 2 $ prettyList $ map (text . show) args
-        ]
-      -- The list args can take a long time to compute, but contains
-      -- small elements, and is stored in the interface (right?), so
-      -- it is computed deep-strictly.
-      setArgOccurrences q $!! args
+    -- Set the polarity of the arguments to a couple of definitions
+    setArgOccs :: Set QName -> [QName] -> Graph Node Occurrence -> TCM ()
+    setArgOccs qset qs g = do
+      -- Compute a map from each name in q to the maximal argument index
+      let maxs = Map.fromListWith max
+           [ (q, i) | ArgNode q i <- Set.toList $ Graph.sourceNodes g, q `Set.member` qset ]
+      forM_ qs $ \ q -> inConcreteOrAbstractMode q $ do
+        reportSDoc "tc.pos.args" 10 $ text "checking args of" <+> prettyTCM q
+        n <- getDefArity =<< getConstInfo q
+        -- If there is no outgoing edge @ArgNode q i@, all @n@ arguments are @Unused@.
+        -- Otherwise, we obtain the occurrences from the Graph.
+        let findOcc i = fromMaybe Unused $ Graph.lookup (ArgNode q i) (DefNode q) g
+            args = caseMaybe (Map.lookup q maxs) (replicate n Unused) $ \ m ->
+              map findOcc [0 .. max m (n - 1)]
+        reportSDoc "tc.pos.args" 10 $ sep
+          [ text "args of" <+> prettyTCM q <+> text "="
+          , nest 2 $ prettyList $ map (text . show) args
+          ]
+        -- The list args can take a long time to compute, but contains
+        -- small elements, and is stored in the interface (right?), so
+        -- it is computed deep-strictly.
+        setArgOccurrences q $!! args
 
 getDefArity :: Definition -> TCM Int
 getDefArity def = case theDef def of
-  Function{ funClauses = cs, funProjection = proj } -> do
-    let dropped = maybe 0 (subtract 1 . projIndex) proj
+  defn@Function{} -> do
+    let dropped = projectionArgs defn
     subtract dropped . arity <$> instantiateFull (defType def)
   Datatype{ dataPars = n } -> return n
   Record{ recPars = n }    -> return n
@@ -166,41 +215,7 @@ getDefArity def = case theDef def of
 
 -- Specification of occurrences -------------------------------------------
 
--- | 'Occurrence' is a complete lattice with least element 'Mixed'
---   and greatest element 'Unused'.
---
---   It forms a commutative semiring where 'oplus' is meet (glb)
---   and 'otimes' is composition. Both operations are idempotent.
---
---   For 'oplus', 'Unused' is neutral (zero) and 'Mixed' is dominant.
---   For 'otimes', 'StrictPos' is neutral (one) and 'Unused' is dominant.
-
-instance SemiRing Occurrence where
-  oplus Mixed _           = Mixed     -- dominant
-  oplus _ Mixed           = Mixed
-  oplus Unused o          = o         -- neutral
-  oplus o Unused          = o
-  oplus JustNeg  JustNeg  = JustNeg
-  oplus JustNeg  o        = Mixed     -- negative and any form of positve
-  oplus o        JustNeg  = Mixed
-  oplus GuardPos o        = o         -- second-rank neutral
-  oplus o GuardPos        = o
-  oplus StrictPos o       = o         -- third-rank neutral
-  oplus o StrictPos       = o
-  oplus JustPos JustPos   = JustPos
-
-  otimes Unused _            = Unused     -- dominant
-  otimes _ Unused            = Unused
-  otimes Mixed _             = Mixed      -- second-rank dominance
-  otimes _ Mixed             = Mixed
-  otimes JustNeg JustNeg     = JustPos
-  otimes JustNeg _           = JustNeg    -- third-rank dominance
-  otimes _ JustNeg           = JustNeg
-  otimes JustPos _           = JustPos    -- fourth-rank dominance
-  otimes _ JustPos           = JustPos
-  otimes GuardPos _          = GuardPos   -- _ `elem` [StrictPos, GuardPos]
-  otimes _ GuardPos          = GuardPos
-  otimes StrictPos StrictPos = StrictPos  -- neutral
+-- See also Agda.TypeChecking.Positivity.Occurrence.
 
 -- | Description of an occurrence.
 data OccursWhere
@@ -216,7 +231,7 @@ data OccursWhere
   | InDefOf QName OccursWhere    -- ^ in the definition of a constant
   | Here
   | Unknown                      -- ^ an unknown position (treated as negative)
-  deriving (Show, Eq)
+  deriving (Show, Eq, Ord)
 
 (>*<) :: OccursWhere -> OccursWhere -> OccursWhere
 Here            >*< o  = o
@@ -240,14 +255,12 @@ instance PrettyTCM OccursWhere where
       nth 2 = pwords "third"
       nth n = pwords $ show (n + 1) ++ "th"
 
-      uniq (x:y:xs)
-        | x == y  = uniq (x:xs)
-      uniq (x:xs) = x : uniq xs
-      uniq []     = []
+      -- remove consecutive duplicates
+      uniq = map head . group
 
       prettyOs [] = __IMPOSSIBLE__
       prettyOs [o] = prettyO o <> text "."
-      prettyOs (o:os) = prettyO o <> text ", which occurs" <+> prettyOs os
+      prettyOs (o:os) = prettyO o <> text ", which occurs" $$ prettyOs os
 
       prettyO o = case o of
         Here           -> empty
@@ -354,8 +367,12 @@ withExtendedOccEnv :: Maybe Item -> OccM a -> OccM a
 withExtendedOccEnv i = local $ \ e -> e { vars = i : vars e }
 
 -- | Running the monad
-getOccurrences :: ComputeOccurrences a => [Maybe Item] -> a -> TCM Occurrences
+getOccurrences
+  :: (Show a, PrettyTCM a, ComputeOccurrences a)
+  => [Maybe Item] -> a -> TCM Occurrences
 getOccurrences vars a = do
+  reportSDoc "tc.pos.occ" 70 $ text "computing occurrences in " <+> text (show a)
+  reportSDoc "tc.pos.occ" 20 $ text "computing occurrences in " <+> prettyTCM a
   kit <- coinductionKit
   return $ runReader (occurrences a) $ OccEnv vars $ fmap nameOfInf kit
 
@@ -364,21 +381,15 @@ class ComputeOccurrences a where
 
 instance ComputeOccurrences Clause where
   occurrences cl = do
-    let ps = map unArg $ clausePats cl
+    let ps = clausePats cl
     (concatOccurs (mapMaybe matching (zip [0..] ps)) >+<) <$>
       walk (patItems ps) (clauseBody cl)
     where
       matching (i, p)
-        | properlyMatching p = Just $ occursAs Matched $ here $ AnArg i
-        | otherwise          = Nothing
-{-
-      matching (i, VarP{}) = Nothing
-      matching (i, DotP{}) = Nothing
-      matching (i, ConP _ Just{} _) = Nothing -- record patterns are not matches
-      matching (i, _     ) = Just (occursAs Matched $ here (AnArg i))
--}
+        | properlyMatching (unArg p) = Just $ occursAs Matched $ here $ AnArg i
+        | otherwise                  = Nothing
 
-      walk _         NoBody     = return $ Map.empty
+      walk _         NoBody     = return empty
       walk []        (Body v)   = occurrences v
       walk (i : pis) (Bind b)   = withExtendedOccEnv i $ walk pis $ absBody b
       walk []        Bind{}     = __IMPOSSIBLE__
@@ -390,24 +401,23 @@ instance ComputeOccurrences Clause where
       patItems ps = concat $ zipWith patItem [0..] ps
 
       -- @patItem i p@ replicates index @i@ as often as there are
-      -- pattern variables in @p@
-      patItem i p = replicate (nVars p) (Just (AnArg i))
-
-      -- @nVars p@ computes the number of variables bound by pattern @p@
-      nVars p = case p of
-        VarP{}      -> 1
-        DotP{}      -> 1
-        ConP _ _ ps -> sum $ map (nVars . namedArg) ps
-        LitP{}      -> 0
-        ProjP{}     -> 0
+      -- pattern variables in @p@ (dot patterns count as variable)
+      patItem :: Int -> I.Arg Pattern -> [Maybe Item]
+      patItem i p = map (const $ Just $ AnArg i) $ patternVars p
 
 instance ComputeOccurrences Term where
-  occurrences v = case v of
+  occurrences v = case unSpine v of
     Var i args -> do
       vars <- asks vars
       occs <- occurrences args
-      return $ maybe Map.empty here (index vars i)
-               >+< occursAs VarArg occs
+      -- Apparently some development version of GHC chokes if the
+      -- following line is replaced by vars ! i.
+      let mi | i < length vars = vars !! i
+             | otherwise       = flip trace __IMPOSSIBLE__ $
+                 "impossible: occurrence of de Bruijn index " ++ show i ++
+                 " in vars " ++ show vars ++ " is unbound"
+      return $ maybe empty here mi >+< occursAs VarArg occs
+
     Def d args   -> do
       inf <- asks inf
       let occsAs = if Just d /= inf then occursAs . DefArg d else \ n ->
@@ -424,37 +434,31 @@ instance ComputeOccurrences Term where
       return $ occursAs LeftOfArrow oa >+< ob
     Lam _ b      -> occurrences b
     Level l      -> occurrences l
-    Lit{}        -> return $ Map.empty
-    Sort{}       -> return $ Map.empty
-    DontCare _   -> return $ Map.empty -- Andreas, 2011-09-09: do we need to check for negative occurrences in irrelevant positions?
+    Lit{}        -> return empty
+    Sort{}       -> return empty
+    DontCare _   -> return empty -- Andreas, 2011-09-09: do we need to check for negative occurrences in irrelevant positions?
     Shared p     -> occurrences $ derefPtr p
     ExtLam{}     -> __IMPOSSIBLE__
-    where
-      -- Apparently some development version of GHC chokes if the
-      -- following line is replaced by vs ! i.
-      index vs i
-        | i < length vs = vs !! i
-        | otherwise     = __IMPOSSIBLE__
 
 instance ComputeOccurrences Level where
   occurrences (Max as) = occurrences as
 
 instance ComputeOccurrences PlusLevel where
-  occurrences ClosedLevel{} = return $ Map.empty
+  occurrences ClosedLevel{} = return empty
   occurrences (Plus _ l)    = occurrences l
 
 instance ComputeOccurrences LevelAtom where
   occurrences l = case l of
     MetaLevel _ vs   -> occursAs MetaArg <$> occurrences vs
     BlockedLevel _ v -> occurrences v
-    NeutralLevel v   -> occurrences v
+    NeutralLevel _ v -> occurrences v
     UnreducedLevel v -> occurrences v
 
 instance ComputeOccurrences Type where
   occurrences (El _ v) = occurrences v
 
 instance ComputeOccurrences a => ComputeOccurrences (Tele a) where
-  occurrences EmptyTel        = return $ Map.empty
+  occurrences EmptyTel        = return empty
   occurrences (ExtendTel a b) = occurrences (a, b)
 
 instance ComputeOccurrences a => ComputeOccurrences (Abs a) where
@@ -462,7 +466,7 @@ instance ComputeOccurrences a => ComputeOccurrences (Abs a) where
   occurrences (NoAbs _ b) = occurrences b
 
 instance ComputeOccurrences a => ComputeOccurrences (Elim' a) where
-  occurrences Proj{}    = return $ Map.empty
+  occurrences Proj{}    = __IMPOSSIBLE__
   occurrences (Apply a) = occurrences a
 
 instance ComputeOccurrences a => ComputeOccurrences (I.Arg a) where
@@ -482,7 +486,11 @@ instance (ComputeOccurrences a, ComputeOccurrences b) => ComputeOccurrences (a,
 
 -- | Compute the occurrences in a given definition.
 computeOccurrences :: QName -> TCM Occurrences
-computeOccurrences q = do
+computeOccurrences q = inConcreteOrAbstractMode q $ do
+  reportSDoc "tc.pos" 25 $ do
+    a <- defAbstract <$> getConstInfo q
+    m <- asks envAbstractMode
+    text "computeOccurrences" <+> prettyTCM q <+> text (show a) <+> text (show m)
   def <- getConstInfo q
   occursAs (InDefOf q) <$> case theDef def of
     Function{funClauses = cs} -> do
@@ -515,22 +523,25 @@ computeOccurrences q = do
       getOccurrences vars =<< instantiateFull tel'
 
     -- Arguments to other kinds of definitions are hard-wired.
-    Constructor{} -> return Map.empty
-    Axiom{}       -> return Map.empty
-    Primitive{}   -> return Map.empty
+    Constructor{} -> return empty
+    Axiom{}       -> return empty
+    Primitive{}   -> return empty
 
 -- | Eta expand a clause to have the given number of variables.
---   Warning: doesn't update telescope or permutation!
+--   Warning: doesn't put correct types in telescope!
 --   This is used instead of special treatment of lambdas
 --   (which was unsound: issue 121)
 etaExpandClause :: Nat -> Clause -> Clause
-etaExpandClause n c@Clause{ namedClausePats = ps, clauseBody = b }
+etaExpandClause n c@Clause{ clauseTel = tel, clausePerm = perm, namedClausePats = ps, clauseBody = b }
   | m <= 0    = c
-  | otherwise = c { namedClausePats = ps ++ genericReplicate m (defaultArg $ unnamed $ VarP underscore)
-                  , clauseBody      = liftBody m b
-                  , clauseTel       = __IMPOSSIBLE__
-                  , clausePerm      = __IMPOSSIBLE__
-                  }
+  | otherwise = c
+      { namedClausePats = ps ++ genericReplicate m (defaultArg $ unnamed $ VarP underscore)
+      , clauseBody      = liftBody m b
+      , clauseTel       = telFromList $
+          telToList tel ++ (replicate m $ (underscore,) <$> dummyDom)
+          -- dummyDom, not __IMPOSSIBLE__, because of debug printing.
+      , clausePerm      = Perm.liftP m perm -- Andreas, 2015-06-28 this is probably correct.
+      }
   where
     m = n - genericLength ps
 
@@ -558,40 +569,25 @@ instance PrettyTCM Node where
   prettyTCM (DefNode q)   = prettyTCM q
   prettyTCM (ArgNode q i) = prettyTCM q <> text ("." ++ show i)
 
-instance PrettyTCM Occurrence where
-  prettyTCM GuardPos  = text "-[g+]->"
-  prettyTCM StrictPos = text "-[++]->"
-  prettyTCM JustPos   = text "-[+]->"
-  prettyTCM JustNeg   = text "-[-]->"
-  prettyTCM Mixed     = text "-[*]->"
-  prettyTCM Unused    = text "-[ ]->"
-
--- | Pairing something with a node (for printing only).
-data WithNode n a = WithNode n a
-
 instance PrettyTCM n => PrettyTCM (WithNode n Edge) where
   prettyTCM (WithNode n (Edge o w)) =
     prettyTCM o <+> prettyTCM n <+> fsep (pwords $ show w)
 
-instance PrettyTCM n => PrettyTCM (WithNode n Occurrence) where
-  prettyTCM (WithNode n o) = prettyTCM o <+> prettyTCM n
-
-instance (PrettyTCM n, PrettyTCM (WithNode n e)) => PrettyTCM (Graph n e) where
-  prettyTCM g = vcat $ map pr $ Map.assocs $ Graph.unGraph g
-    where
-      pr (n, es) = sep
-        [ prettyTCM n
-        , nest 2 $ vcat $ map (prettyTCM . uncurry WithNode) $ Map.assocs es
-        ]
-
 -- | Edge labels for the positivity graph.
 data Edge = Edge Occurrence OccursWhere
-  deriving (Show)
+  deriving (Eq, Ord, Show)
+
+instance Null Edge where
+  null (Edge o _) = null o
+  empty = Edge empty Unknown
 
 -- | These operations form a semiring if we quotient by the relation
 -- \"the 'Occurrence' components are equal\".
 
 instance SemiRing Edge where
+  ozero = Edge ozero Unknown
+  oone  = Edge oone  Unknown
+
   oplus _                    e@(Edge Mixed _)     = e -- dominant
   oplus e@(Edge Mixed _) _                        = e
   oplus (Edge Unused _)      e                    = e -- neutral
@@ -607,17 +603,26 @@ instance SemiRing Edge where
 
   otimes (Edge o1 w1) (Edge o2 w2) = Edge (otimes o1 o2) (w1 >*< w2)
 
+-- | As 'OccursWhere' does not have an 'oplus' we cannot do something meaningful
+--   for the @OccursWhere@ here.
+--
+--   E.g. @ostar (Edge JustNeg w) = Edge Mixed (w `oplus` (w >*< w))@
+--   would probably more sense, if we could do it.
+instance StarSemiRing Edge where
+  ostar (Edge o w) = Edge (ostar o) w
+
 buildOccurrenceGraph :: Set QName -> TCM (Graph Node Edge)
-buildOccurrenceGraph qs = Graph.unions <$> mapM defGraph (Set.toList qs)
+buildOccurrenceGraph qs = Graph.unionsWith oplus <$> mapM defGraph (Set.toList qs)
   where
     defGraph :: QName -> TCM (Graph Node Edge)
     defGraph q = do
       occs <- computeOccurrences q
-      let onItem (item, occs) = do
-            es <- mapM (computeEdge qs) occs
-            return $ Graph.unions $
-                map (\(b, w) -> Graph.singleton (itemToNode item) b w) es
-      Graph.unions <$> mapM onItem (Map.assocs occs)
+      Graph.unionsWith oplus <$> do
+        forM (Map.assocs occs) $ \ (item, occs) -> do
+          let src = itemToNode item
+          es <- mapM (computeEdge qs) occs
+          return $ Graph.unionsWith oplus $
+            for es $ \ (tgt, w) -> Graph.singleton src tgt w
       where
         itemToNode (AnArg i) = ArgNode q i
         itemToNode (ADef q)  = DefNode q
@@ -666,3 +671,72 @@ computeEdge muts o = do
         -- D: (A B -> C) generates a positive edge B --> A.1
         -- even though the context is negative.
         inArg d i = mkEdge (ArgNode d i) StrictPos
+
+------------------------------------------------------------------------
+-- * Generators and tests
+------------------------------------------------------------------------
+
+instance Arbitrary OccursWhere where
+  arbitrary = sized arbitraryS
+    where
+    arbitraryS n = oneof $
+      [ return Here
+      , return Unknown
+      ] ++
+      if n <= 0 then [] else
+        [ LeftOfArrow <$> arb
+        , DefArg <$> arbitrary <*> arbitrary <*> arb
+        , UnderInf <$> arb
+        , VarArg <$> arb
+        , MetaArg <$> arb
+        , ConArgType <$> arbitrary <*> arb
+        , IndArgType <$> arbitrary <*> arb
+        , InClause <$> arbitrary <*> arb
+        , Matched <$> arb
+        , InDefOf <$> arbitrary <*> arb
+        ]
+      where arb = arbitraryS (n - 1)
+
+  shrink x = replaceConstructor x ++ genericShrink x
+    where
+    replaceConstructor Here    = []
+    replaceConstructor Unknown = []
+    replaceConstructor _       = [Here, Unknown]
+
+    genericShrink (LeftOfArrow a)  = a : [ LeftOfArrow a  | a <- shrink a ]
+    genericShrink (DefArg a b c)   = c : [ DefArg a b c   | c <- shrink c ]
+    genericShrink (UnderInf a)     = a : [ UnderInf a     | a <- shrink a ]
+    genericShrink (VarArg a)       = a : [ VarArg a       | a <- shrink a ]
+    genericShrink (MetaArg a)      = a : [ MetaArg a      | a <- shrink a ]
+    genericShrink (ConArgType a b) = b : [ ConArgType a b | b <- shrink b ]
+    genericShrink (IndArgType a b) = b : [ IndArgType a b | b <- shrink b ]
+    genericShrink (InClause a b)   = b : [ InClause a b   | b <- shrink b ]
+    genericShrink (Matched a)      = a : [ Matched a      | a <- shrink a ]
+    genericShrink (InDefOf a b)    = b : [ InDefOf a b    | b <- shrink b ]
+    genericShrink Here             = []
+    genericShrink Unknown          = []
+
+instance CoArbitrary OccursWhere where
+  coarbitrary (LeftOfArrow a)  = variant  0 . coarbitrary a
+  coarbitrary (DefArg a b c)   = variant  1 . coarbitrary (a, b, c)
+  coarbitrary (UnderInf a)     = variant  2 . coarbitrary a
+  coarbitrary (VarArg a)       = variant  3 . coarbitrary a
+  coarbitrary (MetaArg a)      = variant  4 . coarbitrary a
+  coarbitrary (ConArgType a b) = variant  5 . coarbitrary (a, b)
+  coarbitrary (IndArgType a b) = variant  6 . coarbitrary (a, b)
+  coarbitrary (InClause a b)   = variant  7 . coarbitrary (a, b)
+  coarbitrary (Matched a)      = variant  8 . coarbitrary a
+  coarbitrary (InDefOf a b)    = variant  9 . coarbitrary (a, b)
+  coarbitrary Here             = variant 10
+  coarbitrary Unknown          = variant 11
+
+instance Arbitrary Edge where
+  arbitrary = Edge <$> arbitrary <*> arbitrary
+
+  shrink (Edge o w) = [ Edge o w | o <- shrink o ] ++
+                      [ Edge o w | w <- shrink w ]
+
+instance CoArbitrary Edge where
+  coarbitrary (Edge o w) = coarbitrary (o, w)
+
+-- properties moved to Agda.TypeChecking.Positivity.Tests
