@@ -1,7 +1,10 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP               #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards     #-}
+
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
 
 module Agda.TypeChecking.ProjectionLike where
 
@@ -55,16 +58,18 @@ unProjView pv =
     NoProjection v          -> v
 
 -- | Top-level 'ProjectionView' (no reduction).
-projView :: Term -> TCM ProjectionView
+{-# SPECIALIZE projView :: Term -> TCM ProjectionView #-}
+projView :: HasConstInfo m => Term -> m ProjectionView
 projView v = do
   let fallback = return $ NoProjection v
   case ignoreSharing v of
     Def f es -> caseMaybeM (isProjection f) fallback $ \ isP -> do
-      case es of
-        []           -> return $ LoneProjectionLike f $ projArgInfo isP
-        Apply a : es -> return $ ProjectionView f a es
-        -- Since a projection is a function, it cannot be projected itself.
-        Proj{}  : _  -> __IMPOSSIBLE__
+      if projIndex isP <= 0 then fallback else do
+        case es of
+          []           -> return $ LoneProjectionLike f $ projArgInfo isP
+          Apply a : es -> return $ ProjectionView f a es
+          -- Since a projection is a function, it cannot be projected itself.
+          Proj{}  : _  -> __IMPOSSIBLE__
     _ -> fallback
 
 -- | Reduce away top-level projection like functions.
@@ -85,7 +90,7 @@ reduceProjectionLike v = do
 --   This does just one layer, such that the top spine contains
 --   the projection-like functions as projections.
 --   Used in 'compareElims' in @TypeChecking.Conversion@
---   and in 'Agda.TypeChecking.CheckInternal'.
+--   and in "Agda.TypeChecking.CheckInternal".
 --
 --   If the 'Bool' is 'True', a lone projection like function will be
 --   turned into a lambda-abstraction, expecting the principal argument.
@@ -109,18 +114,6 @@ elimView loneProjToLambda v = do
       | otherwise         -> return v
     ProjectionView f a es -> (`applyE` (Proj f : es)) <$> elimView loneProjToLambda (unArg a)
 
-{- Andreas, 2013-11-01: Use of unLevel no longer necessary, since we do not reduce!
-  case ignoreSharing v of
-    Def f (Apply (Arg _ rv) : es) -> do
-      flip (maybeM (return v)) (isProjection f) $ \ _ -> do
-        (`applyE` (Proj f : es)) <$> do elimView =<< unLevel =<< reduce rv
-          -- domi 2012-7-24: Add unLevel to handle neutral levels.
-          -- The problem is that reduce turns @suc (neutral)@
-          -- into @Level (Max [Plus 1 (NeutralLevel neutral)])@
-          -- which the below pattern match does not handle.
-    _ -> return v
--}
-
 -- | Which @Def@types are eligible for the principle argument
 --   of a projection-like function?
 eligibleForProjectionLike :: QName -> TCM Bool
@@ -158,54 +151,54 @@ makeProjection x = inTopContext $ do
       reportSLn "tc.proj.like" 30 $ if null ps0 then "  no candidates found"
                                                 else "  candidates: " ++ show ps0
       unless (null ps0) $ do
-      -- Andreas 2012-09-26: only consider non-recursive functions for proj.like.
-      -- Issue 700: problems with recursive funs. in term.checker and reduction
-      ifM recursive (reportSLn "tc.proj.like" 30 $ "  recursive functions are not considered for projection-likeness") $ do
-      ps <- return $ filter (checkOccurs cls . snd) ps0
-      when (null ps) $
-        reportSLn "tc.proj.like" 50 $
-          "  occurs check failed\n    clauses = " ++ show cls
-      case reverse ps of
-        []         -> return ()
-        (d, n) : _ -> do
-          reportSDoc "tc.proj.like" 10 $ sep
-            [ prettyTCM x <+> text " : " <+> prettyTCM t
-            , text $ " is projection like in argument " ++ show n ++ " for type " ++ show d
-            ]
+        -- Andreas 2012-09-26: only consider non-recursive functions for proj.like.
+        -- Issue 700: problems with recursive funs. in term.checker and reduction
+        ifM recursive (reportSLn "tc.proj.like" 30 $ "  recursive functions are not considered for projection-likeness") $ do
+          ps <- return $ filter (checkOccurs cls . snd) ps0
+          when (null ps) $
+            reportSLn "tc.proj.like" 50 $
+              "  occurs check failed\n    clauses = " ++ show cls
+          case reverse ps of
+            []         -> return ()
+            (d, n) : _ -> do
+              reportSDoc "tc.proj.like" 10 $ sep
+                [ prettyTCM x <+> text " : " <+> prettyTCM t
+                , text $ " is projection like in argument " ++ show n ++ " for type " ++ show d
+                ]
 {-
-          reportSLn "tc.proj.like" 10 $
-            show (defName defn) ++ " is projection like in argument " ++
-            show n ++ " for type " ++ show d
+              reportSLn "tc.proj.like" 10 $
+                show (defName defn) ++ " is projection like in argument " ++
+                show n ++ " for type " ++ show d
 -}
-          let cls' = map (dropArgs n) cls
-              cc   = dropArgs n cc0
-          -- cc <- compileClauses (Just (x, __IMPOSSIBLE__)) cls'
-          reportSLn "tc.proj.like" 60 $ "  rewrote clauses to\n    " ++ show cc
+              let cls' = map (dropArgs n) cls
+                  cc   = dropArgs n cc0
+              -- cc <- compileClauses (Just (x, __IMPOSSIBLE__)) cls'
+              reportSLn "tc.proj.like" 60 $ "  rewrote clauses to\n    " ++ show cc
 
-          -- Andreas, 2013-10-20 build parameter dropping function
-          let (ptel, Dom ai _ : _) = splitAt n $ telToList $ theTel $ telView' t
-              -- leading lambdas are to ignore parameter applications
-              proj = teleNoAbs ptel $ Def x []
-              -- proj = foldr (\ (Dom ai (y, _)) -> Lam ai . NoAbs y) (Def x []) ptel
+              -- Andreas, 2013-10-20 build parameter dropping function
+              let (ptel, Dom ai _ : _) = splitAt n $ telToList $ theTel $ telView' t
+                  -- leading lambdas are to ignore parameter applications
+                  proj = teleNoAbs ptel $ Def x []
+                  -- proj = foldr (\ (Dom ai (y, _)) -> Lam ai . NoAbs y) (Def x []) ptel
 
-          let projection = Projection
-                { projProper   = Nothing
-                , projFromType = d
-                , projIndex    = n + 1
-                , projDropPars = proj
-                , projArgInfo  = ai
-                }
-          let newDef = def
-                       { funProjection     = Just projection
-                       , funClauses        = cls'
-                       , funCompiled       = cc
-                       , funInv            = dropArgs n $ funInv def
-                       }
-          addConstant x $ defn { defPolarity       = drop n $ defPolarity defn
-                               , defArgOccurrences = drop n $ defArgOccurrences defn
-                               , defDisplay        = []
-                               , theDef            = newDef
-                               }
+              let projection = Projection
+                    { projProper   = Nothing
+                    , projFromType = d
+                    , projIndex    = n + 1
+                    , projDropPars = proj
+                    , projArgInfo  = ai
+                    }
+              let newDef = def
+                           { funProjection     = Just projection
+                           , funClauses        = cls'
+                           , funCompiled       = cc
+                           , funInv            = dropArgs n $ funInv def
+                           }
+              addConstant x $ defn { defPolarity       = drop n $ defPolarity defn
+                                   , defArgOccurrences = drop n $ defArgOccurrences defn
+                                   , defDisplay        = []
+                                   , theDef            = newDef
+                                   }
     Function{funInv = Inverse{}} ->
       reportSLn "tc.proj.like" 30 $ "  injective functions can't be projections"
     Function{funAbstr = AbstractDef} ->
@@ -282,17 +275,3 @@ makeProjection x = inTopContext $ do
       where
         candidateRec NoAbs{}   = []
         candidateRec (Abs x t) = candidateArgs (var (size vs) : vs) t
-
-{-
-    candidateArgs :: [Term] -> Term -> [(QName,Int)]
-    candidateArgs vs (Shared p) = candidateArgs vs $ derefPtr p
-    candidateArgs vs (Pi (Dom info (El _ def)) b)
-      | Def d es <- ignoreSharing def,
-        Just us <- allApplyElims es,
-        vs == map unArg us    = (d, length vs) : candidateRec vs b
-    candidateArgs vs (Pi _ b) = candidateRec vs b
-    candidateArgs _ _ = []
-
-    candidateRec vs NoAbs{} = []
-    candidateRec vs b       = candidateArgs (var (size vs) : vs) (unEl $ absBody b)
--}
