@@ -15,12 +15,13 @@ module Agda.TypeChecking.RecordPatterns
   ) where
 
 import Control.Applicative
-import Control.Arrow ((***))
+import Control.Arrow (first, second)
 import Control.Monad.Fix
 import Control.Monad.Reader
 import Control.Monad.State
 
 import Data.List
+import Data.Maybe
 import qualified Data.Map as Map
 import qualified Data.Traversable as Trav
 
@@ -31,7 +32,7 @@ import Agda.TypeChecking.Coverage.SplitTree
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Pretty hiding (pretty)
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
@@ -42,6 +43,8 @@ import Agda.Utils.List
 import qualified Agda.Utils.Map as Map
 import Agda.Utils.Maybe
 import Agda.Utils.Permutation hiding (dropFrom)
+import Agda.Utils.Pretty (Pretty(..))
+import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
 
 #include "undefined.h"
@@ -60,15 +63,16 @@ import Agda.Utils.Impossible
 recordPatternToProjections :: Pattern -> TCM [Term -> Term]
 recordPatternToProjections p =
   case p of
-    VarP{}             -> return [ \ x -> x ]
-    LitP{}             -> typeError $ ShouldBeRecordPattern p
-    DotP{}             -> typeError $ ShouldBeRecordPattern p
-    ConP c Nothing  ps -> typeError $ ShouldBeRecordPattern p
-    ConP c (Just (_, t)) ps -> do
-      t <- reduce t
+    VarP{}       -> return [ \ x -> x ]
+    LitP{}       -> typeError $ ShouldBeRecordPattern p
+    DotP{}       -> typeError $ ShouldBeRecordPattern p
+    ConP c ci ps -> do
+      whenNothing (conPRecord ci) $
+        typeError $ ShouldBeRecordPattern p
+      t <- reduce $ fromMaybe __IMPOSSIBLE__ $ conPType ci
       fields <- getRecordTypeFields (unArg t)
       concat <$> zipWithM comb (map proj fields) (map namedArg ps)
-    ProjP{}            -> __IMPOSSIBLE__ -- copattern cannot appear here
+    ProjP{}      -> __IMPOSSIBLE__ -- copattern cannot appear here
   where
     proj p = (`applyE` [Proj $ unArg p])
 --    proj p = \ x -> Def (unArg p) [defaultArg x]
@@ -125,9 +129,10 @@ translateCompiledClauses cc = snd <$> loop cc
     loops :: Int                  -- split variable
           -> Case CompiledClauses -- original split tree
           -> TCM ([Bool], CompiledClauses)
-    loops i cs@(Branches { conBranches = conMap
-                         , litBranches = litMap
-                         , catchAllBranch = catchAll }) = do
+    loops i cs@Branches{ projPatterns   = cop
+                       , conBranches    = conMap
+                       , litBranches    = litMap
+                       , catchAllBranch = catchAll } = do
 
       -- recurse on and compute variable status of catch-all clause
       (xssa, catchAll) <- unzipMaybe <$> Trav.mapM loop catchAll
@@ -169,7 +174,8 @@ translateCompiledClauses cc = snd <$> loop cc
       case concat $ Map.elems ccs of
         -- case: no record pattern was translated
         []   -> return (xs, Case i $ Branches
-                  { conBranches = conMap
+                  { projPatterns = cop
+                  , conBranches = conMap
                   , litBranches = litMap
                   , catchAllBranch = catchAll })
 
@@ -244,10 +250,12 @@ replaceByProjections i projs cc =
         Fail -> Fail
 
       loops :: Int -> Case CompiledClauses -> Case CompiledClauses
-      loops i Branches{ conBranches    = conMap
+      loops i Branches{ projPatterns   = cop
+                      , conBranches    = conMap
                       , litBranches    = litMap
                       , catchAllBranch = catchAll } =
-        Branches{ conBranches    = fmap (\ (WithArity n c) -> WithArity n $ loop (i + n - 1) c) conMap
+        Branches{ projPatterns   = cop
+                , conBranches    = fmap (\ (WithArity n c) -> WithArity n $ loop (i + n - 1) c) conMap
                 , litBranches    = fmap (loop (i - 1)) litMap
                 , catchAllBranch = fmap (loop i) catchAll
                 }
@@ -411,14 +419,6 @@ isRecordSplit _ = return Nothing
 -- projection functions. Does not remove record constructor patterns
 -- which have sub-patterns containing non-record constructor or
 -- literal patterns.
---
--- If the input clause contains dot patterns inside record patterns,
--- then the translation may yield clauses which are not type-correct.
--- However, we believe that it is safe to use the output as input to
--- 'Agda.TypeChecking.CompiledClause.Compile.compileClauses'. Perhaps
--- it would be better to perform record pattern translation on the
--- compiled clauses instead, but the code below has already been
--- implemented and seems to work.
 
 translateRecordPatterns :: Clause -> TCM Clause
 translateRecordPatterns clause = do
@@ -464,20 +464,19 @@ translateRecordPatterns clause = do
       -- The new telescope, still flattened, with types in the context
       -- of the new RHS, in textual left-to-right order, and with
       -- Nothing in place of dot patterns.
+      substTel = map . fmap . second . applySubst
       newTel' =
-        map (fmap (id *** applySubst rhsSubst')) $
+        substTel rhsSubst' $
         translateTel cs $
         flattenedOldTel
 
       -- Permutation taking the new variable and dot patterns to the
       -- new telescope.
       newPerm = adjustForDotPatterns $
-                  reorderTel_ $ map (maybe dummy snd) newTel'
-        where
-        -- It is important that dummy does not mention any variable
+                  reorderTel_ $ map (maybe dummyDom snd) newTel'
+        -- It is important that dummyDom does not mention any variable
         -- (see the definition of reorderTel).
-        dummy = dummyDom -- defaultArg (El Prop (Sort Prop))
-
+        where
         isDotP n = case genericIndex cs n of
                      Left DotP{} -> True
                      _           -> False
@@ -487,7 +486,7 @@ translateRecordPatterns clause = do
 
       -- Substitution used to convert terms in the new RHS's context
       -- to terms in the new telescope's context.
-      lhsSubst' = permToSubst (reverseP newPerm)
+      lhsSubst' = renaming (reverseP newPerm)
 
       -- Substitution used to convert terms in the old telescope's
       -- context to terms in the new telescope's context.
@@ -496,9 +495,9 @@ translateRecordPatterns clause = do
       -- The new telescope.
       newTel =
         uncurry unflattenTel . unzip $
-        map (maybe __IMPOSSIBLE__ id) $
+        map (fromMaybe __IMPOSSIBLE__) $
         permute newPerm $
-        map (fmap (id *** applySubst lhsSubst')) $
+        substTel lhsSubst' $
         newTel'
 
       -- New clause.
@@ -509,6 +508,30 @@ translateRecordPatterns clause = do
             , clauseBody      = translateBody cs rhsSubst $ clauseBody clause
             }
 
+  reportSDoc "tc.lhs.recpat" 20 $ vcat
+      [ text "Original clause:"
+      , nest 2 $ inTopContext $ vcat
+        [ text "delta =" <+> prettyTCM (clauseTel clause)
+        , text "perm  =" <+> prettyTCM (clausePerm clause)
+        , text "pats  =" <+> text (show $ clausePats clause)
+        ]
+      , text "Intermediate results:"
+      , nest 2 $ vcat
+        [ text "ps        =" <+> text (show ps)
+        , text "s         =" <+> prettyList (map prettyTCM s)
+        , text "cs        =" <+> prettyList (map prettyTCM cs)
+        , text "flattenedOldTel =" <+> (text . show) flattenedOldTel
+        , text "newTel'   =" <+> (text . show) newTel'
+        , text "newPerm   =" <+> prettyTCM newPerm
+        ]
+      ]
+
+  reportSDoc "tc.lhs.recpat" 20 $ vcat
+        [ text "lhsSubst' =" <+> (text . show) lhsSubst'
+        , text "lhsSubst  =" <+> (text . show) lhsSubst
+        , text "newTel  =" <+> prettyTCM newTel
+        ]
+
   reportSDoc "tc.lhs.recpat" 10 $
     escapeContext (size $ clauseTel clause) $ vcat
       [ text "Translated clause:"
@@ -555,7 +578,7 @@ nextVar = RecPatM $ do
   n <- lift get
   lift $ put $ succ n
   noVars <- lift ask
-  return (VarP "r", Var (noVars - n - 1) [])
+  return (VarP "r", var $ noVars - n - 1)
 
 ------------------------------------------------------------------------
 -- Types used to record changes to a clause
@@ -572,7 +595,20 @@ data Kind = VarPat | DotPat
 -- patterns, should be removed, and a new variable, with the name @x@,
 -- inserted instead. The type of the new variable is @t@.
 
-type Changes = [Either Pattern (Kind -> Nat, ArgName, I.Dom Type)]
+type Change  = Either Pattern (Kind -> Nat, ArgName, I.Dom Type)
+type Changes = [Change]
+
+instance Pretty (Kind -> Nat) where
+  pretty f =
+    P.text "(VarPat:" P.<+> P.text (show $ f VarPat) P.<+>
+    P.text "DotPat:"  P.<+> P.text (show $ f DotPat) P.<> P.text ")"
+
+instance PrettyTCM (Kind -> Nat) where
+  prettyTCM = return . pretty
+
+instance PrettyTCM Change where
+  prettyTCM (Left  p) = prettyTCM p
+  prettyTCM (Right (f, x, t)) = text "Change" <+> prettyTCM f <+> text x <+> prettyTCM t
 
 -- | Record pattern trees.
 
@@ -600,8 +636,7 @@ projections (Leaf (DotP{})) = [(id, DotPat)]
 projections (Leaf (VarP{})) = [(id, VarPat)]
 projections (Leaf _)        = __IMPOSSIBLE__
 projections (RecCon _ args) =
-  concatMap (\(p, t) -> map (\(p', k) -> (p' . p, k))
-                            (projections t))
+  concatMap (\ (p, t) -> map (first (. p)) $ projections t)
             args
 
 -- | Converts a record tree to a single pattern along with information
@@ -646,18 +681,20 @@ removeTree tree = do
 -- This function assumes that literals are never of record type.
 
 translatePattern :: Pattern -> RecPatM (Pattern, [Term], Changes)
-translatePattern (ConP c Nothing ps) = do
-  (ps, s, cs) <- translatePatterns ps
-  return (ConP c Nothing ps, s, cs)
-translatePattern p@(ConP _ (Just _) _) = do
-  r <- recordTree p
-  case r of
-    Left  r -> r
-    Right t -> removeTree t
+translatePattern p@(ConP c ci ps)
+  -- Andreas, 2015-05-28 only translate implicit record patterns
+  | Just ConPImplicit <- conPRecord ci = do
+      r <- recordTree p
+      case r of
+        Left  r -> r
+        Right t -> removeTree t
+  | otherwise = do
+      (ps, s, cs) <- translatePatterns ps
+      return (ConP c ci ps, s, cs)
 translatePattern p@VarP{} = removeTree (Leaf p)
 translatePattern p@DotP{} = removeTree (Leaf p)
 translatePattern p@LitP{} = return (p, [], [])
-translatePattern p@ProjP{}= __IMPOSSIBLE__
+translatePattern p@ProjP{}= return (p, [], [])
 
 translatePatterns :: [I.NamedArg Pattern] -> RecPatM ([I.NamedArg Pattern], [Term], Changes)
 translatePatterns ps = do
@@ -681,8 +718,9 @@ translatePatterns ps = do
 recordTree ::
   Pattern ->
   RecPatM (Either (RecPatM (Pattern, [Term], Changes)) RecordTree)
-recordTree p@(ConP _ Nothing _) = return $ Left $ translatePattern p
-recordTree (ConP c ci@(Just (_, t)) ps) = do
+-- Andreas, 2015-05-28 only translate implicit record patterns
+recordTree (ConP c ci ps) | Just ConPImplicit <- conPRecord ci = do
+  let t = fromMaybe __IMPOSSIBLE__ $ conPType ci
   rs <- mapM (recordTree . namedArg) ps
   case allRight rs of
     Nothing ->
@@ -696,10 +734,11 @@ recordTree (ConP c ci@(Just (_, t)) ps) = do
 --      let proj p = \x -> Def (unArg p) [defaultArg x]
       let proj p = (`applyE` [Proj $ unArg p])
       return $ Right $ RecCon t $ zip (map proj fields) ts
+recordTree p@(ConP _ ci _) = return $ Left $ translatePattern p
 recordTree p@VarP{} = return (Right (Leaf p))
 recordTree p@DotP{} = return (Right (Leaf p))
 recordTree p@LitP{} = return $ Left $ translatePattern p
-recordTree p@ProjP{}= __IMPOSSIBLE__
+recordTree p@ProjP{}= return $ Left $ translatePattern p
 
 ------------------------------------------------------------------------
 -- Translation of the clause telescope and body
@@ -740,16 +779,6 @@ translateBody _               _ _          = __IMPOSSIBLE__
 ------------------------------------------------------------------------
 -- Helper functions
 
--- | Turns a permutation into a substitution.
-
-permToSubst :: Permutation -> Substitution
-permToSubst (Perm n is) =
-  [ makeVar i | i <- [0..n - 1] ] ++# raiseS (size is)
-  where
-  makeVar i = case genericElemIndex i is of
-    Nothing -> __IMPOSSIBLE__
-    Just k  -> var k
-
 -- | @dropBinds n b@ drops the initial @n@ occurrences of 'Bind' from @b@.
 --
 -- Precondition: @b@ has to start with @n@ occurrences of 'Bind'.
