@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP           #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE TupleSections #-}
 
@@ -26,10 +26,12 @@ import Agda.TypeChecking.Reduce.Monad ()
 import Agda.TypeChecking.Telescope
 
 import Agda.Utils.Either
-import Agda.Utils.List
 import Agda.Utils.Functor (for, ($>))
+import Agda.Utils.Lens
+import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Size
@@ -41,16 +43,13 @@ import Agda.Utils.Impossible
 --   Use the second argument for missing fields.
 orderFields :: QName -> a -> [C.Name] -> [(C.Name, a)] -> TCM [a]
 orderFields r def xs fs = do
-  shouldBeNull (ys \\ nub ys) $ DuplicateFields . nub
-  shouldBeNull (ys \\ xs)     $ TooManyFields r
+  unlessNull (ys \\ nub ys) $ typeError . DuplicateFields . nub
+  unlessNull (ys \\ xs)     $ typeError . TooManyFields r
   -- shouldBeNull (xs \\ ys)     $ TooFewFields r
   return $ order xs fs
   where
     ys = map fst fs
 
-    shouldBeNull [] err = return ()
-    shouldBeNull xs err = typeError $ err xs
-
     -- invariant: the first list contains at least the fields of the second list
     order [] [] = []
     order [] _  = __IMPOSSIBLE__
@@ -85,7 +84,7 @@ recordFieldNames = map (fmap (nameConcrete . qnameName)) . recFields
 -- | Find all records with at least the given fields.
 findPossibleRecords :: [C.Name] -> TCM [QName]
 findPossibleRecords fields = do
-  defs <- (HMap.union `on` sigDefinitions) <$> getSignature <*> getImportedSignature
+  defs <- (HMap.union `on` (^. sigDefinitions)) <$> getSignature <*> getImportedSignature
   let possible def = case theDef def of
         Record{ recFields = fs } -> Set.isSubsetOf given inrecord
           where inrecord = Set.fromList $ map (nameConcrete . qnameName . unArg) fs
@@ -109,6 +108,12 @@ getRecordTypeFields t =
         _ -> __IMPOSSIBLE__
     _ -> __IMPOSSIBLE__
 
+-- | Get the original name of the projection
+--   (the current one could be from a module application).
+getOriginalProjection :: QName -> TCM QName
+getOriginalProjection q = do
+  proj <- fromMaybe __IMPOSSIBLE__ <$> isProjection q
+  return $ fromMaybe __IMPOSSIBLE__ $ projProper proj
 
 -- | Get the type of the record constructor.
 getRecordConstructorType :: QName -> TCM Type
@@ -149,35 +154,17 @@ tryRecordType t = ifBlockedType t (\ _ _ -> return $ Left Nothing) $ \ t -> do
       caseMaybeM (isRecord r) no $ \ def -> return $ Right (r,vs,def)
     _ -> no
 
--- | The analogue of 'piApply'.  If @v@ is a value of record type @T@
---   with field @f@, then @projectType T f@ returns the type of @f v@.
-projectType :: Type -> QName -> TCM (Maybe Type)
-projectType t f = do
-  res <- isRecordType t
-  case res of
-    Nothing -> return Nothing
-    Just (_r, ps, _rdef) -> do
-      def <- getConstInfo f
-      if (isProperProjection $ theDef def)
-        then return $ Just $ defType def `apply` ps
-        else return Nothing
-
-{- DEPRECATED, use Signature.getDefType instead!
-
--- | @projectionType t f@ returns the type of projection or
---   projection-like function @t@ applied to the parameters,
---   which are read of @t@.
---   It fails if @t@ is not a @Def@ applied to the parameters.
-projectionType :: Type -> QName -> TCM (Maybe Type)
-projectionType t f = do
-  t <- reduce t
-  case ignoreSharing $ unEl t of
-    Def _ es -> do
-      flip (maybe $ return Nothing) (allApplyElims es) $ \ pars -> do
-        ft <- defType <$> getConstInfo f  -- type of projection(like) function
-        return $ Just $ ft `piApply` pars
-    _ -> return Nothing
--}
+-- | The analogue of 'piApply'.  If @v@ is a value of record type @t@
+--   with field @f@, then @projectTyped v t f@ returns the type of @f v@.
+--
+--   Works also for projection-like definitions @f@.
+--
+--   Precondition: @t@ is reduced.
+projectTyped :: Term -> Type -> QName -> TCM (Maybe (Term, Type))
+projectTyped v t f = caseMaybeM (getDefType f t) (return Nothing) $ \ tf -> do
+  (dom, b) <- mustBePi tf
+  u <- f `applyDef` (argFromDom dom $> v)
+  return $ Just (u, b `absApp` v)
 
 -- | Check if a name refers to an eta expandable record.
 {-# SPECIALIZE isEtaRecord :: QName -> TCM Bool #-}
@@ -206,8 +193,8 @@ isEtaRecordType a = case ignoreSharing $ unEl a of
 
 -- | Check if a name refers to a record constructor.
 --   If yes, return record definition.
-isRecordConstructor :: QName -> TCM (Maybe (QName, Defn))
-isRecordConstructor c = do
+isRecordConstructor :: MonadTCM tcm => QName -> tcm (Maybe (QName, Defn))
+isRecordConstructor c = liftTCM $ do
   def <- theDef <$> getConstInfo c
   case def of
     Constructor{ conData = r } -> fmap (r,) <$> isRecord r
@@ -259,51 +246,78 @@ Precondition: The current context is @Γ = Γ₁, x:R pars, Γ₂@ where
 Postcondition: @Δ = Γ₁, Γ', Γ₂[c Γ']@ and @Γ ⊢ σ : Δ@ and @Δ ⊢ τ : Γ@.
 -}
 etaExpandBoundVar :: Int -> TCM (Maybe (Telescope, Substitution, Substitution))
-etaExpandBoundVar i = do
-  -- Get the context with last variable added first in list.
-  gamma <- getContext
-  -- Extract type of @i@th variable.
-  let (gamma2, dom@(Dom ai (x, a)) : gamma1) = splitAt i gamma
+etaExpandBoundVar i = fmap (\ (delta, sigma, tau, _) -> (delta, sigma, tau)) <$> do
+  expandRecordVar i =<< getContextTelescope
+
+-- | @expandRecordVar i Γ = (Δ, σ, τ, Γ')@
+--
+--   Precondition: @Γ = Γ₁, x:R pars, Γ₂@ where
+--     @|Γ₂| = i@ and @R@ is a eta-expandable record type
+--     with constructor @c@ and fields @Γ'@.
+--
+--   Postcondition: @Δ = Γ₁, Γ', Γ₂[c Γ']@ and @Γ ⊢ σ : Δ@ and @Δ ⊢ τ : Γ@.
+
+expandRecordVar :: Int -> Telescope -> TCM (Maybe (Telescope, Substitution, Substitution, Telescope))
+expandRecordVar i gamma0 = do
+  -- Get the context with last variable added last in list.
+  let gamma = telToList gamma0
+  -- Convert the de Bruijn index i to a de Bruijn level
+      l     = size gamma - 1 - i
+  -- Extract type of @i@th de Bruijn index.
+  -- Γ = Γ₁, x:a, Γ₂
+  let (gamma1, dom@(Dom ai (x, a)) : gamma2) = splitAt l gamma
   -- This must be a eta-expandable record type.
   let failure = do
         reportSDoc "tc.meta.assign.proj" 25 $
-          text "failed to eta-expand variable " <+> prettyTCM x <+>
+          text "failed to eta-expand variable " <+> pretty x <+>
           text " since its type " <+> prettyTCM a <+>
           text " is not a record type"
         return Nothing
   caseMaybeM (isRecordType a) failure $ \ (r, pars, def) -> do
-  if not (recEtaEquality def) then return Nothing else Just <$> do
-  -- Get the record fields @Γ₁ ⊢ tel@ (@tel = Γ'@).
-  -- TODO: compose argInfo ai with tel.
-  let tel = recTel def `apply` pars
-      m   = size tel
-      fs  = recFields def
-  -- Construct the record pattern @Γ₁, Γ' ⊢ u := c ys@.
-      ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
-      u   = Con (recConHead def) ys
-  -- @Γ₁, Γ' ⊢ τ₀ : Γ₁, x:_@
-      tau0 = u :# raiseS m
-  -- @Γ₁, Γ', Γ₂ ⊢ τ₀ : Γ₁, x:_, Γ₂@
-      tau  = liftS (size gamma2) tau0
+    if not (recEtaEquality def) then return Nothing else Just <$> do
+      -- Get the record fields @Γ₁ ⊢ tel@ (@tel = Γ'@).
+      -- TODO: compose argInfo ai with tel.
+      let tel = recTel def `apply` pars
+          m   = size tel
+          fs  = recFields def
+      -- Construct the record pattern @Γ₁, Γ' ⊢ u := c ys@.
+          ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
+          u   = Con (recConHead def) ys
+      -- @Γ₁, Γ' ⊢ τ₀ : Γ₁, x:_@
+          tau0 = consS u $ raiseS m
+      -- @Γ₁, Γ', Γ₂ ⊢ τ₀ : Γ₁, x:_, Γ₂@
+          tau  = liftS (size gamma2) tau0
 
-  --  Fields are in order first-first.
-      zs  = for fs $ fmap $ \ f -> Var 0 [Proj f]
-  --  We need to reverse the field sequence to build the substitution.
-  -- @Γ₁, x:_ ⊢ σ₀ : Γ₁, Γ'@
-      sigma0 = parallelS $ reverse $ map unArg zs
-  -- @Γ₁, x:_, Γ₂ ⊢ σ₀ : Γ₁, Γ', Γ₂@
-      sigma  = liftS (size gamma2) sigma0
+      --  Fields are in order first-first.
+          zs  = for fs $ fmap $ \ f -> Var 0 [Proj f]
+      --  We need to reverse the field sequence to build the substitution.
+      -- @Γ₁, x:_ ⊢ σ₀ : Γ₁, Γ'@
+          sigma0 = reverse (map unArg zs) ++# raiseS 1
+      -- @Γ₁, x:_, Γ₂ ⊢ σ₀ : Γ₁, Γ', Γ₂@
+          sigma  = liftS (size gamma2) sigma0
 
-  -- Construct @Δ@ as telescope.
-  -- Note @Γ₁, x:_ ⊢ Γ₂@, thus, @Γ₁, Γ' ⊢ [τ₀]Γ₂@
+      -- Construct @Δ@ as telescope.
+      -- Note @Γ₁, x:_ ⊢ Γ₂@, thus, @Γ₁, Γ' ⊢ [τ₀]Γ₂@
 
-      rev   = foldl (\ l (Dom ai (n, t)) -> Dom ai (nameToArgName n, t) : l) []
-      -- Use "f(x)" as variable name for the projection f(x).
-      s     = prettyShow x
-      tel'  = mapAbsNames (\ f -> stringToArgName $ argNameToString f ++ "(" ++ s ++ ")") tel
-      delta = telFromList $ rev gamma1 ++ telToList tel' ++ rev (applySubst tau0 gamma2)
+          -- Use "f(x)" as variable name for the projection f(x).
+          s     = prettyShow x
+          tel'  = mapAbsNames (\ f -> stringToArgName $ argNameToString f ++ "(" ++ s ++ ")") tel
+          delta = telFromList $ gamma1 ++ telToList tel' ++ applySubst tau0 gamma2
 
-  return (delta, sigma, tau)
+      return (delta, sigma, tau, tel)
+
+-- | Precondition: variable list is ordered descendingly.  Can be empty.
+expandRecordVarsRecursively :: [Int] -> Telescope -> TCM (Telescope, Substitution, Substitution)
+expandRecordVarsRecursively [] gamma = return (gamma, idS, idS)
+expandRecordVarsRecursively (i : is) gamma = do
+  caseMaybeM (expandRecordVar i gamma) (expandRecordVarsRecursively is gamma)
+  $ \ (gamma1, sigma1, tau1, tel) -> do
+    -- Γ ⊢ σ₁ : Γ₁  and  Γ₁ ⊢ τ₁ : Γ
+    let n = size tel
+        newis = take n $ downFrom $ i + n
+    (gamma2, sigma2, tau2) <- expandRecordVarsRecursively (newis ++ is) gamma1
+    -- Γ₁ ⊢ σ₂ : Γ₂  and  Γ₂ ⊢ τ₂ : Γ₁
+    return (gamma2, applySubst sigma1 sigma2, applySubst tau2 tau1)
 
 -- | @curryAt v (Γ (y : R pars) -> B) n =
 --     ( \ v -> λ Γ ys → v Γ (c ys)            {- curry   -}
