@@ -1,8 +1,11 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE TupleSections        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TupleSections #-}
+
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
 
 module Agda.TypeChecking.Reduce where
 
@@ -18,6 +21,7 @@ import Data.Hashable
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Scope.Base (Scope)
 import Agda.Syntax.Literal
@@ -30,13 +34,17 @@ import Agda.TypeChecking.Monad.Builtin hiding (getPrimitive, constructorForm)
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.EtaContract
-import {-# SOURCE #-} Agda.TypeChecking.Pretty
-
-import {-# SOURCE #-} Agda.TypeChecking.Patterns.Match
-import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Match
 
 import Agda.TypeChecking.Reduce.Monad
 
+import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Match
+import {-# SOURCE #-} Agda.TypeChecking.Patterns.Match
+import {-# SOURCE #-} Agda.TypeChecking.Pretty
+import {-# SOURCE #-} Agda.TypeChecking.Rewriting
+
+import Agda.Utils.Either
+import Agda.Utils.Function
+import Agda.Utils.Functor
 import Agda.Utils.Monad
 import Agda.Utils.HashMap (HashMap)
 import Agda.Utils.Maybe
@@ -181,6 +189,7 @@ instance Instantiate Constraint where
   instantiate' (UnBlock m)          = return $ UnBlock m
   instantiate' (FindInScope m args) = FindInScope m <$> mapM instantiate' args
   instantiate' (IsEmpty r t)        = IsEmpty r <$> instantiate' t
+  instantiate' (CheckSizeLtSat t)   = CheckSizeLtSat <$> instantiate' t
 
 instance (Ord k, Instantiate e) => Instantiate (Map k e) where
     instantiate' = traverse instantiate'
@@ -190,14 +199,17 @@ instance (Ord k, Instantiate e) => Instantiate (Map k e) where
 -- * Reduction to weak head normal form.
 ---------------------------------------------------------------------------
 
+-- | Case on whether a term is blocked on a meta (or is a meta).
+--   That means it can change its shape when the meta is instantiated.
 ifBlocked :: MonadTCM tcm =>  Term -> (MetaId -> Term -> tcm a) -> (Term -> tcm a) -> tcm a
 ifBlocked t blocked unblocked = do
   t <- liftTCM $ reduceB t
   case ignoreSharing <$> t of
-    Blocked m _            -> blocked m (ignoreBlocking t)
-    NotBlocked (MetaV m _) -> blocked m (ignoreBlocking t)
-    NotBlocked _           -> unblocked (ignoreBlocking t)
+    Blocked m _              -> blocked m (ignoreBlocking t)
+    NotBlocked _ (MetaV m _) -> blocked m (ignoreBlocking t)
+    NotBlocked{}             -> unblocked (ignoreBlocking t)
 
+-- | Case on whether a type is blocked on a meta (or is a meta).
 ifBlockedType :: MonadTCM tcm => Type -> (MetaId -> Type -> tcm a) -> (Type -> tcm a) -> tcm a
 ifBlockedType (El s t) blocked unblocked =
   ifBlocked t (\ m v -> blocked m $ El s v) (\ v -> unblocked $ El s v)
@@ -228,6 +240,7 @@ instance Reduce Sort where
             Prop       -> return s
             Type s'    -> levelSort <$> reduce' s'
             Inf        -> return Inf
+            SizeUniv   -> return SizeUniv
 
 instance Reduce Elim where
   reduce' (Apply v) = Apply <$> reduce' v
@@ -244,7 +257,7 @@ instance Reduce PlusLevel where
 instance Reduce LevelAtom where
   reduceB' l = case l of
     MetaLevel m vs   -> fromTm (MetaV m vs)
-    NeutralLevel v   -> return $ NotBlocked $ NeutralLevel v
+    NeutralLevel r v -> return $ NotBlocked r $ NeutralLevel r v
     BlockedLevel m v ->
       ifM (isInstantiatedMeta m) (fromTm v) (return $ Blocked m $ BlockedLevel m v)
     UnreducedLevel v -> fromTm v
@@ -253,9 +266,9 @@ instance Reduce LevelAtom where
         bv <- reduceB' v
         let v = ignoreBlocking bv
         case ignoreSharing <$> bv of
-          NotBlocked (MetaV m vs) -> return $ NotBlocked $ MetaLevel m vs
-          Blocked m _             -> return $ Blocked m  $ BlockedLevel m v
-          NotBlocked _            -> return $ NotBlocked $ NeutralLevel v
+          NotBlocked r (MetaV m vs) -> return $ NotBlocked r $ MetaLevel m vs
+          Blocked m _               -> return $ Blocked m    $ BlockedLevel m v
+          NotBlocked r _            -> return $ NotBlocked r $ NeutralLevel r v
 
 
 instance (Subst t, Reduce t) => Reduce (Abs t) where
@@ -285,7 +298,7 @@ instance (Reduce a, Reduce b,Reduce c) => Reduce (a,b,c) where
     reduce' (x,y,z) = (,,) <$> reduce' x <*> reduce' y <*> reduce' z
 
 instance Reduce Term where
-  reduceB' v = {-# SCC "reduce'<Term>" #-} do
+  reduceB' = {-# SCC "reduce'<Term>" #-} rewriteAfter $ \ v -> do
     v <- instantiate' v
     let done = return $ notBlocked v
     case v of
@@ -329,6 +342,24 @@ instance Reduce Term where
               _                -> Con c [defaultArg w]
       reduceNat v = return v
 
+rewriteAfter :: (Term -> ReduceM (Blocked Term)) -> Term -> ReduceM (Blocked Term)
+rewriteAfter f = trampolineM $ rewrite <=< f
+
+-- Andreas, 2013-03-20 recursive invokations of unfoldCorecursion
+-- need also to instantiate metas, see Issue 826.
+unfoldCorecursionE :: Elim -> ReduceM (Blocked Elim)
+unfoldCorecursionE e@(Proj f)           = return $ notBlocked e
+unfoldCorecursionE (Apply (Common.Arg info v)) = fmap (Apply . Common.Arg info) <$>
+  unfoldCorecursion v
+
+unfoldCorecursion :: Term -> ReduceM (Blocked Term)
+unfoldCorecursion = rewriteAfter $ \ v -> do
+  v <- instantiate' v
+  case v of
+    Def f es -> unfoldDefinitionE True unfoldCorecursion (Def f []) f es
+    Shared{} -> fmap shared <$> unfoldCorecursion (ignoreSharing v) -- don't update when unfolding corecursion!
+    _          -> reduceB' v
+
 -- | If the first argument is 'True', then a single delayed clause may
 -- be unfolded.
 unfoldDefinition ::
@@ -374,14 +405,14 @@ unfoldDefinition' unfoldDelayed keepGoing v0 f es =
           reduceNormalE keepGoing v0 f (map notReduced es)
                        dontUnfold
                        (defClauses info) (defCompiled info)
-        else retSimpl $ notBlocked v
+        else retSimpl $ notBlocked v  -- Andrea(s), 2014-12-05 OK?
 
   where
     retSimpl v = (,v) <$> getSimplification
 
     reducePrimitive x v0 f es pf dontUnfold cls mcc
       | genericLength es < ar
-                  = retSimpl $ notBlocked $ v0 `applyE` es -- not fully applied
+                  = retSimpl $ NotBlocked Underapplied $ v0 `applyE` es -- not fully applied
       | otherwise = {-# SCC "reducePrimitive" #-} do
           let (es1,es2) = genericSplitAt ar es
               args1     = fromMaybe __IMPOSSIBLE__ $ mapM isApplyElim es1
@@ -408,7 +439,7 @@ unfoldDefinition' unfoldDelayed keepGoing v0 f es =
     reduceNormalE :: (Term -> ReduceM (Simplification, Blocked Term)) -> Term -> QName -> [MaybeReduced Elim] -> Bool -> [Clause] -> Maybe CompiledClauses -> ReduceM (Simplification, Blocked Term)
     reduceNormalE keepGoing v0 f es dontUnfold def mcc = {-# SCC "reduceNormal" #-} do
       case def of
-        _ | dontUnfold -> defaultResult
+        _ | dontUnfold -> defaultResult -- non-terminating or delayed
         [] -> defaultResult -- no definition for head
         cls -> do
             ev <- appDefE_ f v0 cls mcc es
@@ -419,13 +450,13 @@ unfoldDefinition' unfoldDelayed keepGoing v0 f es =
                   , text "    es =  " <+> sep (map (prettyTCM . ignoreReduced) es)
 --                  [ text "*** tried to reduce' " <+> prettyTCM vfull
                   , text "    stuck on" <+> prettyTCM (ignoreBlocking v) ]) $ do
-                retSimpl v
+                  retSimpl v
               YesReduction simpl v -> performedSimplification' simpl $ do
                 traceSDoc "tc.reduce'"  90 (text "*** reduced definition: " <+> prettyTCM f) $ do
-                traceSDoc "tc.reduce'"  95 (text "    result" <+> prettyTCM v) $ do
-                traceSDoc "tc.reduce'" 100 (text "    raw   " <+> text (show v)) $ do
-                keepGoing v
-      where defaultResult = retSimpl $ notBlocked $ vfull
+                  traceSDoc "tc.reduce'"  95 (text "    result" <+> prettyTCM v) $ do
+                    traceSDoc "tc.reduce'" 100 (text "    raw   " <+> text (show v)) $ do
+                      keepGoing v
+      where defaultResult = retSimpl $ NotBlocked AbsurdMatch vfull
             vfull         = v0 `applyE` map ignoreReduced es
 
 -- | Reduce a non-primitive definition if it is a copy linking to another def.
@@ -520,25 +551,24 @@ appDefE' v cls es = goCls cls $ map ignoreReduced es
     goCls :: [Clause] -> [Elim] -> ReduceM (Reduced (Blocked Term) Term)
     goCls cl es = do
       reportSLn "tc.reduce'" 95 $ "Reduce.goCls tries reduction, #clauses = " ++ show (length cl)
-      let cantReduce es = return $ NoReduction $ notBlocked $ v `applyE` es
       case cl of
         -- Andreas, 2013-10-26  In case of an incomplete match,
-        -- we just do not reduce'.  This allows adding single function
+        -- we just do not reduce.  This allows adding single function
         -- clauses after they have been type-checked, to type-check
         -- the remaining clauses (see Issue 907).
-        [] -> cantReduce es -- WAS: typeError $ IncompletePatternMatching v es
-        cl @ Clause { clauseBody = body } : cls -> do
-          let pats = namedClausePats cl
-              n    = length pats
+        -- Andrea(s), 2014-12-05:  We return 'MissingClauses' here, since this
+        -- is the most conservative reason.
+        [] -> return $ NoReduction $ NotBlocked MissingClauses $ v `applyE` es
+        Clause{ namedClausePats = pats, clauseBody = body } : cls -> do
+          let n = length pats
           -- if clause is underapplied, skip to next clause
           if length es < n then goCls cls es else do
             let (es0, es1) = splitAt n es
             (m, es0) <- matchCopatterns pats es0
             es <- return $ es0 ++ es1
             case m of
-              No                -> goCls cls es
-              DontKnow Nothing  -> cantReduce es
-              DontKnow (Just m) -> return $ NoReduction $ blocked m $ v `applyE` es
+              No         -> goCls cls es
+              DontKnow b -> return $ NoReduction $ b $> v `applyE` es
               Yes simpl vs -- vs is the subst. for the variables bound in body
                 | isJust (getBodyUnraised body)  -- clause has body?
                                 -> return $ YesReduction simpl $
@@ -546,7 +576,7 @@ appDefE' v cls es = goCls cls $ map ignoreReduced es
                     -- of the original arguments!
                     -- Andreas, 2013-05-19 isn't this done now?
                     app vs body EmptyS `applyE` es1
-                | otherwise     -> cantReduce es
+                | otherwise     -> return $ NoReduction $ NotBlocked AbsurdMatch $ v `applyE` es
 
     -- Build an explicit substitution from arguments
     -- and execute it using parallel substitution.
@@ -560,7 +590,7 @@ appDefE' v cls es = goCls cls $ map ignoreReduced es
     --   Δ.A ⊢ b : B
     app :: [Term] -> ClauseBody -> Substitution -> Term
     app []       (Body v)           sigma = applySubst sigma v
-    app (v : vs) (Bind (Abs   _ b)) sigma = app vs b (v :# sigma) -- CBN
+    app (v : vs) (Bind (Abs   _ b)) sigma = app vs b $ consS v sigma -- CBN
     app (v : vs) (Bind (NoAbs _ b)) sigma = app vs b sigma
     app  _        NoBody            sigma = __IMPOSSIBLE__
     app (_ : _)  (Body _)           sigma = __IMPOSSIBLE__
@@ -588,6 +618,7 @@ instance Reduce Constraint where
   reduce' (UnBlock m)           = return $ UnBlock m
   reduce' (FindInScope m cands) = FindInScope m <$> mapM reduce' cands
   reduce' (IsEmpty r t)         = IsEmpty r <$> reduce' t
+  reduce' (CheckSizeLtSat t)    = CheckSizeLtSat <$> reduce' t
 
 instance (Ord k, Reduce e) => Reduce (Map k e) where
     reduce' = traverse reduce'
@@ -606,7 +637,7 @@ instance Simplify Term where
     v <- instantiate' v
     case v of
       Def f vs   -> do
-        let keepGoing v = (,NotBlocked v) <$> getSimplification
+        let keepGoing v = (,notBlocked v) <$> getSimplification  -- Andrea(s), 2014-12-05 OK?
         (simpl, v) <- unfoldDefinition' False keepGoing (Def f []) f vs
         reportSDoc "tc.simplify'" 20 $
           text ("simplify': unfolding definition returns " ++ show simpl)
@@ -627,8 +658,8 @@ instance Simplify Term where
       Shared{}   -> __IMPOSSIBLE__ -- updateSharedTerm simplify' v
 
 simplifyBlocked' :: Simplify t => Blocked t -> ReduceM t
-simplifyBlocked' (Blocked _ t)  = return t
-simplifyBlocked' (NotBlocked t) = simplify' t
+simplifyBlocked' (Blocked _ t) = return t
+simplifyBlocked' (NotBlocked _ t) = simplify' t  -- Andrea(s), 2014-12-05 OK?
 
 instance Simplify Type where
     simplify' (El s t) = El <$> simplify' s <*> simplify' t
@@ -644,6 +675,7 @@ instance Simplify Sort where
         Type s     -> levelSort <$> simplify' s
         Prop       -> return s
         Inf        -> return s
+        SizeUniv   -> return s
 
 instance Simplify Level where
   simplify' (Max as) = levelMax <$> simplify' as
@@ -658,7 +690,7 @@ instance Simplify LevelAtom where
     case l of
       MetaLevel m vs   -> MetaLevel m <$> simplify' vs
       BlockedLevel m v -> BlockedLevel m <$> simplify' v
-      NeutralLevel v   -> NeutralLevel   <$> simplify' v -- ??
+      NeutralLevel r v -> NeutralLevel r <$> simplify' v -- ??
       UnreducedLevel v -> UnreducedLevel <$> simplify' v -- ??
 
 instance (Subst t, Simplify t) => Simplify (Abs t) where
@@ -717,17 +749,23 @@ instance Simplify Constraint where
   simplify' (UnBlock m)           = return $ UnBlock m
   simplify' (FindInScope m cands) = FindInScope m <$> mapM simplify' cands
   simplify' (IsEmpty r t)         = IsEmpty r <$> simplify' t
+  simplify' (CheckSizeLtSat t)    = CheckSizeLtSat <$> simplify' t
 
 instance Simplify Bool where
   simplify' = return
 
-instance Simplify Pattern where
-  simplify' p = case p of
-    VarP _       -> return p
-    LitP _       -> return p
-    ConP c mt ps -> ConP c <$> simplify' mt <*> simplify' ps
-    DotP v       -> DotP <$> simplify' v
-    ProjP _      -> return p
+-- UNUSED
+-- instance Simplify ConPatternInfo where
+--   simplify' (ConPatternInfo mr mt) = ConPatternInfo mr <$> simplify' mt
+
+-- UNUSED
+-- instance Simplify Pattern where
+--   simplify' p = case p of
+--     VarP _       -> return p
+--     LitP _       -> return p
+--     ConP c ci ps -> ConP c <$> simplify' ci <*> simplify' ps
+--     DotP v       -> DotP <$> simplify' v
+--     ProjP _      -> return p
 
 instance Simplify ClauseBody where
     simplify' (Body   t) = Body   <$> simplify' t
@@ -753,6 +791,7 @@ instance Normalise Sort where
         Prop       -> return s
         Type s     -> levelSort <$> normalise' s
         Inf        -> return Inf
+        SizeUniv   -> return SizeUniv
 
 instance Normalise Type where
     normalise' (El s t) = El <$> normalise' s <*> normalise' t
@@ -791,7 +830,7 @@ instance Normalise LevelAtom where
     case l of
       MetaLevel m vs   -> MetaLevel m <$> normalise' vs
       BlockedLevel m v -> BlockedLevel m <$> normalise' v
-      NeutralLevel v   -> NeutralLevel <$> normalise' v
+      NeutralLevel r v -> NeutralLevel r <$> normalise' v
       UnreducedLevel{} -> __IMPOSSIBLE__    -- I hope
 
 instance Normalise ClauseBody where
@@ -850,10 +889,14 @@ instance Normalise Constraint where
   normalise' (UnBlock m)           = return $ UnBlock m
   normalise' (FindInScope m cands) = FindInScope m <$> mapM normalise' cands
   normalise' (IsEmpty r t)         = IsEmpty r <$> normalise' t
+  normalise' (CheckSizeLtSat t)    = CheckSizeLtSat <$> normalise' t
 
 instance Normalise Bool where
   normalise' = return
 
+instance Normalise ConPatternInfo where
+  normalise' (ConPatternInfo mr mt) = ConPatternInfo mr <$> normalise' mt
+
 instance Normalise Pattern where
   normalise' p = case p of
     VarP _       -> return p
@@ -894,8 +937,9 @@ instance InstantiateFull Sort where
             Prop       -> return s
             DLub s1 s2 -> dLub <$> instantiateFull' s1 <*> instantiateFull' s2
             Inf        -> return s
+            SizeUniv   -> return s
 
-instance InstantiateFull Type where
+instance (InstantiateFull a) => InstantiateFull (Type' a) where
     instantiateFull' (El s t) =
       El <$> instantiateFull' s <*> instantiateFull' t
 
@@ -931,7 +975,7 @@ instance InstantiateFull LevelAtom where
       case ignoreSharing v of
         MetaV m vs -> return $ MetaLevel m vs
         _          -> return $ UnreducedLevel v
-    NeutralLevel v -> NeutralLevel <$> instantiateFull' v
+    NeutralLevel r v -> NeutralLevel r <$> instantiateFull' v
     BlockedLevel m v ->
       ifM (isInstantiatedMeta m)
           (UnreducedLevel <$> instantiateFull' v)
@@ -946,12 +990,15 @@ instance InstantiateFull Substitution where
       Wk   n sigma         -> Wk   n         <$> instantiateFull' sigma
       Lift n sigma         -> Lift n         <$> instantiateFull' sigma
       Strengthen bot sigma -> Strengthen bot <$> instantiateFull' sigma
-      t :# sigma           -> (:#) <$> instantiateFull' t
-                                   <*> instantiateFull' sigma
+      t :# sigma           -> consS <$> instantiateFull' t
+                                    <*> instantiateFull' sigma
 
 instance InstantiateFull Bool where
     instantiateFull' = return
 
+instance InstantiateFull ConPatternInfo where
+  instantiateFull' (ConPatternInfo mr mt) = ConPatternInfo mr <$> instantiateFull' mt
+
 instance InstantiateFull Pattern where
     instantiateFull' v@VarP{}       = return v
     instantiateFull' (DotP t)       = DotP <$> instantiateFull' t
@@ -1011,8 +1058,9 @@ instance InstantiateFull Constraint where
     UnBlock m           -> return $ UnBlock m
     FindInScope m cands -> FindInScope m <$> mapM instantiateFull' cands
     IsEmpty r t         -> IsEmpty r <$> instantiateFull' t
+    CheckSizeLtSat t    -> CheckSizeLtSat <$> instantiateFull' t
 
-instance InstantiateFull Elim where
+instance (InstantiateFull a) => InstantiateFull (Elim' a) where
   instantiateFull' (Apply v) = Apply <$> instantiateFull' v
   instantiateFull' (Proj f)  = pure $ Proj f
 
@@ -1029,10 +1077,10 @@ instance InstantiateFull Scope where
     instantiateFull' = return
 
 instance InstantiateFull Signature where
-  instantiateFull' (Sig a b) = uncurry Sig <$> instantiateFull' (a, b)
+  instantiateFull' (Sig a b c) = uncurry3 Sig <$> instantiateFull' (a, b, c)
 
 instance InstantiateFull Section where
-  instantiateFull' (Section tel n) = flip Section n <$> instantiateFull' tel
+  instantiateFull' (Section tel) = Section <$> instantiateFull' tel
 
 instance (Subst a, InstantiateFull a) => InstantiateFull (Tele a) where
   instantiateFull' EmptyTel = return EmptyTel
@@ -1042,9 +1090,18 @@ instance InstantiateFull Char where
     instantiateFull' = return
 
 instance InstantiateFull Definition where
-    instantiateFull' (Defn rel x t pol occ df i c rew inst d) = do
-      (t, (df, d, rew)) <- instantiateFull' (t, (df, d, rew))
-      return $ Defn rel x t pol occ df i c rew inst d
+    instantiateFull' (Defn rel x t pol occ df i c inst d) = do
+      (t, df, d) <- instantiateFull' (t, df, d)
+      return $ Defn rel x t pol occ df i c inst d
+
+instance InstantiateFull NLPat where
+  instantiateFull' (PVar x)   = return $ PVar x
+  instantiateFull' (PWild)    = return PWild
+  instantiateFull' (PDef x y) = PDef <$> instantiateFull' x <*> instantiateFull' y
+  instantiateFull' (PLam x y) = PLam x <$> instantiateFull' y
+  instantiateFull' (PPi x y)  = PPi <$> instantiateFull' x <*> instantiateFull' y
+  instantiateFull' (PBoundVar x y) = PBoundVar x <$> instantiateFull' y
+  instantiateFull' (PTerm x)  = PTerm <$> instantiateFull' x
 
 instance InstantiateFull RewriteRule where
   instantiateFull' (RewriteRule q gamma lhs rhs t) =
@@ -1064,7 +1121,7 @@ instance InstantiateFull DisplayTerm where
   instantiateFull' (DTerm v)       = DTerm <$> instantiateFull' v
   instantiateFull' (DDot  v)       = DDot  <$> instantiateFull' v
   instantiateFull' (DCon c vs)     = DCon c <$> instantiateFull' vs
-  instantiateFull' (DDef c vs)     = DDef c <$> instantiateFull' vs
+  instantiateFull' (DDef c es)     = DDef c <$> instantiateFull' es
   instantiateFull' (DWithApp v vs ws) = uncurry3 DWithApp <$> instantiateFull' (v, vs, ws)
 
 instance InstantiateFull Defn where
@@ -1095,10 +1152,11 @@ instance InstantiateFull a => InstantiateFull (WithArity a) where
   instantiateFull' (WithArity n a) = WithArity n <$> instantiateFull' a
 
 instance InstantiateFull a => InstantiateFull (Case a) where
-  instantiateFull' (Branches cs ls m) =
-    Branches <$> instantiateFull' cs
-             <*> instantiateFull' ls
-             <*> instantiateFull' m
+  instantiateFull' (Branches cop cs ls m) =
+    Branches cop
+      <$> instantiateFull' cs
+      <*> instantiateFull' ls
+      <*> instantiateFull' m
 
 instance InstantiateFull CompiledClauses where
   instantiateFull' Fail        = return Fail
@@ -1133,4 +1191,3 @@ instance InstantiateFull QName where
 
 instance InstantiateFull a => InstantiateFull (Maybe a) where
   instantiateFull' = mapM instantiateFull'
-
