@@ -1,7 +1,9 @@
-{-# LANGUAGE CPP              #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE TupleSections    #-}
-{-# LANGUAGE RankNTypes       #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE RankNTypes #-}
+
+{-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.TypeChecking.Reduce.Monad
   ( constructorForm
@@ -12,15 +14,19 @@ module Agda.TypeChecking.Reduce.Monad
   , lookupMeta
   , reportSDoc, reportSLn
   , traceSLn, traceSDoc
+  , askR, applyWhenVerboseS
   ) where
 
+import Prelude hiding (null)
+
 import Control.Arrow ((***), first, second)
-import Control.Applicative
+import Control.Applicative hiding (empty)
 import Control.Monad.Reader
 import Control.Monad.Identity
 
 import qualified Data.Map as Map
 import Data.Maybe
+import Data.Monoid
 
 import Debug.Trace
 import System.IO.Unsafe
@@ -38,6 +44,7 @@ import Agda.Interaction.Options
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Lens
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Pretty
 
 #include "undefined.h"
@@ -160,9 +167,10 @@ traceSLn :: HasOptions m => VerboseKey -> Int -> String -> m a -> m a
 traceSLn k n s = applyWhenVerboseS k n (trace s)
 
 instance HasConstInfo ReduceM where
+  getRewriteRulesFor = defaultGetRewriteRulesFor (gets id)
   getConstInfo q = ReduceM $ ReaderT $ \(ReduceEnv env st) -> Identity $
-    let defs  = sigDefinitions $ st^.stSignature
-        idefs = sigDefinitions $ st^.stImports
+    let defs  = st^.(stSignature . sigDefinitions)
+        idefs = st^.(stImports . sigDefinitions)
     in case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
         []  -> trace ("Unbound name: " ++ show q ++ " " ++ showQNameId q) __IMPOSSIBLE__
         [d] -> mkAbs env d
@@ -181,8 +189,4 @@ instance HasConstInfo ReduceM where
             _                 -> q
 
           dropLastModule q@QName{ qnameModule = m } =
-            q{ qnameModule = mnameFromList $ init' $ mnameToList m }
-
-          init' [] = {-'-} __IMPOSSIBLE__
-          init' xs = init xs
-
+            q{ qnameModule = mnameFromList $ ifNull (mnameToList m) __IMPOSSIBLE__ init }
