@@ -19,9 +19,11 @@ import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Rules.Builtin
 import Agda.TypeChecking.Rules.Term
 
@@ -67,27 +69,27 @@ bindBuiltinInf e = bindPostulatedName builtinInf e $ \inf _ ->
 bindBuiltinSharp :: A.Expr -> TCM ()
 bindBuiltinSharp e =
   bindPostulatedName builtinSharp e $ \sharp sharpDefn -> do
-    sharpE    <- instantiateFull =<< checkExpr (A.Def sharp) =<< typeOfSharp
+    sharpType <- typeOfSharp
+    TelV fieldTel _ <- telView sharpType
+    sharpE    <- instantiateFull =<< checkExpr (A.Def sharp) sharpType
     Def inf _ <- ignoreSharing <$> primInf
     infDefn   <- getConstInfo inf
     addConstant (defName infDefn) $
       infDefn { defPolarity       = [] -- not monotone
               , defArgOccurrences = [Unused, StrictPos]
-              , theDef = Datatype
-                  { dataPars           = 2
-                  , dataSmallPars      = Perm 2 []
-                  , dataNonLinPars     = Drop 0 $ Perm 2 []
-                  , dataIxs            = 0
-                  , dataInduction      = CoInductive
-                  , dataClause         = Nothing
-                  , dataCons           = [sharp]
-                  , dataSort           = varSort 1
-{-
-                  , dataPolarity       = [Invariant, Invariant]
-                  , dataArgOccurrences = [Unused, StrictPos]
--}
-                  , dataMutual         = []
-                  , dataAbstr          = ConcreteDef
+              , theDef = Record
+                  { recPars           = 2
+                  , recInduction      = Just CoInductive
+                  , recClause         = Nothing
+                  , recConHead        = ConHead sharp CoInductive []  -- flat is added later
+                  , recConType        = sharpType
+                  , recNamedCon       = True
+                  , recFields         = []  -- flat is added later
+                  , recTel            = fieldTel
+                  , recRecursive      = False
+                  , recEtaEquality    = False
+                  , recMutual         = []
+                  , recAbstr          = ConcreteDef
                   }
               }
     addConstant sharp $
@@ -122,23 +124,27 @@ bindBuiltinFlat e =
                    ExtendTel (domH $ sort $ varSort 0)       $ Abs "A" $
                    ExtendTel (domN $ El (varSort 1) $ var 0) $ Abs "x" $
                    EmptyTel
+        infA     = El (varSort 2) $ Def inf [ Apply $ defaultArg $ var 1 ]
+        cpi      = ConPatternInfo Nothing $ Just $ defaultArg infA
     let clause   = Clause
           { clauseRange     = noRange
           , clauseTel       = tel
           , clausePerm      = idP 1
-          , namedClausePats = [ argN $ Named Nothing $ ConP sharpCon Nothing [ argN $ Named Nothing $ VarP "x" ] ]
+          , namedClausePats = [ argN $ Named Nothing $
+              ConP sharpCon cpi [ argN $ Named Nothing $ VarP "x" ] ]
           , clauseBody      = Bind $ Abs "x" $ Body $ var 0
           , clauseType      = Just $ defaultArg $ El (varSort 2) $ var 1
           }
-        cc = Case 0 $ Branches (Map.singleton sharp
+        cc = Case 0 $ Branches False
+                               (Map.singleton sharp
                                  $ WithArity 1 $ Done [defaultArg "x"] $ var 0)
                                Map.empty
                                Nothing
         projection = Projection
-          { projProper   = Nothing
+          { projProper   = Just flat
           , projFromType = inf
           , projIndex    = 3
-          , projDropPars = teleNoAbs (take 2 $ telToList tel) $ Def flat []
+          , projDropPars = teleNoAbs (take 2 $ telToList tel) $ Lam defaultArgInfo $ Abs "x" $ Var 0 [Proj flat]
           , projArgInfo  = defaultArgInfo
           }
     addConstant flat $
@@ -164,6 +170,8 @@ bindBuiltinFlat e =
     -- register flat as record field for constructor sharp
     modifySignature $ updateDefinition sharp $ updateTheDef $ \ def ->
       def { conSrcCon = sharpCon }
+    modifySignature $ updateDefinition inf $ updateTheDef $ \ def ->
+      def { recConHead = sharpCon, recFields = [defaultArg flat] }
     return flatE
 
 -- The coinductive primitives.
