@@ -1,5 +1,9 @@
 {-# LANGUAGE CPP #-}
 
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
+
 module Agda.TypeChecking.Rules.Data where
 
 import Control.Applicative
@@ -47,7 +51,7 @@ import Agda.Utils.Impossible
 --   checked.
 checkDataDef :: Info.DefInfo -> QName -> [A.LamBinding] -> [A.Constructor] -> TCM ()
 checkDataDef i name ps cs =
-    traceCall (CheckDataDef (getRange i) (qnameName name) ps cs) $ do -- TODO!! (qnameName)
+    traceCall (CheckDataDef (getRange name) (qnameName name) ps cs) $ do -- TODO!! (qnameName)
         let countPars A.DomainFree{} = 1
             countPars (A.DomainFull (A.TypedBindings _ (Arg _ b))) = case b of
               A.TLet{}       -> 0
@@ -55,7 +59,7 @@ checkDataDef i name ps cs =
             npars = sum $ map countPars ps
 
         -- Add the datatype module
-        addSection (qnameToMName name) 0
+        addSection (qnameToMName name)
 
         -- Look up the type of the datatype.
         t <- instantiateFull =<< typeOfConst name
@@ -144,7 +148,7 @@ checkDataDef i name ps cs =
         -- Prop contain at most one element.
         do  proofIrr <- proofIrrelevance
             case (proofIrr, s, cs) of
-                (True, Prop, _:_:_) -> setCurrentRange (getRange cons) $
+                (True, Prop, _:_:_) -> setCurrentRange cons $
                                          typeError PropMustBeSingleton
                 _                   -> return ()
 
@@ -193,8 +197,13 @@ smallParams tel s = do
 -- | Type check a constructor declaration. Checks that the constructor targets
 --   the datatype and that it fits inside the declared sort.
 --   Returns the non-linear parameters.
-checkConstructor :: QName -> Telescope -> Nat -> Sort
-                 -> A.Constructor -> TCM [Int]
+checkConstructor
+  :: QName         -- ^ Name of data type.
+  -> Telescope     -- ^ Parameter telescope.
+  -> Nat           -- ^ Number of indices of the data type.
+  -> Sort          -- ^ Sort of the data type.
+  -> A.Constructor -- ^ Constructor declaration (type signature).
+  -> TCM [Int]     -- ^ Non-linear parameters.
 checkConstructor d tel nofIxs s (A.ScopedDecl scope [con]) = do
   setScope scope
   checkConstructor d tel nofIxs s con
@@ -265,7 +274,7 @@ checkConstructor _ _ _ _ _ = __IMPOSSIBLE__ -- constructors are axioms
 bindParameters :: [A.LamBinding] -> Type -> (Telescope -> Type -> TCM a) -> TCM a
 bindParameters [] a ret = ret EmptyTel a
 bindParameters (A.DomainFull (A.TypedBindings _ (Arg info (A.TBind _ xs _))) : bs) a ret =
-  bindParameters ([ A.DomainFree info x | x <- xs ] ++ bs) a ret
+  bindParameters (map (mergeHiding . fmap (A.DomainFree info)) xs ++ bs) a ret
 bindParameters (A.DomainFull (A.TypedBindings _ (Arg info (A.TLet _ lbs))) : bs) a ret =
   __IMPOSSIBLE__
 bindParameters ps0@(A.DomainFree info x : ps) (El _ (Pi arg@(Dom info' a) b)) ret
@@ -299,7 +308,8 @@ fitsIn t s = do
   t <- reduce t
   case ignoreSharing $ unEl t of
     Pi dom b -> do
-      getSort (unDom dom) `leqSort` s
+      sa <- reduce $ getSort dom
+      unless (sa == SizeUniv) $ sa `leqSort` s
       addContext (absName b, dom) $ fitsIn (absBody b) (raise 1 s)
     _ -> return () -- getSort t `leqSort` s  -- Andreas, 2013-04-13 not necessary since constructor type ends in data type
 
@@ -373,10 +383,10 @@ constructs nofPars t q = constrT 0 t
         nonLinearParams n pars ixs =
           -- compute the free de Bruijn indices in the data indices
           -- ALT: Ignore all sorts?
-          let fv = freeVarsIgnore IgnoreInAnnotations ixs
+          let fv = allRelevantVarsIgnoring IgnoreInAnnotations ixs
           -- keep relevant ones, convert to de Bruijn levels
           -- note: xs is descending list
-              xs = map ((n-1) -) $ VarSet.toList $ relevantVars fv
+              xs = map ((n-1) -) $ VarSet.toList fv
           -- keep those that correspond to parameters
           -- in ascending list
           in  reverse $ filter (< size pars) xs
