@@ -1,6 +1,11 @@
-{-# LANGUAGE CPP           #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE TupleSections #-}
 
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
+
 module Agda.TypeChecking.Rules.Decl where
 
 import Control.Monad
@@ -28,17 +33,17 @@ import Agda.Syntax.Translation.InternalToAbstract
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Benchmark (billTop, reimburseTop)
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Injectivity
+import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Positivity
 import Agda.TypeChecking.Polarity
 import Agda.TypeChecking.Pretty
-import Agda.TypeChecking.Primitive hiding (Nat)
+import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.ProjectionLike
 import Agda.TypeChecking.Quote
 import Agda.TypeChecking.Records
@@ -53,7 +58,7 @@ import Agda.TypeChecking.Rules.Term
 import Agda.TypeChecking.Rules.Data    ( checkDataDef )
 import Agda.TypeChecking.Rules.Record  ( checkRecDef )
 import Agda.TypeChecking.Rules.Def     ( checkFunDef, useTerPragma )
-import Agda.TypeChecking.Rules.Builtin ( bindBuiltin )
+import Agda.TypeChecking.Rules.Builtin
 
 import Agda.Termination.TermCheck
 
@@ -69,6 +74,7 @@ import Agda.Utils.Impossible
 -- | Type check a sequence of declarations.
 checkDecls :: [A.Declaration] -> TCM ()
 checkDecls ds = do
+  reportSLn "tc.decl" 45 $ "Checking " ++ show (length ds) ++ " declarations..."
   mapM_ checkDecl ds
   -- Andreas, 2011-05-30, unfreezing moved to Interaction/Imports
   -- whenM onTopLevel unfreezeMetas
@@ -76,14 +82,14 @@ checkDecls ds = do
 -- | Type check a single declaration.
 
 checkDecl :: A.Declaration -> TCM ()
-checkDecl d = traceCall (SetRange (getRange d)) $ do
-    reportSDoc "tc.decl" 10 $ vcat
-      [ text "checking declaration"
-      , prettyA d
-      ]
+checkDecl d = setCurrentRange d $ do
+    reportSDoc "tc.decl" 10 $ text "checking declaration"
+    debugPrintDecl d
+    reportSDoc "tc.decl" 90 $ (text . show) d
+    reportSDoc "tc.decl" 10 $ prettyA d  -- Might loop, see e.g. Issue 1597
 
     -- Issue 418 fix: freeze metas before checking an abstract thing
-    when isAbstract freezeMetas
+    when_ isAbstract freezeMetas
 
     let -- What kind of final checks/computations should be performed
         -- if we're not inside a mutual block?
@@ -126,7 +132,7 @@ checkDecl d = traceCall (SetRange (getRange d)) $ do
                                   -- highlighting purposes.
       A.UnquoteDecl mi i x e   -> checkUnquoteDecl mi i x e
 
-    unlessM (isJust <$> asks envMutualBlock) $ do
+    whenNothingM (asks envMutualBlock) $ do
 
       -- Syntax highlighting.
       highlight_ d
@@ -135,8 +141,7 @@ checkDecl d = traceCall (SetRange (getRange d)) $ do
       whenJust finalChecks $ \ theMutualChecks -> do
         solveSizeConstraints
         wakeupConstraints_   -- solve emptiness constraints
-        freezeMetas
-
+        _ <- freezeMetas
         theMutualChecks
 
     where
@@ -149,6 +154,7 @@ checkDecl d = traceCall (SetRange (getRange d)) $ do
 
     check x i m = do
       reportSDoc "tc.decl" 5 $ text "Checking" <+> prettyTCM x <> text "."
+      reportSLn "tc.decl.abstract" 25 $ show (Info.defAbstract i)
       r <- abstract (Info.defAbstract i) m
       reportSDoc "tc.decl" 5 $ text "Checked" <+> prettyTCM x <> text "."
       return r
@@ -170,10 +176,19 @@ mutualChecks i d ds names = do
   -- to avoid making the injectivity checker loop.
   checkTermination_        d
   checkPositivity_         names
-  checkCoinductiveRecords  ds
+  -- Andreas, 2015-03-26 Issue 1470:
+  -- Restricting coinduction to recursive does not solve the
+  -- actual problem, and prevents interesting sound applications
+  -- of sized types.
+  -- checkCoinductiveRecords  ds
   -- Andreas, 2012-09-11:  Injectivity check stores clauses
   -- whose 'Relevance' is affected by polarity computation,
-  -- so do it here.
+  -- so do it here (again).
+  -- Andreas, 2015-07-01:  In particular, 'UnusedArg's of local functions
+  -- are only recognized after the polarity computation.
+  -- See Issue 1366 for an example where injectivity of a local function
+  -- is used to solve metas.  It fails if we do injectivity analysis
+  -- before polarity only.
   checkInjectivity_        names
   checkProjectionLikeness_ names
 
@@ -195,13 +210,13 @@ checkUnquoteDecl mi i x e = do
       -- Add x to signature, otherwise reification gets unhappy.
       addConstant x =<< do
         useTerPragma $ defaultDefn defaultArgInfo x a emptyFunction
-      a <- reifyUnquoted a
+      a <- reifyUnquoted $ killRange a
       reportSDoc "tc.unquote.decl" 10 $
         vcat [ text "unquoteDecl" <+> prettyTCM x <+> text "-->"
              , prettyTCM x <+> text ":" <+> prettyA a ]
       tel <- getContextTelescope
       let tel' = replaceEmptyName "r" $ killRange tel
-      cs <- mapM (reifyUnquoted . QNamed x . abstract tel) cs
+      cs <- mapM (reifyUnquoted . QNamed x . killRange . abstract tel) cs
       reportSDoc "tc.unquote.decl" 10 $ vcat $ map prettyA cs
       let ds = [ A.Axiom A.FunSig i defaultArgInfo x a   -- TODO other than defaultArg
                , A.FunDef i x NotDelayed cs ]
@@ -239,7 +254,7 @@ instantiateDefinitionType q = do
 highlight_ :: A.Declaration -> TCM ()
 highlight_ d = do
   let highlight d = generateAndPrintSyntaxInfo d Full
-  reimburseTop Bench.Typing $ billTop Bench.Highlighting $ case d of
+  Bench.billTo [Bench.Highlighting] $ case d of
     A.Axiom{}                -> highlight d
     A.Field{}                -> __IMPOSSIBLE__
     A.Primitive{}            -> highlight d
@@ -270,7 +285,7 @@ highlight_ d = do
 
 -- | Termination check a declaration.
 checkTermination_ :: A.Declaration -> TCM ()
-checkTermination_ d = reimburseTop Bench.Typing $ billTop Bench.Termination $ do
+checkTermination_ d = Bench.billTo [Bench.Termination] $ do
   reportSLn "tc.decl" 20 $ "checkDecl: checking termination..."
   whenM (optTerminationCheck <$> pragmaOptions) $ do
     case d of
@@ -283,7 +298,7 @@ checkTermination_ d = reimburseTop Bench.Typing $ billTop Bench.Termination $ do
 
 -- | Check a set of mutual names for positivity.
 checkPositivity_ :: Set QName -> TCM ()
-checkPositivity_ names = reimburseTop Bench.Typing $ billTop Bench.Positivity $ do
+checkPositivity_ names = Bench.billTo [Bench.Positivity] $ do
   -- Positivity checking.
   reportSLn "tc.decl" 20 $ "checkDecl: checking positivity..."
   checkStrictlyPositive names
@@ -291,65 +306,77 @@ checkPositivity_ names = reimburseTop Bench.Typing $ billTop Bench.Positivity $
   -- Andreas, 2012-02-13: Polarity computation uses info from
   -- positivity check, so it needs happen after positivity
   -- check.
-  let -- | Do we need to compute polarity information for the
-      -- definition corresponding to the given name?
-      relevant q = do
-        def <- theDef <$> getConstInfo q
-        return $ case def of
-          Function{}    -> Just q
-          Datatype{}    -> Just q
-          Record{}      -> Just q
-          Axiom{}       -> Nothing
-          Constructor{} -> Nothing
-          Primitive{}   -> Nothing
-  mapM_ computePolarity =<< do mapMaybeM relevant $ Set.toList names
+  mapM_ computePolarity $ Set.toList names
 
 -- | Check that all coinductive records are actually recursive.
 --   (Otherwise, one can implement invalid recursion schemes just like
 --   for the old coinduction.)
 checkCoinductiveRecords :: [A.Declaration] -> TCM ()
 checkCoinductiveRecords ds = forM_ ds $ \ d -> case d of
-  A.RecDef _ q (Just (Ranged r CoInductive)) _ _ _ _ -> traceCall (SetRange r) $ do
+  A.RecDef _ q (Just (Ranged r CoInductive)) _ _ _ _ -> setCurrentRange r $ do
     unlessM (isRecursiveRecord q) $ typeError $ GenericError $
       "Only recursive records can be coinductive"
   _ -> return ()
 
 -- | Check a set of mutual names for constructor-headedness.
 checkInjectivity_ :: Set QName -> TCM ()
-checkInjectivity_ names = reimburseTop Bench.Typing $ billTop Bench.Injectivity $ do
+checkInjectivity_ names = Bench.billTo [Bench.Injectivity] $ do
   reportSLn "tc.decl" 20 $ "checkDecl: checking injectivity..."
-
-  -- OLD CODE, REFACTORED using for-loop
-  -- let checkInj (q, def@Defn{ theDef = d@Function{ funClauses = cs, funTerminates = Just True }}) = do
-  --       inv <- checkInjectivity q cs
-  --       modifySignature $ updateDefinition q $ const $
-  --         def { theDef = d { funInv = inv }}
-  --     checkInj _ = return ()
-  -- namesDefs <- mapM (\ q -> (q,) <$> getConstInfo q) $ Set.toList names
-  -- mapM_ checkInj namesDefs
-
-  Fold.forM_ names $ \ q -> do
+  -- Andreas, 2015-07-01, see Issue1366b:
+  -- Injectivity check needs also to be run for abstract definitions.
+  -- Fold.forM_ names $ \ q -> ignoreAbstractMode $ do -- NOT NECESSARY after all
+  Fold.forM_ names $ \ q -> inConcreteOrAbstractMode q $ do
+    -- For abstract q, we should be inAbstractMode,
+    -- otherwise getConstInfo returns Axiom.
+    --
+    -- Andreas, 2015-07-01:
+    -- Quite surprisingly, inAbstractMode does not allow us to look
+    -- at a local definition (@where@ block) of an abstract definition.
+    -- This is because the local definition is defined in a strict submodule.
+    -- We can only see through abstract definitions in the current module
+    -- or super modules inAbstractMode.
+    -- I changed that in Monad.Signature.treatAbstractly', so we can see
+    -- our own local definitions.
     def <- getConstInfo q
     case theDef def of
-      d@Function{ funClauses = cs, funTerminates = Just True } -> do
-        inv <- checkInjectivity q cs
-        modifySignature $ updateDefinition q $ const $
-          def { theDef = d { funInv = inv }}
-      _ -> return ()
+      d@Function{ funClauses = cs, funTerminates = term } -> do
+        case term of
+          Just True -> do
+            inv <- checkInjectivity q cs
+            modifySignature $ updateDefinition q $ const $
+              def { theDef = d { funInv = inv }}
+          _ -> reportSLn "tc.inj.check" 20 $
+             show q ++ " is not verified as terminating, thus, not considered for injectivity"
+      _ -> do
+        abstr <- asks envAbstractMode
+        reportSLn "tc.inj.check" 20 $
+          "we are in " ++ show abstr ++ " and " ++
+             show q ++ " is abstract or not a function, thus, not considered for injectivity"
 
 -- | Check a set of mutual names for projection likeness.
+--
+--   Only a single, non-abstract function can be projection-like.
+--   Making an abstract function projection-like would break the
+--   invariant that the type of the principle argument of a projection-like
+--   function is always inferable.
+
 checkProjectionLikeness_ :: Set QName -> TCM ()
-checkProjectionLikeness_ names = reimburseTop Bench.Typing $ billTop Bench.ProjectionLikeness $ do
+checkProjectionLikeness_ names = Bench.billTo [Bench.ProjectionLikeness] $ do
       -- Non-mutual definitions can be considered for
       -- projection likeness
-      reportSLn "tc.decl" 20 $ "checkDecl: checking projection-likeness..."
-      case Set.toList names of
+      let ds = Set.toList names
+      reportSLn "tc.proj.like" 20 $ "checkDecl: checking projection-likeness of " ++ show ds
+      case ds of
         [d] -> do
           def <- getConstInfo d
+          -- For abstract identifiers, getConstInfo returns Axiom.
+          -- Thus, abstract definitions are not considered for projection-likeness.
           case theDef def of
             Function{} -> makeProjection (defName def)
-            _          -> return ()
-        _ -> return ()
+            _          -> reportSLn "tc.proj.like" 25 $
+              show d ++ " is abstract or not a function, thus, not considered for projection-likeness"
+        _ -> reportSLn "tc.proj.like" 25 $
+               "mutual definitions are not considered for projection-likeness"
 
 -- | Type check an axiom.
 checkAxiom :: A.Axiom -> Info.DefInfo -> A.ArgInfo -> QName -> A.Expr -> TCM ()
@@ -359,11 +386,18 @@ checkAxiom funSig i info0 x e = do
   rel <- max (getRelevance info0) <$> asks envRelevance
   let info = setRelevance rel $ convColor info0
   -- rel <- ifM ((Irrelevant ==) <$> asks envRelevance) (return Irrelevant) (return rel0)
-  t <- isType_ e
+  t <- workOnTypes $ isType_ e
   reportSDoc "tc.decl.ax" 10 $ sep
     [ text $ "checked type signature"
     , nest 2 $ prettyTCM rel <> prettyTCM x <+> text ":" <+> prettyTCM t
+    , nest 2 $ text "of sort " <+> prettyTCM (getSort t)
     ]
+  -- Andreas, 2015-03-17 Issue 1428: Do not postulate sizes in parametrized
+  -- modules!
+  when (funSig == A.NoFunSig) $ do
+    whenM ((== SizeUniv) <$> do reduce $ getSort t) $ do
+      whenM ((> 0) <$> getContextSize) $ do
+        typeError $ GenericError $ "We don't like postulated sizes in parametrized modules."
   -- Not safe. See Issue 330
   -- t <- addForcingAnnotations t
   addConstant x =<< do
@@ -379,7 +413,7 @@ checkAxiom funSig i info0 x e = do
   traceCall (IsType_ e) $ solveSizeConstraints  -- need Range for error message
 
   -- Andreas, 2011-05-31, that freezing below is probably wrong:
-  -- when (Info.defAbstract i == AbstractDef) $ freezeMetas
+  -- when_ (Info.defAbstract i == AbstractDef) $ freezeMetas
 
 -- | Type check a primitive function declaration.
 checkPrimitive :: Info.DefInfo -> QName -> A.Expr -> TCM ()
@@ -399,6 +433,7 @@ checkPragma :: Range -> A.Pragma -> TCM ()
 checkPragma r p =
     traceCall (CheckPragma r p) $ case p of
         A.BuiltinPragma x e -> bindBuiltin x e
+        A.BuiltinNoDefPragma b x -> bindBuiltinNoDef b x
         A.RewritePragma q   -> addRewriteRule q
         A.CompiledTypePragma x hs -> do
           def <- getConstInfo x
@@ -474,8 +509,8 @@ checkPragma r p =
           if not correct
             then typeError $ GenericError "COMPILED_EXPORT directive only works on functions"
             else do
-          ty <- haskellType $ defType def
-          addHaskellExport x ty hs
+              ty <- haskellType $ defType def
+              addHaskellExport x ty hs
         A.CompiledEpicPragma x ep -> do
           def <- getConstInfo x
           case theDef def of
@@ -493,7 +528,7 @@ checkPragma r p =
             _          -> typeError $ GenericError "STATIC directive only works on functions"
         A.OptionsPragma{} -> typeError $ GenericError $ "OPTIONS pragma only allowed at beginning of file, before top module declaration"
         A.EtaPragma r -> do
-          unlessM (isJust <$> isRecord r) $
+          whenNothingM (isRecord r) $
             typeError $ GenericError $ "ETA pragma is only applicable to records"
           modifySignature $ updateDefinition r $ updateTheDef $ setEta
           where
@@ -534,8 +569,8 @@ checkTypeSignature _ = __IMPOSSIBLE__   -- type signatures are always axioms
 -- | Type check a module.
 checkSection :: Info.ModuleInfo -> ModuleName -> A.Telescope -> [A.Declaration] -> TCM ()
 checkSection i x tel ds =
-  checkTelescope_ tel $ \tel' -> do
-    addSection x (size tel')
+  checkTelescope tel $ \ tel' -> do
+    addSection x
     verboseS "tc.mod.check" 10 $ do
       dx   <- prettyTCM x
       dtel <- mapM prettyAs tel
@@ -608,52 +643,54 @@ checkSectionApplication' i m1 (A.SectionApp ptel m2 args) rd rm = do
   -- Module applications can appear in lets, in which case we treat
   -- lambda-bound variables as additional parameters to the module.
   extraParams <- do
-    mfv <- getModuleFreeVars =<< currentModule
+    mfv <- getCurrentModuleFreeVars
     fv  <- size <$> getContextTelescope
     return (fv - mfv)
   when (extraParams > 0) $ reportSLn "tc.mod.apply" 30 $ "Extra parameters to " ++ show m1 ++ ": " ++ show extraParams
   -- Type-check the LHS (ptel) of the module macro.
-  checkTelescope_ ptel $ \ ptel -> do
-  -- We are now in the context @ptel@.
-  -- Get the correct parameter telescope of @m2@.
-  tel <- lookupSection m2
-  vs  <- freeVarsToApply $ mnameToQName m2
-  let tel'  = apply tel vs
-      args' = convColor args
-  -- Compute the remaining parameter telescope after stripping of
-  -- the initial parameters that are determined by the @args@.
-  -- Warning: @etaTel@ is not well-formed in @ptel@, since
-  -- the actual application has not happened.
-  etaTel <- checkModuleArity m2 tel' args'
-  -- Take the module parameters that will be instantiated by @args@.
-  let tel'' = telFromList $ take (size tel' - size etaTel) $ telToList tel'
-  reportSDoc "tc.mod.apply" 15 $ vcat
-    [ text "applying section" <+> prettyTCM m2
-    , nest 2 $ text "args =" <+> sep (map prettyA args)
-    , nest 2 $ text "ptel =" <+> escapeContext (size ptel) (prettyTCM ptel)
-    , nest 2 $ text "tel  =" <+> prettyTCM tel
-    , nest 2 $ text "tel' =" <+> prettyTCM tel'
-    , nest 2 $ text "tel''=" <+> prettyTCM tel''
-    , nest 2 $ text "eta  =" <+> escapeContext (size ptel) (addContext tel'' $ prettyTCM etaTel)
-    ]
-  -- Now, type check arguments.
-  ts <- noConstraints $ checkArguments_ DontExpandLast (getRange i) args' tel''
-  -- Perform the application of the module parameters.
-  let aTel = tel' `apply` ts
-  reportSDoc "tc.mod.apply" 15 $ vcat
-    [ nest 2 $ text "aTel =" <+> prettyTCM aTel
-    ]
-  -- Andreas, 2014-04-06, Issue 1094:
-  -- Add the section with well-formed telescope.
-  addCtxTel aTel $ addSection m1 (size ptel + size aTel + extraParams)
+  checkTelescope ptel $ \ ptel -> do
+    -- We are now in the context @ptel@.
+    -- Get the correct parameter telescope of @m2@.
+    tel <- lookupSection m2
+    vs  <- freeVarsToApply $ mnameToQName m2
+    let tel'  = apply tel vs
+        args' = convColor args
+    -- Compute the remaining parameter telescope after stripping of
+    -- the initial parameters that are determined by the @args@.
+    -- Warning: @etaTel@ is not well-formed in @ptel@, since
+    -- the actual application has not happened.
+    etaTel <- checkModuleArity m2 tel' args'
+    -- Take the module parameters that will be instantiated by @args@.
+    let tel'' = telFromList $ take (size tel' - size etaTel) $ telToList tel'
+    reportSDoc "tc.mod.apply" 15 $ vcat
+      [ text "applying section" <+> prettyTCM m2
+      , nest 2 $ text "args =" <+> sep (map prettyA args)
+      , nest 2 $ text "ptel =" <+> escapeContext (size ptel) (prettyTCM ptel)
+      , nest 2 $ text "tel  =" <+> prettyTCM tel
+      , nest 2 $ text "tel' =" <+> prettyTCM tel'
+      , nest 2 $ text "tel''=" <+> prettyTCM tel''
+      , nest 2 $ text "eta  =" <+> escapeContext (size ptel) (addContext tel'' $ prettyTCM etaTel)
+      ]
+    -- Now, type check arguments.
+    ts <- (noConstraints $ checkArguments_ DontExpandLast (getRange i) args' tel') >>= \case
+      (ts, etaTel') | (size etaTel == size etaTel') -> return ts
+      _ -> __IMPOSSIBLE__
+    -- Perform the application of the module parameters.
+    let aTel = tel' `apply` ts
+    reportSDoc "tc.mod.apply" 15 $ vcat
+      [ nest 2 $ text "aTel =" <+> prettyTCM aTel
+      ]
+    -- Andreas, 2014-04-06, Issue 1094:
+    -- Add the section with well-formed telescope.
+    addCtxTel aTel $ addSection m1
 
-  reportSDoc "tc.mod.apply" 20 $ vcat
-    [ sep [ text "applySection", prettyTCM m1, text "=", prettyTCM m2, fsep $ map prettyTCM (vs ++ ts) ]
-    , nest 2 $ text "  defs:" <+> text (show rd)
-    , nest 2 $ text "  mods:" <+> text (show rm)
-    ]
-  args <- instantiateFull $ vs ++ ts
-  applySection m1 ptel m2 args rd rm
+    reportSDoc "tc.mod.apply" 20 $ vcat
+      [ sep [ text "applySection", prettyTCM m1, text "=", prettyTCM m2, fsep $ map prettyTCM (vs ++ ts) ]
+      , nest 2 $ text "  defs:" <+> text (show rd)
+      , nest 2 $ text "  mods:" <+> text (show rm)
+      ]
+    args <- instantiateFull $ vs ++ ts
+    applySection m1 ptel m2 args rd rm
 
 checkSectionApplication' i m1 (A.RecordModuleIFS x) rd rm = do
   let name = mnameToQName x
@@ -711,3 +748,46 @@ checkSectionApplication' i m1 (A.RecordModuleIFS x) rd rm = do
 --   the work is done when scope checking.
 checkImport :: Info.ModuleInfo -> ModuleName -> TCM ()
 checkImport i x = return ()
+
+------------------------------------------------------------------------
+-- * Debugging
+------------------------------------------------------------------------
+
+class ShowHead a where
+  showHead :: a -> String
+
+instance ShowHead A.Declaration where
+  showHead d =
+    case d of
+      A.Axiom        {} -> "Axiom"
+      A.Field        {} -> "Field"
+      A.Primitive    {} -> "Primitive"
+      A.Mutual       {} -> "Mutual"
+      A.Section      {} -> "Section"
+      A.Apply        {} -> "Apply"
+      A.Import       {} -> "Import"
+      A.Pragma       {} -> "Pragma"
+      A.Open         {} -> "Open"
+      A.FunDef       {} -> "FunDef"
+      A.DataSig      {} -> "DataSig"
+      A.DataDef      {} -> "DataDef"
+      A.RecSig       {} -> "RecSig"
+      A.RecDef       {} -> "RecDef"
+      A.PatternSynDef{} -> "PatternSynDef"
+      A.UnquoteDecl  {} -> "UnquoteDecl"
+      A.ScopedDecl   {} -> "ScopedDecl"
+
+debugPrintDecl :: A.Declaration -> TCM ()
+debugPrintDecl d = do
+    verboseS "tc.decl" 45 $ do
+      reportSLn "tc.decl" 45 $ "checking a " ++ showHead d
+      case d of
+        A.Section info mname tel ds -> do
+          reportSLn "tc.decl" 45 $
+            "section " ++ show mname ++ " has "
+              ++ show (length tel) ++ " parameters and "
+              ++ show (length ds) ++ " declarations"
+          reportSDoc "tc.decl" 45 $ prettyA $ A.Section info mname tel []
+          forM_ ds $ \ d -> do
+            reportSDoc "tc.decl" 45 $ prettyA d
+        _ -> return ()
