@@ -1,4 +1,7 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.TypeChecking.Rules.LHS where
 
@@ -9,6 +12,7 @@ import Data.Maybe
 import Control.Applicative
 import Control.Monad hiding (mapM)
 import Control.Monad.State hiding (mapM)
+import Control.Monad.Trans.Maybe
 
 import Data.Traversable
 
@@ -20,7 +24,7 @@ import Agda.Syntax.Internal.Pattern
 import Agda.Syntax.Abstract (IsProjP(..))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Views (asView)
-import Agda.Syntax.Common
+import Agda.Syntax.Common as Common
 import Agda.Syntax.Info
 import Agda.Syntax.Position
 
@@ -48,7 +52,10 @@ import Agda.TypeChecking.Rules.LHS.Implicit
 import Agda.TypeChecking.Rules.LHS.Instantiate
 import Agda.TypeChecking.Rules.Data
 
-import Agda.Utils.Functor (($>))
+import Agda.Utils.Except (MonadError(..))
+import Agda.Utils.Functor
+import Agda.Utils.List
+import Agda.Utils.ListT
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
 import Agda.Utils.Size
@@ -57,23 +64,53 @@ import Agda.Utils.Size
 import Agda.Utils.Impossible
 
 -- | Compute the set of flexible patterns in a list of patterns. The result is
---   the deBruijn indices of the flexible patterns. A pattern is flexible if it
---   is dotted or implicit.
+--   the deBruijn indices of the flexible patterns.
 flexiblePatterns :: [A.NamedArg A.Pattern] -> TCM FlexibleVars
-flexiblePatterns nps = map setFlex <$> filterM (flexible . namedArg . snd) (zip [0..] $ reverse nps)
-  where
-    setFlex (i, Arg ai p)    = FlexibleVar (getHiding ai) (classify $ namedThing p) i
-    classify A.DotP{}        = DotFlex
-    classify A.ImplicitP{}   = ImplicitFlex
-    classify A.ConP{}        = RecordFlex
-    classify _               = __IMPOSSIBLE__
-    flexible (A.DotP _ _)    = return True
-    flexible (A.ImplicitP _) = return True
-    flexible (A.ConP _ (A.AmbQ [c]) qs) =
-      ifM (isJust <$> isRecordConstructor c)
-          (andM $ map (flexible . namedArg) qs)
-          (return False)
-    flexible _               = return False
+flexiblePatterns nps = do
+  forMaybeM (zip (downFrom $ length nps) nps) $ \ (i, Arg ai p) -> do
+    runMaybeT $ (\ f -> FlexibleVar (getHiding ai) f i) <$> maybeFlexiblePattern p
+
+-- | A pattern is flexible if it is dotted or implicit, or a record pattern
+--   with only flexible subpatterns.
+class IsFlexiblePattern a where
+  maybeFlexiblePattern :: a -> MaybeT TCM FlexibleVarKind
+
+  isFlexiblePattern :: a -> TCM Bool
+  isFlexiblePattern p = isJust <$> runMaybeT (maybeFlexiblePattern p)
+
+instance IsFlexiblePattern A.Pattern where
+  maybeFlexiblePattern p =
+    case p of
+      A.DotP{}
+        -> return DotFlex
+      A.WildP{}
+        -> return ImplicitFlex
+      A.ConP _ (A.AmbQ [c]) qs
+        -> ifM (isNothing <$> isRecordConstructor c) mzero {-else-}
+             (maybeFlexiblePattern qs)
+      _ -> mzero
+
+instance IsFlexiblePattern (I.Pattern' a) where
+  maybeFlexiblePattern p =
+    case p of
+      I.DotP{}  -> return DotFlex
+      I.ConP _ i ps
+        | Just ConPImplicit <- conPRecord i -> return ImplicitFlex  -- expanded from ImplicitP
+        | Just _            <- conPRecord i -> maybeFlexiblePattern ps
+        | otherwise -> mzero
+      I.VarP{}  -> mzero
+      I.LitP{}  -> mzero
+      I.ProjP{} -> mzero
+
+-- | Lists of flexible patterns are 'RecordFlex'.
+instance IsFlexiblePattern a => IsFlexiblePattern [a] where
+  maybeFlexiblePattern ps = RecordFlex <$> mapM maybeFlexiblePattern ps
+
+instance IsFlexiblePattern a => IsFlexiblePattern (Common.Arg c a) where
+  maybeFlexiblePattern = maybeFlexiblePattern . unArg
+
+instance IsFlexiblePattern a => IsFlexiblePattern (Common.Named name a) where
+  maybeFlexiblePattern = maybeFlexiblePattern . namedThing
 
 -- | Compute the dot pattern instantiations.
 dotPatternInsts :: [A.NamedArg A.Pattern] -> Substitution -> [I.Dom Type] -> TCM [DotPatternInst]
@@ -88,7 +125,7 @@ dotPatternInsts ps s as = dpi (map namedArg ps) (reverse s) as
     dpi (p : ps) (Just u : s) (a : as) =
       case p of
         A.DotP _ e    -> (DPI e u a :) <$> dpi ps s as
-        A.ImplicitP _ -> dpi ps s as
+        A.WildP _     -> dpi ps s as
         -- record pattern
         A.ConP _ (A.AmbQ [c]) qs -> do
           Def r es   <- ignoreSharing <$> reduce (unEl $ unDom a)
@@ -152,6 +189,7 @@ instantiatePattern sub perm ps
     merge (Just qs) ps = zipWith mergeA qs ps
       where
         mergeA a1 a2 = fmap (mergeP (namedArg a1) (namedArg a2) <$) a1
+
         mergeP (DotP s)  (DotP t)
           | s == t                    = DotP s
           | otherwise                 = __IMPOSSIBLE__
@@ -161,8 +199,9 @@ instantiatePattern sub perm ps
         -- the rest is homomorphical
         mergeP (DotP _)  _            = __IMPOSSIBLE__
         mergeP _         (DotP _)     = __IMPOSSIBLE__
-        mergeP (ConP c1 mt1 ps) (ConP c2 mt2 qs)
-          | c1 == c2                  = ConP (c1 `withRangeOf` c2) (mplus mt1 mt2) $ zipWith mergeA ps qs
+        mergeP (ConP c1 i1 ps) (ConP c2 i2 qs)
+          | c1 == c2                  = ConP (c1 `withRangeOf` c2) (mergeCPI i1 i2) $
+                                          zipWith mergeA ps qs
           | otherwise                 = __IMPOSSIBLE__
         mergeP (LitP l1) (LitP l2)
           | l1 == l2                  = LitP (l1 `withRangeOf` l2)
@@ -182,6 +221,9 @@ instantiatePattern sub perm ps
         mergeP ProjP{} _              = __IMPOSSIBLE__
         mergeP _       ProjP{}        = __IMPOSSIBLE__
 
+        mergeCPI (ConPatternInfo mr1 mt1) (ConPatternInfo mr2 mt2) =
+          ConPatternInfo (mplus mr1 mr2) (mplus mt1 mt2)
+
 
 -- | In an internal pattern, replace some pattern variables
 --   by dot patterns, according to the given substitution.
@@ -213,7 +255,10 @@ instantiatePattern' sub perm ps = evalState (mapM goArg ps) 0
       ProjP{}      -> return p
     replace p = do
       i <- next
-      return $ fromMaybe p $ DotP <$> sub' !! i
+      let s = case sub' !!! i of
+                Nothing -> __IMPOSSIBLE__
+                Just s  -> s
+      return $ fromMaybe p $ DotP <$> s
 
 
 
@@ -222,12 +267,18 @@ isSolvedProblem :: Problem -> Bool
 isSolvedProblem problem = null (restPats $ problemRest problem) &&
     all (isSolved . snd . asView . namedArg) (problemInPat problem)
   where
-    isSolved (A.DefP _ _ []) = False  -- projection pattern
-    isSolved (A.VarP _)      = True
-    isSolved (A.WildP _)     = True
-    isSolved (A.ImplicitP _) = True
-    isSolved (A.AbsurdP _)   = True
-    isSolved _               = False
+    -- need further splitting:
+    isSolved A.ConP{}        = False
+    isSolved A.DotP{}        = False
+    isSolved A.LitP{}        = False
+    isSolved A.DefP{}        = False  -- projection pattern
+    -- solved:
+    isSolved A.VarP{}        = True
+    isSolved A.WildP{}       = True
+    isSolved A.AbsurdP{}     = True
+    -- impossible:
+    isSolved A.AsP{}         = __IMPOSSIBLE__  -- removed by asView
+    isSolved A.PatternSynP{} = __IMPOSSIBLE__  -- expanded before
 
 -- | For each user-defined pattern variable in the 'Problem', check
 -- that the corresponding data type (if any) does not contain a
@@ -237,19 +288,18 @@ isSolvedProblem problem = null (restPats $ problemRest problem) &&
 -- Precondition: The problem has to be solved.
 
 noShadowingOfConstructors
-  :: (Maybe r -> Call) -- ^ Trace, e.g., @CheckPatternShadowing clause@
+  :: Call -- ^ Trace, e.g., @CheckPatternShadowing clause@
   -> Problem -> TCM ()
 noShadowingOfConstructors mkCall problem =
   traceCall mkCall $ do
     let pat = map (snd . asView . namedArg) $
                   problemInPat problem
         tel = map (unEl . snd . unDom) $ telToList $ problemTel problem
-    zipWithM noShadowing pat tel -- TODO: does not work for flexible arity and projection patterns
+    zipWithM_ noShadowing pat tel -- TODO: does not work for flexible arity and projection patterns
     return ()
   where
   noShadowing (A.WildP     {}) t = return ()
   noShadowing (A.AbsurdP   {}) t = return ()
-  noShadowing (A.ImplicitP {}) t = return ()
   noShadowing (A.ConP      {}) t = return ()  -- only happens for eta expanded record patterns
   noShadowing (A.DefP      {}) t = return ()  -- projection pattern
   noShadowing (A.AsP       {}) t = __IMPOSSIBLE__
@@ -270,7 +320,7 @@ noShadowingOfConstructors mkCall problem =
           Datatype { dataCons = cs } -> do
             case filter ((A.nameConcrete x ==) . A.nameConcrete . A.qnameName) cs of
               []      -> return ()
-              (c : _) -> setCurrentRange (getRange x) $
+              (c : _) -> setCurrentRange x $
                 typeError $ PatternShadowsConstructor x c
           Axiom       {} -> return ()
           Function    {} -> return ()
@@ -309,8 +359,11 @@ checkDotPattern (DPI e v (Dom info a)) =
     -- Should be ok to do noConstraints here
     noConstraints $ equalTerm a u v
 
--- | Bind the variables in a left hand side. Precondition: the patterns should
---   all be 'A.VarP', 'A.WildP', 'A.AbsurdP', or 'A.ImplicitP' and the
+-- | Bind the variables in a left hand side and check that 'Hiding' of
+--   the patterns matches the hiding info in the type.
+--
+--   Precondition: the patterns should
+--   all be 'A.VarP', 'A.WildP', or 'A.AbsurdP' and the
 --   telescope should have the same size as the pattern list.
 --   There could also be 'A.ConP's resulting from eta expanded implicit record
 --   patterns.
@@ -321,11 +374,14 @@ bindLHSVars []        tel@ExtendTel{}  _   = do
   __IMPOSSIBLE__
 bindLHSVars (_ : _)   EmptyTel         _   = __IMPOSSIBLE__
 bindLHSVars []        EmptyTel         ret = ret
-bindLHSVars (p : ps) (ExtendTel a tel) ret =
+bindLHSVars (p : ps) (ExtendTel a tel) ret = do
+  unless (getHiding p == getHiding a) $ typeError WrongHidingInLHS
   case namedArg p of
     A.VarP x      -> addContext (x, a) $ bindLHSVars ps (absBody tel) ret
     A.WildP _     -> bindDummy (absName tel)
-    A.ImplicitP _ -> bindDummy (absName tel)
+                 -- @bindDummy underscore@ does not fix issue 819, but
+                 -- introduces unwanted underscores in error messages
+                 -- (Andreas, 2015-05-28)
     A.AbsurdP pi  -> do
       -- Andreas, 2012-03-15: allow postponement of emptyness check
       isEmptyType (getRange pi) $ unDom a
@@ -361,27 +417,24 @@ bindAsPatterns (AsB x v a : asb) ret = do
         ]
   addLetBinding defaultArgInfo x v a $ bindAsPatterns asb ret
 
-
+-- | Result of checking the LHS of a clause.
 data LHSResult = LHSResult
-  { lhsPatternTele :: Maybe Telescope   -- ^ Γ: The types of the patterns.
-                                        -- 'Nothing' if more patterns than domain types in @a@.
-                                        -- Used only to construct a @with@ function; see 'stripwithClausePatterns'.
-  , lhsVarTele :: Telescope             -- ^ Δ : The types of the pattern variables.
-  , lhsSubstitution :: S.Substitution   -- ^ σ : The patterns in form of a substitution Δ ⊢ σ : Γ
-  , lhsVarNames :: [String]             -- ^ Names for the variables in Δ, for binding the body.
-  , lhsPatterns :: [I.NamedArg Pattern] -- ^ The patterns in internal syntax.
-  , lhsBodyType :: I.Arg Type           -- ^ The type of the body. Is @bσ@ if @Γ@ is defined.
-                                        -- 'Irrelevant' to indicate the rhs must be checked
-                                        -- in irrelevant mode.
-  , lhsPermutation :: Permutation       -- ^ The permutation from pattern vars to @Δ@.
+  { lhsVarTele      :: Telescope
+    -- ^ Δ : The types of the pattern variables, in internal dependency order.
+    -- Corresponds to 'clauseTel'.
+  , lhsPatterns     :: [I.NamedArg Pattern]
+    -- ^ The patterns in internal syntax.
+  , lhsBodyType     :: I.Arg Type
+    -- ^ The type of the body. Is @bσ@ if @Γ@ is defined.
+    -- 'Irrelevant' to indicate the rhs must be checked in irrelevant mode.
+  , lhsPermutation  :: Permutation
+    -- ^ The permutation from pattern vars to @Δ@.
+    -- Corresponds to 'clausePerm'.
   }
 
 instance InstantiateFull LHSResult where
-  instantiateFull' (LHSResult mtel tel sub xs ps t perm) = LHSResult
-    <$> instantiateFull' mtel
-    <*> instantiateFull' tel
-    <*> instantiateFull' sub
-    <*> return xs
+  instantiateFull' (LHSResult tel ps t perm) = LHSResult
+    <$> instantiateFull' tel
     <*> instantiateFull' ps
     <*> instantiateFull' t
     <*> return perm
@@ -393,7 +446,7 @@ instance InstantiateFull LHSResult where
 --   if successful.
 
 checkLeftHandSide
-  :: (Maybe r -> Call)
+  :: Call
      -- ^ Trace, e.g. @CheckPatternShadowing clause@
   -> Maybe QName
      -- ^ The name of the definition we are checking.
@@ -409,7 +462,6 @@ checkLeftHandSide c f ps a ret = do
   -- Andreas, 2013-03-15 deactivating the following test allows
   -- flexible arity
   -- unless (noProblemRest problem) $ typeError $ TooManyArgumentsInLHS a
-  let mgamma = if noProblemRest problem0 then Just $ problemTel problem0 else Nothing
 
   -- doing the splits:
   LHSState problem@(Problem ps (perm, qs) delta rest) sigma dpi asb
@@ -441,9 +493,7 @@ checkLeftHandSide c f ps a ret = do
     -- Check dot patterns
     mapM_ checkDotPattern dpi
 
-    let rho = renamingR perm -- I'm not certain about this...
-        xs  = [ stringToArgName $ "h" ++ show n | n <- [0..permRange perm - 1] ]
-    lhsResult <- return $ LHSResult mgamma delta rho xs qs b' perm
+    lhsResult <- return $ LHSResult delta qs b' perm
     applyRelevanceToContext (getRelevance b') $ ret lhsResult
 
 -- | The loop (tail-recursive): split at a variable in the problem until problem is solved
@@ -458,21 +508,19 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
   -- since we might insert eta expanded record patterns.
   if isSolvedProblem problem then return $ st { lhsProblem = problem } else do
 
-  unlessM (optPatternMatching <$> gets getPragmaOptions) $
-    typeError $ GenericError $ "Pattern matching is disabled"
+    unlessM (optPatternMatching <$> gets getPragmaOptions) $
+      typeError $ GenericError $ "Pattern matching is disabled"
 
-  sp <- splitProblem f problem
-  reportSDoc "tc.lhs.split" 20 $ text "splitting completed"
-  case sp of
-    Left NothingToSplit   -> do
+    foldListT trySplit nothingToSplit $ splitProblem f problem
+  where
+
+    nothingToSplit = do
       reportSLn "tc.lhs.split" 50 $ "checkLHS: nothing to split in problem " ++ show problem
       nothingToSplitError problem
-    Left (SplitPanic err) -> do
-      reportSLn "impossible" 10 $ "checkLHS: panic: " ++ err
-      __IMPOSSIBLE__
 
-    -- Split problem rest (projection pattern)
-    Right (SplitRest projPat projType) -> do
+    -- Split problem rest (projection pattern, does not fail as there is no call to unifier)
+
+    trySplit (SplitRest projPat projType) _ = do
 
       -- Compute the new problem
       let Problem ps1 (iperm, ip) delta (ProblemRest (p:ps2) b) = problem
@@ -488,8 +536,9 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
       applyRelevanceToContext (getRelevance projPat) $ do
         checkLHS f st'
 
-    -- Split on literal pattern
-    Right (Split p0 xs (Arg _ (LitFocus lit iph hix a)) p1) -> do
+    -- Split on literal pattern (does not fail as there is no call to unifier)
+
+    trySplit (Split p0 xs (Arg _ (LitFocus lit iph hix a)) p1) _ = do
 
       -- plug the hole with a lit pattern
       let ip    = plugHole (LitP lit) iph
@@ -498,7 +547,9 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
       -- substitute the literal in p1 and sigma and dpi and asb
       let delta1 = problemTel p0
           delta2 = absApp (fmap problemTel p1) (Lit lit)
-          rho    = liftS (size delta2) $ singletonS (Lit lit)
+          rho    = singletonS (size delta2) (Lit lit)
+          -- Andreas, 2015-06-13 Literals are closed, so need to raise them!
+          -- rho    = liftS (size delta2) $ singletonS 0 (Lit lit)
           -- rho    = [ var i | i <- [0..size delta2 - 1] ]
           --       ++ [ raise (size delta2) $ Lit lit ]
           --       ++ [ var i | i <- [size delta2 ..] ]
@@ -516,10 +567,35 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
       st' <- updateProblemRest (LHSState problem' sigma' dpi' asb')
       checkLHS f st'
 
-    -- Split on constructor pattern
-    Right (Split p0 xs (Arg info
-            ( Focus { focusCon      = c
-                    , focusImplicit = impl
+    -- Split on constructor pattern (unifier might fail)
+
+    trySplit (Split p0 xs focus@(Arg info Focus{}) p1) tryNextSplit = do
+      res <- trySplitConstructor p0 xs focus p1
+      case res of
+        -- Success.  Continue checking LHS.
+        Unifies st'    -> checkLHS f st'
+        -- Mismatch.  Report and abort.
+        NoUnify  tcerr -> throwError tcerr
+        -- Unclear situation.  Try next split.
+        -- If no split works, give error from first split.
+        -- This is conservative, but might not be the best behavior.
+        -- It might be better to collect all the errors and print all of them.
+        DontKnow tcerr -> tryNextSplit `catchError` \ _ -> throwError tcerr
+
+    whenUnifies
+      :: UnificationResult' a
+      -> (a -> TCM (UnificationResult' b))
+      -> TCM (UnificationResult' b)
+    whenUnifies res cont = do
+      case res of
+        Unifies a      -> cont a
+        NoUnify  tcerr -> return $ NoUnify  tcerr
+        DontKnow tcerr -> return $ DontKnow tcerr
+
+    trySplitConstructor p0 xs (Arg info LitFocus{}) p1 = __IMPOSSIBLE__
+    trySplitConstructor p0 xs (Arg info
+             (Focus { focusCon      = c
+                    , focusPatOrigin= porigin
                     , focusConArgs  = qs
                     , focusRange    = r
                     , focusOutPat   = iph
@@ -529,247 +605,250 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
                     , focusIndices  = ws
                     , focusType     = a
                     }
-            )) p1
-          ) -> traceCall (CheckPattern (A.ConP (ConPatInfo impl $ PatRange r) (A.AmbQ [c]) qs)
+             )) p1 = do
+      traceCall (CheckPattern (A.ConP (ConPatInfo porigin $ PatRange r) (A.AmbQ [c]) qs)
                                        (problemTel p0)
                                        (El Prop $ Def d $ map Apply $ vs ++ ws)) $ do
 
-      let delta1 = problemTel p0
-      let typeOfSplitVar = Arg info a
+        let delta1 = problemTel p0
+        let typeOfSplitVar = Arg info a
 
-      reportSDoc "tc.lhs.split" 10 $ sep
-        [ text "checking lhs"
-        , nest 2 $ text "tel =" <+> prettyTCM (problemTel problem)
-        , nest 2 $ text "rel =" <+> (text $ show $ argInfoRelevance info)
-        ]
-
-      reportSDoc "tc.lhs.split" 15 $ sep
-        [ text "split problem"
-        , nest 2 $ vcat
-          [ text "delta1 = " <+> prettyTCM delta1
-          , text "typeOfSplitVar =" <+> prettyTCM typeOfSplitVar
-          , text "focusOutPat =" <+> (text . show) iph
-          , text "delta2 = " <+> prettyTCM (problemTel $ absBody p1)
+        reportSDoc "tc.lhs.split" 10 $ sep
+          [ text "checking lhs"
+          , nest 2 $ text "tel =" <+> prettyTCM (problemTel problem)
+          , nest 2 $ text "rel =" <+> (text $ show $ argInfoRelevance info)
           ]
-        ]
 
-{-
-      c <- conSrcCon . theDef <$> getConstInfo c
-      Con c' [] <- ignoreSharing <$> (constructorForm =<< normalise (Con c []))
-      c  <- return $ c' `withRangeOf` c
--}
-      c <- (`withRangeOf` c) <$> getConForm c
-      ca <- defType <$> getConInfo c
+        reportSDoc "tc.lhs.split" 15 $ sep
+          [ text "split problem"
+          , nest 2 $ vcat
+            [ text "delta1 = " <+> prettyTCM delta1
+            , text "typeOfSplitVar =" <+> prettyTCM typeOfSplitVar
+            , text "focusOutPat =" <+> (text . show) iph
+            , text "delta2 = " <+> prettyTCM (problemTel $ absBody p1)
+            ]
+          ]
 
-      reportSDoc "tc.lhs.split" 20 $ nest 2 $ vcat
-        [ text "ca =" <+> prettyTCM ca
-        , text "vs =" <+> prettyList (map prettyTCM vs)
-        ]
+        c <- (`withRangeOf` c) <$> getConForm c
+        ca <- defType <$> getConInfo c
 
-      -- Lookup the type of the constructor at the given parameters
-      let a = ca `piApply` vs
+        reportSDoc "tc.lhs.split" 20 $ nest 2 $ vcat
+          [ text "ca =" <+> prettyTCM ca
+          , text "vs =" <+> prettyList (map prettyTCM vs)
+          ]
 
-      -- It will end in an application of the datatype
-      (gamma', ca, d', us) <- do
-        TelV gamma' ca@(El _ def) <- telView a
-        let Def d' es = ignoreSharing def
-            Just us   = allApplyElims es
-        return (gamma', ca, d', us)
+        -- Lookup the type of the constructor at the given parameters
+        let a = ca `piApply` vs
 
-      -- This should be the same datatype as we split on
-      unless (d == d') $ typeError $ ShouldBeApplicationOf ca d'
+        -- It will end in an application of the datatype
+        (gamma', ca, d', us) <- do
+          TelV gamma' ca@(El _ def) <- telView a
+          let Def d' es = ignoreSharing def
+              Just us   = allApplyElims es
+          return (gamma', ca, d', us)
 
-      -- reportSDoc "tc.lhs.top" 20 $ nest 2 $ vcat
-      --   [ text "gamma' =" <+> text (show gamma')
-      --   ]
+        -- This should be the same datatype as we split on
+        unless (d == d') $ typeError $ ShouldBeApplicationOf ca d'
 
-      -- Andreas 2010-09-07  propagate relevance info to new vars
-      -- Andreas 2014-11-25  clear 'Forced' and 'Unused'
-      let updRel = ignoreForced . composeRelevance (getRelevance info)
-      gamma' <- return $ mapRelevance updRel <$> gamma'
+        -- reportSDoc "tc.lhs.top" 20 $ nest 2 $ vcat
+        --   [ text "gamma' =" <+> text (show gamma')
+        --   ]
 
-      -- Insert implicit patterns
-      qs' <- insertImplicitPatterns ExpandLast qs gamma'
+        -- Andreas 2010-09-07  propagate relevance info to new vars
+        let updRel = composeRelevance (getRelevance info)
+        gamma' <- return $ mapRelevance updRel <$> gamma'
 
-      unless (size qs' == size gamma') $
-        typeError $ WrongNumberOfConstructorArguments (conName c) (size gamma') (size qs')
+        -- Insert implicit patterns
+        qs' <- insertImplicitPatterns ExpandLast qs gamma'
 
-      let gamma = useNamesFromPattern qs' gamma'
+        unless (size qs' == size gamma') $
+          typeError $ WrongNumberOfConstructorArguments (conName c) (size gamma') (size qs')
 
-      -- Get the type of the datatype.
-      da <- (`piApply` vs) . defType <$> getConstInfo d
+        let gamma = useNamesFromPattern qs' gamma'
 
-      -- Compute the flexible variables
-      flex <- flexiblePatterns (problemInPat p0 ++ qs')
+        -- Get the type of the datatype.
+        da <- (`piApply` vs) . defType <$> getConstInfo d
 
-      -- Compute the constructor indices by dropping the parameters
-      let us' = drop (size vs) us
+        -- Compute the flexible variables
+        flex <- flexiblePatterns (problemInPat p0 ++ qs')
 
-      reportSDoc "tc.lhs.top" 15 $ addCtxTel delta1 $
-        sep [ text "preparing to unify"
+        -- Compute the constructor indices by dropping the parameters
+        let us' = drop (size vs) us
+
+        reportSDoc "tc.lhs.top" 15 $ addCtxTel delta1 $
+          sep [ text "preparing to unify"
+              , nest 2 $ vcat
+                [ text "c      =" <+> prettyTCM c <+> text ":" <+> prettyTCM a
+                , text "d      =" <+> prettyTCM d <+> text ":" <+> prettyTCM da
+                , text "gamma  =" <+> prettyTCM gamma
+                , text "gamma' =" <+> prettyTCM gamma'
+                , text "vs     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM vs)
+                , text "us'    =" <+> brackets (fsep $ punctuate comma $ map prettyTCM us')
+                , text "ws     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM ws)
+                ]
+              ]
+
+        -- Unify constructor target and given type (in Δ₁Γ)
+        res <- addCtxTel (delta1 `abstract` gamma) $
+                unifyIndices flex (raise (size gamma) da) us' (raise (size gamma) ws)
+        whenUnifies res $ \ sub0 -> do
+
+          -- Andreas 2014-11-25  clear 'Forced' and 'Unused'
+          -- Andreas 2015-01-19  ... only after unification
+          gamma <- return $ mapRelevance ignoreForced <$> gamma
+
+          -- We should substitute c ys for x in Δ₂ and sigma
+          let ys     = teleArgs gamma
+              delta2 = absApp (raise (size gamma) $ fmap problemTel p1) (Con c ys)
+              rho0   = liftS (size delta2) $ consS (Con c ys) $ raiseS (size gamma)
+              -- rho0 = [ var i | i <- [0..size delta2 - 1] ]
+              --     ++ [ raise (size delta2) $ Con c ys ]
+              --     ++ [ var i | i <- [size delta2 + size gamma ..] ]
+              sigma0 = applySubst rho0 sigma
+              dpi0   = applySubst rho0 dpi
+              asb0   = applySubst rho0 asb
+              rest0  = applySubst rho0 (problemRest problem)
+
+          reportSDoc "tc.lhs.top" 15 $ addCtxTel (delta1 `abstract` gamma) $ nest 2 $ vcat
+            [ text "delta2 =" <+> prettyTCM delta2
+            , text "sub0   =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) sub0)
+            ]
+          reportSDoc "tc.lhs.top" 15 $ addCtxTel (delta1 `abstract` gamma `abstract` delta2) $
+            nest 2 $ vcat
+              [ text "dpi0 = " <+> brackets (fsep $ punctuate comma $ map prettyTCM dpi0)
+              , text "asb0 = " <+> brackets (fsep $ punctuate comma $ map prettyTCM asb0)
+              ]
+
+          -- Andreas, 2010-09-09, save the type.
+          -- It is relative to delta1, but it should be relative to
+          -- all variables which will be bound by patterns.
+          -- Thus, it has to be raised by 1 (the "hole" variable)
+          -- plus the length of delta2 (the variables coming after the hole).
+          let storedPatternType = raise (1 + size delta2) typeOfSplitVar
+          -- Also remember if we are a record pattern and from an implicit pattern.
+          isRec <- isRecord d
+          let cpi = ConPatternInfo (isRec $> porigin) (Just storedPatternType)
+
+          -- Plug the hole in the out pattern with c ys
+          let ysp = map (argFromDom . fmap (namedVarP . fst)) $ telToList gamma
+              ip  = plugHole (ConP c cpi ysp) iph
+              ip0 = applySubst rho0 ip
+
+          -- Δ₁Γ ⊢ sub0, we need something in Δ₁ΓΔ₂
+          -- Also needs to be padded with Nothing's to have the right length.
+          let pad n xs x = xs ++ replicate (max 0 $ n - size xs) x
+              newTel = problemTel p0 `abstract` (gamma `abstract` delta2)
+              sub    = replicate (size delta2) Nothing ++
+                       pad (size delta1 + size gamma) (raise (size delta2) sub0) Nothing
+
+          reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
+            [ text "newTel =" <+> prettyTCM newTel
+            , addCtxTel newTel $ text "sub =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) sub)
+            , text "ip   =" <+> do prettyList $ map (prettyTCM . namedArg) ip
+            , text "ip0  =" <+> do prettyList $ map (prettyTCM . namedArg) ip0
+            ]
+          reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
+            [ text "rho0 =" <+> text (show rho0)
+            ]
+
+          -- Instantiate the new telescope with the given substitution
+          (delta', perm, rho, instTypes) <- instantiateTel sub newTel
+
+
+          reportSDoc "tc.lhs.inst" 12 $
+            vcat [ sep [ text "instantiateTel"
+                       , nest 4 $ brackets $ fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) sub
+                       , nest 4 $ prettyTCM newTel
+                       ]
+                 , nest 2 $ text "delta' =" <+> prettyTCM delta'
+                 , nest 2 $ text "perm   =" <+> text (show perm)
+                 , nest 2 $ text "itypes =" <+> fsep (punctuate comma $ map prettyTCM instTypes)
+                 ]
+
+  {-      -- Andreas, 2010-09-09
+          -- temporary error message to find non-id perms
+          let sorted (Perm _ xs) = xs == List.sort xs
+          unless (sorted (perm)) $ typeError $ GenericError $ "detected proper permutation " ++ show perm
+  -}
+          -- Compute the new dot pattern instantiations
+          let ps0'   = problemInPat p0 ++ qs' ++ problemInPat (absBody p1)
+
+          reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
+            [ text "subst rho sub =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) (applySubst rho sub))
+            , text "ps0'  =" <+> brackets (fsep $ punctuate comma $ map prettyA ps0')
+            ]
+
+          newDpi <- dotPatternInsts ps0' (applySubst rho sub) instTypes
+
+          -- The final dpis and asbs are the new ones plus the old ones substituted by ρ
+          let dpi' = applySubst rho dpi0 ++ newDpi
+              asb' = applySubst rho $ asb0 ++ raise (size delta2) (map (\x -> AsB x (Con c ys) ca) xs)
+
+          reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
+            [ text "dpi' = " <+> brackets (fsep $ punctuate comma $ map prettyTCM dpi')
+            , text "asb' = " <+> brackets (fsep $ punctuate comma $ map prettyTCM asb')
+            ]
+
+          -- Apply the substitution to the type
+          let sigma'   = applySubst rho sigma0
+              rest'    = applySubst rho rest0
+
+          reportSDoc "tc.lhs.inst" 15 $
+            nest 2 $ text "ps0 = " <+> brackets (fsep $ punctuate comma $ map prettyA ps0')
+
+          -- Permute the in patterns
+          let ps'  = permute perm ps0'
+
+          -- Compute the new permutation of the out patterns. This is the composition of
+          -- the new permutation with the expansion of the old permutation to
+          -- reflect the split.
+          let perm'  = expandP hix (size gamma) $ fst (problemOutPat problem)
+              iperm' = perm `composeP` perm'
+
+          -- Instantiate the out patterns
+          let ip'    = instantiatePattern sub perm' ip0
+              newip  = applySubst rho ip'
+
+          -- Construct the new problem
+          let problem' = Problem ps' (iperm', newip) delta' rest'
+
+          reportSDoc "tc.lhs.top" 12 $ sep
+            [ text "new problem"
             , nest 2 $ vcat
-              [ text "c      =" <+> prettyTCM c <+> text ":" <+> prettyTCM a
-              , text "d      =" <+> prettyTCM d <+> text ":" <+> prettyTCM da
-              , text "gamma  =" <+> prettyTCM gamma
-              , text "gamma' =" <+> prettyTCM gamma'
-              , text "vs     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM vs)
-              , text "us'    =" <+> brackets (fsep $ punctuate comma $ map prettyTCM us')
-              , text "ws     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM ws)
+              [ text "ps'    = " <+> fsep (map prettyA ps')
+              , text "delta' = " <+> prettyTCM delta'
               ]
             ]
 
-      -- Unify constructor target and given type (in Δ₁Γ)
-      sub0 <- addCtxTel (delta1 `abstract` gamma) $
-              unifyIndices_ flex (raise (size gamma) da) us' (raise (size gamma) ws)
+          reportSDoc "tc.lhs.top" 14 $ nest 2 $ vcat
+            [ text "perm'  =" <+> text (show perm')
+            , text "iperm' =" <+> text (show iperm')
+            ]
+          reportSDoc "tc.lhs.top" 14 $ nest 2 $ vcat
+            [ text "ip'    =" <+> prettyList (map (prettyTCM . namedArg) ip')
+            , text "newip  =" <+> prettyList (map (prettyTCM . namedArg) newip)
+            ]
+          reportSDoc "tc.lhs.top" 60 $ nest 2 $ vcat
+            [ text "ip'    =" <+> text (show ip')
+            , text "newip  =" <+> text (show newip)
+            ]
 
-      -- We should substitute c ys for x in Δ₂ and sigma
-      let ys     = teleArgs gamma
-          delta2 = absApp (raise (size gamma) $ fmap problemTel p1) (Con c ys)
-          rho0   = liftS (size delta2) $ Con c ys :# raiseS (size gamma)
-          -- rho0 = [ var i | i <- [0..size delta2 - 1] ]
-          --     ++ [ raise (size delta2) $ Con c ys ]
-          --     ++ [ var i | i <- [size delta2 + size gamma ..] ]
-          sigma0 = applySubst rho0 sigma
-          dpi0   = applySubst rho0 dpi
-          asb0   = applySubst rho0 asb
-          rest0  = applySubst rho0 (problemRest problem)
+          -- if rest type reduces,
+          -- extend the split problem by previously not considered patterns
+          st'@(LHSState problem'@(Problem ps' (iperm', ip') delta' rest')
+                        sigma' dpi' asb')
+            <- updateProblemRest $ LHSState problem' sigma' dpi' asb'
 
-      reportSDoc "tc.lhs.top" 15 $ addCtxTel (delta1 `abstract` gamma) $ nest 2 $ vcat
-        [ text "delta2 =" <+> prettyTCM delta2
-        , text "sub0   =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) sub0)
-        ]
-      reportSDoc "tc.lhs.top" 15 $ addCtxTel (delta1 `abstract` gamma `abstract` delta2) $
-        nest 2 $ vcat
-          [ text "dpi0 = " <+> brackets (fsep $ punctuate comma $ map prettyTCM dpi0)
-          , text "asb0 = " <+> brackets (fsep $ punctuate comma $ map prettyTCM asb0)
-          ]
-
-      -- Andreas, 2010-09-09, save the type a of record pattern.
-      -- It is relative to delta1, but it should be relative to
-      -- all variables which will be bound by patterns.
-      -- Thus, it has to be raised by 1 (the "hole" variable)
-      -- plus the length of delta2 (the variables coming after the hole).
-      storedPatternType <- ifM (isJust <$> isRecord d)
-        (return $ Just (impl, raise (1 + size delta2) typeOfSplitVar))
-        (return $ Nothing)
-
-      -- Plug the hole in the out pattern with c ys
-      let ysp = map (argFromDom . fmap (namedVarP . fst)) $ telToList gamma
-          ip  = plugHole (ConP c storedPatternType ysp) iph
-          ip0 = applySubst rho0 ip
-
-      -- Δ₁Γ ⊢ sub0, we need something in Δ₁ΓΔ₂
-      -- Also needs to be padded with Nothing's to have the right length.
-      let pad n xs x = xs ++ replicate (max 0 $ n - size xs) x
-          newTel = problemTel p0 `abstract` (gamma `abstract` delta2)
-          sub    = replicate (size delta2) Nothing ++
-                   pad (size delta1 + size gamma) (raise (size delta2) sub0) Nothing
-
-      reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
-        [ text "newTel =" <+> prettyTCM newTel
-        , addCtxTel newTel $ text "sub =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) sub)
-        , text "ip   =" <+> text (show ip)
-        , text "ip0  =" <+> text (show ip0)
-        ]
-      reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
-        [ text "rho0 =" <+> text (show rho0)
-        ]
-
-      -- Instantiate the new telescope with the given substitution
-      (delta', perm, rho, instTypes) <- instantiateTel sub newTel
-
-
-      reportSDoc "tc.lhs.inst" 12 $
-        vcat [ sep [ text "instantiateTel"
-                   , nest 4 $ brackets $ fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) sub
-                   , nest 4 $ prettyTCM newTel
-                   ]
-             , nest 2 $ text "delta' =" <+> prettyTCM delta'
-             , nest 2 $ text "perm   =" <+> text (show perm)
-             , nest 2 $ text "itypes =" <+> fsep (punctuate comma $ map prettyTCM instTypes)
-             ]
-
-{-          -- Andreas, 2010-09-09
-      -- temporary error message to find non-id perms
-      let sorted (Perm _ xs) = xs == List.sort xs
-      unless (sorted (perm)) $ typeError $ GenericError $ "detected proper permutation " ++ show perm
--}
-      -- Compute the new dot pattern instantiations
-      let ps0'   = problemInPat p0 ++ qs' ++ problemInPat (absBody p1)
-
-      reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
-        [ text "subst rho sub =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) (applySubst rho sub))
-        , text "ps0'  =" <+> brackets (fsep $ punctuate comma $ map prettyA ps0')
-        ]
-
-      newDpi <- dotPatternInsts ps0' (applySubst rho sub) instTypes
-
-      -- The final dpis and asbs are the new ones plus the old ones substituted by ρ
-      let dpi' = applySubst rho dpi0 ++ newDpi
-          asb' = applySubst rho $ asb0 ++ raise (size delta2) (map (\x -> AsB x (Con c ys) ca) xs)
-
-      reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
-        [ text "dpi' = " <+> brackets (fsep $ punctuate comma $ map prettyTCM dpi')
-        , text "asb' = " <+> brackets (fsep $ punctuate comma $ map prettyTCM asb')
-        ]
-
-      -- Apply the substitution to the type
-      let sigma'   = applySubst rho sigma0
-          rest'    = applySubst rho rest0
-
-      reportSDoc "tc.lhs.inst" 15 $
-        nest 2 $ text "ps0 = " <+> brackets (fsep $ punctuate comma $ map prettyA ps0')
-
-      -- Permute the in patterns
-      let ps'  = permute perm ps0'
-
-     -- Compute the new permutation of the out patterns. This is the composition of
-      -- the new permutation with the expansion of the old permutation to
-      -- reflect the split.
-      let perm'  = expandP hix (size gamma) $ fst (problemOutPat problem)
-          iperm' = perm `composeP` perm'
-
-      -- Instantiate the out patterns
-      let ip'    = instantiatePattern sub perm' ip0
-          newip  = applySubst rho ip'
-
-      -- Construct the new problem
-      let problem' = Problem ps' (iperm', newip) delta' rest'
-
-      reportSDoc "tc.lhs.top" 12 $ sep
-        [ text "new problem"
-        , nest 2 $ vcat
-          [ text "ps'    = " <+> fsep (map prettyA ps')
-          , text "delta' = " <+> prettyTCM delta'
-          ]
-        ]
-
-      reportSDoc "tc.lhs.top" 14 $ nest 2 $ vcat
-        [ text "perm'  =" <+> text (show perm')
-        , text "iperm' =" <+> text (show iperm')
-        ]
-      reportSDoc "tc.lhs.top" 14 $ nest 2 $ vcat
-        [ text "ip'    =" <+> text (show ip')
-        , text "newip  =" <+> text (show newip)
-        ]
-
-      -- if rest type reduces,
-      -- extend the split problem by previously not considered patterns
-      st'@(LHSState problem'@(Problem ps' (iperm', ip') delta' rest')
-                    sigma' dpi' asb')
-        <- updateProblemRest $ LHSState problem' sigma' dpi' asb'
-
-      reportSDoc "tc.lhs.top" 12 $ sep
-        [ text "new problem from rest"
-        , nest 2 $ vcat
-          [ text "ps'    = " <+> fsep (map prettyA ps')
-          , text "delta' = " <+> prettyTCM delta'
-          , text "ip'    =" <+> text (show ip')
-          , text "iperm' =" <+> text (show iperm')
-          ]
-        ]
-      -- Continue splitting
-      checkLHS f st'
+          reportSDoc "tc.lhs.top" 12 $ sep
+            [ text "new problem from rest"
+            , nest 2 $ vcat
+              [ text "ps'    = " <+> fsep (map prettyA ps')
+              , text "delta' = " <+> prettyTCM delta'
+              , text "ip'    = " <+> prettyList (map (prettyTCM . namedArg) ip')
+              , text "iperm' = " <+> text (show iperm')
+              ]
+            ]
+          return $ Unifies st'
 
 
 -- | Ensures that we are not performing pattern matching on codata.
