@@ -1,6 +1,6 @@
 -- {-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveFoldable    #-}
-{-# LANGUAGE DeriveFunctor     #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE DeriveTraversable #-}
 {-# LANGUAGE FlexibleInstances #-}
 
@@ -8,7 +8,7 @@ module Agda.TypeChecking.Rules.LHS.Problem where
 
 import Prelude hiding (null)
 
-import Data.Foldable
+import Data.Foldable ( Foldable )
 import Data.Traversable
 
 import Agda.Syntax.Common
@@ -19,7 +19,7 @@ import Agda.Syntax.Internal.Pattern
 import qualified Agda.Syntax.Abstract as A
 
 import Agda.TypeChecking.Substitute as S
-import Agda.TypeChecking.Pretty hiding (empty)
+import Agda.TypeChecking.Pretty
 
 import Agda.Utils.Except ( Error(noMsg, strMsg) )
 import Agda.Utils.Null
@@ -32,8 +32,10 @@ type FlexibleVars   = [FlexibleVar Nat]
 --   The alternatives are ordered such that we will assign the higher one first,
 --   i.e., first we try to assign a @DotFlex@, then...
 data FlexibleVarKind
-  = RecordFlex   -- ^ From a record pattern ('ConP').
-  | ImplicitFlex -- ^ From a hidden formal argument ('ImplicitP').
+  = RecordFlex [FlexibleVarKind]
+      -- ^ From a record pattern ('ConP').
+      --   Saves the 'FlexibleVarKind' of its subpatterns.
+  | ImplicitFlex -- ^ From a hidden formal argument or underscore ('WildP').
   | DotFlex      -- ^ From a dot pattern ('DotP').
   deriving (Eq, Ord, Show)
 
@@ -117,7 +119,7 @@ data ProblemRest = ProblemRest
 data Focus
   = Focus
     { focusCon      :: QName
-    , focusImplicit :: Bool -- ^ Do we come from an implicit record pattern?
+    , focusPatOrigin:: ConPOrigin -- ^ Do we come from an implicit or record pattern?
     , focusConArgs  :: [A.NamedArg A.Pattern]
     , focusRange    :: Range
     , focusOutPat   :: OneHolePatterns
@@ -165,11 +167,6 @@ consSplitProblem p x dom s@Split{ splitLPats = ps } = s{ splitLPats = consProble
   consProblem' (Problem ps () tel pr) =
     Problem (p:ps) () (ExtendTel dom $ Abs x tel) pr
 
-data SplitError
-  = NothingToSplit
-  | SplitPanic String
-    -- ^ __IMPOSSIBLE__, only there to make this instance of 'Error'.
-
 data DotPatternInst = DPI A.Expr Term (I.Dom Type)
 data AsBinding      = AsB Name Term Type
 
@@ -207,10 +204,6 @@ instance PrettyTCM AsBinding where
         , nest 2 $ text ":" <+> prettyTCM a
         ]
 
-instance Error SplitError where
-  noMsg  = NothingToSplit
-  strMsg = SplitPanic
-
 instance Null ProblemRest where
   null  = null . restPats
   empty = ProblemRest { restPats = [], restType = defaultArg typeDontCare }
@@ -218,4 +211,3 @@ instance Null ProblemRest where
 instance Null a => Null (Problem' a) where
   null p = null (problemInPat p) && null (problemRest p)
   empty  = Problem empty empty empty empty
-
