@@ -1,5 +1,5 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE PatternGuards       #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.TypeChecking.Rules.LHS.Split
@@ -10,6 +10,7 @@ import Prelude hiding (null)
 
 import Control.Applicative hiding (empty)
 import Control.Monad.Trans ( lift )
+import Control.Monad.Trans.Maybe
 
 import Data.Maybe (fromMaybe)
 import Data.List hiding (null)
@@ -28,7 +29,7 @@ import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Views (asView)
 import qualified Agda.Syntax.Info as A
 
-import Agda.TypeChecking.Monad hiding (SplitError)
+import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 
 import Agda.TypeChecking.Constraints
@@ -38,21 +39,16 @@ import Agda.TypeChecking.Free
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Patterns.Abstract
-import Agda.TypeChecking.Pretty hiding (empty)
+import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 
 import Agda.TypeChecking.Rules.LHS.Problem
 
-import Agda.Utils.Except
-  ( ExceptT
-  , MonadError(throwError)
-  , runExceptT
-  )
-
 import Agda.Utils.Functor ((<.>))
 import Agda.Utils.List
+import Agda.Utils.ListT
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
@@ -75,8 +71,9 @@ import Agda.Utils.Impossible
 splitProblem ::
   Maybe QName -- ^ The definition we are checking at the moment.
   -> Problem  -- ^ The current state of the lhs patterns.
-  -> TCM (Either SplitError SplitProblem)
+  -> ListT TCM SplitProblem
 splitProblem mf (Problem ps (perm, qs) tel pr) = do
+  lift $ do
     reportSLn "tc.lhs.split" 20 $ "initiating splitting"
       ++ maybe "" ((" for definition " ++) . show) mf
     reportSDoc "tc.lhs.split" 30 $ sep
@@ -85,15 +82,17 @@ splitProblem mf (Problem ps (perm, qs) tel pr) = do
       , nest 2 $ text "perm =" <+> prettyTCM perm
       , nest 2 $ text "tel  =" <+> prettyTCM tel
       ]
-    runExceptT $
-      splitP ps (permute perm $ zip [0..] $ allHoles qs) tel
+  splitP ps (permute perm $ zip [0..] $ allHoles qs) tel
   where
     -- Result splitting
-    splitRest :: ProblemRest -> ExceptT SplitError TCM SplitProblem
+    splitRest :: ProblemRest -> ListT TCM SplitProblem
     splitRest (ProblemRest (p : ps) b) | Just f <- mf = do
-      let failure   = lift $ typeError $ CannotEliminateWithPattern p $ unArg b
-          notProjP  = lift $ typeError $ NotAProjectionPattern p
+      let failure   = typeError $ CannotEliminateWithPattern p $ unArg b
+          notProjP  = typeError $ NotAProjectionPattern p
           notRecord = failure -- lift $ typeError $ ShouldBeRecordType $ unArg b
+          wrongHiding :: MonadTCM tcm => QName -> tcm a
+          wrongHiding d = typeError . GenericDocError =<< do
+            liftTCM $ text "Wrong hiding used for projection " <+> prettyTCM d
       lift $ reportSDoc "tc.lhs.split" 20 $ sep
         [ text "splitting problem rest"
         , nest 2 $ text "pattern         p =" <+> prettyA p
@@ -103,13 +102,15 @@ splitProblem mf (Problem ps (perm, qs) tel pr) = do
       -- Probably then there were too many arguments.
       caseMaybe (isProjP p) failure $ \ d -> do
         -- So it is a projection pattern (d = projection name), is it?
-        caseMaybeM (lift $ isProjection d) notProjP $ \p -> case p of
-          Projection{projProper = Nothing} -> __IMPOSSIBLE__
-          Projection{projProper = Just d, projFromType = _, projIndex = n} -> do
+        caseMaybeM (lift $ isProjection d) notProjP $ \ proj -> case proj of
+          -- Andreas, 2015-05-06 issue 1413 projProper=Nothing is not impossible
+          Projection{projProper = Nothing} -> notProjP
+          Projection{projProper = Just d, projIndex = n, projArgInfo = ai} -> do
             -- If projIndex==0, then the projection is already applied
             -- to the record value (like in @open R r@), and then it
             -- is no longer a projection but a record field.
             unless (n > 0) notProjP
+            unless (getHiding p == getHiding ai) $ wrongHiding d
             lift $ reportSLn "tc.lhs.split" 90 "we are a projection pattern"
             -- If the target is not a record type, that's an error.
             -- It could be a meta, but since we cannot postpone lhs checking, we crash here.
@@ -138,7 +139,7 @@ splitProblem mf (Problem ps (perm, qs) tel pr) = do
                 return $ SplitRest argd $ dType `apply` (vs ++ [self])
               _ -> __IMPOSSIBLE__
     -- if there are no more patterns left in the problem rest, there is nothing to split:
-    splitRest _ = throwError $ NothingToSplit
+    splitRest _ = mzero
 
     -- | In @splitP aps iqs tel@,
     --   @aps@ are the user patterns on which we are splitting (inPats),
@@ -148,7 +149,7 @@ splitProblem mf (Problem ps (perm, qs) tel pr) = do
     splitP :: [A.NamedArg A.Pattern]
            -> [(Int, OneHolePatterns)]
            -> Telescope
-           -> ExceptT SplitError TCM SplitProblem
+           -> ListT TCM SplitProblem
 
     -- the next two cases violate the one-to-one correspondence of qs and tel
     splitP _        []           (ExtendTel _ _)         = __IMPOSSIBLE__
@@ -198,12 +199,13 @@ splitProblem mf (Problem ps (perm, qs) tel pr) = do
           -- Succeed if the split type is (already) equal to the type of the literal.
           ifNotM (lift $ tryConversion $ equalType a =<< litType lit)
             {- then -} keepGoing $
-            {- else -} return $ Split
+            {- else -} return Split
               { splitLPats   = empty
               , splitAsNames = xs
               , splitFocus   = Arg ai $ LitFocus lit q i a
               , splitRPats   = Abs x  $ Problem ps () tel __IMPOSSIBLE__
               }
+              `mplus` keepGoing
 
         -- Case: constructor pattern.
         (xs, p@(A.ConP ci (A.AmbQ cs) args)) -> do
@@ -221,79 +223,79 @@ splitProblem mf (Problem ps (perm, qs) tel pr) = do
                 | otherwise = keepGoing
           -- ifBlockedType reduces the type
           ifBlockedType a (const tryInstantiate) $ \ a' -> do
-          case ignoreSharing $ unEl a' of
+            case ignoreSharing $ unEl a' of
 
-            -- Subcase: split type is a Def.
-            Def d es    -> do
+              -- Subcase: split type is a Def.
+              Def d es    -> do
 
-              def <- liftTCM $ theDef <$> getConstInfo d
+                def <- liftTCM $ theDef <$> getConstInfo d
 
-              -- We cannot split on (shape-)irrelevant non-records.
-              -- Andreas, 2011-10-04 unless allowed by option
-              unless (defIsRecord def) $
-                when (unusableRelevance $ getRelevance ai) $
-                unlessM (liftTCM $ optExperimentalIrrelevance <$> pragmaOptions) $
-                typeError $ SplitOnIrrelevant p dom
+                -- We cannot split on (shape-)irrelevant non-records.
+                -- Andreas, 2011-10-04 unless allowed by option
+                unless (defIsRecord def) $
+                  when (unusableRelevance $ getRelevance ai) $
+                  unlessM (liftTCM $ optExperimentalIrrelevance <$> pragmaOptions) $
+                  typeError $ SplitOnIrrelevant p dom
 
-              -- Check that we are at record or data type and return
-              -- the number of parameters.
-              let mp = case def of
-                        Datatype{dataPars = np} -> Just np
-                        Record{recPars = np}    -> Just np
-                        _                       -> Nothing
-              case mp of
-                Nothing -> keepGoing
-                Just np -> do
-                  let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-                  liftTCM $ traceCall (CheckPattern p EmptyTel a) $ do  -- TODO: wrong telescope
-                  -- Check that we construct something in the right datatype
-                  c <- do
-                      cs' <- mapM canonicalName cs
-                      d'  <- canonicalName d
-                      let cons def = case theDef def of
-                            Datatype{dataCons = cs} -> cs
-                            Record{recConHead = c}      -> [conName c]
-                            _                       -> __IMPOSSIBLE__
-                      cs0 <- cons <$> getConstInfo d'
-                      case [ c | (c, c') <- zip cs cs', elem c' cs0 ] of
-                        [c]   -> do
-                          -- If constructor pattern was ambiguous,
-                          -- remember our choice for highlighting info.
-                          when (length cs >= 2) $ storeDisambiguatedName c
-                          return c
-                        []    -> typeError $ ConstructorPatternInWrongDatatype (head cs) d
-                        cs    -> -- if there are more than one we give up (they might have different types)
-                          typeError $ CantResolveOverloadedConstructorsTargetingSameDatatype d cs
+                -- Check that we are at record or data type and return
+                -- the number of parameters.
+                let mp = case def of
+                          Datatype{dataPars = np} -> Just np
+                          Record{recPars = np}    -> Just np
+                          _                       -> Nothing
+                case mp of
+                  Nothing -> keepGoing
+                  Just np -> do
+                    let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+                    traceCall (CheckPattern p EmptyTel a) $ do  -- TODO: wrong telescope
+                      -- Check that we construct something in the right datatype
+                      c <- lift $ do
+                          cs' <- mapM canonicalName cs
+                          d'  <- canonicalName d
+                          let cons def = case theDef def of
+                                Datatype{dataCons = cs} -> cs
+                                Record{recConHead = c}      -> [conName c]
+                                _                       -> __IMPOSSIBLE__
+                          cs0 <- cons <$> getConstInfo d'
+                          case [ c | (c, c') <- zip cs cs', elem c' cs0 ] of
+                            [c]   -> do
+                              -- If constructor pattern was ambiguous,
+                              -- remember our choice for highlighting info.
+                              when (length cs >= 2) $ storeDisambiguatedName c
+                              return c
+                            []    -> typeError $ ConstructorPatternInWrongDatatype (head cs) d
+                            cs    -> -- if there are more than one we give up (they might have different types)
+                              typeError $ CantResolveOverloadedConstructorsTargetingSameDatatype d cs
 
-                  let (pars, ixs) = genericSplitAt np vs
-                  reportSDoc "tc.lhs.split" 10 $ vcat
-                    [ sep [ text "splitting on"
-                          , nest 2 $ fsep [ prettyA p, text ":", prettyTCM dom ]
-                          ]
-                    , nest 2 $ text "pars =" <+> fsep (punctuate comma $ map prettyTCM pars)
-                    , nest 2 $ text "ixs  =" <+> fsep (punctuate comma $ map prettyTCM ixs)
-                    ]
+                      let (pars, ixs) = genericSplitAt np vs
+                      lift $ reportSDoc "tc.lhs.split" 10 $ vcat
+                        [ sep [ text "splitting on"
+                              , nest 2 $ fsep [ prettyA p, text ":", prettyTCM dom ]
+                              ]
+                        , nest 2 $ text "pars =" <+> fsep (punctuate comma $ map prettyTCM pars)
+                        , nest 2 $ text "ixs  =" <+> fsep (punctuate comma $ map prettyTCM ixs)
+                        ]
 
-                  -- Andreas, 2013-03-22 fixing issue 279
-                  -- To resolve ambiguous constructors, Agda always looks up
-                  -- their original definition and reconstructs the parameters
-                  -- from the type @Def d vs@ we check against.
-                  -- However, the constructor could come from a module instantiation
-                  -- with some of the parameters already fixed.
-                  -- Agda did not make sure the two parameter lists coincide,
-                  -- so we add a check here.
-                  -- I guess this issue could be solved more systematically,
-                  -- but the extra check here is non-invasive to the existing code.
-                  checkParsIfUnambiguous cs d pars
+                      -- Andreas, 2013-03-22 fixing issue 279
+                      -- To resolve ambiguous constructors, Agda always looks up
+                      -- their original definition and reconstructs the parameters
+                      -- from the type @Def d vs@ we check against.
+                      -- However, the constructor could come from a module instantiation
+                      -- with some of the parameters already fixed.
+                      -- Agda did not make sure the two parameter lists coincide,
+                      -- so we add a check here.
+                      -- I guess this issue could be solved more systematically,
+                      -- but the extra check here is non-invasive to the existing code.
+                      checkParsIfUnambiguous cs d pars
 
-                  return $ Split
-                    { splitLPats   = empty
-                    , splitAsNames = xs
-                    , splitFocus   = Arg ai $ Focus c (A.patImplicit ci) args (getRange p) q i d pars ixs a
-                    , splitRPats   = Abs x  $ Problem ps () tel __IMPOSSIBLE__
-                    }
-            -- Subcase: split type is not a Def.
-            _   -> keepGoing
+                      (return Split
+                        { splitLPats   = empty
+                        , splitAsNames = xs
+                        , splitFocus   = Arg ai $ Focus c (A.patOrigin ci) args (getRange p) q i d pars ixs a
+                        , splitRPats   = Abs x  $ Problem ps () tel __IMPOSSIBLE__
+                        }) `mplus` keepGoing
+              -- Subcase: split type is not a Def.
+              _   -> keepGoing
 
         -- Case: neither literal nor constructor pattern.
         _ -> keepGoing
@@ -302,8 +304,8 @@ splitProblem mf (Problem ps (perm, qs) tel pr) = do
 -- | @checkParsIfUnambiguous [c] d pars@ checks that the data/record type
 --   behind @c@ is has initial parameters (coming e.g. from a module instantiation)
 --   that coincide with an prefix of @pars@.
-checkParsIfUnambiguous :: [QName] -> QName -> Args -> TCM ()
-checkParsIfUnambiguous [c] d pars = do
+checkParsIfUnambiguous :: MonadTCM tcm => [QName] -> QName -> Args -> tcm ()
+checkParsIfUnambiguous [c] d pars = liftTCM $ do
   dc <- getConstructorData c
   a  <- reduce (Def dc [])
   case ignoreSharing a of
