@@ -1,11 +1,17 @@
-{-# LANGUAGE CPP           #-}
-{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE CPP                      #-}
+{-# LANGUAGE NondecreasingIndentation #-}
+{-# LANGUAGE PatternGuards            #-}
 
 module Agda.TypeChecking.SizedTypes where
 
+import Prelude hiding (null)
+
+import Control.Monad.Writer
+
 import Data.Function
-import Data.List
+import Data.List hiding (null)
 import qualified Data.Map as Map
+import Data.Monoid
 
 import Agda.Interaction.Options
 
@@ -23,9 +29,10 @@ import {-# SOURCE #-} Agda.TypeChecking.Conversion
 import {-# SOURCE #-} Agda.TypeChecking.Constraints
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
-import Agda.Utils.List
+import Agda.Utils.List as List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 
@@ -38,6 +45,122 @@ import Agda.Utils.Impossible
 -- * SIZELT stuff
 ------------------------------------------------------------------------
 
+-- | Check whether a type is either not a SIZELT or a SIZELT that is non-empty.
+checkSizeLtSat :: Type -> TCM ()
+checkSizeLtSat t = do
+  reportSDoc "tc.size" 10 $ sep
+    [ text "checking that " <+> prettyTCM t <+> text " is not an empty type of sizes"
+    , text "in context " <+> do inTopContext . prettyTCM =<< getContextTelescope
+    ]
+  let postpone t = addConstraint $ CheckSizeLtSat t
+  ifBlockedType t (const postpone) $ \ t -> do
+    caseMaybeM (isSizeType t) (return ()) $ \ b -> do
+      case b of
+        BoundedNo -> return ()
+        BoundedLt b -> do
+          ifBlocked b (\ _ _ -> postpone t) $ \ b -> do
+            catchConstraint (CheckSizeLtSat t) $ do
+              unlessM (checkSizeNeverZero b) $ do
+                typeError . GenericDocError =<< do
+                  text "Possibly empty type of sizes " <+> prettyTCM t
+
+-- | Precondition: Term is reduced and not blocked.
+--   Throws a 'patternViolation' if undecided
+checkSizeNeverZero :: Term -> TCM Bool
+checkSizeNeverZero u = do
+  v <- sizeView u
+  case v of
+    SizeInf     -> return True  -- OK, infty is never 0.
+    SizeSuc{}   -> return True  -- OK, a + 1 is never 0.
+    OtherSize u ->
+      case ignoreSharing u of
+        Var i [] -> checkSizeVarNeverZero i
+        -- neutral sizes cannot be guaranteed > 0
+        _ -> return False
+
+-- -- | A size variable is never zero if it is the strict upper bound of
+-- --   some other size variable in context.
+-- --   Eg. @i : Size, j : Size< i@ |- i is never zero.
+-- --   Throws a 'patternViolation' if undecided.
+-- checkSizeVarNeverZero :: Int -> TCM Bool
+-- checkSizeVarNeverZero i = do
+--   -- Looking for a variable j : Size< i, we can restrict to the last i
+--   -- entries, as this variable necessarily has been defined later than i.
+--   doms <- take i <$> getContext
+--   -- We raise each type to make sense in the current context.
+--   let ts = zipWith raise [1..] $ map (snd . unDom) doms
+--   reportSDoc "tc.size" 15 $ sep
+--     [ text "checking that size " <+> prettyTCM (var i) <+> text " is never 0"
+--     , text "in context " <+> do sep $ map prettyTCM ts
+--     ]
+--   foldr f (return False) ts
+--   where
+--   f t cont = do
+--     -- If we encounter a blocked type in the context, we cannot
+--     -- definitely say no.
+--     let yes     = return True
+--         no      = cont
+--         perhaps = cont >>= \ res -> if res then return res else patternViolation
+--     ifBlockedType t (\ _ _ -> perhaps) $ \ t -> do
+--       caseMaybeM (isSizeType t) no $ \ b -> do
+--         case b of
+--           BoundedNo -> no
+--           BoundedLt u -> ifBlocked u (\ _ _ -> perhaps) $ \ u -> do
+--             case ignoreSharing u of
+--                Var i' [] | i == i' -> yes
+--                _ -> no
+
+
+-- | Checks that a size variable is ensured to be @> 0@.
+--   E.g. variable @i@ cannot be zero in context
+--   @(i : Size) (j : Size< ↑ ↑ i) (k : Size< j) (k' : Size< k)@.
+--   Throws a 'patternViolation' if undecided.
+checkSizeVarNeverZero :: Int -> TCM Bool
+checkSizeVarNeverZero i = do
+  reportSDoc "tc.size" 20 $ text "checkSizeVarNeverZero" <+> prettyTCM (var i)
+  -- Looking for the minimal value for size variable i,
+  -- we can restrict to the last i
+  -- entries, as only these can contain i in an upper bound.
+  ts <- map (snd . unDom) . take i <$> getContext
+  -- If we encountered a blocking meta in the context, we cannot
+  -- say ``no'' for sure.
+  (n, Any meta) <- runWriterT $ minSizeValAux ts $ repeat 0
+  if n > 0 then return True else
+    if meta then patternViolation else return False
+  where
+  -- Compute the least valuation for size context ts above the
+  -- given valuation and return its last value.
+  minSizeValAux :: [Type] -> [Int] -> WriterT Any TCM Int
+  minSizeValAux _        []      = __IMPOSSIBLE__
+  minSizeValAux []       (n : _) = return n
+  minSizeValAux (t : ts) (n : ns) = do
+    reportSDoc "tc.size" 60 $
+       text ("minSizeVal (n:ns) = " ++ show (take (length ts + 2) $ n:ns) ++
+             " t =") <+> (text . show) t  -- prettyTCM t  -- Wrong context!
+    -- n is the min. value for variable 0 which has type t.
+    let cont = minSizeValAux ts ns
+        perhaps = tell (Any True) >> cont
+    -- If we encounter a blocked type in the context, we cannot
+    -- give a definite answer.
+    ifBlockedType t (\ _ _ -> perhaps) $ \ t -> do
+      caseMaybeM (liftTCM $ isSizeType t) cont $ \ b -> do
+        case b of
+          BoundedNo -> cont
+          BoundedLt u -> ifBlocked u (\ _ _ -> perhaps) $ \ u -> do
+            reportSLn "tc.size" 60 $ "minSizeVal upper bound u = " ++ show u
+            v <- liftTCM $ deepSizeView u
+            case v of
+              -- Variable 0 has bound @(< j + m)@
+              -- meaning that @minval(j) > n - m@, i.e., @minval(j) >= n+1-m@.
+              -- Thus, we update the min value for @j@ with function @(max (n+1-m))@.
+              DSizeVar j m -> do
+                reportSLn "tc.size" 60 $ "minSizeVal upper bound v = " ++ show v
+                let ns' = List.updateAt j (max $ n+1-m) ns
+                reportSLn "tc.size" 60 $ "minSizeVal ns' = " ++ show (take (length ts + 1) ns')
+                minSizeValAux ts ns'
+              DSizeMeta{} -> perhaps
+              _ -> cont
+
 -- | Check whether a variable in the context is bounded by a size expression.
 --   If @x : Size< a@, then @a@ is returned.
 isBounded :: MonadTCM tcm => Nat -> tcm BoundedSize
@@ -107,13 +230,13 @@ deepSizeView v = do
   Def inf [] <- ignoreSharing <$> primSizeInf
   Def suc [] <- ignoreSharing <$> primSizeSuc
   let loop v = do
-      v <- reduce v
-      case ignoreSharing v of
-        Def x []        | x == inf -> return $ DSizeInf
-        Def x [Apply u] | x == suc -> sizeViewSuc_ suc <$> loop (unArg u)
-        Var i []                   -> return $ DSizeVar i 0
-        MetaV x us                 -> return $ DSizeMeta x us 0
-        _                          -> return $ DOtherSize v
+        v <- reduce v
+        case ignoreSharing v of
+          Def x []        | x == inf -> return $ DSizeInf
+          Def x [Apply u] | x == suc -> sizeViewSuc_ suc <$> loop (unArg u)
+          Var i []                   -> return $ DSizeVar i 0
+          MetaV x us                 -> return $ DSizeMeta x us 0
+          _                          -> return $ DOtherSize v
   loop v
 
 sizeMaxView :: Term -> TCM SizeMaxView
@@ -122,14 +245,14 @@ sizeMaxView v = do
   suc <- getBuiltinDefName builtinSizeSuc
   max <- getBuiltinDefName builtinSizeMax
   let loop v = do
-      v <- reduce v
-      case v of
-        Def x []                   | Just x == inf -> return $ [DSizeInf]
-        Def x [Apply u]            | Just x == suc -> maxViewSuc_ (fromJust suc) <$> loop (unArg u)
-        Def x [Apply u1, Apply u2] | Just x == max -> maxViewMax <$> loop (unArg u1) <*> loop (unArg u2)
-        Var i []                      -> return $ [DSizeVar i 0]
-        MetaV x us                    -> return $ [DSizeMeta x us 0]
-        _                             -> return $ [DOtherSize v]
+        v <- reduce v
+        case v of
+          Def x []                   | Just x == inf -> return $ [DSizeInf]
+          Def x [Apply u]            | Just x == suc -> maxViewSuc_ (fromJust suc) <$> loop (unArg u)
+          Def x [Apply u1, Apply u2] | Just x == max -> maxViewMax <$> loop (unArg u1) <*> loop (unArg u2)
+          Var i []                      -> return $ [DSizeVar i 0]
+          MetaV x us                    -> return $ [DSizeMeta x us 0]
+          _                             -> return $ [DOtherSize v]
   loop v
 
 ------------------------------------------------------------------------
@@ -265,7 +388,7 @@ getSizeMetas interactionMetas = do
             caseMaybe (test b) no $ \ _ -> do
               let yes = return $ Just (m, a, tel)
               if interactionMetas then yes else do
-              ifM (isJust <$> isInteractionMeta m) no yes
+                ifM (isJust <$> isInteractionMeta m) no yes
           _ -> no
 
 {- ROLLED BACK
@@ -430,51 +553,51 @@ solveSizeConstraints = whenM haveSizedTypes $ do
   ms <- getSizeMetas True -- get all size metas, also interaction metas
 
   when (not (null cs) || not (null ms)) $ do
-  reportSLn "tc.size.solve" 10 $ "Solving size constraints " ++ show cs
+    reportSLn "tc.size.solve" 10 $ "Solving size constraints " ++ show cs
 
-  cs <- return $ mapMaybe canonicalizeSizeConstraint cs
-  reportSLn "tc.size.solve" 10 $ "Canonicalized constraints: " ++ show cs
+    cs <- return $ mapMaybe canonicalizeSizeConstraint cs
+    reportSLn "tc.size.solve" 10 $ "Canonicalized constraints: " ++ show cs
 
-  let -- Error for giving up
-      cannotSolve = typeError . GenericDocError =<<
-        vcat (text "Cannot solve size constraints" : map prettyTCM cs0)
+    let -- Error for giving up
+        cannotSolve = typeError . GenericDocError =<<
+          vcat (text "Cannot solve size constraints" : map prettyTCM cs0)
 
-      -- Size metas in constraints.
-      metas0 :: [(MetaId, Int)]  -- meta id + arity
-      metas0 = nub $ map (mapSnd length) $ concatMap flexibleVariables cs
+        -- Size metas in constraints.
+        metas0 :: [(MetaId, Int)]  -- meta id + arity
+        metas0 = nub $ map (mapSnd length) $ concatMap flexibleVariables cs
 
-      -- Unconstrained size metas that do not occur in constraints.
-      metas1 :: [(MetaId, Int)]
-      metas1 = forMaybe ms $ \ (m, _, tel) ->
-        maybe (Just (m, size tel)) (const Nothing) $
-          lookup m metas0
+        -- Unconstrained size metas that do not occur in constraints.
+        metas1 :: [(MetaId, Int)]
+        metas1 = forMaybe ms $ \ (m, _, tel) ->
+          maybe (Just (m, size tel)) (const Nothing) $
+            lookup m metas0
 
-      -- All size metas
-      metas = metas0 ++ metas1
+        -- All size metas
+        metas = metas0 ++ metas1
 
-  reportSLn "tc.size.solve" 15 $ "Metas: " ++ show metas0 ++ ", " ++ show metas1
+    reportSLn "tc.size.solve" 15 $ "Metas: " ++ show metas0 ++ ", " ++ show metas1
 
-  verboseS "tc.size.solve" 20 $
-      -- debug print the type of all size metas
-      forM_ metas $ \ (m, _) ->
-          reportSDoc "tc.size.solve" 20 $ prettyTCM =<< mvJudgement <$> lookupMeta m
+    verboseS "tc.size.solve" 20 $
+        -- debug print the type of all size metas
+        forM_ metas $ \ (m, _) ->
+            reportSDoc "tc.size.solve" 20 $ prettyTCM =<< mvJudgement <$> lookupMeta m
 
-  -- Run the solver.
-  unlessM (oldSolver metas cs) cannotSolve
+    -- Run the solver.
+    unlessM (oldSolver metas cs) cannotSolve
 
-  -- Double-checking the solution.
+    -- Double-checking the solution.
 
-  -- Andreas, 2012-09-19
-  -- The returned solution might not be consistent with
-  -- the hypotheses on rigid vars (j : Size< i).
-  -- Thus, we double check that all size constraints
-  -- have been solved correctly.
-  flip catchError (const cannotSolve) $
-    noConstraints $
-      forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
+    -- Andreas, 2012-09-19
+    -- The returned solution might not be consistent with
+    -- the hypotheses on rigid vars (j : Size< i).
+    -- Thus, we double check that all size constraints
+    -- have been solved correctly.
+    flip catchError (const cannotSolve) $
+      noConstraints $
+        forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
 
 
--- | Old solver for size constraints using 'Agda.Utils.Warshall'.
+-- | Old solver for size constraints using "Agda.Utils.Warshall".
 oldSolver
   :: [(MetaId, Int)]   -- ^ Size metas and their arity.
   -> [SizeConstraint]  -- ^ Size constraints (in preprocessed form).
@@ -511,7 +634,7 @@ oldSolver metas cs = do
                 v = term e
 
             reportSDoc "tc.size.solve" 20 $ sep
-              [ text (show m) <+> text ":="
+              [ pretty m <+> text ":="
               , nest 2 $ prettyTCM v
               ]
 
