@@ -1,10 +1,9 @@
-{-# LANGUAGE CPP                   #-}
-{-# LANGUAGE FlexibleInstances     #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE PatternGuards         #-}
-{-# LANGUAGE ScopedTypeVariables   #-}
-{-# LANGUAGE TupleSections         #-}
-{-# LANGUAGE TypeSynonymInstances  #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TupleSections #-}
 
 -- | Solving size constraints under hypotheses.
 --
@@ -56,21 +55,21 @@
 
 module Agda.TypeChecking.SizedTypes.Solve where
 
-import Control.Monad ( forM )
-import Data.Foldable (Foldable, foldMap)
+import Control.Monad (unless)
+import Data.Foldable (Foldable, foldMap, forM_)
 import Data.Function
 import Data.List
 import Data.Monoid (mappend)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
-import Data.Traversable (Traversable)
+import Data.Traversable (Traversable, forM)
 
 import Agda.Interaction.Options
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
-import Agda.TypeChecking.Monad as TCM
+import Agda.TypeChecking.Monad as TCM hiding (Offset)
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
@@ -92,7 +91,13 @@ import Agda.Utils.Cluster
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Function
 import Agda.Utils.Functor
+
+#if MIN_VERSION_base(4,8,0)
+import Agda.Utils.List hiding ( uncons )
+#else
 import Agda.Utils.List
+#endif
+
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Size
@@ -243,18 +248,18 @@ getSizeHypotheses :: Context -> TCM [(CtxId, SizeConstraint)]
 getSizeHypotheses gamma = inTopContext $ modifyContext (const gamma) $ do
   (_, msizelt) <- getBuiltinSize
   caseMaybe msizelt (return []) $ \ sizelt -> do
-  -- Traverse the context from newest to oldest de Bruijn Index
-  catMaybes <$> do
-    forM (zip [0..] gamma) $ \ (i, ce) -> do
-      -- Get name and type of variable i.
-      let xt = unDom $ ctxEntry ce
-          x  = show $ fst xt
-      t <- reduce . raise (1 + i) . unEl . snd $ xt
-      case ignoreSharing t of
-        Def d [Apply u] | d == sizelt -> do
-          caseMaybeM (sizeExpr $ unArg u) (return Nothing) $ \ a ->
-            return $ Just $ (ctxId ce, Constraint (Rigid (NamedRigid x i) 0) Lt a)
-        _ -> return Nothing
+    -- Traverse the context from newest to oldest de Bruijn Index
+    catMaybes <$> do
+      forM (zip [0..] gamma) $ \ (i, ce) -> do
+        -- Get name and type of variable i.
+        let xt = unDom $ ctxEntry ce
+            x  = show $ fst xt
+        t <- reduce . raise (1 + i) . unEl . snd $ xt
+        case ignoreSharing t of
+          Def d [Apply u] | d == sizelt -> do
+            caseMaybeM (sizeExpr $ unArg u) (return Nothing) $ \ a ->
+              return $ Just $ (ctxId ce, Constraint (Rigid (NamedRigid x i) 0) Lt a)
+          _ -> return Nothing
 
 -- | Convert size constraint into form where each meta is applied
 --   to indices @0,1,..,n-1@ where @n@ is the arity of that meta.
@@ -392,20 +397,20 @@ computeSizeConstraint :: Closure TCM.Constraint -> TCM (Maybe HypSizeConstraint)
 computeSizeConstraint c = do
   let cxt = envContext $ clEnv c
   inTopContext $ modifyContext (const cxt) $ do
-  case clValue c of
-    ValueCmp CmpLeq _ u v -> do
-      reportSDoc "tc.size.solve" 50 $ sep $
-        [ text "converting size constraint"
-        , prettyTCM c
-        ]
-      ma <- sizeExpr u
-      mb <- sizeExpr v
-      (hids, hs) <- unzip <$> getSizeHypotheses cxt
-      let mk a b = HypSizeConstraint cxt hids hs $ Size.Constraint a Le b
-      -- We only create a size constraint if both terms can be
-      -- parsed to our format of size expressions.
-      return $ mk <$> ma <*> mb
-    _ -> __IMPOSSIBLE__
+    case clValue c of
+      ValueCmp CmpLeq _ u v -> do
+        reportSDoc "tc.size.solve" 50 $ sep $
+          [ text "converting size constraint"
+          , prettyTCM c
+          ]
+        ma <- sizeExpr u
+        mb <- sizeExpr v
+        (hids, hs) <- unzip <$> getSizeHypotheses cxt
+        let mk a b = HypSizeConstraint cxt hids hs $ Size.Constraint a Le b
+        -- We only create a size constraint if both terms can be
+        -- parsed to our format of size expressions.
+        return $ mk <$> ma <*> mb
+      _ -> __IMPOSSIBLE__
 
 -- | Turn a term into a size expression.
 --
@@ -419,7 +424,7 @@ sizeExpr u = do
   s <- sizeView u
   case s of
     SizeInf     -> return $ Just Infty
-    SizeSuc u   -> fmap (`plus` (1 :: Int)) <$> sizeExpr u
+    SizeSuc u   -> fmap (`plus` (1 :: Offset)) <$> sizeExpr u
     OtherSize u -> case ignoreSharing u of
       Var i []    -> (\ x -> Just $ Rigid (NamedRigid x i) 0) . show <$> nameOfBV i
 --      MetaV m es  -> return $ Just $ Flex (SizeMeta m es) 0
@@ -436,11 +441,11 @@ sizeExpr u = do
 unSizeExpr :: DBSizeExpr -> TCM Term
 unSizeExpr a =
   case a of
-    Infty                  -> primSizeInf
-    Rigid r n              -> do
+    Infty         -> primSizeInf
+    Rigid r (O n) -> do
       unless (n >= 0) __IMPOSSIBLE__
       sizeSuc n $ var $ rigidIndex r
-    Flex (SizeMeta x es) n -> do
+    Flex (SizeMeta x es) (O n) -> do
       unless (n >= 0) __IMPOSSIBLE__
       sizeSuc n $ MetaV x $ map (Apply . defaultArg . var) es
-    Const{}                -> __IMPOSSIBLE__
+    Const{} -> __IMPOSSIBLE__
