@@ -1,12 +1,12 @@
-{-# LANGUAGE DeriveFoldable            #-}
-{-# LANGUAGE DeriveFunctor             #-}
-{-# LANGUAGE DeriveTraversable         #-}
-{-# LANGUAGE FlexibleInstances         #-}
-{-# LANGUAGE FunctionalDependencies    #-}
-{-# LANGUAGE MultiParamTypeClasses     #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FunctionalDependencies #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
-{-# LANGUAGE TypeSynonymInstances      #-}
-{-# LANGUAGE UndecidableInstances      #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 -- | Syntax of size expressions and constraints.
 
@@ -19,12 +19,28 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable (Traversable)
 
+import Test.QuickCheck
+
 import Agda.TypeChecking.SizedTypes.Utils
 
 -- * Syntax
 
 -- | Constant finite sizes @n >= 0@.
-type Offset = Int
+newtype Offset = O Int
+  deriving (Eq, Ord, Num, Show, Enum)
+
+instance MeetSemiLattice Offset where
+  meet = min
+
+instance Plus Offset Offset Offset where
+  plus (O x) (O y) = O (plus x y)
+
+instance Arbitrary Offset where
+  arbitrary = do
+    NonNegative x <- arbitrary
+    return x
+
+  shrink (O x) = map O $ filter (>= 0) (shrink x)
 
 -- | Fixed size variables @i@.
 newtype Rigid  = RigidId { rigidId :: String }
@@ -54,6 +70,10 @@ data Cmp
   | Le  -- ^ @≤@.
   deriving (Eq, Bounded, Enum)
 
+instance Dioid Cmp where
+  compose     = min
+  unitCompose = Le
+
 -- | Comparison operator is ordered @'Lt' < 'Le'@.
 instance Ord Cmp where
   Lt <= x  = True
@@ -66,6 +86,9 @@ instance MeetSemiLattice Cmp where
 instance Top Cmp where
   top = Le
 
+instance Arbitrary Cmp where
+  arbitrary = arbitraryBoundedEnum
+
 -- | Constraint: an inequation between size expressions,
 --   e.g. @X < ∞@ or @i + 3 ≤ j@.
 data Constraint' rigid flex = Constraint
