@@ -24,6 +24,9 @@ class Eq a => Top a where
 class Plus a b c where
   plus :: a -> b -> c
 
+instance Plus Int Int Int where
+  plus = (+)
+
 class MeetSemiLattice a where
   meet :: a -> a -> a
 
