@@ -1,17 +1,20 @@
-{-# LANGUAGE CPP                       #-}
-{-# LANGUAGE FlexibleInstances         #-}
-{-# LANGUAGE MultiParamTypeClasses     #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
-{-# LANGUAGE PatternGuards             #-}
-{-# LANGUAGE ScopedTypeVariables       #-}
-{-# LANGUAGE TupleSections             #-}
-{-# LANGUAGE TypeSynonymInstances      #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TupleSections #-}
+
+#if __GLASGOW_HASKELL__ >= 710
+{-# LANGUAGE FlexibleContexts #-}
+#endif
 
 module Agda.TypeChecking.SizedTypes.WarshallSolver where
 
 import Prelude hiding (truncate)
 
--- import Control.Applicative hiding (Const)-- ((<*>))
+import Control.Applicative hiding (Const)
 import Control.Monad
 
 import Data.Function (on)
@@ -24,6 +27,8 @@ import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Traversable (for)
 
+import Test.QuickCheck (Arbitrary(..), frequency, choose)
+
 import Agda.TypeChecking.SizedTypes.Syntax
 import Agda.TypeChecking.SizedTypes.Utils
 
@@ -94,13 +99,10 @@ transClos g = setFoldl step g $ allNodes ns
         , l2 <- maybeToList $ lookupEdge g v w
       ]
 
-instance MeetSemiLattice Offset where
-  meet = min
-
 -- * Edge weights
 
 data Weight
-  = Offset Int
+  = Offset Offset
   | Infinity
   deriving (Eq)
 
@@ -124,8 +126,8 @@ instance Enum Weight where
   succ (Infinity) = Infinity
   pred (Offset x) = Offset (pred x)
   pred (Infinity) = Infinity
-  toEnum   = Offset
-  fromEnum (Offset x) = x
+  toEnum = Offset . toEnum
+  fromEnum (Offset x) = fromEnum x
   fromEnum (Infinity) = __IMPOSSIBLE__
 
 -- | Partial implementation of @Num@.
@@ -146,6 +148,12 @@ instance Num Weight where
 instance Plus Weight Offset Weight where
   plus w k = w + (Offset k)
 
+instance Arbitrary Weight where
+  arbitrary = frequency
+    [ (1, return Infinity)
+    , (5, Offset . O <$> choose (0, 200))
+    ]
+
 -- | Test for negativity, used to detect negative cycles.
 class Negative a where
   negative :: a -> Bool
@@ -158,6 +166,9 @@ instance (Ord a, Num a) => Negative a where
 instance Negative Int where
   negative = (< 0)
 
+instance Negative Offset where
+  negative (O x) = negative x
+
 instance Negative Weight where
   negative Infinity = False
   negative (Offset x) = negative x
@@ -216,16 +227,18 @@ instance Top Label where
   isTop Label{}       = False
   isTop LInf          = True
 
+instance Arbitrary Label where
+  arbitrary = frequency
+    [ (1, return LInf)
+    , (5, Label <$> arbitrary <*> arbitrary)
+    ]
+
 -- * Semiring with idempotent '+' == dioid
 
 instance Dioid Weight where
   compose     = (+)
   unitCompose = 0
 
-instance Dioid Cmp where
-  compose     = min
-  unitCompose = Le
-
 instance Dioid Label where
   compose (Label Lt w) (Label Lt w')    = Label Lt $ pred $ w + w'
   compose (Label cmp w) (Label cmp' w') = Label (compose cmp cmp') $ w + w'
@@ -541,9 +554,6 @@ instance (Ord r, Ord f) => SetToInfty f (ConGraph r f) where
 
 -- * Compute solution from constraint graph.
 
-instance Plus Int Int Int where
-  plus = (+)
-
 instance Plus Offset Weight Weight where
   plus e Infinity   = Infinity
   plus e (Offset x) = Offset $ plus e x
@@ -834,38 +844,38 @@ solveGraph pols hg g = do
   -- iterate over all flexible variables
   xas <- catMaybes <$> do
     forM xs $ \ x -> do
-    -- get lower and upper bounds for flexible x
-    let lx = Set.toList $ Map.findWithDefault Set.empty x lbs
-        ux = Set.toList $ Map.findWithDefault Set.empty x ubs
-    traceM ("lower bounds for " ++ show x ++ ": " ++ show lx)
-    traceM ("upper bounds for " ++ show x ++ ": " ++ show ux)
-    -- compute maximum of lower bounds
-    lb <- do
-      case lx of
-        []     -> return $ Nothing
-        (a:as) -> do
-          case foldM (lub hg) a as of
-            Nothing -> Left $ "inconsistent lower bound for " ++ show x
-            Just l  -> return $ Just $ truncateOffset l
-    -- compute minimum of upper bounds
-    ub <- do
-      case ux of
-        []     -> return $ Nothing
-        (a:as) -> do
-          case foldM (glb hg) a as of
-            Nothing -> Left $ "inconsistent upper bound for " ++ show x
-            Just l | validOffset l -> return $ Just l
-                   | otherwise     -> return $ findRigidBelow hg l
-    case (lb, ub) of
-      (Just l, Nothing) -> return $ Just (x, l)  -- solve x = lower bound
-      (Nothing, Just u) -> return $ Just (x, u)  -- solve x = upper bound
-      (Just l,  Just u) -> do
-        traceM ("lower bound for " ++ show x ++ ": " ++ show l)
-        traceM ("upper bound for " ++ show x ++ ": " ++ show u)
-        case getPolarity pols x of
-          Least    -> return $ Just (x, l)
-          Greatest -> return $ Just (x, u)
-      _ -> return Nothing
+      -- get lower and upper bounds for flexible x
+      let lx = Set.toList $ Map.findWithDefault Set.empty x lbs
+          ux = Set.toList $ Map.findWithDefault Set.empty x ubs
+      traceM ("lower bounds for " ++ show x ++ ": " ++ show lx)
+      traceM ("upper bounds for " ++ show x ++ ": " ++ show ux)
+      -- compute maximum of lower bounds
+      lb <- do
+        case lx of
+          []     -> return $ Nothing
+          (a:as) -> do
+            case foldM (lub hg) a as of
+              Nothing -> Left $ "inconsistent lower bound for " ++ show x
+              Just l  -> return $ Just $ truncateOffset l
+      -- compute minimum of upper bounds
+      ub <- do
+        case ux of
+          []     -> return $ Nothing
+          (a:as) -> do
+            case foldM (glb hg) a as of
+              Nothing -> Left $ "inconsistent upper bound for " ++ show x
+              Just l | validOffset l -> return $ Just l
+                     | otherwise     -> return $ findRigidBelow hg l
+      case (lb, ub) of
+        (Just l, Nothing) -> return $ Just (x, l)  -- solve x = lower bound
+        (Nothing, Just u) -> return $ Just (x, u)  -- solve x = upper bound
+        (Just l,  Just u) -> do
+          traceM ("lower bound for " ++ show x ++ ": " ++ show l)
+          traceM ("upper bound for " ++ show x ++ ": " ++ show u)
+          case getPolarity pols x of
+            Least    -> return $ Just (x, l)
+            Greatest -> return $ Just (x, u)
+        _ -> return Nothing
   return $ Map.fromList xas
 
 -- | Solve a forest of constraint graphs relative to a hypotheses graph.
