@@ -1,12 +1,17 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE DeriveDataTypeable   #-}
-{-# LANGUAGE DeriveFunctor        #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
+{-# LANGUAGE CPP                #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFunctor      #-}
+{-# LANGUAGE FlexibleContexts   #-}
+{-# LANGUAGE FlexibleInstances  #-}
+{-# LANGUAGE PatternGuards      #-}
+{-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE TupleSections      #-}
+
+#if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE OverlappingInstances #-}
-{-# LANGUAGE StandaloneDeriving   #-}
-{-# LANGUAGE TupleSections        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+#endif
+
+{-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.TypeChecking.Substitute
   ( module Agda.TypeChecking.Substitute
@@ -20,6 +25,7 @@ import Data.Functor
 import Data.List hiding (sort, drop)
 import qualified Data.List as List
 import Data.Map (Map)
+import Data.Monoid
 import Data.Typeable (Typeable)
 
 import Debug.Trace (trace)
@@ -29,9 +35,10 @@ import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Position
 
-import Agda.TypeChecking.Monad.Base as Base
+import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Free as Free
 import Agda.TypeChecking.CompiledClause
+import Agda.TypeChecking.Positivity.Occurrence as Occ
 
 import Agda.Utils.Empty
 import Agda.Utils.Functor
@@ -58,6 +65,10 @@ class Apply t where
   applyE t es  = apply  t $ map argFromElim es
     -- precondition: all @es@ are @Apply@s
 
+-- | Apply to a single argument.
+apply1 :: Apply t => t -> Term -> t
+apply1 t u = apply t [ defaultArg u ]
+
 instance Apply Term where
   applyE m [] = m
   applyE m es =
@@ -150,17 +161,31 @@ instance Subst a => Apply (Tele a) where
   apply (ExtendTel _ tel) (t : ts) = lazyAbsApp tel (unArg t) `apply` ts
 
 instance Apply Definition where
-  apply (Defn info x t pol occ df m c rew inst d) args =
-    Defn info x (piApply t args) (apply pol args) (apply occ args) df m c (apply rew args) inst (apply d args)
+  apply (Defn info x t pol occ df m c inst d) args =
+    Defn info x (piApply t args) (apply pol args) (apply occ args) df m c inst (apply d args)
 
 instance Apply RewriteRule where
-  apply (RewriteRule q gamma lhs rhs t) args =
-    RewriteRule q (apply gamma args) lhs rhs t
+  apply r args = RewriteRule
+    { rewName    = rewName r
+    , rewContext = apply (rewContext r) args
+    , rewLHS     = applySubst sub (rewLHS r)
+    , rewRHS     = applySubst sub (rewRHS r)
+    , rewType    = applySubst sub (rewType r)
+    }
+    where sub = parallelS (map unArg args)
 
-instance Apply [Base.Occurrence] where
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} Apply [Occ.Occurrence] where
+#else
+instance Apply [Occ.Occurrence] where
+#endif
   apply occ args = List.drop (length args) occ
 
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} Apply [Polarity] where
+#else
 instance Apply [Polarity] where
+#endif
   apply pol args = List.drop (length args) pol
 
 instance Apply Projection where
@@ -181,20 +206,23 @@ instance Apply Defn where
         }
 
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
-            , funProjection = Just p} ->
-      case p `apply` args of
+            , funProjection = Just p0} ->
+      case p0 `apply` args of
         p@Projection{ projIndex = n }
           | n < 0     -> __IMPOSSIBLE__
           -- case: applied only to parameters
           | n > 0     -> d { funProjection = Just p }
-          -- case: applied also to record value
+          -- case: applied also to record value (n == 0)
           | otherwise ->
               d { funClauses        = apply cs args'
                 , funCompiled       = apply cc args'
                 , funInv            = apply inv args'
-                , funProjection     = Nothing -- WAS: Just $ p { projIndex = 0 }
+                , funProjection     = if isVar0 then Just p{ projIndex = 0 } else Nothing
                 }
-              where args' = [last args]  -- the record value
+              where
+                larg  = last args -- the record value
+                args' = [larg]
+                isVar0 = case ignoreSharing $ unArg larg of Var 0 [] -> True; _ -> False
 {-
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
             , funProjection = Just p@Projection{ projIndex = n } }
@@ -262,10 +290,10 @@ instance Apply a => Apply (WithArity a) where
   applyE (WithArity n a) es   = WithArity n $ applyE a es
 
 instance Apply a => Apply (Case a) where
-  apply (Branches cs ls m) args =
-    Branches (apply cs args) (apply ls args) (apply m args)
-  applyE (Branches cs ls m) es =
-    Branches (applyE cs es) (applyE ls es) (applyE m es)
+  apply (Branches cop cs ls m) args =
+    Branches cop (apply cs args) (apply ls args) (apply m args)
+  applyE (Branches cop cs ls m) es =
+    Branches cop (applyE cs es) (applyE ls es) (applyE m es)
 
 instance Apply FunctionInverse where
   apply NotInjective  args = NotInjective
@@ -287,10 +315,14 @@ instance Apply DisplayTerm where
   apply (DTerm v)          args = DTerm $ apply v args
   apply (DDot v)           args = DDot  $ apply v args
   apply (DCon c vs)        args = DCon c $ vs ++ map (fmap DTerm) args
-  apply (DDef c vs)        args = DDef c $ vs ++ map (fmap DTerm) args
+  apply (DDef c es)        args = DDef c $ es ++ map (Apply . fmap DTerm) args
   apply (DWithApp v ws args') args = DWithApp v ws $ args' ++ args
 
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPABLE #-} Apply t => Apply [t] where
+#else
 instance Apply t => Apply [t] where
+#endif
   apply  ts args = map (`apply` args) ts
   applyE ts es   = map (`applyE` es) ts
 
@@ -364,8 +396,8 @@ instance Abstract Telescope where
   ExtendTel arg xtel `abstract` tel = ExtendTel arg $ xtel <&> (`abstract` tel)
 
 instance Abstract Definition where
-  abstract tel (Defn info x t pol occ df m c rews inst d) =
-    Defn info x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c (abstract tel rews) inst (abstract tel d)
+  abstract tel (Defn info x t pol occ df m c inst d) =
+    Defn info x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c inst (abstract tel d)
 
 -- | @tel ⊢ (Γ ⊢ lhs ↦ rhs : t)@ becomes @tel, Γ ⊢ lhs ↦ rhs : t)@
 --   we do not need to change lhs, rhs, and t since they live in Γ.
@@ -374,11 +406,19 @@ instance Abstract RewriteRule where
   abstract tel (RewriteRule q gamma lhs rhs t) =
     RewriteRule q (abstract tel gamma) lhs rhs t
 
-instance Abstract [Base.Occurrence] where
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} Abstract [Occ.Occurrence] where
+#else
+instance Abstract [Occ.Occurrence] where
+#endif
   abstract tel []  = []
   abstract tel occ = replicate (size tel) Mixed ++ occ -- TODO: check occurrence
 
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} Abstract [Polarity] where
+#else
 instance Abstract [Polarity] where
+#endif
   abstract tel []  = []
   abstract tel pol = replicate (size tel) Invariant ++ pol -- TODO: check polarity
 
@@ -386,6 +426,8 @@ instance Abstract Projection where
   abstract tel p = p
     { projIndex    = size tel + projIndex p
     , projDropPars = abstract tel $ projDropPars p
+    , projArgInfo  = if projIndex p > 0 then projArgInfo p else
+       domInfo $ last $ telToList tel
     }
 
 instance Abstract Defn where
@@ -397,8 +439,19 @@ instance Abstract Defn where
         , funCompiled = abstract tel cc
         , funInv      = abstract tel inv
         }
-    Function{ funProjection = Just p } ->
-      d { funProjection = Just $ abstract tel p }
+    Function{ funClauses = cs, funCompiled = cc, funInv = inv
+            , funProjection = Just p } ->
+      -- Andreas, 2015-05-11 if projection was applied to Var 0
+      -- then abstract over last element of tel (the others are params).
+      if projIndex p > 0 then d' else
+        d' { funClauses  = abstract tel1 cs
+           , funCompiled = abstract tel1 cc
+           , funInv      = abstract tel1 inv
+           }
+        where
+          d' = d { funProjection = Just $ abstract tel p }
+          tel1 = telFromList $ drop (size tel - 1) $ telToList tel
+
     Datatype{ dataPars = np, dataSmallPars = sps, dataNonLinPars = nlps, dataClause = cl } ->
       d { dataPars       = np + size tel
         , dataSmallPars  = abstract tel sps
@@ -436,8 +489,8 @@ instance Abstract a => Abstract (WithArity a) where
   abstract tel (WithArity n a) = WithArity n $ abstract tel a
 
 instance Abstract a => Abstract (Case a) where
-  abstract tel (Branches cs ls m) =
-    Branches (abstract tel cs) (abstract tel ls) (abstract tel m)
+  abstract tel (Branches cop cs ls m) =
+    Branches cop (abstract tel cs) (abstract tel ls) (abstract tel m)
 
 telVars :: Telescope -> [Arg Pattern]
 telVars = map (fmap namedThing) . namedTelVars
@@ -456,7 +509,11 @@ instance Abstract ClauseBody where
   abstract EmptyTel          b = b
   abstract (ExtendTel _ tel) b = Bind $ fmap (`abstract` b) tel
 
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPABLE #-} Abstract t => Abstract [t] where
+#else
 instance Abstract t => Abstract [t] where
+#endif
   abstract tel = map (abstract tel)
 
 instance Abstract t => Abstract (Maybe t) where
@@ -496,9 +553,12 @@ consS (Var n []) (Wk m rho)
   | n + 1 == m = wkS (m - 1) (liftS 1 rho)
 consS u rho = seq u (u :# rho)
 
-singletonS :: Term -> Substitution
-singletonS u = consS u idS
+-- | To replace index @n@ by term @u@, do @applySubst (singletonS n u)@.
+singletonS :: Int -> Term -> Substitution
+singletonS n u = map var [0..n-1] ++# consS u (raiseS n)
+  -- ALT: foldl (\ s i -> var i `consS` s) (consS u $ raiseS n) $ downFrom n
 
+-- | Lift a substitution under k binders.
 liftS :: Int -> Substitution -> Substitution
 liftS 0 rho          = rho
 liftS k IdS          = IdS
@@ -557,6 +617,12 @@ parallelS us = us ++# idS
 compactS :: Empty -> [Maybe Term] -> Substitution
 compactS err us = prependS err us idS
 
+-- | Γ ⊢ (strengthenS ⊥ |Δ|) : Γ,Δ
+strengthenS :: Empty -> Int -> Substitution
+strengthenS err n
+  | n < 0     = __IMPOSSIBLE__
+  | otherwise = iterate (Strengthen err) idS !! n
+
 lookupS :: Substitution -> Nat -> Term
 lookupS rho i = case rho of
   IdS                    -> var i
@@ -596,14 +662,17 @@ raise = raiseFrom 0
 raiseFrom :: Subst t => Nat -> Nat -> t -> t
 raiseFrom n k = applySubst (liftS n $ raiseS k)
 
-subst :: Subst t => Term -> t -> t
-subst u t = substUnder 0 u t
+-- | Replace de Bruijn index i by a 'Term' in something.
+subst :: Subst t => Int -> Term -> t -> t
+subst i u = applySubst $ singletonS i u
 
 strengthen :: Subst t => Empty -> t -> t
 strengthen err = applySubst (compactS err [Nothing])
 
+-- | Replace what is now de Bruijn index 0, but go under n binders.
+--   @substUnder n u == subst n (raise n u)@.
 substUnder :: Subst t => Nat -> Term -> t -> t
-substUnder n u = applySubst (liftS n (singletonS u))
+substUnder n u = applySubst (liftS n (singletonS 0 u))
 
 instance Subst Substitution where
   applySubst rho sgm = composeS rho sgm
@@ -627,7 +696,7 @@ instance Subst Term where
 instance Subst a => Subst (Ptr a) where
   applySubst rho = fmap (applySubst rho)
 
-instance Subst Type where
+instance (Subst a) => Subst (Type' a) where
   applySubst rho (El s t) = applySubst rho s `El` applySubst rho t
 
 instance Subst Sort where
@@ -635,6 +704,7 @@ instance Subst Sort where
     Type n     -> levelSort $ sub n
     Prop       -> Prop
     Inf        -> Inf
+    SizeUniv   -> SizeUniv
     DLub s1 s2 -> DLub (sub s1) (sub s2)
     where sub x = applySubst rho x
 
@@ -648,15 +718,25 @@ instance Subst PlusLevel where
 instance Subst LevelAtom where
   applySubst rho (MetaLevel m vs)   = MetaLevel m    $ applySubst rho vs
   applySubst rho (BlockedLevel m v) = BlockedLevel m $ applySubst rho v
-  applySubst rho (NeutralLevel v)   = UnreducedLevel $ applySubst rho v
+  applySubst rho (NeutralLevel _ v) = UnreducedLevel $ applySubst rho v
   applySubst rho (UnreducedLevel v) = UnreducedLevel $ applySubst rho v
 
 instance Subst Bool where
   applySubst rho = id
 
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} Subst String where
+#else
+instance Subst String where
+#endif
+  applySubst rho = id
+
 instance Subst Name where
   applySubst rho = id
 
+instance Subst ConPatternInfo where
+  applySubst rho (ConPatternInfo mr mt) = ConPatternInfo mr $ applySubst rho mt
+
 instance Subst Pattern where
   applySubst rho p = case p of
     ConP c mt ps -> ConP c (applySubst rho mt) $ applySubst rho ps
@@ -665,6 +745,16 @@ instance Subst Pattern where
     LitP l       -> p
     ProjP _      -> p
 
+instance Subst NLPat where
+  applySubst rho p = case p of
+    PVar i -> p
+    PWild  -> p
+    PDef f es -> PDef f $ applySubst rho es
+    PLam i u -> PLam i $ applySubst rho u
+    PPi a b -> PPi (applySubst rho a) (applySubst rho b)
+    PBoundVar i es -> PBoundVar i $ applySubst rho es
+    PTerm u -> PTerm $ applySubst rho u
+
 instance Subst t => Subst (Blocked t) where
   applySubst rho b = fmap (applySubst rho) b
 
@@ -677,7 +767,7 @@ instance Subst DisplayTerm where
   applySubst rho (DTerm v)        = DTerm $ applySubst rho v
   applySubst rho (DDot v)         = DDot  $ applySubst rho v
   applySubst rho (DCon c vs)      = DCon c $ applySubst rho vs
-  applySubst rho (DDef c vs)      = DDef c $ applySubst rho vs
+  applySubst rho (DDef c es)      = DDef c $ applySubst rho es
   applySubst rho (DWithApp v vs ws) = uncurry3 DWithApp $ applySubst rho (v, vs, ws)
 
 instance Subst a => Subst (Tele a) where
@@ -694,14 +784,15 @@ instance Subst Constraint where
     LevelCmp cmp l1 l2       -> LevelCmp cmp (rf l1) (rf l2)
     Guarded c cs             -> Guarded (rf c) cs
     IsEmpty r a              -> IsEmpty r (rf a)
+    CheckSizeLtSat t         -> CheckSizeLtSat (rf t)
     FindInScope m cands      -> FindInScope m (rf cands)
     UnBlock{}                -> c
     where
       rf x = applySubst rho x
 
-instance Subst Elim where
+instance Subst a => Subst (Elim' a) where
   applySubst rho e = case e of
-    Apply v -> Apply (applySubst rho v)
+    Apply v -> Apply $ applySubst rho v
     Proj{}  -> e
 
 instance Subst a => Subst (Abs a) where
@@ -781,9 +872,26 @@ bindsToTel' f (x:xs) t = fmap (f x,) t : bindsToTel' f xs (raise 1 t)
 bindsToTel :: [Name] -> Dom Type -> ListTel
 bindsToTel = bindsToTel' nameToArgName
 
+-- | Turn a typed binding @(x1 .. xn : A)@ into a telescope.
+bindsWithHidingToTel' :: (Name -> a) -> [WithHiding Name] -> Dom Type -> ListTel' a
+bindsWithHidingToTel' f []                    t = []
+bindsWithHidingToTel' f (WithHiding h x : xs) t =
+  fmap (f x,) (mapHiding (mappend h) t) : bindsWithHidingToTel' f xs (raise 1 t)
+
+bindsWithHidingToTel :: [WithHiding Name] -> Dom Type -> ListTel
+bindsWithHidingToTel = bindsWithHidingToTel' nameToArgName
+
+-- | Takes off all exposed function domains from the given type.
+--   This means that it does not reduce to expose @Pi@-types.
 telView' :: Type -> TelView
-telView' t = case ignoreSharing $ unEl t of
-  Pi a b  -> absV a (absName b) $ telView' (absBody b)
+telView' = telView'UpTo (-1)
+
+-- | @telView'UpTo n t@ takes off the first @n@ exposed function types of @t@.
+-- Takes off all (exposed ones) if @n < 0@.
+telView'UpTo :: Int -> Type -> TelView
+telView'UpTo 0 t = TelV EmptyTel t
+telView'UpTo n t = case ignoreSharing $ unEl t of
+  Pi a b  -> absV a (absName b) $ telView'UpTo (n - 1) (absBody b)
   _       -> TelV EmptyTel t
   where
     absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
@@ -803,7 +911,7 @@ telePi' reAbs = telePi where
   telePi (ExtendTel u tel) t = el $ Pi u $ reAbs b
     where
       b  = (`telePi` t) <$> tel
-      s1 = getSort $ unDom u
+      s1 = getSort u
       s2 = getSort <$> b
       el = El $ dLub s1 s2
 
@@ -846,7 +954,7 @@ instance TeleNoAbs Telescope where
 --   @dLub s1 \i.s2 = \omega@ if @i@ appears in the rigid variables of @s2@.
 dLub :: Sort -> Abs Sort -> Sort
 dLub s1 (NoAbs _ s2) = sLub s1 s2
-dLub s1 b@(Abs _ s2) = case occurrence 0 $ freeVars s2 of
+dLub s1 b@(Abs _ s2) = case occurrence 0 s2 of
   Flexible      -> DLub s1 b
   Irrelevantly  -> DLub s1 b
   NoOccurrence  -> sLub s1 (noabsApp __IMPOSSIBLE__ b)
@@ -863,14 +971,14 @@ dLub s1 b@(Abs _ s2) = case occurrence 0 $ freeVars s2 of
 
 -- | Instantiate an abstraction. Strict in the term.
 absApp :: Subst t => Abs t -> Term -> t
-absApp (Abs   _ v) u = subst u v
+absApp (Abs   _ v) u = subst 0 u v
 absApp (NoAbs _ v) _ = v
 
 -- | Instantiate an abstraction. Lazy in the term, which allow it to be
 --   __IMPOSSIBLE__ in the case where the variable shouldn't be used but we
 --   cannot use 'noabsApp'. Used in Apply.
 lazyAbsApp :: Subst t => Abs t -> Term -> t
-lazyAbsApp (Abs   _ v) u = applySubst (u :# IdS) v
+lazyAbsApp (Abs   _ v) u = applySubst (u :# IdS) v  -- Note: do not use consS here!
 lazyAbsApp (NoAbs _ v) _ = v
 
 -- | Instantiate an abstraction that doesn't use its argument.
@@ -951,24 +1059,27 @@ instance GetBody Clause where
 -- * Syntactic equality and order
 ---------------------------------------------------------------------------
 
-deriving instance (Subst a, Eq a) => Eq (Tele a)
-deriving instance (Subst a, Ord a) => Ord (Tele a)
-
 deriving instance Eq Substitution
 deriving instance Ord Substitution
 
 deriving instance Eq Sort
 deriving instance Ord Sort
-deriving instance Eq Type
-deriving instance Ord Type
 deriving instance Eq Level
 deriving instance Ord Level
 deriving instance Eq PlusLevel
 deriving instance Ord LevelAtom
-deriving instance Eq Elim
-deriving instance Ord Elim
+deriving instance Eq NotBlocked
+deriving instance Ord NotBlocked
+deriving instance Eq t => Eq (Blocked t)
+deriving instance Ord t => Ord (Blocked t)
+
+deriving instance (Subst a, Eq a) => Eq (Elim' a)
+deriving instance (Subst a, Ord a) => Ord (Elim' a)
+deriving instance (Subst a, Eq a) => Eq (Tele a)
+deriving instance (Subst a, Ord a) => Ord (Tele a)
 
 deriving instance Eq Constraint
+deriving instance Eq Section
 
 instance Ord PlusLevel where
   compare ClosedLevel{} Plus{}            = LT
@@ -980,7 +1091,14 @@ instance Ord PlusLevel where
 instance Eq LevelAtom where
   (==) = (==) `on` unLevelAtom
 
--- | Syntactic equality, ignores stuff below @DontCare@.
+-- | Syntactic 'Type' equality, ignores sort annotations.
+instance Eq a => Eq (Type' a) where
+  (==) = (==) `on` unEl
+
+instance Ord a => Ord (Type' a) where
+  compare = compare `on` unEl
+
+-- | Syntactic 'Term' equality, ignores stuff below @DontCare@ and sharing.
 instance Eq Term where
   Var x vs   == Var x' vs'   = x == x' && vs == vs'
   Lam h v    == Lam h' v'    = h == h' && v  == v'
@@ -1046,13 +1164,28 @@ instance (Subst a, Ord a) => Ord (Abs a) where
 -- * Level stuff
 ---------------------------------------------------------------------------
 
+-- | The ``rule'', if Agda is considered as a functional
+--   pure type system (pts).
+--
+--   TODO: This needs to be properly implemented, requiring
+--   refactoring of Agda's handling of levels.
+--   Without impredicativity or 'SizeUniv', Agda's pts rule is
+--   just the least upper bound, which is total and commutative.
+--   The handling of levels relies on this simplification.
+pts :: Sort -> Sort -> Sort
+pts = sLub
+
 sLub :: Sort -> Sort -> Sort
 sLub s Prop = s
 sLub Prop s = s
 sLub Inf _ = Inf
 sLub _ Inf = Inf
+sLub SizeUniv s = s         -- one can freely quantify over sizes in any Set
+sLub _ SizeUniv = SizeUniv  -- but everything resulting in a size lives in the SizeUniv
 sLub (Type (Max as)) (Type (Max bs)) = Type $ levelMax (as ++ bs)
-sLub (DLub a b) c = DLub (sLub a c) b
+-- sLub (DLub a b) c = DLub (sLub a c) b -- no longer commutative!
+sLub (DLub a NoAbs{}) c = __IMPOSSIBLE__
+sLub (DLub a (Abs x b)) c = DLub a $ Abs x $ sLub b $ raise 1 c
 sLub a (DLub b c) = DLub (sLub a b) c
 
 lvlView :: Term -> Level
@@ -1081,7 +1214,7 @@ levelMax as0 = Max $ ns ++ List.sort bs
 
     expandAtom l = case l of
       BlockedLevel _ v -> expandTm v
-      NeutralLevel v   -> expandTm v
+      NeutralLevel _ v -> expandTm v
       UnreducedLevel v -> expandTm v
       MetaLevel{}      -> [Plus 0 l]
       where
@@ -1107,17 +1240,18 @@ sortTm s        = Sort s
 
 levelSort :: Level -> Sort
 levelSort (Max as)
-  | List.any isInf as = Inf
+  | List.any (levelIs Inf     ) as = Inf
+  | List.any (levelIs SizeUniv) as = SizeUniv
   where
-    isInf ClosedLevel{}        = False
-    isInf (Plus _ l)           = infAtom l
-    infAtom (NeutralLevel a)   = infTm a
-    infAtom (UnreducedLevel a) = infTm a
-    infAtom MetaLevel{}        = False
-    infAtom BlockedLevel{}     = False
-    infTm (Sort Inf)           = True
-    infTm (Shared p)           = infTm $ derefPtr p
-    infTm _                    = False
+    levelIs s ClosedLevel{}     = False
+    levelIs s (Plus _ l)        = atomIs s l
+    atomIs s (NeutralLevel _ a) = tmIs s a
+    atomIs s (UnreducedLevel a) = tmIs s a
+    atomIs s MetaLevel{}        = False
+    atomIs s BlockedLevel{}     = False
+    tmIs s (Sort s')            = s == s'
+    tmIs s (Shared p)           = tmIs s $ derefPtr p
+    tmIs s _                    = False
 levelSort l =
   case ignoreSharing $ levelTm l of
     Sort s -> s
@@ -1131,26 +1265,6 @@ levelTm l =
 
 unLevelAtom :: LevelAtom -> Term
 unLevelAtom (MetaLevel x es)   = MetaV x es
-unLevelAtom (NeutralLevel v)   = v
+unLevelAtom (NeutralLevel _ v) = v
 unLevelAtom (UnreducedLevel v) = v
 unLevelAtom (BlockedLevel _ v) = v
-
----------------------------------------------------------------------------
--- * Boring instances
----------------------------------------------------------------------------
-
-instance Sized Substitution where
-  size IdS                = 1
-  size EmptyS             = 1
-  size (Wk _ rho)         = 1 + size rho
-  size (t :# rho)         = 1 + size t + size rho
-  size (Strengthen _ rho) = 1 + size rho
-  size (Lift _ rho)       = 1 + size rho
-
-instance KillRange Substitution where
-  killRange IdS                  = IdS
-  killRange EmptyS               = EmptyS
-  killRange (Wk n rho)           = killRange1 (Wk n) rho
-  killRange (t :# rho)           = killRange2 (:#) t rho
-  killRange (Strengthen err rho) = killRange1 (Strengthen err) rho
-  killRange (Lift n rho)         = killRange1 (Lift n) rho
