@@ -1,8 +1,8 @@
 {-# LANGUAGE CPP                    #-}
+{-# LANGUAGE FlexibleContexts       #-}
 {-# LANGUAGE FlexibleInstances      #-}
 {-# LANGUAGE FunctionalDependencies #-}
 {-# LANGUAGE MultiParamTypeClasses  #-}
-{-# LANGUAGE TypeSynonymInstances   #-}
 {-# LANGUAGE UndecidableInstances   #-}
 
 module Agda.TypeChecking.Test.Generators where
@@ -433,8 +433,8 @@ instance ShrinkC a b => ShrinkC (I.Dom a) (I.Dom b) where
   noShrink = fmap noShrink
 
 instance ShrinkC a b => ShrinkC (Blocked a) (Blocked b) where
-  shrinkC conf (Blocked m x)  = Blocked m <$> shrinkC conf x
-  shrinkC conf (NotBlocked x) = NotBlocked <$> shrinkC conf x
+  shrinkC conf (Blocked m x)    = Blocked m <$> shrinkC conf x
+  shrinkC conf (NotBlocked r x) = NotBlocked r <$> shrinkC conf x
   noShrink = fmap noShrink
 
 instance ShrinkC a b => ShrinkC (Elim' a) (Elim' b) where
@@ -449,6 +449,7 @@ instance ShrinkC Sort Sort where
     Type n     -> [] -- No Level instance yet -- Type <$> shrinkC conf n
     Prop       -> __IMPOSSIBLE__
     Inf        -> []
+    SizeUniv   -> []
     DLub s1 s2 -> __IMPOSSIBLE__
   noShrink = id
 
@@ -548,8 +549,8 @@ instance (KillVar a, KillVar b) => KillVar (a, b) where
 
 -- Tests ------------------------------------------------------------------
 
-isWellScoped :: Free a => TermConfiguration -> a -> Bool
-isWellScoped conf t = allVars (freeVars t) `Set.isSubsetOf` Set.fromList (tcFreeVariables conf)
+isWellScoped :: FreeVS a => TermConfiguration -> a -> Bool
+isWellScoped conf t = allFreeVars t `Set.isSubsetOf` Set.fromList (tcFreeVariables conf)
 
 -- | Check that the generated terms don't have any out of scope variables.
 prop_wellScopedVars :: TermConfiguration -> Property
