@@ -1,3 +1,5 @@
+{-# LANGUAGE CPP #-}
+
 module Agda.TypeChecking.Tests where
 
 import qualified Agda.Utils.VarSet as Set
@@ -11,6 +13,9 @@ import Agda.Utils.Size
 import Agda.Utils.Permutation
 import Agda.Utils.TestHelpers
 
+#include "undefined.h"
+import Agda.Utils.Impossible
+
 ---------------------------------------------------------------------------
 -- * Tests for "Agda.Utils.Permutation"
 ---------------------------------------------------------------------------
@@ -46,7 +51,7 @@ prop_reorderTelStable conf =
 -- | The result of splitting a telescope is well-scoped.
 prop_splitTelescopeScope :: TermConfiguration -> Property
 prop_splitTelescopeScope conf =
-  forAll (genC conf)			    $ \tel ->
+  forAll (genC conf)                        $ \tel ->
   forAll (listOfElements [0..size tel - 1]) $ \vs ->
   let SplitTel tel1 tel2 perm = splitTelescope (Set.fromList vs) tel
       tel' = telFromList (telToList tel1 ++ telToList tel2)
@@ -55,21 +60,21 @@ prop_splitTelescopeScope conf =
 -- | The permutation generated when splitting a telescope preserves scoping.
 prop_splitTelescopePermScope :: TermConfiguration -> Property
 prop_splitTelescopePermScope conf =
-      forAllShrink (genC conf) (shrinkC conf)		     $ \tel ->
+      forAllShrink (genC conf) (shrinkC conf)                $ \tel ->
       forAllShrink (listOfElements [0..size tel - 1]) shrink $ \vs ->
   let SplitTel tel1 tel2 perm = splitTelescope (Set.fromList vs) tel
       conf1 = extendWithTelConf tel1 conf
       conf2 = conf1 { tcFreeVariables = map (size tel2 +) (tcFreeVariables conf1) }
       conf' = conf  { tcFreeVariables = map (size tel +) (tcFreeVariables conf) ++ vs }
   in  forAllShrink (genC conf') (shrinkC conf') $ \t ->
-      isWellScoped conf2 (applySubst (renamingR $ invertP perm) (t :: Term))
+      isWellScoped conf2 (applySubst (renamingR $ invertP __IMPOSSIBLE__ perm) (t :: Term))
 
 {-
 -- | The permutation generated when splitting a telescope correctly translates
 --   between the old and the new telescope.
 prop_splitTelescopePermInv :: TermConfiguration -> Property
 prop_splitTelescopePermInv conf =
-      forAll (wellScopedTel conf)		$ \tel ->
+      forAll (wellScopedTel conf)               $ \tel ->
       forAll (listOfElements [0..size tel - 1]) $ \vs ->
   let SplitTel tel1 tel2 perm = splitTelescope (Set.fromList vs) tel
       tel' = telFromList (telToList tel1 ++ telToList tel2)
@@ -77,8 +82,8 @@ prop_splitTelescopePermInv conf =
       conf2 = extendWithTelConf tel' conf
   in  forAll (wellScopedTerm conf1) $ \t1 ->
       forAll (wellScopedTerm conf2) $ \t2 ->
-  let t1' = rename (invertP perm) $ rename perm t1
-      t2' = rename perm $ rename (invertP perm) t2
+  let t1' = rename (invertP __IMPOSSIBLE__ perm) $ rename perm t1
+      t2' = rename perm $ rename (invertP __IMPOSSIBLE__ perm) t2
   in  t1 == t1' && t2 == t2'
 -}
 
