@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP           #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE TupleSections #-}
 
 -- | Utilities for functors.
@@ -9,16 +9,20 @@ module Agda.Utils.Functor
   )
   where
 
+import Control.Applicative ( Const(Const), getConst )
+
 import Data.Functor
 import Data.Functor.Identity
 import Data.Functor.Compose
-import Data.Functor.Constant
+import Data.Functor.Classes
 
--- ASR (15 October 2014): See issue 1304.
-#if !MIN_VERSION_transformers(0,4,1)
-instance Eq a => Eq (Identity a) where
-  Identity x == Identity x' = x == x'
-#endif
+-- Andreas, 2015-05-06, see issue 1501.
+-- This instance can be obtained from transformers-compat.
+-- -- ASR (15 October 2014): See issue 1304.
+-- #if !MIN_VERSION_transformers(0,4,1)
+-- instance Eq a => Eq (Identity a) where
+--   Identity x == Identity x' = x == x'
+-- #endif
 
 infixr 4 $>
 
@@ -65,9 +69,9 @@ class Functor t => Decoration t where
 dmap :: Decoration t => (a -> b) -> t a -> t b
 dmap f = runIdentity . traverseF (Identity . f)
 
--- | Any decoration is a lens.  @set@ is a special case of @fmap@.
+-- | Any decoration is a lens.  @set@ is a special case of @dmap@.
 dget :: Decoration t => t a -> a
-dget = getConstant . traverseF Constant
+dget = getConst . traverseF Const
 
 -- | The identity functor is a decoration.
 instance Decoration Identity where
