@@ -8,10 +8,11 @@ module Agda.Utils.Lens
   , (<&>) -- reexported from Agda.Utils.Functor
   ) where
 
+import Control.Applicative ( Const(Const), getConst )
 import Control.Monad.State
 import Control.Monad.Reader
+import Control.Monad.Writer
 
-import Data.Functor.Constant
 import Data.Functor.Identity
 
 import Agda.Utils.Functor ((<&>))
@@ -28,7 +29,7 @@ type Lens' i o = forall f. Functor f => (i -> f i) -> o -> f o
 infixl 8 ^.
 -- | Get inner part @i@ of structure @o@ as designated by @Lens' i o@.
 (^.) :: o -> Lens' i o -> i
-o ^. l = getConstant $ l Constant o
+o ^. l = getConst $ l Const o
 
 -- | Set inner part @i@ of structure @o@ as designated by @Lens' i o@.
 set :: Lens' i o -> i -> o -> o
@@ -45,16 +46,29 @@ over l f o = runIdentity $ l (Identity . f) o
 use :: MonadState o m => Lens' i o -> m i
 use l = gets (^. l)
 
-infix  4 .=
+infix 4 .=
 -- | Write a part of the state.
 (.=) :: MonadState o m => Lens' i o -> i -> m ()
 l .= i = modify $ set l i
 
-infix  4 %=
+infix 4 %=
 -- | Modify a part of the state.
 (%=) :: MonadState o m => Lens' i o -> (i -> i) -> m ()
 l %= f = modify $ over l f
 
+infix 4 %==
+-- | Modify a part of the state monadically.
+(%==) :: (MonadState o m, Functor m) => Lens' i o -> (i -> m i) -> m ()
+l %== f = put =<< l f =<< get
+
+infix 4 %%=
+-- | Modify a part of the state monadically, and return some result.
+(%%=) :: (MonadState o m, Functor m) => Lens' i o -> (i -> m (i, r)) -> m r
+l %%= f = do
+  o <- get
+  (o', r) <- runWriterT $ l (WriterT . f) o
+  put o'
+  return r
 
 -- * Read-only state accessors and modifiers.
 
@@ -65,4 +79,3 @@ view l = asks (^. l)
 -- | Modify a part of the state in a subcomputation.
 locally :: MonadReader o m => Lens' i o -> (i -> i) -> m a -> m a
 locally l = local . over l
-
