@@ -1,4 +1,4 @@
--- {-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP #-}
 
 -- | Extend 'Data.Maybe' by common operations for the 'Maybe' type.
 --
@@ -11,7 +11,14 @@ module Agda.Utils.Maybe
     , module Data.Maybe
     ) where
 
+import Control.Monad
+import Control.Monad.Trans.Maybe
+
 import Data.Maybe
+import Data.Functor
+
+#include "undefined.h"
+import Agda.Utils.Impossible
 
 -- * Collection operations.
 
@@ -48,9 +55,9 @@ forMaybe = flip mapMaybe
 --   in the 'Just' case, but only a default action in the 'Nothing'
 --   case.  Then, the argument ordering of @caseMaybe@ is preferable.
 --
---   @caseMaybe m err f = flip (maybe err) m f@
+--   @caseMaybe m d f = flip (maybe d) m f@
 caseMaybe :: Maybe a -> b -> (a -> b) -> b
-caseMaybe m err f = maybe err f m
+caseMaybe m d f = maybe d f m
 
 -- * Monads and Maybe.
 
@@ -67,7 +74,7 @@ fromMaybeM m mm = maybeM m return mm
 -- | Monadic version of 'caseMaybe'.
 --   That is, 'maybeM' with a different argument ordering.
 caseMaybeM :: Monad m => m (Maybe a) -> m b -> (a -> m b) -> m b
-caseMaybeM mm err f = maybeM  err f mm
+caseMaybeM mm d f = maybeM d f mm
 
 -- | 'caseMaybeM' with flipped branches.
 ifJustM :: Monad m => m (Maybe a) -> (a -> m b) -> m b -> m b
@@ -78,6 +85,39 @@ ifJustM mm = flip (caseMaybeM mm)
 whenJust :: Monad m => Maybe a -> (a -> m ()) -> m ()
 whenJust m k = caseMaybe m (return ()) k
 
+-- | 'caseMaybe' without the 'Just' case.
+whenNothing :: Monad m => Maybe a -> m () -> m ()
+whenNothing m d = caseMaybe m d (\_ -> return ())
+
 -- | 'caseMaybeM' without the 'Nothing' case.
 whenJustM :: Monad m => m (Maybe a) -> (a -> m ()) -> m ()
 whenJustM c m = c >>= (`whenJust` m)
+
+-- | 'caseMaybeM' without the 'Just' case.
+whenNothingM :: Monad m => m (Maybe a) -> m () -> m ()
+whenNothingM mm d = mm >>= (`whenNothing` d)
+
+-- | Lazy version of @allJust <.> sequence@.
+--   (@allJust = mapM@ for the @Maybe@ monad.)
+--   Only executes monadic effect while @isJust@.
+allJustM :: Monad m => [m (Maybe a)] -> m (Maybe [a])
+allJustM = runMaybeT . mapM MaybeT
+-- allJustM []         = return $ Just []
+-- allJustM (mm : mms) = caseMaybeM mm (return Nothing) $ \ a ->
+--   fmap (a:) <$> allJust mms
+
+-- | Precondition: list not empty.
+--   @
+--     allJustsOrNothings [Nothing, Nothing] = Just Nothing
+--     allJustsOrNothings [Just 0, Just 1]   = Just $ Just [0,1]
+--     allJustsOrNothings [Just 0, Nothing]  = Nothing
+--   @
+allJustsOrNothings :: [Maybe a] -> Maybe (Maybe [a])
+allJustsOrNothings [] = __IMPOSSIBLE__
+allJustsOrNothings [ma] = return $ (:[]) <$> ma
+allJustsOrNothings (ma : mas) = do
+  res <- allJustsOrNothings mas
+  case (ma, res) of
+    (Nothing, Nothing) -> return Nothing
+    (Just a, Just as)  -> return $ Just (a : as)
+    _                  -> mzero
