@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP                #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 
 #if __GLASGOW_HASKELL__ >= 706
@@ -6,7 +6,7 @@
 #endif
 
 {-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE StandaloneDeriving    #-}
+{-# LANGUAGE StandaloneDeriving #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -28,25 +28,24 @@ module Agda.Utils.Maybe.Strict
   , module Agda.Utils.Maybe.Strict
   ) where
 
-import Data.Strict.Maybe
-
 -- The following code is copied from
 -- http://hackage.haskell.org/package/strict-base-types-0.3.0/docs/src/Data-Maybe-Strict.html
 
-import           Prelude             hiding (Maybe (..), maybe)
+import           Prelude             hiding (Maybe (..), maybe, null)
 import qualified Prelude             as Lazy
 
 import           Control.Applicative (pure, (<$>))
 import           Control.DeepSeq     (NFData (..))
 import           Data.Binary         (Binary (..))
-#if MIN_VERSION_base(4,7,0)
-import           Data.Data           (Data (..), Typeable)
-#else
-import           Data.Data           (Data (..), Typeable1 (..))
-#endif
+import           Data.Data           (Data (..))
 import           Data.Monoid         (Monoid (..))
 import           Data.Foldable       (Foldable (..))
 import           Data.Traversable    (Traversable (..))
+#if MIN_VERSION_base(4,7,0)
+import           Data.Typeable       (Typeable)
+#else
+import           Data.Typeable       (Typeable1)
+#endif
 import           Data.Strict.Maybe   (Maybe (Nothing, Just), fromJust,
                                       fromMaybe, isJust, isNothing, maybe)
 #if __GLASGOW_HASKELL__ >= 706
@@ -54,6 +53,7 @@ import           GHC.Generics        (Generic (..))
 #endif
 import           Test.QuickCheck     (Arbitrary (..))
 
+import Agda.Utils.Null
 
 toStrict :: Lazy.Maybe a -> Maybe a
 toStrict Lazy.Nothing  = Nothing
@@ -74,6 +74,10 @@ deriving instance Typeable1 Maybe
 deriving instance Generic  (Maybe a)
 #endif
 
+instance Null (Maybe a) where
+  empty = Nothing
+  null = isNothing
+
 instance Monoid a => Monoid (Maybe a) where
   mempty = Nothing
 
