@@ -1,19 +1,19 @@
-{-# LANGUAGE CPP                #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE DeriveFoldable     #-}
-{-# LANGUAGE DeriveFunctor      #-}
-{-# LANGUAGE DeriveTraversable  #-}
-{-# LANGUAGE FlexibleInstances  #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE TupleSections      #-}
+{-# LANGUAGE TupleSections #-}
 
 module Agda.Utils.Permutation where
 
-import Prelude hiding (drop)
+import Prelude hiding (drop, null)
 
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
-import Data.List hiding (drop)
+import Data.List hiding (drop, null)
 import qualified Data.List as List
 import Data.Maybe
 
@@ -21,8 +21,12 @@ import Data.Foldable (Foldable)
 import Data.Traversable (Traversable)
 import Data.Typeable (Typeable)
 
+import Test.QuickCheck
+
+import Agda.Syntax.Position (KillRange(..))
 import Agda.Utils.Functor
 import Agda.Utils.List ((!!!))
+import Agda.Utils.Null
 import Agda.Utils.Size
 
 #include "undefined.h"
@@ -53,6 +57,13 @@ instance Show Permutation where
 instance Sized Permutation where
   size (Perm _ xs) = size xs
 
+instance Null Permutation where
+  empty = Perm 0 []
+  null (Perm _ picks) = null picks
+
+instance KillRange Permutation where
+  killRange = id
+
 -- | @permute [1,2,0] [x0,x1,x2] = [x1,x2,x0]@
 --   More precisely, @permute indices list = sublist@, generates @sublist@
 --   from @list@ by picking the elements of list as indicated by @indices@.
@@ -96,6 +107,9 @@ instance InversePermute [Maybe a] [Maybe a] where
   inversePermute p@(Perm n _) = tabulate . inversePermute p
     where tabulate m = for [0..n-1] $ \ i -> IntMap.lookup i m
 
+instance InversePermute (Int -> a) [Maybe a] where
+  inversePermute (Perm n xs) f = for [0..n-1] $ \ x -> f <$> findIndex (x ==) xs
+
 -- | Identity permutation.
 idP :: Int -> Permutation
 idP n = Perm n [0..n - 1]
@@ -160,8 +174,23 @@ compactP (Perm n xs) = Perm m $ map adjust xs
 --     == reverse $ permute (Perm 4 [1,3,0]) [x3,x2,x1,x0]
 --     == reverse $ permute (Perm 4 [1,3,0]) $ reverse [x0,x1,x2,x3]
 --   @
+--
+--   With @reverseP@, you can convert a permutation on de Bruijn indices
+--   to one on de Bruijn levels, and vice versa.
 reverseP :: Permutation -> Permutation
 reverseP (Perm n xs) = Perm n $ map ((n - 1) -) $ reverse xs
+                  -- = flipP $ Perm n $ reverse xs
+
+-- | @permPicks (flipP p) = permute p (downFrom (permRange p))@
+--   or
+--   @permute (flipP (Perm n xs)) [0..n-1] = permute (Perm n xs) (downFrom n)@
+--
+--   Can be use to turn a permutation from (de Bruijn) levels to levels
+--   to one from levels to indices.
+--
+--   See 'Agda.Syntax.Internal.Patterns.numberPatVars'.
+flipP :: Permutation -> Permutation
+flipP (Perm n xs) = Perm n $ map (n - 1 -) xs
 
 -- | @expandP i n π@ in the domain of @π@ replace the /i/th element by /n/ elements.
 expandP :: Int -> Int -> Permutation -> Permutation
@@ -203,6 +232,9 @@ data Drop a = Drop
   }
   deriving (Eq, Ord, Show, Typeable, Functor, Foldable, Traversable)
 
+instance KillRange a => KillRange (Drop a) where
+  killRange = fmap killRange
+
 -- | Things that support delayed dropping.
 class DoDrop a where
 
@@ -226,5 +258,15 @@ instance DoDrop Permutation where
   unDrop m = dropMore (-m) -- allow picking up more than dropped
 
 ------------------------------------------------------------------------
--- * Properties, see 'Agda.Utils.Permutation.Tests'
+-- * Test data generator
+------------------------------------------------------------------------
+
+instance Arbitrary Permutation where
+  arbitrary = do
+    is <- nub . map getNonNegative <$> arbitrary
+    NonNegative n <- arbitrary
+    return $ Perm (if null is then n else maximum is + n + 1) is
+
+------------------------------------------------------------------------
+-- * Properties, see "Agda.Utils.Permutation.Tests".
 ------------------------------------------------------------------------
