@@ -14,7 +14,7 @@ import Data.Hashable
 import Data.IORef
 import Data.Traversable
 import System.IO.Unsafe
-import Data.Typeable
+import Data.Typeable ( Typeable )
 
 data Ptr a = Ptr { ptrTag :: !Integer
                  , ptrRef :: !(IORef a) }
@@ -73,4 +73,4 @@ instance Ord (Ptr a) where
 instance Hashable (Ptr a) where
   hashWithSalt salt = (hashWithSalt salt) . ptrTag
 
-instance NFData (Ptr a) where
+instance NFData (Ptr a) where rnf x = seq x ()
