@@ -1,5 +1,5 @@
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
 
 {-| Pretty printing functions.
 -}
@@ -10,7 +10,11 @@ module Agda.Utils.Pretty
 
 import Data.Function
 import Data.Int ( Int32 )
-import Text.PrettyPrint hiding (TextDetails(Str))
+import Data.Monoid
+
+import Text.PrettyPrint hiding (TextDetails(Str), empty)
+
+import Agda.Utils.Null
 
 -- * Pretty class
 
@@ -22,42 +26,40 @@ import Text.PrettyPrint hiding (TextDetails(Str))
 --   implement 'pretty' as @pretty a = text $ ... a ...@.
 
 class Pretty a where
-    pretty      :: a -> Doc
-    prettyPrec  :: Int -> a -> Doc
+  pretty      :: a -> Doc
+  prettyPrec  :: Int -> a -> Doc
 
-    pretty      = prettyPrec 0
-    prettyPrec  = const pretty
+  pretty      = prettyPrec 0
+  prettyPrec  = const pretty
 
 -- | Use instead of 'show' when printing to world.
 
 prettyShow :: Pretty a => a -> String
 prettyShow = render . pretty
 
+-- | Space separated list of pretty things.
+
+prettyList :: Pretty a => [a] -> Doc
+prettyList = sep . map (prettyPrec 10000)
+
 -- * Pretty instances
 
-instance Pretty Int where
-    pretty = text . show
-
-instance Pretty Int32 where
-    pretty = text . show
-
-instance Pretty Integer where
-    pretty = text . show
+instance Pretty Bool    where pretty = text . show
+instance Pretty Int     where pretty = text . show
+instance Pretty Int32   where pretty = text . show
+instance Pretty Integer where pretty = text . show
 
 instance Pretty Char where
-    pretty c = text [c]
-
-instance Pretty String where
-    pretty = text
+  pretty c = text [c]
 
 instance Pretty Doc where
-    pretty = id
+  pretty = id
+
+instance Pretty String where
+  pretty = text
 
 -- * 'Doc' utilities
 
-instance Eq Doc where
-  (==) = (==) `on` render
-
 pwords :: String -> [Doc]
 pwords = map text . words
 
