@@ -1,51 +1,36 @@
-{-# LANGUAGE CPP #-}
-
-#if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ >= 702
-{-# OPTIONS_GHC -fno-warn-identities #-}
-#endif
--- To avoid warning on derived Integral instance for CPUTime
-
+{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
+-- To avoid warning on derived Integral instance for CPUTime.
+{-# OPTIONS_GHC -fno-warn-identities #-}
+
 -- | Time-related utilities.
 
 module Agda.Utils.Time
   ( ClockTime
   , getClockTime
+  , getCPUTime
   , measureTime
   , CPUTime(..)
   ) where
 
 import Control.Monad.Trans
-import System.CPUTime
+import qualified System.CPUTime as CPU
 
-#if MIN_VERSION_directory(1,1,1)
+import Data.Functor
 import qualified Data.Time
-#else
-import qualified System.Time
-#endif
 
 import Agda.Utils.Pretty
 import Agda.Utils.String
 
 -- | Timestamps.
 
-type ClockTime =
-#if MIN_VERSION_directory(1,1,1)
-  Data.Time.UTCTime
-#else
-  System.Time.ClockTime
-#endif
+type ClockTime = Data.Time.UTCTime
 
 -- | The current time.
 
 getClockTime :: IO ClockTime
-getClockTime =
-#if MIN_VERSION_directory(1,1,1)
-  Data.Time.getCurrentTime
-#else
-  System.Time.getClockTime
-#endif
+getClockTime = Data.Time.getCurrentTime
 
 -- | CPU time in pico (10^-12) seconds.
 
@@ -58,11 +43,15 @@ instance Pretty CPUTime where
   pretty (CPUTime ps) =
     text $ showThousandSep (div ps 1000000000) ++ "ms"
 
--- | Measure the time of a computation. Returns the
+{-# SPECIALIZE getCPUTime :: IO CPUTime #-}
+getCPUTime :: MonadIO m => m CPUTime
+getCPUTime = liftIO $ CPUTime <$> CPU.getCPUTime
+
+-- | Measure the time of a computation.
+--   Of course, does not work with exceptions.
 measureTime :: MonadIO m => m a -> m (a, CPUTime)
 measureTime m = do
   start <- liftIO $ getCPUTime
   x     <- m
   stop  <- liftIO $ getCPUTime
-  return (x, CPUTime $ stop - start)
-
+  return (x, stop - start)
