@@ -1,10 +1,10 @@
-{-# LANGUAGE CPP                        #-}
-{-# LANGUAGE BangPatterns               #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE PatternGuards              #-}
-{-# LANGUAGE TupleSections              #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TupleSections #-}
 
--- | Strict tries (based on 'Data.Map.Strict' and 'Agda.Utils.Maybe.Strict').
+-- | Strict tries (based on "Data.Map.Strict" and "Agda.Utils.Maybe.Strict").
 
 module Agda.Utils.Trie
   ( Trie
@@ -13,24 +13,20 @@ module Agda.Utils.Trie
   , lookupPath
   ) where
 
-import Prelude hiding (Maybe(..), maybe)
+import Prelude hiding (Maybe(..), maybe, null)
 import qualified Prelude as Lazy
 
 import Data.Function
 import Data.Functor
 import Data.List (nubBy, sortBy, isPrefixOf)
-#if MIN_VERSION_containers(0,5,0)
 import Data.Map.Strict (Map)
 import qualified Data.Map.Strict as Map
-#else
-import Data.Map (Map)
-import qualified Data.Map as Map
-#endif
 
 import Test.QuickCheck
 
 import Agda.Utils.List (mcons)
 import Agda.Utils.Maybe.Strict
+import Agda.Utils.Null
 
 -- | Finite map from @[k]@ to @v@.
 --
@@ -39,8 +35,9 @@ data Trie k v = Trie !(Maybe v) !(Map k (Trie k v))
   deriving Show
 
 -- | Empty trie.
-empty :: Trie k v
-empty = Trie Nothing Map.empty
+instance Null (Trie k v) where
+  empty = Trie Nothing Map.empty
+  null (Trie v t) = null v && null t
 
 -- | Singleton trie.
 singleton :: [k] -> v -> Trie k v
