@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 -- | Var field implementation of sets of (small) natural numbers.
 
@@ -20,11 +19,6 @@ type VarSet = IntSet
 subtract :: Int -> VarSet -> VarSet
 subtract n = IntSet.map (Prelude.subtract n)
 
-#if !MIN_VERSION_containers(0,5,0)
-toDescList :: VarSet -> [Int]
-toDescList = reverse . toList
-#endif
-
 {-
 import Data.Bits
 
